/**
 * app.js
 */

const Koa        = require('koa');
const debug      = require('debug')('light:app');
const CONST      = require('./lib/constant');
const developer  = require('./lib/developer');
const helper     = require('./lib/helper');
const middleware = require('./lib/middleware');
const setting    = require('./lib/setting');
const job        = require('./lib/job');
const log        = require('./lib/log');
const rider      = require('./lib/model/datarider');
const socket     = require('./lib/socket');
const udp        = require('./lib/udp');
const Model      = require('./lib/db/mongo/model');
const pool       = require('./lib/db/mongo/pool');

class App {

  /**
   * 初始化运行环境
   */
  static async init() {

    debug('init web app');

    developer.loadConfig();         // 加载开发环境配置 config.yaml
    if (helper.isCron()) {          // 定时处理
      return await this.initCron();
    }

    developer.preLoadController();  // 校验controller的基本语法
    developer.collectCoverage();    // 收集代码覆盖率
    developer.watch();              // 监视文件变化

    await this.initServer();

  }

  /**
   * 初始化定时处理用的服务
   */
  static async initCron() {

    try {
      await setting.init();
      rider.init();
      job.init();
    } catch (err) {
      log.error(err);
      process.exit(-1);
    }

  }

  /**
   * 初始化Web服务（包括http，websocket，udp）
   */
  static async initServer() {

    try {

      await setting.init();

      const model = new Model(process.env.APPNAME, CONST.SYSTEM_DB_PREFIX);
      this.koa = new Koa();
      this.koa.context.db = await model.db;

      rider.init();
      middleware.init(this.koa);

      this.server = this.koa.listen(process.env.APPPORT || 7000);
      socket.listen(this.server, this.koa.context.db);
      udp.listen();

      log.info(`Light server listening on port ${process.env.APPPORT || 7000}`);
    } catch (err) {
      log.error(err);
      process.exit(-1);
    }

  }

  static async destory() {
    await pool.destroy();
  }

}

if (!module.parent) App.init();
module.exports = App;
