/**
 * dump.js
 * 下载数据
 *
 * 用法：
 *  $ node bin/dump --system
 *  注 - 需要配置config.yaml
 */

'use strict';

const fs      = require('fs');
const path    = require('path');
const light   = require('../lib/model/light');
const helper  = require('../lib/helper');
const CONST   = require('../lib/constant');
const Context = require('../lib/context');
const debug   = require('debug')('light:bin');

class Dump {

  constructor(scope = [], system = false, folder = 'setting') {

    debug('scope', scope);
    debug('is system', system);
    debug('folder', folder);

    this.folder = folder;
    this.scope = scope;
    this.system = system;
  }

  start() {

    const handler = new Context(CONST.DEFAULT_USER, undefined, CONST.SYSTEM_DB_PREFIX);

    return new Promise(resolve => {

      fs.access(this.folder, fs.constants.F_OK, async err => {
        if (err) {
          fs.mkdirSync(this.folder);
        }

        if (this.scope.includes('struct')) {
          await this.dumpStructure(handler);
        }
        if (this.scope.includes('board')) {
          await this.dumpBoard(handler);
        }
        if (this.scope.includes('config')) {
          await this.dumpConfig(handler);
        }
        if (this.scope.includes('validator')) {
          await this.dumpValidator(handler);
        }
        if (this.scope.includes('route')) {
          await this.dumpRoute(handler);
        }
        if (this.scope.includes('function')) {
          await this.dumpFunction(handler);
        }
        if (this.scope.includes('i18n')) {
          await this.dumpI18n(handler);
        }
        if (this.scope.includes('job')) {
          await this.dumpJob(handler);
        }

        resolve();
      });

    });
  }

  async dumpStructure(handler) {

    debug('dump structure');

    const { items } = await light.structure.list(handler);
    const result = items.map(item => {

      this.deleteCommonField(item);
      this.deleteIfZero(item, 'public');
      this.deleteIfZero(item, 'lock');
      this.deleteIfEmpty(item, 'extend');
      this.deleteIfBlank(item, 'version');

      return item;
    });

    debug('Output struct');
    const data = result.filter(item => (this.system ? [2, 3] : [0, 1]).includes(item.kind));
    fs.writeFileSync(
      path.join(this.folder, `${CONST.SYSTEM_DB_STRUCTURE}.yml`),
      helper.yamlDumper(data)
    );
    debug(`A total of ${data.length} records.`);
    console.info('dump structure :', `${data.length} records.`);
  }

  async dumpBoard(handler) {

    debug('dump board');

    const { items } = await light.board.list(handler);
    const result = items.map(item => {
      this.deleteCommonField(item);
      this.deleteIfEmpty(item, 'sorts');
      this.deleteIfEmpty(item, 'filters');
      this.deleteIfBlank(item, 'path');
      return item;
    });

    debug('Output board');
    const data = result.filter(item => (this.system ? [2, 3] : [0, 1]).includes(item.kind));
    fs.writeFileSync(
      path.join(this.folder, `${CONST.SYSTEM_DB_BOARD}.yml`),
      helper.yamlDumper(data)
    );
    debug(`A total of ${data.length} records.`);
    console.info('dump board :', `${data.length} records.`);
  }

  async dumpRoute(handler) {

    debug('dump route');

    const { items } = await light.route.list(handler);
    const data = items.map(item => {
      this.deleteCommonField(item);
      this.deleteIfBlank(item, 'class');
      this.deleteIfBlank(item, 'action');
      return item;
    });

    fs.writeFileSync(
      path.join(this.folder, `${CONST.SYSTEM_DB_ROUTE}.yml`),
      helper.yamlDumper(data)
    );
    debug(`A total of ${data.length} records.`);
    console.info('dump route :', `${data.length} records.`);
  }

  async dumpConfig(handler) {

    debug('dump config');

    const { items } = await light.configuration.list(handler);
    const data = items.map(item => {
      this.deleteCommonField(item);
      this.deleteIfBlank(item, 'description');
      this.deleteIfZero(item, 'displayType');
      this.deleteIfEmpty(item, 'options');
      return item;
    });

    fs.writeFileSync(
      path.join(this.folder, `${CONST.SYSTEM_DB_CONFIG}.yml`),
      helper.yamlDumper(data)
    );
    debug(`Download config. A total of ${data.length} records.`);
    console.info('dump config :', `${data.length} records.`);
  }

  async dumpFunction(handler) {

    debug('dump function');

    const { items } = await light.function.list(handler);
    const data = items.map(item => {
      this.deleteCommonField(item);
      return item;
    });

    fs.writeFileSync(
      path.join(this.folder, `${CONST.SYSTEM_DB_FUNCTION}.yml`),
      helper.yamlDumper(data)
    );
    debug(`Download function. A total of ${data.length} records.`);
    console.info('dump function :', `${data.length} records.`);
  }

  async dumpValidator(handler) {

    debug('dump validator');

    const { items } = await light.validator.list(handler);
    const data = items.map(item => {
      this.deleteCommonField(item);
      this.deleteIfBlank(item, 'sanitize');
      this.deleteIfBlank(item, 'class');
      this.deleteIfBlank(item, 'action');
      this.deleteIfBlank(item, 'description');
      return item;
    });

    fs.writeFileSync(
      path.join(this.folder, `${CONST.SYSTEM_DB_VALIDATOR}.yml`),
      helper.yamlDumper(data)
    );
    debug(`Download validator. A total of ${data.length} records.`);
    console.info('dump validator :', `${data.length} records.`);
  }

  async dumpI18n(handler) {

    debug('dump i18n');

    const { items } = await light.i18n.list(handler);
    const data = items.map(item => {
      this.deleteCommonField(item);
      return item;
    });

    fs.writeFileSync(
      path.join(this.folder, `${CONST.SYSTEM_DB_I18N}.yml`),
      helper.yamlDumper(data)
    );
    debug(`Download i18n. A total of ${data.length} records.`);
    console.info('dump i18n :', `${data.length} records.`);
  }

  async dumpJob(handler) {

    debug('dump job');

    const { items } = await light.job.list(handler);
    const data = items.map(item => {
      this.deleteCommonField(item, { keepId: true });
      return item;
    });

    fs.writeFileSync(
      path.join(this.folder, `${CONST.SYSTEM_DB_JOB}.yml`),
      helper.yamlDumper(data)
    );
    debug(`Download job. A total of ${data.length} records.`);
    console.info('dump job :', `${data.length} records.`);
  }

  deleteCommonField(item, option = { keepId: false }) {

    // 如果是保持_id，将_id转成字符串
    if (option.keepId) {
      item._id = item._id.toString();
    } else {
      delete item._id;
    }
    delete item.valid;
    delete item.createAt;
    delete item.createBy;
    delete item.updateAt;
    delete item.updateBy;
    delete item.version;
  }

  deleteIfBlank(item, key) {
    if (!item[key]) {
      delete item[key];
    }
  }

  deleteIfZero(item, key) {
    if (item[key] === 0) {
      delete item[key];
    }
  }

  deleteIfEmpty(item, key) {
    const value = item[key];
    if (!value) {
      return delete item[key];
    }

    if (Array.isArray(value)) {
      if (value.length <= 0) {
        delete item[key];
      }
      return;
    }

    const keys = Object.keys(value);
    if (!keys || keys.length <= 0) {
      delete item[key];
    }
  }

}

if (!module.parent) {

  const Client  = require('../client');
  const system = process.argv.includes('--system');
  const scope = system
    ? ['struct', 'board', 'config', 'validator']
    : ['struct', 'board', 'config', 'validator', 'route', 'function', 'i18n', 'job'];

  Client.init('client')
    .then(async () => {
      const dumper = new Dump(scope, system);
      await dumper.start();
    })
    .catch(err => console.error(err))
    .finally(async () => {
      await Client.destory();
      process.exit();
    });
}

module.exports = Dump;
