/**
 * generate.js
 *
 * 生成项目
 */

'use strict';

const fs     = require('fs');
const assert = require('assert').strict;
const path   = require('path');
const debug  = require('debug')('light:bin');
const pkg    = require('../package.json');
const helper = require('../lib/helper');

class Generator {

  constructor(project, type = 'app') {
    this.project = project;
    this.type = type;
  }

  start() {

    assert(this.project, 'The project name cannot be empty');

    const folder = path.join(process.cwd(), this.project);
    debug('folder', folder);

    // 检查文件夹是否存在
    if (fs.existsSync(folder)) {

      // 检查文件夹是否为空
      const existPackage = fs.existsSync(path.join(folder, 'package.json'));
      const existConfig = fs.existsSync(path.join(folder, 'config.yml'));

      if (existPackage || existConfig) {
        throw new Error('The project folder is not empty');
      }

      debug('Use existing folder');
    } else {

      debug('The folder does not exist, create it!');

      // 没有就创建
      fs.mkdirSync(folder, { recursive: true });
    }

    if (this.type === 'app') {
      this.createPackage(folder);
      this.createConfig(folder);
      this.createViews(folder);
      this.createControllers(folder, 'account');
      this.createSetting(folder);
      this.createApp(folder);
    }

    if (this.type === 'service') {
      this.createPackage(folder);
      this.createConfig(folder);
      this.createControllers(folder, 'service');
      this.createApp(folder);
    }

  }

  createPackage(folder) {

    const file = path.join(__dirname, 'template', this.type, 'tmpl.package.json');
    const data = {
      name       : path.basename(folder),
      version    : '0.0.1',
      coreVersion: pkg.version
    };
    const content = helper.ejsParser(file, data);

    fs.writeFileSync(path.join(folder, 'package.json'), content, 'utf-8');

  }

  createConfig(folder) {

    const file = path.join(__dirname, 'template', this.type, 'tmpl.config.yml');
    const data = {
      name: path.basename(folder),
      lang: 'node'
    };
    const content = helper.ejsParser(file, data);

    fs.writeFileSync(path.join(folder, 'config.yml'), content, 'utf-8');

  }

  createViews(folder) {

    fs.mkdirSync(path.join(folder, 'views'), { recursive: true });

    const file = path.join(__dirname, 'template', this.type, 'tmpl.views.index.html');
    const data = {};
    const content = helper.ejsParser(file, data);

    fs.writeFileSync(path.join(folder, 'views', 'index.html'), content, 'utf-8');

  }

  createControllers(folder, clazz) {

    fs.mkdirSync(path.join(folder, 'controllers'), { recursive: true });

    const file = path.join(__dirname, 'template', this.type, `tmpl.controllers.${clazz}.js`);
    const data = {
      name: path.basename(folder)
    };
    const content = helper.ejsParser(file, data);

    fs.writeFileSync(path.join(folder, 'controllers', `${clazz}.js`), content, 'utf-8');

  }

  createSetting(folder) {

    fs.mkdirSync(path.join(folder, 'setting'), { recursive: true });

    const configuration = path.join(__dirname, 'template', this.type, 'tmpl.setting.configuration.yaml');
    const data = {
      stamp        : helper.randomGUID8(),
      hmackey      : `light.hmk.${this.appDomain}.${helper.randomGUID4()}`,
      cookieSecret : `light.cs.${this.appDomain}.${helper.randomGUID4()}`,
      sessionSecret: `light.ss.${this.appDomain}.${helper.randomGUID4()}`,
      version      : pkg.version
    };
    const configurationContent = helper.ejsParser(configuration, data);

    fs.writeFileSync(path.join(folder, 'setting', 'configuration.yml'), configurationContent, 'utf-8');

    const route = path.join(__dirname, 'template', this.type, 'tmpl.setting.route.yaml');
    const routeContent = helper.ejsParser(route, {});

    fs.writeFileSync(path.join(folder, 'setting', 'route.yml'), routeContent, 'utf-8');

  }

  createApp(folder) {

    const file = path.join(__dirname, 'template', this.type, 'tmpl.app.js');
    const content = helper.ejsParser(file, {});

    fs.writeFileSync(path.join(folder, 'app.js'), content, 'utf-8');

  }

}

if (!module.parent) {

  assert(process.argv.length > 2, 'Incorrect number of parameters');
  const name = process.argv.pop();
  const type = process.argv.includes('--service') ? 'service' : 'app';

  try {
    const dumper = new Generator(name, type);
    dumper.start();
  } catch (err) {
    console.error(err);
  }

  process.exit();
}

module.exports = Generator;
