/**
 * @file 上传代码
 * @module push
 * @author r2space@gmail.com
 * @version 1.0.0
 */

'use strict';

const path    = require('path');
const fs      = require('fs');
const ignore  = require('ignore');
const log     = require('../lib/log');
const Context = require('../lib/context');
const helper  = require('../lib/helper');
const light   = require('../lib/model/light');
const CONST   = require('../lib/constant');

class Push {

  async start(root = process.cwd()) {

    this.root = root;

    // 获取代码类型说明文件，包含非上传对象文件，二进制文件的定义
    const { lang = 'node', ignore = [], binary = [] } = this.getConfig();

    this.lang = lang;
    this.ignore = this.getIgnore(this.root, ignore);
    this.binary = this.getBinary(binary);

    const handler = new Context(CONST.DEFAULT_USER, process.env.APPNAME, CONST.SYSTEM_DB_PREFIX);

    try {

      const changed = await this.getChanged(handler);
      await this.removeCode(handler, changed.remove);
      await this.updateCode(handler, changed.update);
      await this.addCode(handler, changed.add);

      return changed;
    } catch (err) {
      console.error(err);
    }
  }

  /**
   * 比较数据库与本地文件的md5，找出有变更的文件
   * @param {*} handler
   */
  async getChanged(handler) {

    const changed = { add: [], update: [], remove: [] };
    const remote = await light.code.list(handler, { select: 'name, md5', condition: { lang: this.lang } });
    const local = helper.tree(this.root, this.ignore).map(item => item.file.replace(this.root, ''));

    // 遍历所有本地文件，对比md5
    local.forEach(name => {

      const code = remote.items.find(item => item.name === name);
      const md5 = helper.fileMd5(path.join(this.root, name));

      // 不存在则添加文件
      if (!code) {
        return changed.add.push({ file: name, md5 });
      }

      // md5不同则更新文件
      if (code.md5 !== md5) {
        return changed.update.push({ file: name, md5 });
      }
    });

    // 遍历所有数据库中内容，确定被删除的文件
    remote.items.forEach(({ name }) => {
      if (!local.includes(name)) {
        changed.remove.push({ file: name });
      }
    });

    return changed;
  }

  /**
   * 读取工程目录的 config.yml 文件设置内容
   * @param {*} root
   */
  getConfig() {

    let config = {};

    // 查看根目录
    if (fs.existsSync(path.join(this.root, 'config.yml'))) {
      config = helper.yamlLoader('config.yml', this.root);
    }

    // 尝试查看Java的资源目录
    if (fs.existsSync(path.join(this.root, 'src/main/resources', 'config.yml'))) {
      config = helper.yamlLoader('config.yml', path.join(this.root, 'src/main/resources'));
    }

    return config;
  }

  /**
   * 添加
   * @param {*} handler
   * @param {*} items
   */
  async addCode(handler, items) {

    for (const item of items) {
      const isBinary = this.binary.ignores(item.file.substr(1));
      const data = {
        name: item.file,
        type: isBinary ? 'binary' : 'code',
        app : handler.domain,
        md5 : item.md5,
        lang: this.lang
      };

      log.debug(`>> add ${item.file}`);
      console.info(`add ${data.type}: ${item.file}`);

      if (isBinary) {
        const [result] = await this.addFile(handler, path.join(this.root, item.file));
        data.source = result._id;
        await light.code.add(handler, { data });
      } else {
        data.source = fs.readFileSync(path.join(this.root, item.file), 'utf8');
        await light.code.add(handler, { data });
      }
    }

  }

  /**
   * 更新
   * @param {*} handler
   * @param {*} items
   */
  async updateCode(handler, items) {

    for (const item of items) {
      const isBinary = this.binary.ignores(item.file.substr(1));
      log.debug(`>> update ${item.file}`);
      console.info(`update ${isBinary ? 'binary' : 'code'}: ${item.file}`);

      if (isBinary) {
        const [result] = await this.addFile(handler, path.join(this.root, item.file));
        await light.code.update(handler, {
          data     : { source: result._id, md5: item.md5 },
          condition: { name: item.file, lang: this.lang }
        });
      } else {
        await light.code.update(handler, {
          condition: { name: item.file, lang: this.lang },
          data     : {
            source: fs.readFileSync(path.join(this.root, item.file), 'utf8'),
            md5   : item.md5
          }
        });
      }
    }
  }

  /**
   * 删除
   * @param {*} handler
   * @param {*} items
   */
  async removeCode(handler, items) {
    for (const item of items) {
      log.debug(`>> remove ${item.file}`);
      console.info(`remove       : ${item.file}`);
      await light.code.remove(handler, { condition: { name: item.file, lang: this.lang } });
    }
  }

  /**
   * 添加代码文件到gridfs
   * @param {*} handler
   * @param {*} name
   */
  async addFile(handler, name) {
    const data = {
      name: path.basename(name),
      type: 'application/octet-stream',
      path: name
    };

    return await light.file.add(handler.copy({ files: [data] }));
  }

  /**
   * 代码上传时，忽略的本地文件或文件夹
   * @param folder
   * @param exclude
   * @returns {Array.<string>}
   */
  getIgnore(folder, exclude) {

    const gitignore = path.join(folder, '.gitignore');

    if (fs.existsSync(gitignore)) {
      const content = fs.readFileSync(gitignore, 'utf8');
      return content.split('\n')
        .filter(item => item)
        .concat(exclude)
        .concat('.git');
    }

    return exclude.condat(['.git']);
  }

  /**
   * 代码上传时，作为二进制文件上传的文件
   * @param binary
   * @returns {Array.<string>}
   */
  getBinary(binary = []) {

    const image = ['*.png', '*.jpg', '*.jpeg', '*.ico', '*.gif'];
    const font = ['*.eot', '*.svg', '*.ttf', '*.woff', '*.woff2'];
    const office = ['*.xlsx', '*.docx', '*.pptx'];
    const pdf = ['*.pdf'];
    const pattern = binary.concat(pdf)
      .concat(image)
      .concat(font)
      .concat(office);

    return ignore().add(pattern);
  }

}


if (!module.parent) {
  const Client  = require('../client');
  Client.init('client')
    .then(async () => {
      const push = new Push();
      await push.start();
    })
    .catch(err => console.error(err))
    .finally(async () => {
      await Client.destory();
      process.exit();
    });
}
module.exports = Push;
