/**
 * context.js
 *
 *   - handler.params
 *   - handler.uid          用户ID
 *   - handler.user         登陆用户
 *   - handler.corp         账户识别号
 *   - handler.domain       应用识别号
 *   - handler.code         客户识别号
 *   - handler.db           应用数据库账户
 *   - handler.timezone     时区
 *   - handler.strict       是否是严格模式，默认为true
 *   - handler.req
 *   - handler.res
 *   - handler.ctx
 */

'use strict';

const _      = require('lodash');
const assert = require('assert').strict;
const debug  = require('debug')('light:context');
const CONST  = require('./constant');
const config = require('./configuration');

class Context {

  constructor(uid, domain = process.env.APPNAME, code = CONST.DEFAULT_TENANT) {

    debug('init, uid: %s, domain: %s, code: %s', uid, domain, code);

    this._uid = uid;
    this._domain = domain;
    this._code = code;
    this._attributes = {};
    this._transaction = {
      use    : false, // 是否启用了事物
      session: null   // 管理事物的session实例
    };
  }

  /**
   * 设置参数
   * @param {*} ctx
   */
  bind(ctx) {

    this.ctx = ctx;
    this.req = ctx.request;
    this.res = ctx.response;

    // 拷贝请求中的参数
    this._attributes = {
      ...ctx.request[CONST.SYMBOL.QUERY],
      ...ctx.request.body,
      ...ctx.request[CONST.SYMBOL.XML]
    };

    // 设置上传文件
    // 将文件列表转换成数组 { a: File, b: [File, File] } -> [File, File, File]
    if (ctx.request.files) {
      this._attributes.files = Object
        .values(ctx.request.files)
        .reduce((memo, item) => memo.concat(Array.isArray(item) ? item : [item]), []);
    }

    // 设置domain，user等信息
    this.bindAccount();
    this.bindTimezone();
    this.bindDomain();

    return this;
  }

  /**
   * context的domain,code,user,corp,db值是依赖session内容的
   * 当这些session值被更新是，成员变量的值也会一起更新
   * 如果只想更新session的内容，可以使用ctx.session来修改值
   *
   * 未指定参数时，返回session对象
   * @param {*} key
   * @param {*} value
   */
  session(key, value) {

    // 不指定参数时，返回session对象
    if (!key) {
      return this.ctx.session;
    }

    this.ctx.session[key] = value;
    switch (key) {
      case CONST.SESSION.DOMAIN:
        this._domain = value || process.env.APPNAME;
        break;
      case CONST.SESSION.CODE:
        this._code = value || CONST.DEFAULT_TENAN;
        break;
      case CONST.SESSION.USER:
        this._user = value;
        if (this._user) {
          this._uid =  this._user._id;
        }
        this.bindTimezone();
        break;
      case CONST.SESSION.CORP:
        assert(value, 'corp cannot be empty');
        this._corp = value;
        break;
      case CONST.SESSION.DB:
        assert(value, 'db cannot be empty');
        this._db = value;
        break;
    }
  }

  /**
   * 根据session信息，恢复context的domain,code,user,corp,db值
   * 如果jwt被设定了，那么重jwt数据里恢复
   */
  bindAccount() {

    debug('set account info');

    // 使用session数据
    if (this.ctx.session) {
      debug('use session');
      this._domain = this.ctx.session[CONST.SESSION.DOMAIN] || process.env.APPNAME;
      this._code = this.ctx.session[CONST.SESSION.CODE] || CONST.DEFAULT_TENANT;
      this._user = this.ctx.session[CONST.SESSION.USER];     // TODO: user里有 _id, timezone 信息必须
      this._corp = this.ctx.session[CONST.SESSION.CORP];     // TODO: 明确用途
      this._db = this.ctx.session[CONST.SESSION.DB];         // TODO: 明确用途
      if (this._user) {
        this._uid =  this._user._id;
      }
    }

    // 使用JWT数据
    const jwt = this.ctx[CONST.SYMBOL.JWT];
    if (jwt) {
      debug('use jwt');
      this._user = { _id: jwt[CONST.JWT.ID], timezone: jwt[CONST.JWT.TIMEZONE] };
      this._domain = jwt[CONST.JWT.DOMAIN] || process.env.APPNAME;
      this._code = jwt[CONST.JWT.CODE] || CONST.DEFAULT_TENANT;
      this._uid =  this._user._id;
    }
  }

  /**
   * 设置时区信息
   */
  bindTimezone() {

    debug('set timezone');

    // 设置用户ID
    if (this._user) {
      this._timezone = this._user.timezone;
      debug('try user timezone', this._timezone);
    }

    // 缺省的时区 - 先从用户信息取，没有设置取配置文件里的内容，否则默认UTC
    if (!this._timezone) {
      this._timezone = config.app && config.app.timezone
        ? config.app.timezone
        : 'UTC';
      debug('try config timezone', this._timezone);
    }
  }

  /**
   * 允许切换到特定的domain，或特定的code（如访问特定的API时固定切换到另外的domain时使用）
   */
  bindDomain() {
    const domain = this.ctx[CONST.SYMBOL.DOMAIN];
    debug('special domain', domain);
    if (domain) {
      debug('use special domain');
      this._domain = domain;
    }

    const code = this.ctx[CONST.SYMBOL.CODE];
    debug('special code', code);
    if (code) {
      debug('use special code');
      this._code = code;
    }
  }

  /**
   * 复制一个新的Handler，并将params替换为指定的值
   * @param {*} params
   */
  copy(params) {

    const handler = new Context(this.uid, this.domain, this.code);

    handler.ctx = this.ctx;
    handler.req = this.req;
    handler.res = this.res;
    handler.user = this.user;
    handler.db = this.db;
    handler.strict = this.strict;
    handler.api = this.api;
    handler.sysapi = this.sysapi;

    handler._corp = this.corp;
    handler._timezone = this.timezone;
    handler._transaction = this.transaction;

    if (params) {

      // 指定了参数，使用参数替换当前参数
      handler._attributes = _.cloneDeep(params);
    } else {

      // 变量 files 里可能会保存stream，buffer等信息，不做深拷贝
      handler._attributes = _.cloneDeep({ ...this._attributes, files: [] });
      handler._attributes.files = this._attributes.files;
    }

    return handler;
  }

  /**
   * 获取请求参数（合并了GET，POST方法的请求参数）
   */
  get params() {
    return this._attributes;
  }
  addParams(object = {}) {
    Object.assign(this._attributes, object);
  }
  removeParams(key) {
    delete this._attributes[key];
  }
  clearParams() {
    this._attributes = {};
  }

  /**
   * 获取当前用户的ID
   */
  get uid() {
    if (this._uid && this._uid.constructor.name === 'ObjectID') {
      return this._uid.toString();
    }
    return this._uid;
  }

  /**
   * 保持当前用户信息
   */
  get user() {
    return this._user;
  }
  set user(user) {
    this._user = user;
    if (user) {
      this._uid = user._id;
    }
  }

  /**
   * 保持当前租户信息
   * set方法在，需要对明确指定的code进行数据操作时使用。如超级管理员操作等
   */
  get code() {
    return this._code;
  }
  set code(code) {
    this._code = code;
  }

  /**
   * 保持当前domain信息
   * set方法在，需要对明确指定的domain进行数据操作时使用。如超级管理员操作等。
   */
  get domain() {
    return this._domain;
  }
  set domain(domain) {
    this._domain = domain;
  }

  /**
   * 保存数据库连接账户
   * 用于在App中切换不同的数据库
   */
  get db() {
    return this._db || {};
  }
  set db(db) {
    this._db = db;
  }

  /**
   * 平台提供了一个模式叫 strict，默认状态下 strict = true 为严格模式
   *
   * 非严格模式，即 strict = false 时
   * 1. 更新数据允许指定 valid createAt createBy updateAt updateBy，
   *    如果没有指定继续使用系统默认值
   * 2. select 项目受 board 定义的选择项目限制。
   *    如果指定了select，那么忽略board的限制，返回select指定的所有项目
   */
  get strict() {
    if (typeof this._strict === 'undefined') {
      return true;
    }
    return this._strict;
  }
  set strict(strict) {
    this._strict = strict;
  }

  /**
   * 获取时区
   */
  get timezone() {
    return this._timezone;
  }

  /**
   * 获取事物信息
   */
  get transaction() {
    return this._transaction;
  }

  /**
   * 获取公司信息
   */
  get corp() {
    return this._corp;
  }

}

module.exports = Context;
