/**
 * connect.js
 */

'use strict';

const { ReadPreference, MongoClient } = require('mongodb');
const CONST       = require('../../constant');
const { DBError } = require('../../error/db');
const pool        = require('./pool');
const debug       = require('debug')('light:db:mongo:connection');

class Connection {

  /**
   * 获取数据库连接
   *   使用完DB后，无需明确关闭数据库
   *   没有指定，取环境里的用户名密码
   * @param domain 数据库名
   * @param user 数据库用户
   * @param password 数据库密码
   */
  static connect(
    domain = CONST.SYSTEM_DB,
    user = process.env.LIGHTMONGO_USER,
    password = process.env.LIGHTMONGO_PASS
  ) {

    const host  = process.env.LIGHTMONGO_HOST || CONST.SYSTEM_DB_HOST;
    const port  = process.env.LIGHTMONGO_PORT || CONST.SYSTEM_DB_PORT;
    const hosts = process.env.LIGHTMONGO_HOSTS || `${host}:${port}`;

    return new Promise((resolve, reject) => {

      // 重用连接
      const cache = pool.get(domain);
      if (cache) {
        debug('use cached connection', domain);
        return resolve(cache);
      }

      debug('database hosts:', hosts);
      debug('database name:', domain);
      debug('database user:', user);

      if (!user || !password) {
        return reject(new DBError('The username or password cannot be empty'));
      }

      const option = {
        w                 : 1, // 确保写入的mongod实例数量
        readPreference    : ReadPreference.PRIMARY,
        replicaSet        : process.env.LIGHTMONGO_HOSTS ? CONST.SYSTEM_DB_REPLICA : undefined,
        poolSize          : 6, // AP中如果并行操作数据库，这个值设小会影响性能
        auth              : { user, password },
        authSource        : domain,
        authMechanism     : 'SCRAM-SHA-1',
        useNewUrlParser   : true,
        useUnifiedTopology: true
      };

      // 连接数据库
      MongoClient
        .connect(`mongodb://${hosts}`, option)
        .then(client => {

          const db = client.db(domain);
          db.on('close', () => debug('connection close.'));
          db.on('error', err => console.error(err));
          db.on('reconnect', () => {
            this.reconnect = (this.reconnect || 0) + 1;
            console.warn('connection reconnect.', this.reconnect);
          });

          pool.add(domain, client, db);
          resolve({ client, db });
        })
        .catch(err => reject(err));
    });

  }

}

module.exports = Connection;
