/**
 * controller.js
 */

'use strict';

const assert = require('assert').strict;
const debug    = require('debug')('light:db:mongo:controller');
const CONST    = require('../../constant');
const setting  = require('../../setting');
const Model    = require('./model');
const { UpdateError, RemoveError } = require('../../error/db');

class Controller {

  constructor(handler, table) {

    const account = handler.db || {};
    const params  = handler.params || {};
    const code    = handler.code || CONST.DEFAULT_TENANT;

    this.domain = handler.domain || process.env.APPNAME;

    debug('domain', this.domain);
    debug('account', account);
    debug('code', code);
    debug('uid', handler.uid);
    debug('strict', handler.strict);

    // Model实例
    this.model = new Model(this.domain, code, table, {
      user       : account.user,
      pass       : account.pass,
      tz         : handler.timezone,
      transaction: handler.transaction,
      struct     : handler.code === CONST.SYSTEM_DB_PREFIX ? setting.system.struct : setting.struct
    });

    // 参数
    this.uid = handler.uid;
    this.id = params.id;
    this.condition = params.condition || params.filter;
    this.data = params.data || params;
    this.skip = params.skip || 0;
    this.limit = params.limit;
    this.sort = params.sort || params.order;
    this.select = params.select || params.field;
    this.files = params.files || [];
    this.strict = handler.strict;
  }

  /**
   * 添加数据
   */
  add() {

    assert(this.data, 'data cannot be empty');

    const data = (Array.isArray(this.data) ? this.data : [this.data])
      .map(item => {

        assert(!item._id, '_id is not allowed');

        const now = new Date();

        // Strict mode, the use of system time
        if (this.strict) {
          return Object.assign(item, {
            valid   : CONST.VALID,
            createAt: now,
            createBy: this.uid || CONST.DEFAULT_USER,
            updateAt: now,
            updateBy: this.uid || CONST.DEFAULT_USER
          });
        }

        // In non-strict mode, the user specified time is attempted
        return Object.assign(item, {
          valid   : typeof item.valid === 'undefined' ? CONST.VALID : item.valid,
          createAt: item.createAt || now,
          createBy: item.createBy || this.uid || CONST.DEFAULT_USER,
          updateAt: item.updateAt || now,
          updateBy: item.updateBy || this.uid || CONST.DEFAULT_USER
        });
      });

    return this.model.add(Array.isArray(this.data) ? data : data[0]);
  }

  /**
   * 获取单个数据
   */
  get() {

    const condition = this.id || this.condition;

    debug('get');
    if (!this.id && Object.keys(this.condition || {}).length <= 0) {
      debug('condition is empty');
      return Promise.resolve({});
    }

    debug('condition', condition);
    return this.model.get(condition, this.select);
  }

  /**
   * 获取数据一览
   */
  list(count = true) {

    debug('list');
    debug('fetch count', count);
    debug('condition', this.condition);

    assert(typeof count === 'boolean', 'invalid count type');

    return new Promise((resolve, reject) => {
      Promise
        .all([
          count ? this.model.count(this.condition) : Promise.resolve(0),
          this.model.list(this.condition, this.skip, this.limit, this.sort, this.select)
        ])
        .then(result => {
          const [totalItems, items] = result;
          debug('count', totalItems);
          resolve({ totalItems, items });
        })
        .catch(err => reject(err));
    });
  }

  /**
   * 获取数据件数
   */
  count() {
    debug('count');
    return this.model.count(this.condition);
  }

  /**
   * 删除数据
   */
  remove() {

    debug('remove');
    if (this.strict) {

      // 获取去除 valid 的实际删除条件
      const pure = Object.keys(this.condition || {}).filter(item => item !== 'valid');

      // 判断是否指定了条件（防止误删除所有数据）
      if (!this.id && pure.length <= 0) {
        debug('condition is empty');
        return Promise.reject(new RemoveError('condition cannot be empty'));
      }
    }

    const condition = this.id || this.condition;
    debug('condition', condition);
    return this.model.remove(condition, { updateAt: new Date(), updateBy: this.uid });
  }

  /**
   * 更新数据
   */
  update() {

    debug('update');
    if (this.strict) {

      // 获取去除 valid 的实际删除条件
      const pure = Object.keys(this.condition || {}).filter(item => item !== 'valid');

      // 判断是否指定了条件（防止误删除所有数据）
      if (!this.id && pure.length <= 0) {
        debug('condition is empty');
        return Promise.reject(new UpdateError('condition cannot be empty'));
      }
    }

    const data = {};
    if (this.strict) {

      // Strict mode, the use of system time
      data.updateAt = new Date();
      data.updateBy = this.uid || CONST.DEFAULT_USER;
    } else {

      // In non-strict mode, the user specified time is attempted
      data.updateAt = this.data.updateAt || new Date();
      data.updateBy = this.data.updateBy || this.uid || CONST.DEFAULT_USER;
    }

    const condition = this.id || this.condition;
    debug('condition', condition);
    return this.model.update(condition, Object.assign(this.data, data));
  }

  /**
   * 用给定条件更新，没有指定的条件则插入数据
   * 插入时，返回插入的数据
   * 更新时，返回更新的数据
   */
  upsert() {

    debug('update');
    return new Promise((resolve, reject) => {

      const now = new Date()
        , data = {};

      if (this.strict) {
        data.updateAt = now;
        data.updateBy = this.uid || CONST.DEFAULT_USER;
      } else {
        data.updateAt = this.data.updateAt || now;
        data.updateBy = this.data.updateBy || this.uid || CONST.DEFAULT_USER;
      }

      this.model.count(this.id || this.condition)
        .then(count => {

          // 如果条件未找到数据，设置create信息
          if (count <= 0) {
            if (this.strict) {
              data.valid = CONST.VALID;
              data.createAt = now;
              data.createBy = this.uid || CONST.DEFAULT_USER;
            } else {
              data.valid = typeof this.data.valid === 'undefined' ? CONST.VALID : this.data.valid;
              data.createAt = this.data.createAt || now;
              data.createBy = this.data.createBy || this.uid || CONST.DEFAULT_USER;
            }
          }

          const condition = this.id || this.condition;
          debug('condition', condition);
          this.model.upsert(condition, Object.assign(this.data, data))
            .then(res => resolve(res))
            .catch(err => reject(err));
        })
        .catch(err => reject(err));
    });
  }

  /**
   * 聚合查询
   *
   * @param {*} pipeline
   */
  aggregate(pipeline) {
    debug('aggregate');
    assert(pipeline, 'pipeline cannot be empty');
    return this.model.aggregate(pipeline, {});
  }

  /**
   * 将文件写到GridFS里
   *
   * 1. 如果文件以Buffer形式传过来，则直接把Buffer写入到GridFS里
   * 2. 如果文件以Stream形式传过来，则直接把Stream写入到GridFS里
   * 3. 直接把文件写入GridFS中
   */
  async fileToGrid() {

    debug('fileToGrid');

    const result = { totalItems: 0, items: [] };
    for (const file of this.files) {

      assert(file.name, 'name cannot be empty');
      assert(file.type, 'type cannot be empty');

      if (file.buffer && Buffer.isBuffer(file.buffer)) {
        debug('is file buffer');
        result.items.push(await this.model.bufferToGrid(file));
        result.totalItems += 1;
        continue;
      }

      if (file.stream) {
        debug('is file stream');
        result.items.push(await this.model.streamToGrid(file));
        result.totalItems += 1;
        continue;
      }

      if (file.path) {
        debug('is file name');
        result.items.push(await this.model.fileToGrid(file));
        result.totalItems += 1;
        continue;
      }

      assert.fail('file param is incorrect');
    }

    return result;
  }

  /**
   * 从GridFS中读取到Stream
   * 如果start没有指定，则直接读取整个文件，否则只读取start到end的文件到stream中
   * @param {*} start
   * @param {*} end
   */
  gridToStream(start, end) {

    assert(this.id, 'file id cannot be empty');

    if (typeof start === 'undefined') {
      debug('gridToStream');
      return this.model.gridToStream(this.id);
    }

    debug('gridToStreamInRange');
    return this.model.gridToStreamInRange(this.id, start, end);
  }

  /**
   * 获取数据库进程信息
   */
  statsServer() {
    debug('statsServer');
    return this.model.statsServer();
  }

  /**
   * 获取数据库信息
   */
  statsDb() {
    debug('statsDb');
    return this.model.statsDb();
  }

  /**
   * 创建用户
   */
  addUser() {
    debug('addUser');
    assert(this.data.user, 'data.user cannot be empty');
    assert(this.data.pass, 'data.pass cannot be empty');
    return this.model.addUser(this.data.user, this.data.pass, this.data.domain);
  }

  /**
   * 删除用户
   */
  dropUser() {
    debug('dropUser');
    assert(this.condition.user, 'condition.user cannot be empty');
    return this.model.dropUser(this.condition.user, this.condition.domain);
  }

  /**
   * 删除数据库
   */
  dropDb() {
    debug('dropDb');
    return this.model.dropDb();
  }

  /**
   * 执行指定的数据库命令
   */
  runCommand() {
    debug('runCommand');
    assert(this.condition.command, 'condition.command cannot be empty');
    return this.model.command(this.condition.command);
  }

  /**
   * 取消事物
   */
  abort() {
    debug('abort');
    return this.model.abort();
  }

  /**
   * 提交事务
   */
  commit() {
    debug('commit');
    return this.model.commit();
  }

}

module.exports = Controller;
