/**
 * 数据类型转换
 */

'use strict';

const debug    = require('debug')('light:db:mongo:mapping');
const assert   = require('assert').strict;
const type     = require('./type');
const operator = require('./operator');

class Mapping {

  /**
   * 设置缺省值
   * @param {*} value
   * @param {*} defines
   * @param {*} addition
   */
  static defaults(value, defines, addition) {

    // 没有定义，原值返回
    if (!defines) {
      debug('define is empty');
      return value;
    }

    assert(type.isJson(value), 'invalid defaults value');
    assert(type.isJson(value), 'invalid defaults defines');

    Object.keys(defines).forEach(key => {

      const define = defines[key];

      // 该项没有值且定义有default时，为该项设置默认值
      if (typeof define.default !== 'undefined' && typeof value[key] === 'undefined') {

        // 数组和哈希时，将'undefined'和'null'字符串转换为undefined和null
        if (['array', 'object'].includes(define.type)) {
          if (define.default === 'undefined') {
            return value[key] = undefined;
          }
          if (define.default === 'null') {
            return value[key] = null;
          }
        }

        // 转换default为相应的类型之后付值
        value[key] = type.parse(define.default, define, addition);
        debug(`set default ${key} to ${value[key]}`);
        return;
      }

      // 处理子文档
      if (!type.isJson(define.contents)) {
        return;
      }

      // 数组只对已经有的元素进行设置（数组为空也不添加元素）
      if (define.type === 'array') {
        value[key] = type.parse(value[key], define, addition)
          .map(item => this.defaults(item, define.contents, addition));
      }

      if (define.type === 'object') {
        value[key] = this.defaults(type.parse(value[key], define), define.contents, addition);
      }

      debug(`set sub document ${key} to`, value[key]);

    });

    return value;
  }

  /**
   * 类型转换
   * @param {*} value
   * @param {*} defines
   * @param {*} addition
   */
  static parse(value, defines, addition) {

    // 没有定义，原值返回
    if (!defines) {
      debug('define is empty');
      return value;
    }

    assert(type.isJson(value), 'invalid parse value');
    assert(type.isJson(value), 'invalid parse defines');

    return Object.keys(value).reduce((memo, key) => {

      const item = value[key];

      // 利用操作符对应的方法转换，如果没有找到操作符的定义则不做转换
      if ((/^\$.*/i).test(key)) {
        debug('has operator', key);
        memo[key] = operator.parse(key, item, defines, addition);
        return memo;
      }

      // 删除没有定义的值
      const define = Mapping.find(key, defines);
      if (!define.type) {
        debug('delete element because definition was not found', key);
        return memo;
      }

      memo[key] = this.parseItem(item, define, addition);
      return memo;
    }, {});
  }

  /**
   * 转换单个项目类型
   * @param {*} value
   * @param {*} define
   * @param {*} addition
   */
  static parseItem(value, define, addition) {

    // 如果值中包含操作符，递归转换。查询语句中常见
    //  如 {a: {$elemMatch: {$gt: 1}}} - a的定义是array
    if (this.hasOperator(value)) {
      return this.parse(value, define, addition);
    }

    // 如果是数组则逐项转换，如果包含子文档则递归转换
    if (type.isJson(define.contents)) {

      // 数组值转换, 如果值不是数组先转数组
      //  如 [1, 2, 3] [{a: 1}, {a: 2}]
      if (define.type === 'array') {
        debug('parse array');
        return type.parse(value, define).map(item => this.parse(item, define.contents, addition));
      }

      if (define.type === 'object') {
        debug('parse object');
        return this.parse(type.parse(value, define), define.contents, addition);
      }
    }

    return type.parse(value, define, addition);
  }

  /**
   * 判断值中是否有操作符
   * @param {*} value
   */
  static hasOperator(value) {
    if (!type.isJson(value)) {
      return false;
    }
    const key = Object.keys(value).find(key => (/^\$.*/i).test(key));
    if (key) {
      debug('has operator', key);
    }
    return !!key;
  }

  /**
   * 根据key从数据定义中获取类型描述
   * @param key
   * @param defines
   * @returns {*}
   */
  static find(path, defines) {

    if (!path.includes('.')) {
      return defines[path] || {};
    }

    debug('has dot', path);

    // 过滤掉数组索引，如 xxx.1.xx 中的1
    const keys = String(path)
      .split('.')
      .filter(item => !item.match(/^\d$/g));

    // 如果是xxx.xx.x的形式，则取末端属性对应的类型
    let define = null;
    while (keys.length >= 0) {
      const key = keys.shift();

      // 到达key列表的最后了，立即返回结果
      define = (define || defines)[key];
      if (keys.length <= 0) {
        return define || {};
      }

      // 如果contents未定义，返回空
      if (!define.contents) {
        return {};
      }

      // 查找下一层级
      define = define.contents;
    }
  }

}

module.exports = Mapping;
