/**
 * model.js
 *
 * 注意
 *  删除，更新等操作如果没有指定条件，会更新所有数据
 *
 * TODO
 *  Session删除条件需要再确认
 *  测试0字节文件
 */

'use strict';

const fs                         = require('fs');
const os                         = require('os');
const assert                     = require('assert').strict;
const path                       = require('path');
const { ObjectID, GridFSBucket } = require('mongodb');
const { Readable }               = require('stream');
const CONST                      = require('../../constant');
const connection                 = require('./connection');
const util                       = require('./util');
const mapping                    = require('./mapping');
const debug                      = require('debug')('light:db:mongo:model');
const { AddError, FindError, UpdateError, RemoveError } = require('../../error/db');


class Model {

  /**
   * 构造函数
   * @param {*} domain 数据库名
   * @param {*} code 表前缀
   * @param {*} table 表名
   * @param {*} options 数据库账号，时区等
   */
  constructor(domain, code, table, options = {}) {

    assert(domain, 'domain cannot be empty');

    this._domain = domain;
    this._user = options.user;
    this._pass = options.pass;
    this._addition = { tz: options.tz };

    this._table = table;
    this._code = code;
    this._define = null;

    this._struct = options.struct;  // Define
    this._db = null;                // DB instance
    this._client = null;            // Server instance
    this._write = options.write;    // The write concern.
    this._journal = options.journal;// Specify a journal write concern.
    this._transaction = options.transaction || { use: false, session: null };
  }


  /**
   * 获取服务器连接实例
   */
  get client() {

    debug('get client', this._domain, this._user || '');
    return new Promise((resolve, reject) => {
      if (this._client) {
        return resolve(this._client);
      }
      connection
        .connect(this._domain, this._user, this._pass)
        .then(connection => resolve(connection.client))
        .catch(err => reject(err));
    });
  }


  /**
   * 获取数据库连接实例
   */
  get db() {

    debug('get db', this._domain, this._user || '');
    return new Promise((resolve, reject) => {
      if (this._db) {
        return resolve(this._db);
      }
      connection
        .connect(this._domain, this._user, this._pass)
        .then(connection => resolve(connection.db))
        .catch(err => reject(err));
    });
  }


  /**
   * 获取表实例
   */
  get collection() {

    assert(this._code, 'code cannot be empty');
    assert(this._table, 'table cannot be empty');

    return new Promise((resolve, reject) => {

      connection
        .connect(this._domain, this._user, this._pass)
        .then(({ client, db }) => {

          // 事物有效时开启session
          if (this._transaction.use) {

            debug('use transaction');
            if (!this._transaction.session) {
              debug('start new session');
              this._transaction.session = client.startSession();
            }

            // 判断当前session是否被包含在特定事物中
            if (!this._transaction.session.inTransaction()) {
              debug('start new session (not in transaction)');
              this._transaction.session.startTransaction();
            }
          }

          this._db = db;
          this._client = client;

          // 使用事物时，CRUD操作必须是 existing collections
          // 所以，这里通过 createCollection 获取 collection 的实例
          const option = { w: this._write, j: this._journal };
          if (this._transaction.use) {
            db.collection(this.code, { ...option, strict: true }, (err, res) => {
              if (res) {
                return resolve(res);
              }

              db.createCollection(this.code, option)
                .then(collection => resolve(collection))
                .catch(err => reject(err));
            });
            return;
          }

          resolve(db.collection(this.code, option));
        })
        .catch(err => reject(err));
    });
  }


  /**
   * 获取表名（带有前缀的全名）
   */
  get code() {

    let table = this._table;

    // 获取表定义
    if (this._struct) {

      this._define = this._struct.find(item => item.schema === table);

      // 数据库用户，索引，文件操作时，define为空
      if (this._define) {

        // 如果设有父表，那么使用父表
        table = this._define.parent || table;

        this._define = this._define.items;
      }
    }

    // 表名复数化
    const pluralize = util.collection(table);
    if (!pluralize) {
      return '';
    }

    return this._code ? `${this._code}.${pluralize}` : pluralize;
  }


  /**
   * 提交事物
   */
  commit() {

    debug('>>>>> commit transcation');

    return new Promise((resolve, reject) => {

      debug('sessionid', util.sessionToString(this._transaction));

      if (this._transaction.use && this._transaction.session.inTransaction()) {
        return this._transaction.session.commitTransaction()
          .then(() => {
            this._transaction.use = false;
            this._transaction.session.endSession();
            this._transaction.session = null;

            debug('<<<<< commit transcation');
            resolve();
          })
          .catch(err => reject(err));
      }
      resolve();
    });
  }


  /**
   * 取消事物
   */
  abort() {

    debug('>>>>> abort transcation');

    return new Promise((resolve, reject) => {

      debug('sessionid', util.sessionToString(this._transaction));

      if (this._transaction.use && this._transaction.session.inTransaction()) {
        return this._transaction.session.abortTransaction()
          .then(() => {
            this._transaction.use = false;
            this._transaction.session.endSession();
            this._transaction.session = null;

            debug('<<<<< abort transcation');
            resolve();
          })
          .catch(err => reject(err));
      }
      resolve();
    });
  }


  /**
   * 添加数据
   *  可以指定单个对象，也可以指定一个数组
   *  insert数据会根据define进行类型转换
   *
   *  返回值是插入到数据库中的对象
   *
   * @param {Object | Array} doc 数据
   */
  add(doc) {

    debug('>>>>> add');

    return new Promise((resolve, reject) => {

      const isMany = Array.isArray(doc);
      if ((isMany && doc.length <= 0) || (!isMany && !doc)) {
        return reject(new AddError('Document cannot be empty'));
      }

      this.collection.then(collection => {

        debug('insert to', this._domain, this.code);
        debug('original data %j', doc);
        const parsed = (isMany ? doc : [doc]).map(item => mapping.parse(
          mapping.defaults(item, this._define, this._addition),
          this._define,
          this._addition
        ));

        debug('data %j', parsed);
        collection.insertMany(parsed, { session: this._transaction.session })
          .then(result => {
            debug('insert count', result.insertedCount);
            debug('result %j', result.ops);
            debug('<<<<< add');
            resolve(isMany ? result.ops : result.ops[0]);
          })
          .catch(err => reject(err));

      }).catch(err => reject(err));
    });

  }


  /**
   * 获取单个数据
   *  filter
   *   可以是 字符串的_id值, ObjectID类型的值
   *   可以是 普通的mongodb条件
   *
   *  fields
   *   可以不指定，用于指定需要检索的字段
   *
   *  sort
   *   条件匹配多个时，可以排序后去第一个
   *
   *  doc
   *   如果不为空，会使用 findOneAndUpdate 进行检索并更新
   *   返回更新后的内容
   *
   * @param {String | ObjectID | Object} filter 或 条件
   * @param {String} fields 获取的项目
   * @param {String} sort 条件匹配多个时，可以排序后去第一个
   * @param {String} doc 设置值，则检索并更新
   */
  get(filter, fields = {}, sort, doc) {

    debug('>>>>> get');
    debug('find from', this._domain, this.code);
    debug('original filter', filter);

    return new Promise((resolve, reject) => {

      if (util.isEmpty(filter)) {
        return reject(new FindError('Filter is incorrect'));
      }

      this.collection.then(collection => {

        const parsed = util.filter(filter, this._define, this._addition);
        const options = {
          projection: util.projection(fields),
          sort      : util.sort(sort),
          session   : this._transaction.session
        };

        debug('filter', parsed);
        debug('fields %j', options.projection);
        debug('sort', options.sort);

        if (doc) {

          options.returnOriginal = false;
          const data = util.sets(doc);

          debug('data %j', data);
          return collection.findOneAndUpdate(parsed, data, options)
            .then(result => {
              debug('result', result.value);
              debug('<<<<< get');
              resolve(result.value || {});
            })
            .catch(err => reject(err));
        }

        collection.findOne(parsed, options)
          .then(result => {
            debug('find count', result ? 1 : 0);
            debug('<<<<< get');
            resolve(result || {});
          })
          .catch(err => reject(err));

      }).catch(err => reject(err));

    });
  }


  /**
   * 获取数据一览
   *
   * @param {Object} filter 查询条件
   * @param {Number} skip 跳过的文书数
   * @param {Number} limit 返回的文书的上限数目
   * @param {Object} sort 排序
   * @param {Object} fields 获取的项目
  */
  list(filter, skip = 0, limit = CONST.DB_DEFAULT_LIMIT, sort = {}, fields = {}) {

    debug('>>>>> list');

    return new Promise((resolve, reject) => {
      this.getCursor(filter, skip, limit, sort, fields)
        .then(cursor => {
          cursor.toArray()
            .then(result => {
              debug('list count', result.length);
              debug('<<<<< list');
              resolve(result);
            })
            .catch(err => reject(err));
        })
        .catch(err => reject(err));
    });

  }


  /**
   * 获取游标
   *
   * @param {Object} filter 查询条件
   * @param {Number} skip 跳过的文书数
   * @param {Number} limit 返回的文书的上限数目
   * @param {Object} sort 排序
   * @param {Object} fields 获取的项目
   */
  getCursor(filter, skip = 0, limit = CONST.DB_DEFAULT_LIMIT, sort = {}, fields = {}) {

    return new Promise((resolve, reject) => {
      this.collection.then(collection => {

        debug('find from', this._domain, this.code);
        debug('original filter', filter);
        debug('original fields %j', fields);
        debug('original sort', sort);

        const parsed = util.filter(filter, this._define, this._addition);
        const options = {
          skip      : Number(skip),
          limit     : Number(limit),
          projection: util.projection(fields),
          sort      : util.sort(sort),
          session   : this._transaction.session
        };

        if (Number.isNaN(options.skip)) {
          return reject(new FindError(`Incorrect skip value : ${skip}`));
        }

        if (Number.isNaN(options.limit)) {
          return reject(new FindError(`Incorrect limit value : ${limit}`));
        }

        // mongodb 3.2.7 版开始, limit值如果使用Number.MAX_VALUE
        // 会出 'limit value must be non-negative' 错误，这里改为使用固定的值
        if (options.limit === Number.MAX_VALUE) {
          options.limit = CONST.DB_MILLION_LIMIT;
        }

        debug('filter', parsed);
        debug('fields %j', options.projection);
        debug('sort', options.sort, 'skip', options.skip, 'limit', options.limit);

        resolve(collection.find(parsed, options));
      }).catch(reject);
    });

  }


  /**
   * 更新数据
   *
   * @param {String | ObjectID |  Object} filter
   * @param {Object} doc
   */
  update(filter, doc) {

    debug('>>>>> update');

    return new Promise((resolve, reject) => {

      if (!filter) {
        return reject(new UpdateError('Filter is incorrect'));
      }

      if (!doc || Object.keys(doc).length <= 0) {
        return reject(new UpdateError('Document cannot be empty'));
      }

      this.collection.then(collection => {

        debug('update', this._domain, this.code);
        debug('original data %j', doc);

        const parsed = util.filter(filter, this._define, this._addition);
        const data = util.sets(mapping.parse(doc, this._define, this._addition));
        const options = { session: this._transaction.session };

        debug('filter %j', parsed);
        debug('data %j', data);

        collection.updateMany(parsed, data, options)
          .then(result => {

            debug('matched count', result.matchedCount);
            debug('modified count', result.modifiedCount);

            // 用ID更新时，返回更新后的数据
            if (parsed._id instanceof ObjectID) {
              return collection.findOne(parsed, options)
                .then(result => {
                  debug('update one');
                  debug('<<<<< update');
                  resolve(result);
                })
                .catch(err => reject(err));
            }

            debug('<<<<< update');
            resolve(result.modifiedCount);
          })
          .catch(err => reject(err));

      }).catch(err => reject(err));
    });
  }


  /**
   * 删除数据
   *
   * @param {String | ObjectID |  Object} filter 条件
   * @param {Object} doc 数据
   */
  remove(filter, doc = {}) {

    debug('>>>>> remove');

    return new Promise((resolve, reject) => {

      if (util.isEmpty(filter)) {
        return reject(new RemoveError('Filter is incorrect'));
      }

      this.collection.then(collection => {

        debug('remove', this._domain, this.code);
        doc.valid = CONST.INVALID;

        const parsed = util.filter(filter, this._define, this._addition);
        const data = util.sets(mapping.parse(doc, this._define, this._addition));
        const options = { session: this._transaction.session };

        collection.updateMany(parsed, data, options)
          .then(result => {
            debug('matched count', result.matchedCount);
            debug('remove count', result.modifiedCount);
            debug('<<<<< remove');
            resolve(result.modifiedCount);
          })
          .catch(err => reject(err));
      }).catch(err => reject(err));
    });

  }


  /**
   * 获取数据件数
   * @param {Object} filter 条件
   */
  count(filter) {

    debug('>>>>> count');

    return new Promise((resolve, reject) => {
      this.collection.then(collection => {

        debug('count', this._domain, this.code);

        const parsed = util.filter(filter, this._define, this._addition);
        const options = { session: this._transaction.session };

        debug('filter', parsed);

        collection.countDocuments(parsed, options)
          .then(count => {
            debug('count number', count);
            debug('<<<<< count');
            resolve(count);
          })
          .catch(err => reject(err));
      }).catch(err => reject(err));
    });

  }


  /**
   * 用给定条件更新，没有指定的条件则插入数据
   *  插入时，返回插入的数据
   *  更新时，返回更新的数据
   *
   * 注意
   *  更新条件没有找到数据，则插入一条数据，并返插入的数据的_id
   *  doc不能为空，不能是数组，如果是更新，返回更新的数据件数
   *
   * @param {String} filter 条件
   * @param {Object} doc 数据
   */
  upsert(filter, doc) {

    debug('>>>>> upsert');

    return new Promise((resolve, reject) => {

      if (!filter) {
        return reject(new UpdateError('Filter is incorrect'));
      }

      if (!doc) {
        return reject(new UpdateError('Document cannot be empty'));
      }

      if (Array.isArray(doc)) {
        return reject(new UpdateError('Document cannot be array'));
      }

      this.collection.then(collection => {

        debug('upsert', this._domain, this.code);
        debug('original data %j', doc);

        const parsed = util.filter(filter, this._define, this._addition);
        const data = util.sets(mapping.parse(doc, this._define, this._addition));
        const options = { session: this._transaction.session, upsert: true };

        debug('filter %j', parsed);
        debug('data %j', data);

        collection.updateMany(parsed, data, options)
          .then(result => {

            debug('matched count', result.matchedCount);
            debug('modified count', result.modifiedCount);
            debug('upsert count', result.upsertedCount);

            if (result.upsertedCount === 1 || parsed._id instanceof ObjectID) {

              const condition = result.upsertedId ? { _id: result.upsertedId._id } : parsed;
              debug('fetch condition', condition);

              return collection.findOne(condition, options)
                .then(result => {
                  debug('<<<<< upsert');
                  resolve(result);
                })
                .catch(err => reject(err));
            }

            debug('<<<<< upsert');
            resolve(result.modifiedCount);
          })
          .catch(err => reject(err));
      }).catch(err => reject(err));
    });
  }


  /**
   * 检索，并去掉相同的数据
   * @param {Object} filter 条件
   * @param {String} key Field of the document to find distinct values for.
   */
  distinct(filter, key) {

    debug('>>>>> distinct');

    return new Promise((resolve, reject) => {

      if (!key || typeof key !== 'string') {
        return reject(new FindError('Distinct key is incorrect'));
      }

      debug('find from', this._domain, this.code);
      debug('original filter', filter);

      this.collection.then(collection => {

        const parsed = util.filter(filter, this._define, this._addition);
        const options = { session: this._transaction.session };

        debug('distinct key', key);
        debug('filter', parsed);

        collection.distinct(key, parsed, options)
          .then(result => {
            debug('<<<<< distinct');
            resolve(result);
          })
          .catch(err => reject(err));

      }).catch(err => reject(err));
    });
  }


  /**
   * 执行aggregate
   * @param {Object}} pipeline
   * @param {Object}} option
   */
  aggregate(pipeline, option) {

    debug('>>>>> aggregate');

    return new Promise((resolve, reject) => {

      if (!pipeline) {
        return reject(new Error('Document cannot be empty'));
      }

      this.collection.then(collection => {

        debug('find from', this._domain, this.code);
        debug('pipeline', pipeline);
        debug('option', option);

        const options = Object.assign({ session: this._transaction.session }, option);
        collection.aggregate(pipeline, options).toArray()
          .then(result => {
            debug('<<<<< aggregate');
            resolve(result);
          })
          .catch(err => reject(err));

      }).catch(err => reject(err));
    });
  }


  /**
   * 执行指定的数据库命令
   * @param {Object} command 数据库命令
   */
  command(command) {

    debug('>>>>> command');

    return new Promise((resolve, reject) => {
      this.db.then(db => {

        debug('command', command);

        db.admin().command(command)
          .then(result => {
            debug('success', result.ok);
            debug('<<<<< command');
            resolve(result);
          })
          .catch(err => reject(err));

      }).catch(err => reject(err));
    });
  }


  /**
   * 重命名表
   * @param {String} name 新名称
   */
  rename(name) {

    debug('>>>>> rename');

    return new Promise((resolve, reject) => {
      this.collection.then(collection => {

        debug('rename', this._domain, this.code);

        this._table = name;
        collection.rename(this.code)
          .then(() => {
            debug('new name', this.code);
            debug('<<<<< rename');
            resolve(true);
          })
          .catch(err => reject(err));

      }).catch(err => reject(err));
    });
  }


  /**
   * Buffer写入到GridFS里
   * @param {Object} doc
   *  doc.name          文件逻辑名称
   *  doc.type   文件类型信息
   *  doc.buffer        文件Buffer
   *  doc.metadata      文件扩展字段
   */
  bufferToGrid(doc) {

    debug('>>>>> bufferToGrid');

    return new Promise((resolve, reject) => {

      if (!doc.name) {
        return reject(new AddError('Name cannot be empty'));
      }

      if (!doc.buffer) {
        return reject(new AddError('Buffer cannot be empty'));
      }

      this.db.then(db => {
        const options = {
          contentType: doc.type || CONST.DB_GRIDFS_CTYPE,
          metadata   : doc.metadata || {}
        };
        const bucketName = this._code || CONST.SYSTEM_DB_PREFIX;
        const bucket = new GridFSBucket(db, { bucketName });
        const stream = bucket.openUploadStream(doc.name, options);
        const readable = new Readable();

        debug('gridfs', this._domain, bucketName);
        debug('gridfs content-type', options.contentType);
        debug('gridfs file name', doc.name);

        stream.once('finish', doc => {
          debug('length', doc.length);
          debug('md5', doc.md5);
          debug('<<<<< bufferToGrid');
          resolve(doc);
        });

        stream.once('error', error => reject(error));

        readable.pipe(stream);
        readable.push(doc.buffer);
        readable.push(null);

      }).catch(err => reject(err));
    });
  }


  /**
   * 将文件写到GridFS里
   * @param {Object} doc
   *  doc.name          文件逻辑名称
   *  doc.type          文件类型信息
   *  doc.path          物理文件名称
   *  doc.metadata      文件扩展字段
   */
  fileToGrid(doc) {

    debug('>>>>> fileToGrid');

    return new Promise((resolve, reject) => {

      if (!doc.path) {
        return reject(new AddError('File name cannot be empty'));
      }

      const fullPath = path.resolve(doc.path);

      debug('gridfs', this._domain);
      debug('gridfs file name', doc.name);
      debug('gridfs full path', fullPath);

      doc.stream = fs.createReadStream(fullPath);
      this.streamToGrid(doc)
        .then(res => {
          debug('<<<<< fileToGrid');
          resolve(res);
        })
        .catch(err => reject(err));
    });
  }


  /**
   * 流写到GridFS里
   *
   * @param {Object} doc
   *  doc.name          文件逻辑名称
   *  doc.type   文件类型信息
   *  doc.stream        文件流
   *  doc.metadata      文件扩展字段
   */
  streamToGrid(doc) {

    debug('>>>>> streamToGrid');

    return new Promise((resolve, reject) => {

      if (!doc.name) {
        return reject(new AddError('Name cannot be empty'));
      }

      if (!doc.stream) {
        return reject(new AddError('File stream cannot be empty'));
      }

      this.db.then(db => {
        const options = {
          contentType: doc.type || CONST.DB_GRIDFS_CTYPE,
          metadata   : doc.metadata || {}
        };
        const bucketName = this._code || CONST.SYSTEM_DB_PREFIX;
        const bucket = new GridFSBucket(db, { bucketName });
        const stream = bucket.openUploadStream(doc.name, options);

        debug('gridfs', this._domain, bucketName);
        debug('gridfs content-type', options.contentType);
        debug('gridfs file name', doc.name);

        stream.once('finish', doc => {
          debug('length', doc.length);
          debug('md5', doc.md5);
          debug('<<<<< streamToGrid');
          resolve(doc);
        });

        stream.once('error', err => reject(err));

        doc.stream.pipe(stream);
      }).catch(err => reject(err));
    });
  }


  /**
   * GridFS文件输出到物理文件
   * @param {String | ObjectID} id
   * @param {String} fullPath
   */
  gridToFile(id, fullPath) {

    debug('>>>>> gridToFile');

    return new Promise((resolve, reject) => {

      if (!id) {
        return reject(new FindError('File id cannot be empty'));
      }

      this.db.then(db => {
        const bucketName = this._code || CONST.SYSTEM_DB_PREFIX;
        const bucket = new GridFSBucket(db, { bucketName });
        const filter = util.filter(id)._id;

        assert(filter instanceof ObjectID, 'Invalid file id');
        debug('gridfs', this._domain, bucketName);
        debug('gridfs id', filter);

        // 检查文件是否存在
        bucket.find(filter).next()
          .then(result => {

            if (!result) {
              return reject(new FindError('File not found'));
            }

            const file = fullPath
              ? path.resolve(fullPath)
              : path.join(os.tmpdir(), filter.toHexString());

            debug('gridfs output', file);

            bucket.openDownloadStream(filter)
              .pipe(fs.createWriteStream(file))
              .once('finish', () => {
                debug('<<<<< gridToFile');
                result.path = file;
                resolve(result);
              })
              .once('error', err => reject(err));

          })
          .catch(err => reject(err));
      }).catch(err => reject(err));
    });
  }

  /**
   * 文件读到内存
   * @param {String | ObjectID} id
   */
  gridToBuffer(id) {

    debug('>>>>> gridToBuffer');

    return new Promise((resolve, reject) => {

      if (!id) {
        return reject(new FindError('File id cannot be empty'));
      }

      this.db.then(db => {
        const bucketName = this._code || CONST.SYSTEM_DB_PREFIX;
        const bucket = new GridFSBucket(db, { bucketName });
        const filter = util.filter(id)._id;

        assert(filter instanceof ObjectID, 'Invalid file id');
        debug('gridfs', this._domain, bucketName);
        debug('gridfs id', filter);

        // 检查文件是否存在
        bucket.find(filter).next()
          .then(result => {
            if (!result) {
              return reject(new FindError('File not found'));
            }

            const stream = bucket.openDownloadStream(filter);
            const buffer = [];

            stream.on('error', err => reject(err));
            stream.on('data', data => buffer.push(data));
            stream.on('end', () => {
              debug('<<<<< gridToBuffer');
              result.buffer = Buffer.concat(buffer),
              resolve(result);
            });
          })
          .catch(err => reject(err));
      }).catch(err => reject(err));
    });
  }

  /**
   * 文件读到Stream
   * @param {String | ObjectId} id 文件ID
   */
  gridToStream(id) {

    debug('>>>>> gridToStream');

    return new Promise((resolve, reject) => {
      if (!id) {
        return reject(new FindError('File id cannot be empty'));
      }

      this.db.then(db => {
        const bucketName = this._code || CONST.SYSTEM_DB_PREFIX;
        const bucket = new GridFSBucket(db, { bucketName });
        const filter = util.filter(id)._id;

        assert(filter instanceof ObjectID, 'Invalid file id');
        debug('gridfs', this._domain, bucketName);
        debug('gridfs id', filter);

        // 检查文件是否存在
        bucket.find(filter).next()
          .then(result => {
            if (!result) {
              return reject(new FindError('File not found'));
            }

            debug('<<<<< gridToStream');
            result.stream = bucket.openDownloadStream(filter);
            resolve(result);
          })
          .catch(err => reject(err));
      }).catch(err => reject(err));
    });
  }

  /**
   * 分段读取返回Stream
   * @param {String | ObjectId} id 文件ID
   * @param {number} start 开始位置
   * @param {number} end 结束位置
   */
  gridToStreamInRange(id, start, end) {

    debug('>>>>> gridToStreamInRange');

    return new Promise((resolve, reject) => {

      if (!id) {
        return reject(new FindError('File id cannot be empty'));
      }

      const condition = { start: Number(start), end: Number(end) };
      if (Number.isNaN(condition.start)) {
        return reject(new FindError(`Incorrect start value : ${start}`));
      }
      if (Number.isNaN(condition.end) || condition.end <= 0) {
        return reject(new FindError(`Incorrect end value : ${end}`));
      }

      this.db.then(db => {
        const bucketName = this._code || CONST.SYSTEM_DB_PREFIX;
        const bucket = new GridFSBucket(db, { bucketName });
        const filter = util.filter(id)._id;

        assert(filter instanceof ObjectID, 'Invalid file id');
        debug('gridfs', this._domain, bucketName);
        debug('gridfs id', filter, 'condition', condition);

        // 检查文件是否存在
        bucket.find(filter).next()
          .then(result => {
            if (!result) {
              return reject(new FindError('File not found'));
            }

            debug('<<<<< gridToStreamInRange');
            result.stream = bucket.openDownloadStream(filter, condition),
            result.streamRange = condition;
            resolve(result);
          })
          .catch(err => reject(err));
      }).catch(err => reject(err));
    });
  }

  /**
   * 获取数据库一览
   */
  listDb() {

    debug('>>>>> listDb');

    return new Promise((resolve, reject) => {
      this.db.then(db => {

        debug('list databases', this._domain);

        db.admin()
          .listDatabases({ session: this._transaction.session })
          .then(result => {
            debug('database count', result.databases.length);
            debug('<<<<< listDb');
            resolve(result.databases);
          })
          .catch(err => reject(err));
      }).catch(err => reject(err));
    });
  }


  /**
   * 获取表一览
   */
  listCollection() {

    debug('>>>>> listCollection');

    return new Promise((resolve, reject) => {
      this.db.then(db => {

        debug('list collections', this._domain);

        db.listCollections({}, { session: this._transaction.session }).toArray()
          .then(result => {
            debug('collection count', result.length);
            debug('<<<<< listCollection');
            resolve(result);
          })
          .catch(err => reject(err));
      }).catch(err => reject(err));
    });
  }


  /**
   * 获取索引一览
   */
  listIndex() {

    debug('>>>>> listIndex');

    return new Promise((resolve, reject) => {
      this.collection.then(collection => {

        debug('indexes', this._domain, this.code);

        collection.indexes({ session: this._transaction.session })
          .then(result => {
            debug('index count', result.length);
            debug('<<<<< listIndex');
            resolve(result);
          })
          .catch(err => reject(err));
      }).catch(err => reject(err));
    });
  }


  /**
   * 创建索引
   * @param {Object} index 索引
   * @param {Object} option 索引选项
   */
  addIndex(index, option) {

    debug('>>>>> addIndex');
    assert(index, 'index cannot be empty');

    return new Promise((resolve, reject) => {
      this.db.then(db => {

        debug('createIndex', this._domain, this.code);
        debug('index', index);
        debug('option', option);

        const options = Object.assign({ session: this._transaction.session }, option);
        db.createIndex(this.code, index, options)
          .then(result => {
            debug('<<<<< addIndex');
            resolve(result);
          })
          .catch(err => reject(err));
      }).catch(err => reject(err));
    });
  }

  /**
   * 删除索引
   * @param {String} index
   */
  dropIndex(index) {

    debug('>>>>> dropIndex');
    assert(index, 'index cannot be empty');

    return new Promise((resolve, reject) => {
      this.collection.then(collection => {

        const options = { session: this._transaction.session };
        collection.indexes(options)
          .then(indexes => {

            debug('dropIndex', this._domain, this.code);
            debug('index', index);

            const exist = indexes.find(item => item.name === index);
            if (!exist) {
              debug('index not found');
              debug('<<<<< dropIndex');
              return resolve({ ok: 1 });
            }

            collection.dropIndex(index, options)
              .then(result => {
                debug('<<<<< dropIndex');
                resolve(result);
              })
              .catch(err => reject(err));
          })
          .catch(err => reject(err));
      }).catch(err => reject(err));
    });
  }

  /**
   * 创建用户
   * @param {String} user
   * @param {String} pass
   * @param {String} domain
   * @param {Object} option
   */
  addUser(user, pass, domain, option = { roles: ['dbOwner'] }) {

    debug('>>>>> addUser');
    assert(user, 'user cannot be empty');
    assert(pass, 'pass cannot be empty');

    return new Promise((resolve, reject) => {
      this.client.then(client => {

        const dbName = domain || this._domain;

        debug('addUser', dbName);
        debug('user', user);
        debug('option', option);

        const options = Object.assign({ session: this._transaction.session }, option);
        client.db(dbName).addUser(user, pass, options)
          .then(result => {
            debug('<<<<< addUser');
            resolve(result);
          })
          .catch(err => reject(err));
      }).catch(err => reject(err));
    });
  }


  /**
   * 删除用户
   * @param {String} user
   */
  dropUser(user, domain) {

    debug('>>>>> dropUser');
    assert(user, 'user cannot be empty');

    return new Promise((resolve, reject) => {
      this.client.then(client => {

        const dbName = domain || this._domain;

        debug('removeUser', dbName);
        debug('user', user);

        client.db(dbName).removeUser(user, { session: this._transaction.session })
          .then(result => {
            debug('<<<<< dropUser');
            resolve(result);
          })
          .catch(err => reject(err));
      }).catch(err => reject(err));
    });
  }


  /**
   * 删除表
   */
  dropCollection() {

    debug('>>>>> dropCollection');

    return new Promise((resolve, reject) => {
      this.db.then(db => {

        debug('dropCollection', this._domain, this.code);

        db.listCollections({ name: this.code }, { session: this._transaction.session })
          .next()
          .then(exists => {
            if (!exists) {
              debug('collection not found');
              debug('<<<<< dropCollection');
              return resolve(true);
            }

            db.dropCollection(this.code, { session: this._transaction.session })
              .then(result => {
                debug('<<<<< dropCollection');
                resolve(result);
              })
              .catch(err => reject(err));
          })
          .catch(err => reject(err));
      }).catch(err => reject(err));
    });
  }

  /**
   * 获取表信息
   */
  statsCollection() {

    debug('>>>>> statsCollection');

    return new Promise((resolve, reject) => {
      this.collection.then(collection => {

        debug('stats', this._domain, this.code);

        collection.stats()
          .then(result => {
            debug('<<<<< statsCollection');
            resolve(result);
          })
          .catch(err => reject(err));
      }).catch(err => reject(err));
    });
  }

  /**
   * 获取数据库信息
   */
  statsDb() {

    debug('>>>>> statsDb');

    return new Promise((resolve, reject) => {
      this.db.then(db => {

        debug('stats', this._domain);

        db.stats()
          .then(result => {
            debug('<<<<< statsDb');
            resolve(result);
          })
          .catch(err => reject(err));
      }).catch(err => reject(err));
    });
  }

  /**
   * 获取数据库进程信息
   */
  statsServer() {

    debug('>>>>> statsServer');

    return new Promise((resolve, reject) => {
      this.db.then(db => {

        debug('statsServer', this._domain);

        db.admin().serverStatus()
          .then(result => {
            debug('<<<<< statsServer');
            resolve(result);
          })
          .catch(err => reject(err));
      }).catch(err => reject(err));
    });
  }

  /**
   * 删除库
   */
  dropDb() {

    debug('>>>>> dropDb');

    return new Promise((resolve, reject) => {
      this.db.then(db => {

        debug('dropDb', this._domain);

        db.dropDatabase()
          .then(result => {
            debug('<<<<< dropDb');
            resolve(result);
          })
          .catch(err => reject(err));
      }).catch(err => reject(err));
    });
  }

}

module.exports = Model;
