/**
 * pool.js
 *
 * 由于同时需要连接多个DB，通常的连接池保持机制不适合平台使用。
 * 这里，使用一个逻辑连接池来保持与连接，这些连接会放到内存中。
 * 当APP关闭时，才会释放这些连接。
 *
 * 即，每个AP服务器都会保持多个数据库连接。
 *
 * 由于第一次访问时，会有多个请求同时创建同名的数据库连接，这样无法只保存一份在逻辑池中。
 * 我们用temporary方式，先保存这些，等过一段时间再明确关闭并释放。
 *
 * 可以通过mongodb的mongostat命令，关注数据库的连接及其他状态。
 */

const debug = require('debug')('light:db:mongo:pool');

class Pool {

  /**
   * 获取连接
   * @param key
   * @param keepTime 临时数据库连接的保持最长时间 20秒
   * @returns {*}
   */
  static get(key, keepTime = 2 * 1000) {

    const current = new Date();

    this.temporary = (this.temporary || []).filter(item => {
      if (item.key === key && current - item.time >= keepTime) {
        debug('clear temporary connection', key);
        item.client.close();
        return false;
      }
      return true;
    });

    return (this.connections || {})[key];
  }

  /**
   * 删除
   * @param key
   */
  static remove(key) {
    delete this.connections[key];
  }

  /**
   * 向连接池添加一个连接
   * @param key
   * @param client
   * @param db
   */
  static add(key, client, db) {

    // Connection map
    this.connections = this.connections || {};

    // 临时数据库连接
    this.temporary = this.temporary || [];

    // 如果已经有连接，则存在临时池中
    if (this.connections[key]) {
      debug('add connection to temporary', key);
      this.temporary.push({
        key,
        db,
        client,
        time: new Date()
      });
    } else {
      debug('add connection to pool', key);
      this.connections[key] = { db, client };
    }
  }

  /**
   * 断开所有链接
   * 单元测试时使用
   *
   * @param {Promise} force
   */
  static async destroy(force = false) {

    if (this.connections) {
      debug('destroy all connection', Object.keys(this.connections).length);
      for (const item of Object.values(this.connections)) {
        await item.client.close(force);
      }
      this.connections = {};
    }

    if (this.temporary) {
      debug('destroy all connection', Object.keys(this.temporary).length);
      for (const item of Object.values(this.temporary)) {
        await item.client.close(force);
      }
      this.temporary = [];
    }

  }

}

module.exports = Pool;
