/**
 * @file mongo工具类
 * @author r2space@gmail.com
 * @module light.core.mongo.helper
 * @version 1.0.0
 */

'use strict';

const pluralize    = require('pluralize');
const { ObjectID } = require('mongodb');
const { DBError }  = require('../../error/db');
const type         = require('./type');
const mapping      = require('./mapping');

/**
 * collection名统一用小写
 * 与之前使用的mongoose保持兼容，collection名后面加s
 * @param code
 * @returns {string}
 */
exports.collection = function (code) {
  if (code) {
    return pluralize(code.toLowerCase());
  }

  return code;
};

/**
 * 转换检索条件
 */
exports.filter = function (filter, define, addition) {
  if (filter instanceof ObjectID) {
    return { _id: filter };
  }

  if (typeof filter === 'string' || filter instanceof String) {
    return { _id: new ObjectID(filter) };
  }

  if (type.isJson(filter)) {
    return mapping.parse(filter, define, addition);
  }

  if (!filter) {
    return;
  }

  throw new DBError('Filter type is incorrect');
};

/**
 * 确认条件是否为空，valid:1 不算条件
 */
exports.isEmpty = function (filter) {

  if (!filter) {
    return true;
  }

  if (filter instanceof ObjectID || typeof filter === 'string') {
    return false;
  }

  if (!type.isJson(filter)) {
    return true;
  }

  // 除去valid条件
  return Object.keys(filter).filter(item => item !== 'valid').length <= 0;
};

/**
 * 将字段字符串，转换成mongodb的选择字段格式
 * `col1 col2, col3` => {col1: 1, col2: 1, col3: 1}
 * @param {Object} select 空格或逗号分隔的字符串或select对象
 * @return {Object} mongodb格式
 */
exports.projection = function (select) {
  if (!select) {
    return {};
  }

  if (typeof select === 'string') {
    return select.trim().split(/[ ]+|[ ]*,[ ]*/)
      .reduce((data, item) => {
        data[item] = 1;
        return data;
      }, {});
  }

  return select;
};

/**
 * 格式化排序项目
 * @param param
 * @returns {*}
 */
exports.sort = function (param) {

  if (!param) {
    return {};
  }

  // 数组形式的排序 sort: [['a': 1], ['b': -1]]
  if (Array.isArray(param)) {
    return param;
  }

  // `a b` => [['a': 1], ['b': 1]]
  if (typeof param === 'string') {
    return param.trim().split(/[ ]+|[ ]*,[ ]*/)
      .reduce((data, key) => {
        data.push([key, 1]);
        return data;
      }, []);
  }

  // 对象形式的排序 sort: {valid:1, createAt:1}}。当排序值是文字asc，desc时，转换为数字
  return Object.keys(param).reduce((data, key) => {
    if (typeof param[key] === 'string') {
      data.push([key, param[key].toLowerCase() === 'asc' ? 1 : -1]);
    } else {
      data.push([key, param[key]]);
    }
    return data;
  }, []);
};

/**
 * 为数据添加$set操作符
 * @param {Object}} doc
 */
exports.sets = function (doc) {

  const result = { $set: doc.$set || {} }; // 如果doc里带有$set操作符，直接拷贝

  Object.keys(doc)
    .filter(key => key !== '$set')
    .forEach(key => {
      const [operator] = key;
      if (operator === '$') {
        result[key] = doc[key];
      } else {
        result.$set[key] = doc[key];
      }
    });

  if (Object.keys(result.$set).length <= 0) {
    delete result.$set;
  }

  return result;
};

/**
 * 获取session字符串
 */
exports.sessionToString = function (transaction) {
  return transaction.session.id.id.value(true).toString('base64');
};
