/**
 * MS SQLServer 操作类
 *
 * @file lib/db/sqlserver/controller.js
 */

'use strict';

const _    = require('lodash')
  , fs     = require('fs')
  , stream = require('stream')
  , path   = require('path')
  // , async  = require('async')
  // , TYPES  = require('tedious').TYPES
  , CONST  = require('../../constant')
  , errors = require('../../error')
  , config = require('../../configuration')
  , helper = require('../../helper')
  , Model  = require('./model')
;

class Controller {

  constructor(handler, table) {

    this.handler = handler;
    this.table = table;
    this.uid = handler.uid;
    this.strict = handler.strict;
    this.files = handler.params.files;
    this.parameter = null;  // 用于传递tedious专用的SQL参数，文件操作时使用

    // controller和model会对params进行修改，所以在这儿clone一份
    this.params = _.cloneDeep(handler.params);
    this.params.data = _.cloneDeep(this.params.data || {});
    this.params.condition = this.params.condition || {};

    this.model = new Model(handler.domain, handler.code, table);
  }

  list(callback) {

    this._setValid();

    // list方法内部对condition做了修改，为了不影响下面的count方法，在这儿clone一份condition
    this.model.list(this.params.script, _.cloneDeep(this.params.condition), (err, result) => {
      if (err) {
        return callback(err);
      }

      if (this.params.limit > 0) {
        return this._countList(_.cloneDeep(this.params.condition), (err, count) => {
          if (err) {
            return callback(err);
          }

          callback(err, {totalItems: count, items: result});
        });
      }

      callback(err, {totalItems: result.length, items: result});
    });
  }

  add(callback) {

    const defaults = (data) => {

      // Strict mode, the use of system time
      if (this.strict) {
        return {
          createAt: new Date(),
          updateAt: new Date(),
          valid   : CONST.VALID,
          createBy: this.uid || CONST.DEFAULT_USER,
          updateBy: this.uid || CONST.DEFAULT_USER
        };
      }

      // In non-strict mode, the user specified time is attempted
      return {
        createAt: data.createAt || new Date(),
        updateAt: data.updateAt || new Date(),
        valid   : typeof data.valid === 'undefined' ? CONST.VALID : data.valid,
        createBy: data.createBy || this.uid || CONST.DEFAULT_USER,
        updateBy: data.updateBy || this.uid || CONST.DEFAULT_USER
      };
    };

    if (Array.isArray(this.params.data)) {
      async.map(this.params.data, (item, next) => {
        Object.assign(item, defaults(item));
        this.model.add(this.params.script, item, this.parameter, next);
      }, callback);
    }

    Object.assign(this.params.data, defaults(this.params.data));
    this.model.add(this.params.script, this.params.data, this.parameter, callback);
  }

  remove(callback) {

    this.params.data.updateAt = new Date();
    this.params.data.updateBy = this.uid || CONST.DEFAULT_USER;
    this.params.data.valid = CONST.INVALID;
    this.params.condition.valid = CONST.VALID;

    if (this.params.id) {
      this.params.condition._id = this.params.id;
    }

    this.model.remove(this.params.script, this.params.data, this.params.condition, this.parameter, callback);
  }

  update(callback) {

    if (this.strict) {

      // Strict mode, the use of system time
      this.params.data.updateAt = new Date();
      if (this.uid) {
        this.params.data.updateBy = this.uid;
      }
    } else {

      // In non-strict mode, the user specified time is attempted
      this.params.data.updateAt = this.params.data.updateAt || new Date();
      if (this.uid) {
        this.params.data.updateBy = this.params.data.updateBy || this.uid;
      }
    }

    if (this.params.id) {
      this.params.condition._id = this.params.id;
    }

    this._setValid();
    this.model.update(
      this.params.script, this.params.data, _.cloneDeep(this.params.condition), this.parameter, (err, result) => {
        if (err) {
          return callback(err);
        }

        // 更新对象是一条时，返回更新后的数据
        if (result === 1) {
          const select = this.params.script.replace(/^UPDATE/, 'SELECT * FROM').replace(/SET .* WHERE/, 'WHERE');
          return this.model.get(select, this.params.condition, callback);
        }

        return callback(err, result);
      }
    );
  }

  get(callback) {

    if (this.params.id) {
      this.params.condition._id = this.params.id;
    }

    this._setValid();
    this.model.get(this.params.script, this.params.condition, callback);
  }

  count(callback) {
    this._setValid();
    this.model.count(this.params.script, this.params.condition, callback);
  }

  increase() {
    //TODO:
  }

  // 文件保存到数据库
  fileToGrid(callback) {

    function toBuffer(file, done) {

      if (Buffer.isBuffer(file)) {
        return done(file);
      }

      if (file.fileStream) {
        const buffers = [];
        file.fileStream.on('data', data => buffers.push(data));
        file.fileStream.on('end', () => done(Buffer.concat(buffers)));
        return;
      }

      const base   = path.resolve(process.cwd(), config.app.tmp)
        , fullname = fs.realpathSync(path.join(base, path.basename(file.path)));

      done(fs.readFileSync(fullname));
    }

    async.map(this.files, (file, next) => {

      toBuffer(file, buffer => {

        this.parameter = [
          {name: 'data', type: TYPES.VarBinary, value: buffer},
          {name: 'length', type: TYPES.Decimal, value: buffer.length}
        ];

        if (this.params.id) {
          this.params.script = '' +
            'UPDATE [file] ' +
            'SET [data] = @data, [length] = @length, ' +
            '  [updateAt] =  <%- data.updateAt %>, [updateBy] = <%- data.updateBy %> ' +
            `WHERE [_id] = '${this.params.id}' AND [valid] = 1`;

          return this.update((err, result) => {
            next(err, {
              fileId     : result._id,
              name       : file.originalFilename,
              contentType: file.headers['content-type'],
              length     : buffer.length
            });
          });
        }

        this.params.script = '' +
          'INSERT INTO [file] ( ' +
          '  [data], [length], [_id], [createAt], [createBy], [updateAt], [updateBy], [valid] ' +
          ') VALUES ( ' +
          '  @data, @length, <%- data._id %>, <%- data.createAt %>, <%- data.createBy %>, ' +
          '  <%- data.updateAt %>, <%- data.updateBy %>, <%- data.valid %> ' +
          ')';

        // 多个文件的时候，需要清除_id的值
        delete this.params.data._id;

        this.add((err, result) => {
          next(err, {
            fileId     : result._id,
            name       : file.originalFilename,
            contentType: file.headers['content-type'],
            length     : buffer.length
          });
        });

      });

    }, (err, result) => {
      callback(err, {totalItems: result.length, items: result});
    });
  }

  // 读取数据库文件保存为物理文件
  gridToFile(callback) {

    this._fetchFile((err, result) => {
      if (err) {
        return callback(err);
      }

      const folder = this.params.data.folder || path.resolve(process.cwd(), config.app.tmp)
        , base     = path.join(folder, this.params.data.name || helper.randomGUID8());

      fs.writeFile(base, result.data, 'binary', callback);
    });
  }

  // 读取数据库文件为stream
  readStreamFromGrid(callback) {

    this._fetchFile((err, result) => {
      if (err) {
        return callback(err);
      }

      // buffer转成stream
      const bufferStream = new stream.PassThrough();
      bufferStream.end(result.data);

      // 字段与 mongodb 统和
      callback(null, {
        filename   : result.name,
        contentType: result.contentType,
        length     : result.length,
        uploadDate : result.updateAt,
        fileStream : bufferStream
      });
    });
  }

  _fetchFile(callback) {

    this.params.script = '' +
      'SELECT [name], [contentType], [length], [updateAt], [data] ' +
      'FROM [file] ' +
      'WHERE [_id] = <%- condition._id %> AND [valid] = 1';

    this.get((err, result) => {
      if (err) {
        return callback(err);
      }

      if (!result) {
        return callback(new errors.db.Find('file not found.'));
      }

      callback(null, result);
    });
  }

  readStreamFromGridInRange(start, end, callback) {
    // TODO:
  }

  _countList(condition, callback) {

    let sql = this.params.script;

    // 使用List脚本的条件，获取数据的件数（在原有script基础上删除OFFSET和FETCH NEXT语句，然后检索件数）
    // sql = sql.replace(/[ ]+FETCH NEXT[ ]+\d+[ ]+ROWS ONLY$/, '');
    // sql = sql.replace(/[ ]+OFFSET[ ]+\d+[ ]+ROW$/, '');
    // sql = sql.replace(/[ ]+ORDER BY.*(ASC|DESC)$/, '');
    // sql = `SELECT COUNT(1) AS COUNT FROM (${sql}) AS T`;

    // 在原有script基础上删除ROW_NUMBER语句，和件数限制
    sql = sql.replace(/ROW_NUMBER.*\[_row_\],/, '');
    sql = sql.replace(/WHERE \[T\][^\)]*/, '');
    sql = `SELECT COUNT(1) AS COUNT FROM (${sql}) C`;

    this.model.count(sql, this.params.condition, callback);
  }

  _setValid() {
    if (typeof this.params.condition['valid'] === 'undefined') {
      this.params.condition['valid'] = CONST.VALID;
    }
  }

}

module.exports = Controller;