/**
 * developer.js
 */

const assert   = require('assert').strict;
const debug    = require('debug')('light:developer');
const path     = require('path');
const fs       = require('fs');
const chokidar = require('chokidar');
const Dump     = require('../bin/dump');
const helper   = require('./helper');
const log      = require('./log');
const coverage = require('./helper/coverage');

class Developer {

  /**
   * 预先加载Controller, 校验代码是否正确
   * @param {*} ignore
   */
  static preLoadController(ignore = []) {

    if (process.env.APPDEV !== 'true') {
      return;
    }

    debug('pre load controller');

    if (!fs.existsSync('controllers')) {
      debug('controllers folder does not exist');
      return;
    }

    const pattern = ['*', '!*.js'].concat(ignore);
    helper
      .tree('controllers', pattern)
      .forEach(item => require(item.file));
  }

  /**
   * 收集代码覆盖率
   */
  static collectCoverage() {

    if (process.env.APPDEV !== 'true') {
      return;
    }

    debug('collect coverage');
    coverage.start();
  }

  /**
   * 仅开发时使用！
   */
  static loadServerConfig() {

    // 已经设置了环境变量，就不在配置文件里加载配置项
    if (process.env.APPNAME) {
      return;
    }

    debug('start the service locally!');
    debug('load config.yml');

    const config = helper.yamlLoader('/config.yml');
    if (config === null) {
      log.error('No configuration file found :', path.join(process.cwd(), '/config.yml'));
      return process.exit(-1);
    }

    // 会在启动目录（process.cwd()）加载配置文件
    const { server, mq } = config;

    assert(server, '[server] cannot be empty');
    assert(mq, '[mq] cannot be empty');

    if (server) {
      assert(server.domain, '[server.domain] cannot be empty');
      process.env.APPNAME = server.domain;

      if (server.queue) {
        process.env.SERVER_QUEUE = server.queue;
      }
      if (server.enterpoint) {
        process.env.SERVER_ENTERPOINT = server.enterpoint;
      }
      debug('domain', process.env.APPNAME);
    }

    if (mq) {
      assert(mq.host, '[mongodb.user] cannot be empty');
      assert(mq.port, '[mongodb.pass] cannot be empty');

      process.env.LIGHTMQ_HOST = mq.host;
      process.env.LIGHTMQ_PORT = mq.port;
      if (mq.user) {
        process.env.LIGHTMQ_USER = mq.user;
      }
      if (mq.pass) {
        process.env.LIGHTMQ_PASS = mq.pass;
      }

      debug('host', process.env.LIGHTMQ_HOST);
      debug('port', process.env.LIGHTMQ_PORT);
    }

  }

  /**
   * 仅开发时使用！
   */
  static loadConfig() {

    // 已经设置了环境变量，就不在配置文件里加载配置项
    if (process.env.APPNAME) {
      return;
    }

    debug('start the service locally!');
    debug('load config.yml');

    const config = helper.yamlLoader('/config.yml');
    if (config === null) {
      log.error('No configuration file found :', path.join(process.cwd(), '/config.yml'));
      return process.exit(-1);
    }

    // 会在启动目录（process.cwd()）加载配置文件
    const { app, mongodb, mysql, sqlserver } = config;

    assert(app, '[app] cannot be empty');
    assert(mongodb || mysql || sqlserver, '[mongodb | mysql | sqlserver] cannot be empty');

    if (app) {
      assert(app.domain, '[app.domain] cannot be empty');

      process.env.APPNAME = app.domain;
      process.env.APPPORT = app.port;
      process.env.APPCRON = app.cron;       // 标识是否开启定时器
      process.env.APPDEV = app.dev;         // 开发模式 (Option)
      process.env.APPSETTING = app.setting; // 允许从本地读取配置（不想同步数据库中的配置时可以使用）

      debug('domain', process.env.APPNAME);
      debug('port', process.env.APPPORT);
      debug('cron', process.env.APPCRON);
      debug('dev', process.env.APPDEV);
      debug('setting', process.env.APPSETTING);
    }

    if (mongodb) {
      debug('is mongodb');
      assert(mongodb.user, '[mongodb.user] cannot be empty');
      assert(mongodb.pass, '[mongodb.pass] cannot be empty');

      // hosts优先，通过hosts可以指定多端读取
      if (mongodb.hosts) {
        process.env.LIGHTMONGO_HOSTS = mongodb.hosts;
        debug('hosts', process.env.LIGHTMONGO_HOSTS);
      } else {
        process.env.LIGHTMONGO_HOST = mongodb.host;
        process.env.LIGHTMONGO_PORT = mongodb.port;
        debug('host', process.env.LIGHTMONGO_HOST);
        debug('port', process.env.LIGHTMONGO_PORT);
      }

      process.env.LIGHTMONGO_USER = mongodb.user;
      process.env.LIGHTMONGO_PASS = mongodb.pass;
      process.env.LIGHTMONGO_AUTH = mongodb.auth;
    }

    if (mysql) {
      debug('is mysql');
      assert(mysql.user, '[mysql.user] cannot be empty');
      assert(mysql.pass, '[mysql.pass] cannot be empty');

      process.env.LIGHTMYSQL_HOST = mysql.host;
      process.env.LIGHTMYSQL_PORT = mysql.port;
      process.env.LIGHTMYSQL_USER = mysql.user;
      process.env.LIGHTMYSQL_PASS = mysql.pass;

      debug('host', process.env.LIGHTMYSQL_HOST);
      debug('port', process.env.LIGHTMYSQL_PORT);
    }

    if (sqlserver) {
      debug('is sqlserver');
      assert(sqlserver.user, '[sqlserver.user] cannot be empty');
      assert(sqlserver.pass, '[sqlserver.pass] cannot be empty');

      process.env.LIGHTSQLSERVER_HOST = sqlserver.host;
      process.env.LIGHTSQLSERVER_PORT = sqlserver.port;
      process.env.LIGHTSQLSERVER_USER = sqlserver.user;
      process.env.LIGHTSQLSERVER_PASS = sqlserver.pass;

      debug('host', process.env.LIGHTSQLSERVER_HOST);
      debug('port', process.env.LIGHTSQLSERVER_PORT);
    }

  }

  static asyncSetting() {

    if (process.env.APPDEV !== 'true') {
      return;
    }

    if (!fs.existsSync('setting')) {
      debug('settings folder does not exist');
      return;
    }

    debug('async settings');
    new Dump(['struct', 'board', 'config', 'validator', 'route', 'function', 'i18n', 'job']).start();
  }

  static watch() {

    if (process.env.APPDEV !== 'true') {
      return;
    }

    const target = ['./lib', './bin', './controllers'];
    const ignored = '*.md';
    debug('start watching', target);
    debug('ignored', ignored);

    chokidar.watch(target, { ignored }).on('all', (event, at) => {
      if (event === 'add') {
        debug('watching for', at);
      }

      if (event === 'change') {
        debug('change at', at);

        // 删除js代码缓存
        const source = path.join(process.cwd(), at);
        delete require.cache[source];

        // 重新加载js代码异常时，打印日志（如语法错误等）
        try {
          require(source);
        } catch (err) {
          console.error(err);
        }
      }
    });
  }

}

module.exports = Developer;
