/**
 * aksk.js
 */

const crypto = require('crypto');
const assert = require('assert').strict;

class AkSk {

  /**
   * ak/sk编码请求参数，请求参数必须包含 accesskey, secretkey
   * @param {*} params
   *  accesskey : app.domain
   *  secretkey : app.secret
   *  httpmethod: GET | POST
   */
  static encode(params) {

    const secret = params.secretkey;

    assert(params.accesskey, 'Access key cannot be empty');
    assert(params.secretkey, 'Secret key cannot be empty');
    assert(params.address, 'Address cannot be empty');

    const query = {
      ...params,
      method   : 'HMAC-SHA1',
      nonce    : params.nonce || this.nonce(4),
      timestamp: params.timestamp || this.timestamp(),
      accesskey: params.accesskey
    };

    delete query.secretkey;
    delete query.address;

    const querystring = this.querystring(query, 'encode');
    const signature = this.sign(secret, params.httpmethod || 'POST', query);

    return `${params.address}?${querystring}&signature=${signature}`;
  }

  /**
   * 校验参数签名是否匹配
   * @param {*} secretkey
   * @param {*} data 接受的参数
   * @param {*} method 请求方法
   */
  static verify(secretkey, data, httpmethod = 'POST') {

    assert(secretkey, 'Secret key cannot be empty');

    if (!['GET', 'POST', 'PUT', 'DELETE'].includes(httpmethod)) {
      return false;
    }
    if (!data.method || !data.nonce || !data.timestamp || !data.accesskey) {
      return false;
    }

    const input = encodeURIComponent(data.signature);
    delete data.signature;

    const calculate = this.sign(secretkey, httpmethod, data);
    return input === calculate;
  }

  /**
   * 生成随机数（时间戳+指定位数的随机数）
   * @param {*} count
   */
  static nonce(count = 0) {

    assert(count >= 0, 'Count cannot be less than 0');

    return String(new Date().getTime())
      + Array(count)
        .fill('')
        .map(() => Math.floor(Math.random() * 10))
        .join('');
  }

  /**
   * 签名
   * @param {*} ak - access key
   * @param {*} method - request method (POST | PUT | GET | DELETE)
   * @param {*} query - 签名对象
   */
  static sign(ak, method, query) {

    assert(ak, 'Access key cannot be empty');
    assert(['GET', 'POST', 'PUT', 'DELETE'].includes(method), 'The request method is incorrect');
    assert(query, 'Query key cannot be empty');

    const string = `${method}&%2F&${this.querystring(query)}`;
    const result = crypto.createHmac('sha1', `${ak}&`)
      .update(Buffer.from(string))
      .digest('base64');

    return encodeURIComponent(result);
  }

  /**
   * 生成计算签名用的querystring，只支持单层结构的对象
   * @param {*} query
   * @param {*} type escape | encode
   */
  static querystring(query, type = 'escape') {

    if (type === 'escape') {
      const string = Object.keys(query)
        .map(key => ({ name: key, value: query[key] }))
        .sort((val1, val2) => (val1.name < val2.name ? -1 : 1))
        .map(item => `${this.escape(item.name)}=${this.escape(item.value)}`)
        .join('&');

      return encodeURIComponent(string);
    }

    const querystring = Object.keys(query)
      .map(key => `${key}=${encodeURIComponent(query[key])}`)
      .join('&');

    return querystring;
  }

  /**
   * 转码
   * @param {*} clearString
   */
  static escape(clearString) {
    return encodeURIComponent(clearString)
      .replace(/!/gi, '%21')
      .replace(/'/gi, '%27')
      .replace(/\(/gi, '%28')
      .replace(/\)/gi, '%29')
      .replace(/\*/gi, '%2A');
  }

  /**
   * 生成时间字符串，不带秒
   */
  static timestamp() {
    return new Date()
      .toISOString()
      .replace(/\.\d{3}/, '');
  }

}

module.exports = AkSk;
