/**
 * @file 文件工具类
 */

const debug      = require('debug')('light:helper:file');
const parseRange = require('range-parser');
const response   = require('../response');
const { NotFound, InternalServer } = require('../error').http;

class File {

  /**
   * 下载文件
   *  当 headers 里包含 range 时，执行分段下载
   * @param handler
   * @returns {*}
   */
  static async download(handler) {

    try {

      debug('find file');
      const file = await this.controller(handler).get();
      if (!file._id) {
        return response.sendError(handler, new NotFound('File not found'));
      }

      handler.params.id = file._id;

      // 支持大文件分段下载
      const { headers } = handler.req;
      if (headers.range) {
        debug('range', headers.range, 'file length', file.length);

        // range值不正确
        const range = parseRange(file.length, headers.range);
        debug('parsed range', range);

        if (range === -1 || range === -2 || range.length > 1) {
          handler.ctx.set('Accept-Ranges', 'bytes');
          handler.ctx.set('Content-Range', `bytes */${file.length}`);
          return response.sendFile(handler, 406);
        }

        const [{ start, end }] = range;
        debug('start', start, 'end', end);

        // 返回区域数据
        const data = await this.controller(handler).gridToStream(start, end);
        handler.ctx.set('Accept-Ranges', 'bytes');
        handler.ctx.set('Content-Range', `bytes ${start}-${end}/${data.length}`);
        handler.ctx.set('Content-Length', end - start + 1);

        debug('send range file');
        return response.sendFile(handler, 206, data.stream);
      }

      // 单次完整下载
      const data = await this.controller(handler).gridToStream();
      debug('send file');
      response.sendFile(handler, 200, data);
    } catch (err) {
      debug('download error', err);
      response.sendError(handler, new InternalServer());
    }
  }


  /**
   * 获取文件Stream, 在controllers里调用
   * TODO: light的file模块已经不开放了，所以这个方法需要通过工具类开放出去
   * @param handler
   * @param callback
   */
  static async stream(handler, callback) {
    const _callback = this.defaultCallback(callback);

    try {
      const file = await this.controller(handler).gridToStream();

      return _callback(null, file.stream);
    } catch (err) {
      debug('stream error', err);
      return _callback(err);
    }
  }

  /**
   * 获取默认callback
   * @param {*} callback
   */
  static defaultCallback(callback) {
    return callback || ((err, res) => {
      if (err) throw err;
      return res;
    });
  }

  /**
   * 支持多种数据库
   * @param handler
   */
  static controller(handler, table = 'file') {

    if (process.env.LIGHTMONGO_HOST || process.env.LIGHTMONGO_HOSTS) {
      const Controller = require('../db/mongo/controller');
      return new Controller(handler, table);
    }

    if (process.env.LIGHTSQLSERVER_HOST) {
      const Controller = require('../db/sqlserver/controller');
      return new Controller(handler, table);
    }
    if (process.env.LIGHTMYSQL_HOST) {
      const Controller = require('../db/mysql/controller');
      return new Controller(handler, table);
    }

  }

}

module.exports = File;
