/**
 * @file 通用工具类
 * @module helper
 * @author r2space@gmail.com
 * @version 1.0.0
 */

'use strict';

const assert       = require('assert').strict;
const fs           = require('fs');
const path         = require('path');
const os           = require('os');
const util         = require('util');
const crypto       = require('crypto');
const ejs          = require('ejs');
const xml          = require('xml2js');
const _            = require('lodash');
const Packer       = require('zip-stream');
const QRCode       = require('qrcode');
const numeral      = require('numeral');
const dayjs        = require('dayjs');
const timezone     = require('dayjs/plugin/timezone');
const yaml         = require('js-yaml');
const ignore       = require('ignore');
const pkg          = require('../../package.json');
const Captcha      = require('./captchapng');
const mpath        = require('./mpath');
const file         = require('./file');
const aksk         = require('./aksk');
const debug        = require('debug')('light:helper');
const debugresolve = require('debug')('light:helper:resolve');

dayjs.extend(timezone);

class Helper {

  /**
   * @desc 在对象里取值，或给对象赋值
   * - set 方法使用 lodash 的set方法
   * - get 方法使用 mpath 的get方法
   *
   * mpath的set方法的缺点 - 内嵌文档空时，不能自动创建子文档 set({}, 'a.b.c', 10) 无法生成 {a: {b: {c: 10}}}
   * lodash的get方法的缺点 - 不能获取数组的值 _.get([{a: {b:1}}, {a: {b:2}}], 'a.b') 无法获取 [1, 2]
   */
  static get(object, pathName) {
    return mpath.get(pathName, object);
  }

  static set(...params) {
    return _.set(...params);
  }

  /**
   * @desc 简单生成随机4位字符串
   * @returns {String} 随机4位字符串
   */
  static randomGUID4() {
    return (((1 + Math.random()) * 0x10000) | 0).toString(16).substring(1);
  }

  static randomGUID8() {
    return this.randomGUID4() + this.randomGUID4();
  }

  static randomGUID12() {
    return this.randomGUID4() + this.randomGUID4() + this.randomGUID4();
  }

  /**
   * @desc 生成唯一识别号<br>
   *   - GUID : Universally Unique Identifier
   * @returns {String} uuid
   */
  static uuid() {

    // eslint-disable-next-line id-length
    return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, c => {
      // eslint-disable-next-line one-var,id-length,no-mixed-operators
      const r = Math.random() * 16 | 0, v = c === 'x' ? r : (r & 0x3 | 0x8);
      return v.toString(16);
    });
  }

  /**
   * @desc 读取模板文件，替换参数，生成结果文件，如果没有指定结果文件，则返回解析后的字符串
   * @param {String} templateFile ejs模板文件
   * @param {Object} parameters 模板文件参数对象
   * @param {String} resultFile 结果文件，如果没有指定则以字符串的形式返回解析的内容
   * @returns {String} 解析内容
   */
  static ejsParser(templateFile, parameters, resultFile) {

    debug('template file', templateFile);

    // 读取模板文件
    const template = fs.readFileSync(templateFile, 'utf8');

    // 转换模板文件
    ejs.open = undefined;
    ejs.close = undefined;
    const result = ejs.render(template, parameters);

    // 没有指定输出文件，则返回字符串
    if (!resultFile) {
      return result;
    }

    // 输出文件
    debug('result file', resultFile);
    fs.writeFileSync(resultFile, result);
  }

  /**
   * @deprecated
   */
  static ejsFormat(...param) {
    return ejs.render(...param);
  }

  /**
   * @desc 判断客户端是否是mozilla浏览器
   * @param {Object} req 请求对象
   * @returns {Boolean} 返回是否
   */
  static isBrowser(req) {

    // 可以通过自定义Header的方式，来明确请求端的设备类型
    const clientType = req.headers['x-client-type'];

    if (clientType) {
      debug('x-client-type', clientType);
      if (['wechat', 'iphone', 'android', 'electron'].includes(clientType)) {
        return false;
      }
    }

    const ua = req.headers['user-agent'] || '';

    debug('user-agent', ua);
    return !!ua.toLowerCase().match(/mozilla.*/i);
  }

  /**
   * @desc 判断客户端是否是移动设备
   * @param req
   * @returns {Array|{index: number, input: string}}
   */
  static isMobile(req) {

    const clientType = req.headers['x-client-type'];
    if (clientType) {
      debug('x-client-type', clientType);
      if (['rn', 'wechat'].includes(clientType)) {
        return true;
      }
    }

    if (this.isReactNative(req)) {
      debug('is react native');
      return true;
    }

    const ua = req.headers['user-agent'] || '';
    if (!ua) {
      debug('user-agent is empty');
      return false;
    }

    if (this.isElectron(req)) {
      debug('is electron');
      return true;
    }

    return this.isApple(req) || this.isAndroid(req);
  }

  static isReactNative(req) {
    const rn = req.headers['light-react-native'];
    debug('light-react-native', rn);
    return rn || false;
  }

  static isApple(req) {
    const ua = req.headers['user-agent'] || '';

    debug('user-agent', ua);
    return !!(ua.match(/iPhone/i) || ua.match(/iPod/i) || ua.match(/iPad/i));
  }

  static isAndroid(req) {
    const ua = req.headers['user-agent'] || '';

    debug('user-agent', ua);
    return !!(ua.match(/(?=.*\bAndroid\b)(?=.*\bMobile\b)/i) || ua.match(/Android/i));
  }

  static isElectron(req) {
    const ua = req.headers['user-agent'] || '';

    debug('user-agent', ua);
    return !!(ua.toLowerCase().match(/electron/i));
  }

  /**
   * @desc 获取AP服务器IP地址的数组，获取的IP地址放到global对象中缓存
   * @returns 返回IP地址
   */
  static ip() {

    if (this.addresses) {
      return this.addresses;
    }

    const interfaces = os.networkInterfaces();
    const addresses = [];

    Object.values(interfaces).forEach(item => {
      item.forEach(address => {
        if (address.family === 'IPv4' && !address.internal) {
          addresses.push(address.address);
        }
      });
    });

    this.addresses = addresses;
    return this.addresses;
  }

  /**
   * @desc 获取应用程序情报
   * @returns {Object} 应用程序版本信息等
   */
  static applicationInfo() {
    return {
      version    : pkg.version,
      host       : os.hostname(),
      application: pkg.name,
      time       : new Date()
    };
  }

  /**
   * @desc 判断模块是否可以加载
   *  TODO: cwd方法的目录依赖，会因为启动方式，启动目录不同而不准确
   *  TODO: 是否用代码文件存在来判断更加合理？而不是用异常捕获
   * @ignore
   * @param {String} module 模块名称
   * @param {String} pathName 相对路径
   * @returns {String} 路径
   */
  static resolve(module, pathName) {
    try {
      debug(module, pathName);
      return require(path.join(pathName || process.cwd(), module));
    } catch (err) {
      debugresolve(err.message);
      return undefined;
    }
  }

  /**
   * @desc XML解析器
   * @param {String} name 文件名
   * @param {Function} callback 回调函数，返回解析后结果
   */
  static xmlParser(fileName, pathName) {

    const file = path.join(pathName || process.cwd(), fileName);

    debug('xml file', file);
    return new Promise((resolve, reject) => {
      if (!fs.existsSync(file)) {
        debug('xml file not found');
        return resolve({});
      }

      const data = fs.readFileSync(file);
      new xml.Parser().parseString(data, (err, result) => {
        if (err) {
          return reject(err);
        }
        resolve(result);
      });
    });
  }

  /**
   * 读取yaml文件
   * @param {*} fileName
   * @param {*} pathName
   */
  static yamlLoader(fileName, pathName) {

    const file = path.join(pathName || process.cwd(), fileName);
    debug('load yaml', file);
    if (!fs.existsSync(file)) {
      debug('yaml not exists');
      return [];
    }

    return yaml.safeLoad(fs.readFileSync(file, 'utf8'));
  }

  /**
   * 读取yaml字符串
   */
  static yamlParser(content) {
    return yaml.safeLoad(content);
  }

  /**
   * 写出yaml字符串
   */
  static yamlDumper(json, option = {}) {
    option.flowLevel = 6;
    return yaml.safeDump(json, option);
  }

  /**
   * @desc 压缩指定的文件列表
   * @param {Array} files 文件数组
   * @param {String|Object} out 输出文件名或者输出流
   * @param {Function} callback 回调函数，返回解析后结果
   */
  static async zipFiles(files, out) {

    assert(out, 'output cannot be empty');
    debug('output to', typeof out === 'string' ? out : 'stream');

    function append(archive, data, option) {
      return new Promise((resolve, reject) => {
        archive.entry(data, option, err => {
          if (err) {
            return reject(err);
          }
          resolve();
        });
      });
    }

    const archive = new Packer();
    const output = typeof out === 'string' ? fs.createWriteStream(out) : out;

    if (files && files.length > 0) {

      // 压缩文件列表
      for (const file of files) {
        debug('entry', file);
        await append(archive, fs.createReadStream(file), { name: path.basename(file) });
      }
    } else {

      // 文件列表空时，生成空文件
      debug('file list is empty');
      await append(archive, 'No file.', { name: 'NoFile' });
    }

    archive.pipe(output);
    archive.finish();

    // 输出文件关闭时，返回
    return new Promise(resolve => output.on('close', () => resolve()));
  }

  /**
   * @desc 生成QRcode
   * @param {String} text 信息
   * @param {String|Object} out 输出文件名或者输出流
   * @param {Function} callback 回调函数，返回解析后结果
   */
  static async qrcode(option, out) {

    assert(option.text, 'text cannot be empty');

    // 返回base64
    if (!out) {
      debug('to base64');
      return await QRCode.toDataURL(option.text, option);
    }

    // 写到文件
    if (typeof out === 'string') {
      debug('to file', out);
      return await QRCode.toFile(out, option.text, option);
    }

    // 写到流
    debug('to stream');
    await QRCode.toFileStream(out, option.text, option);
    return new Promise(resolve => out.on('close', () => resolve()));
  }

  /**
   * @desc 删除文件夹
   * @param {String} folder 文件夹
   */
  static rmdir(folder) {
    if (fs.existsSync(folder)) {
      fs.readdirSync(folder).forEach(file => {
        const current = path.join(folder, file);
        if (fs.lstatSync(current).isDirectory()) {
          this.rmdir(current);
        } else {
          debug('remove file', folder);
          fs.unlinkSync(current);
        }
      });
      debug('remove folder', folder);
      fs.rmdirSync(folder);
    }
  }

  /**
   * @desc 创建文件夹, 递归创建
   * @param folder
   * @param includeFile 给定的folder是否包含文件名
   * @deprecated
   */
  static mkdirp(folder, includeFile = true) {
    if (includeFile) {

      // 去除最后的文件名
      return fs.mkdirSync(folder.replace(/\/[^/]*$/, ''), { recursive: true });
    }

    fs.mkdirSync(folder, { recursive: true });
  }

  /**
   * @desc 字符串格式化
   * @param {Object} val 数据
   * @param {String} format 格式
   * @param {String} tz 时区
   * @returns {String} 格式化后结果
   */
  static format(val, format, tz) {
    if (typeof val === 'undefined') {
      debug('format value is undefined');
      return val;
    }

    if (val === null) {
      debug('format value is null');
      return val;
    }

    if (!format) {
      debug('format pattern is empty');
      return val;
    }

    // Format number
    if (typeof val === 'number') {
      debug('format number', val, format);
      return numeral(val).format(format);
    }

    // Format string
    if (typeof val === 'string') {
      debug('format string', val, format);
      return util.format(format, val);
    }

    // Format date
    if (val instanceof Date || dayjs.isDayjs(val) || val.constructor.name === 'Moment') {
      debug('format date', format, tz);
      return dayjs(val).tz(tz || 'UTC')
        .format(format);
    }

    debug('not formated');
    return val;
  }

  /**
   * 获取文件一览，包括子文件夹
   * 允许指定 ignore，ignore遵守gitignore格式
   * 会忽略 SymbolicLink
   * @param folder
   * @param pattern Array 指定对象外文件或文件夹
   * @param root    递归参数 原始的路径
   * @param result  递归参数 结果路径列表
   * @returns {*|Array}
   */
  static tree(folder, pattern = [], root = `${path.resolve(folder)}${path.sep}`, result = []) {

    const ig = Array.isArray(pattern) ? ignore().add(pattern) : pattern;

    fs.readdirSync(folder).forEach(file => {

      const full = path.resolve(folder, file);
      if (ig.ignores(full.replace(root, ''))) {
        return;
      }

      const stats = fs.lstatSync(full);

      // 忽略文件链接
      if (stats.isSymbolicLink()) {
        return;
      }

      // 递归文件夹
      if (stats.isDirectory()) {
        return this.tree(full, ig, root, result);
      }

      return result.push({ file: full, size: stats.size });
    });

    return result;
  }

  /**
   * 获取文件的MD5
   * @param filename
   * @returns {*}
   */
  static fileMd5(filename) {

    const BUFFER_SIZE = 8192;

    const fd = fs.openSync(filename, 'r');
    const hash = crypto.createHash('md5');
    const buffer = Buffer.alloc(BUFFER_SIZE);

    let bytes;

    try {
      do {
        bytes = fs.readSync(fd, buffer, 0, BUFFER_SIZE);
        hash.update(buffer.slice(0, bytes));
      } while (bytes === BUFFER_SIZE);
    } finally {
      fs.closeSync(fd);
    }

    return hash.digest('hex');
  }

  static isMongoDB() {
    return typeof process.env.LIGHTMONGO_HOST !== 'undefined' || process.env.LIGHTMONGO_HOSTS !== 'undefined';
  }

  static isMySQL() {
    return typeof process.env.LIGHTMYSQL_HOST !== 'undefined';
  }

  static isSQLServer() {
    return typeof process.env.LIGHTSQLSERVER_HOST !== 'undefined';
  }

  /**
   * @deprecated
   * 判断当前进程是否为主服务。<br>通过查看环境变量里是否有[ MASTER ] [ APPSLOT ]来判断
   */
  static isMaster() {

    if (process.env.APPMASTER) {
      return process.env.APPMASTER === 'true';
    }

    if (process.env.APPSLOT) {
      return process.env.APPSLOT === '1';
    }

    return false;
  }

  static isCron() {

    if (process.env.APPCRON) {
      return process.env.APPCRON === 'true';
    }

    return false;
  }

  static isUnitTest() {
    return typeof global.it === 'function';
  }

  /**
   * 生成验证码图片
   *  size 尺寸，默认 80x30
   *  bgColor 背景颜色，默认白色
   *  textColor 文字颜色，默认灰色
   *  code 验证码，默认4位随机数
   */
  static captcha({
    size = [80, 30],
    bgColor = [0, 0, 0, 0],
    textColor = [80, 80, 80, 255],
    code = parseInt((Math.random() * 9000) + 1000)
  } = {}) {

    // 生成png
    const png = new Captcha(...size, code);

    // 背景颜色 color: background (red, green, blue, alpha)
    png.color(...bgColor);

    // 文字颜色 color: paint (red, green, blue, alpha)
    png.color(...textColor);

    return { code, png: `data:image/png;base64,${png.getBase64()}` };
  }

  /**
   * 是否是空白
   *  null, undefined, NaN, Infinity, '', ' ', []
   * @param {*} value
   */
  static isBlank(value) {
    if (value === null) {
      return true;
    }

    if (typeof value === 'undefined') {
      return true;
    }

    if (Number.isNaN(value) || value === Infinity) {
      return true;
    }

    // [''] 不视为blank，因为下面的/^\s*$/测试['']时，结果为true
    if (Array.isArray(value)) {
      return value.length <= 0;
    }

    return (/^\s*$/).test(value);
  }

  /**
   * 是否是json对象
   * @param {*} value
   */
  static isJson(value) {
    if (value === null) {
      return false;
    }
    if (typeof value === 'undefined') {
      return false;
    }
    if (value === Infinity) {
      return false;
    }
    if (value.constructor === ({}).constructor) {
      return true;
    }
    return false;
  }

  static download(handler) {
    return file.download(handler);
  }

  static stream(handler, callback) {
    return file.stream(handler, callback);
  }

  static encodeQuery(params) {
    return aksk.encode(params);
  }

  static verifyQuery(secretkey, data, httpmethod) {
    return aksk.verify(secretkey, data, httpmethod);
  }

}

module.exports = Helper;
