/**
 * 脚本启动器，支持定时执行，支持立即执行
 *
 * 为了防止一个DB启动多个AP实例时，定时任务被执行多次
 *  需要 job.domain = process.env.APPNAME
 *
 * 环境变量需要满足下面的条件，脚本才会被启动
 * - 或 process.env.APPCRON = 'true'
 *
 * 异常处理：
 * - script
 *  脚本不存在
 *
 * - action
 *  类不存在
 *  方法不存在
 *  方法内产生异常
 *
 * 产生的异常保存在 extend的error字段中
 *
 *  cron - https://github.com/ncb000gt/node-cron
 * @module lib/job
 * @author r2space@gmail.com
 * @version 1.0.0
 */

'use strict';

const { exec }    = require('child_process');
const debug       = require('debug')('light:developer');
const assert      = require('assert').strict;
const { CronJob } = require('cron');
const path        = require('path');
const Model       = require('./db/mongo/model');
const Context     = require('./context');
const log         = require('./log');
const CONST       = require('./constant');
const helper      = require('./helper');

class Job {

  static catalog = [];

  static load(catalog = []) {
    this.catalog = [].concat(this.catalog).concat(catalog);
  }

  /**
   * 初始化定时任务
   * @param {*} catalog
   */
  static init() {

    debug('init job');
    if (!helper.isCron()) {
      debug('is not master');
      return;
    }

    // Job pool. 保存Job实例，可以用指定的ID停止某个Job
    this.pool = {};

    this.catalog.forEach(item => {
      if (!item.schedule) {
        debug('schedule not set');
        return;
      }
      if (item.run !== CONST.JOB_STARTED) {
        debug('schedule already started');
        return;
      }
      if (item.env !== process.env.APPENV) {
        debug('env mismatch');
        return;
      }

      debug('start job');
      Job.start(item, Job.update);
    });
  }

  /**
   * 更新结果
   * @param {*} result
   */
  static async update(data) {

    if (process.env.APPLOCAL === 'true') {
      return;
    }

    const model = new Model(process.env.APPNAME, CONST.SYSTEM_DB_PREFIX, CONST.SYSTEM_DB_JOB);

    try {
      await model.update({ _id: data.id }, {
        last  : new Date(),
        extend: {
          result: data.error ? data.error : JSON.stringify(data.result)
        },
        status: data.error ? CONST.JOB_FAILED : CONST.JOB_COMPLETED
      });
    } catch (err) {
      log.error(err);
    }

  }

  /**
   * 定时执行
   * 时间格式与cron相同，如
   *  Seconds     : 0-59
   *  Minutes     : 0-59
   *  Hours       : 0-23
   *  Day of Month: 1-31
   *  Months      : 0-11
   *  Day of Week : 0-6
   *  Timezone    : [optional]
   *
   * @param job
   *  id: job的识别号
   *  time: 启动时间 字符串 * * * * * * 格式，或Date型（如果指定日期，则，会在指定日期被执行）
   *  params: 参数
   *  complete: Job停止时，调用的回调函数
   *  step:
   *    index:
   *    class:
   *    action:
   *    script:
   * @param tick 指定时间，脚本被执行以后，回调tick方法
   */
  static start(job, tick) {

    job.id = job.id || job._id.toString();
    debug('job id', job.id);
    debug('job schedule', job.schedule);

    assert(job.id, '[job.id] cannot be empty');
    assert(job.schedule, '[job.schedule] cannot be empty');

    // 保存cron实例
    this.pool[job.id] = new CronJob(

      // job time
      /* eslint-disable-next-line newline-per-chained-call */
      job.schedule.split(' ').slice(0, 6).join(' '),

      // tick
      Job.kick.bind(this, job, tick),

      // complete
      () => job.complete && job.complete(),

      // immediately
      true,

      // timezone
      job.schedule.split(' ')[6]
    );
  }


  /**
   * 立即执行Job
   * @param job
   */
  static immediate(job) {

    debug('job immediate', job);

    return new Promise(resolve => {
      job.id = job.id || job._id.toString();
      Job.kick(job, res => resolve(res));
    });
  }

  /**
   * 停止Job
   * @param id
   */
  static stop(id) {
    if (this.pool[id]) {
      this.pool[id].stop();
      delete this.pool[id];
    }
  }

  /**
   * 停止
   */
  static stopAll() {
    if (!this.pool) {
      return;
    }
    Object.keys(this.pool).forEach(id => Job.stop(id));
  }

  /**
   * 执行脚本或调用指定Class的指定Action
   * @param job
   * @param tick
   */
  static async kick(job, tick) {

    debug('kick', job);

    const result = [];

    try {
      for (const step of job.step) {

        // 指定了类
        if (step.type === 'action') {
          log.debug(`run step : ${step.class}#${step.action}`);
          const res = await Job.invoke(step.class, step.action, step.params);
          result.push(res);
        }

        // 指定了脚本
        if (step.type === 'script') {
          log.debug(`run step : ${step.script}`);
          const res = await new Promise((resolve, reject) => {
            exec(step.script, (err, stdout) => {
              err ? reject(err) : resolve(stdout);
            });
          });
          result.push(res);
        }
      }

      await tick({ id: job.id, result });
    } catch (err) {
      log.error(err);
      await tick({ id: job.id, error: err.toString() });
    }

  }

  /**
   * 调用指定的方法
   * @param {*} clazz
   * @param {*} action
   * @param {*} params
   */
  static async invoke(clazz, action, params) {

    const root = helper.isUnitTest() ? 'test' : '';
    const file = path.join(root, CONST.PATH_CONTROLLER, clazz);
    const inject = helper.resolve(file);

    if (!inject) {
      log.error('class not found.', clazz);
      return 'class not found';
    }

    const func = inject[action];
    if (!func) {
      log.error('action not found.', `${clazz}#${action}`);
      return 'action not found';
    }

    const handler = new Context(CONST.DEFAULT_USER, process.env.APPNAME);
    handler.addParams({ data: params });

    return await func.call(this, handler);
  }

}

module.exports = Job;
