/**
 * @file 输出Log<br>
 *  通过node的console功能打印log，如果需要将日志输出到文件里，<br>
 *  需要重定向node的执行结果<br>
 *
 *  日志的输出级别可以用环境变量的 LIGHTLOG 来设定<br>
 *   DEBUG : 3<br>
 *   WARN  : 2<br>
 *   INFO  : 1<br>
 *   ERROR : 0<br>
 *
 * @author r2space@gmail.com
 * @module lib/log
 * @version 1.0.0
 */

'use strict';

const dayjs  = require('dayjs');
const CONST  = require('./constant');
const helper = require('./helper');


/**
 * @desc debug log
 * @param {*} ...params
 */
exports.debug = function (...params) {
  const [message, user] = parse(params);
  log(toJson(CONST.LOG_TYPE_APPLICATION, CONST.LOG_LEVEL_DEBUG, message, user));
};

/**
 * @desc info log
 * @param {*} ...params
 */
exports.info = function (...params) {
  const [message, user] = parse(params);
  log(toJson(CONST.LOG_TYPE_APPLICATION, CONST.LOG_LEVEL_INFO, message, user));
};

/**
 * @desc warning log
 * @param {*} ...params
 */
exports.warn = function (...params) {
  const [message, user] = parse(params);
  log(toJson(CONST.LOG_TYPE_APPLICATION, CONST.LOG_LEVEL_WARN, message, user));
};

/**
 * @desc error log
 * @param {*} ...params
 */
exports.error = function (...params) {
  const [message, user] = parse(params);
  log(toJson(CONST.LOG_TYPE_APPLICATION, CONST.LOG_LEVEL_ERROR, message, user));
};

/**
 * @desc audit log
 * @param {*} ...params
 */
exports.audit = function (...params) {
  const [message, user] = parse(params);
  log(toJson(CONST.LOG_TYPE_AUDIT, CONST.LOG_LEVEL_INFO, message, user));
};

/**
 * @desc operation log
 * @param {*} ...params
 */
exports.operation = function (...params) {
  const [message, user] = parse(params);
  log(toJson(CONST.LOG_TYPE_OPERATION, CONST.LOG_LEVEL_INFO, message, user));
};

/**
 * @desc Stack info.
 * @param {Function} self this function.
 * @returns {Object} current method stack.
 * @ignore
 */
function stack(self) {

  const original = Error.prepareStackTrace;
  Error.prepareStackTrace = function (_, stack) {
    return stack;
  };

  const error = new Error();
  Error.captureStackTrace(error, self);

  const result = error.stack;
  Error.prepareStackTrace = original;

  return result;
}

/**
 * @desc LineNo,
 *  Defines the number of lines of parent source code.
 * @returns {String} line number.
 * @ignore
 */
function lineNo() {
  return stack(stack)[3].getLineNumber();
}

/**
 * @desc FileName,
 *  Defines the name of the parent file.
 * @returns {String} file name.
 * @ignore
 */
function fileName() {
  const name = stack(stack)[3].getFileName();
  if (process.env.APPDEV === 'true') {
    return name.replace(process.cwd(), '');
  }
  return name;
}

/**
 * @desc Function name
 * @returns {String} function name
 * @ignore
 */
function functionName() {
  return stack(stack)[3].getFunctionName();
}

/**
 * @desc生成json格式的log对象
 * @param {String} logtype log的类别
 * @param {String} loglevel log的级别
 * @param {String} content 输出的log详细
 * @param {String} userid 操作者
 * @returns {object} log对象
 * @ignore
 */
function toJson(logtype, loglevel, content, userid) {

  return {
    sec     : new Date().getTime(),
    type    : logtype,
    level   : loglevel,
    message : content,
    user    : userid ? userid : '-',
    host    : helper.ip(),
    file    : fileName(),
    line    : lineNo(),
    function: functionName()
  };
}

/**
 * 格式化参数
 * @param {*} params
 */
function parse(params) {

  const strings = params.map(param => {

    if (null === param) {
      return 'null';
    }

    if (undefined === param) {
      return 'undefined';
    }

    if (typeof param === 'string') {
      return param;
    }

    if (typeof param === 'bigint') {
      return param.toString();
    }

    if (typeof param === 'symbol') {
      return param.toString();
    }

    if (param instanceof Date) {
      return param.toISOString();
    }

    if (param instanceof Error) {
      if (process.env.APPDEV === 'true') {
        return `${param.name}:${param.message}\n${param.stack}`;
      }
      return `${param.name}:${param.message}`;
    }

    if (dayjs.isDayjs(param)) {
      return param.toISOString();
    }

    if (param.constructor.name === 'Moment') {
      return param.toISOString();
    }

    if (param.constructor.name === 'ObjectID') {
      return param.toHexString();
    }

    return JSON.stringify(param);
  });

  // 如果最后一个参数类型是ObjectID，就将这个参数视为userid
  if (strings.length > 1) {
    const last = params[params.length - 1];
    if (last && last.constructor.name === 'ObjectID') {
      const user = strings.pop();
      return [strings.join(' '), user];
    }
  }

  return [strings.join(' ')];
}

/**
 * @desc 对log对象进行格式化
 * @param {String} body 消息
 * @returns 格式化之后的消息字符串
 * @ignore
 */
function format(body) {

  const datetime = dayjs(body.sec).format('YYYY-MM-DDTHH:mm:ss.SSS');
  const level = ['ERROR', 'INFO', 'WARN', 'DEBUG'][body.level];
  const code = `${body.file}:${body.line}`;

  if (process.env.APPDEV === 'true') {
    return `[${datetime}] [A] [${level}] ${body.message} ${body.user} ${code}`;
  }

  return `[${datetime}] [A] [${level}] ${code} - - ${body.user} - ${body.message}`;
}

/**
 * @desc 输出日志
 * @param body
 * @ignore
 */
function log(body) {

  const level = typeof process.env.APPLOG === 'undefined'
    ? CONST.LOG_LEVEL_DEBUG
    : parseInt(process.env.APPLOG);

  if (body.level > level) {
    return;
  }

  if (body.level === CONST.LOG_LEVEL_ERROR) {
    return console.error.call(this, format(body));
  }

  if (body.level === CONST.LOG_LEVEL_WARN) {
    return console.warn.call(this, format(body));
  }

  if (body.level === CONST.LOG_LEVEL_INFO) {
    return console.info.call(this, format(body));
  }

  // 注释: console.debug 已经不推荐使用，应该使用 console.log
  // eslint-disable-next-line no-console
  console.log.call(this, format(body));
}
