/**
 * CSRF校验
 *  csrf值的保持需要session的支持
 *  校验用的csrf通过header，body，query发送请求，所以需要body-parse的支持
 */

const debug  = require('debug')('light:middleware:csrf-check');
const Csrf   = require('koa-csrf');
const CONST  = require('../constant');
const config = require('../configuration');

class CsrfCheck {

  static use(app) {

    /**
     * 检查csrf token
     */
    app.use(async (ctx, next) => {

      // JWT token 认证
      if (ctx[CONST.SYMBOL.JWT]) {
        debug('auth by jwt');
        return await next();
      }

      const { csrf: ignore = [] } = config.ignore || {};
      debug('ignore', ignore);

      // 检查是否是ignore的路径
      for (const item of ignore) {
        if (ctx.path.match(new RegExp(item, 'i'))) {
          debug('ignore csrf check', ctx.path);
          return await next();
        }
      }

      debug('header csrf token', ctx.get('csrf-token'), ctx.method);
      debug('header x csrf token', ctx.get('x-csrf-token'), ctx.method);
      debug('exclude method', ['GET', 'HEAD', 'OPTIONS']);

      // 校验
      await new Csrf({
        invalidTokenMessage   : 'EBADCSRFTOKEN',
        invalidTokenStatusCode: 403,
        excludedMethods       : ['GET', 'HEAD', 'OPTIONS'],
        disableQuery          : false
      })(ctx, next);
    });

    /**
     * 返回 csrf token 给请求端
     */
    app.use(async (ctx, next) => {

      if (ctx.csrf) {

        const { samesite = [] } = config.ignore || {};
        const client = ctx.get('x-client-type');

        debug('generate csrf: %s, client: %s', ctx.csrf, client);
        ctx.set({ csrftoken: ctx.csrf });
        ctx.cookies.set(CONST.COOKIE_CSRF_KEY, ctx.csrf, {
          sameSite: samesite.includes(client) ? 'none' : 'lax'
        });
      }

      await next();
    });

  }

}

module.exports = CsrfCheck;
