/**
 * index.js
 *
 * 加载顺序为
 * 1. 用户定义middlewre
 * 2. 系统定义middlewre
 *
 * 可以通过config调整加载的 middlewre 及加载的顺序
 */

'use strict';

const path   = require('path');
const debug  = require('debug')('light:middleware');
const CONST  = require('../constant');
const helper = require('../helper');
const log    = require('../log');

class Middleware {

  /**
   * 根据自定义项，调整中间件的加载顺序
   * @param {*} app
   * @param {*} custom
   */
  static init(app, custom = {}) {

    debug('init from', __dirname, 'config.yml');
    const root = helper.isUnitTest() ? 'test' : '';
    const config = this.combine(
      helper.yamlLoader('config.yml', __dirname),
      helper.yamlLoader(path.join(root, CONST.PATH_MIDDLEWARE, 'config.yml'))
    );

    // 如果指定的是数组，那么替换取胜的配置
    if (Array.isArray(custom)) {
      return Middleware.use(app, custom);
    }

    // 否则用参数覆盖缺省的设置
    Object.keys(custom).forEach(key => {
      const row = config.find(item => item.name === key);
      if (row) {
        return Object.assign(row, custom[key]);
      }
      config.push(custom[key]);
    });

    // 用order字段排序
    Middleware.use(app, config.sort((val1, val2) => {
      if (val1.order === val2.order) {
        return 0;
      }
      return val1.order > val2.order ? 1 : -1;
    }));
  }

  /**
   * 加载实际的中间件
   * @param {*} app
   * @param {*} items
   */
  static use(app, items) {

    for (const item of items) {

      debug('middleware', item.name, item.use);

      // 只要没有设置false，就默认加载
      if (!item.name || item.use === false) {
        log.warn(`Disable middleware ${item.name}`);
        continue;
      }

      // 加载用户自定义的middleware
      const root = helper.isUnitTest() ? 'test' : '';
      let clazz = helper.resolve(path.join(root, CONST.PATH_MIDDLEWARE, item.name));
      if (clazz && clazz.use) {
        log.debug(`Use custom middleware ${item.name}`);
        clazz.use(app);
        continue;
      }

      // 加载系统默认的middleware
      clazz = helper.resolve(item.name, __dirname);
      if (clazz && clazz.use) {
        log.debug(`Use system middleware ${item.name}`);
        clazz.use(app);
        continue;
      }

      log.error(`Middleware not found ${item.name}.js`);
    }
  }

  /**
   * 合并中间件定义
   * @param {*} middlewares
   * @param {*} customs
   */
  static combine(middlewares, customs) {

    if (!customs || customs.length <= 0) {
      return middlewares;
    }

    customs.forEach(custom => {
      const middleware = middlewares.find(row => custom.order === row.order);
      if (middleware) {
        Object.assign(middleware, custom);
      } else {
        middlewares.push(custom);
      }
    });

    return middlewares;
  }

}

module.exports = Middleware;
