/**
 * 路径路由器
 */

const path         = require('path');
const ejs          = require('ejs');
const debug        = require('debug')('light:middleware:router');
const CONST        = require('../constant');
const Context      = require('../context');
const log          = require('../log');
const response     = require('../response');
const helper       = require('../helper');
const rider        = require('../model/datarider');
const config       = require('../configuration');
const i18n         = require('../i18n');
const setting      = require('../setting');
const { NotFound } = require('../error').http;

class Router {

  static use(app) {
    app.use(async ctx => {

      debug('request', ctx.path);
      const handler = new Context().bind(ctx);

      // 路由API请求
      if (ctx.path.match(/^\/api\/.*/)) {
        debug('route api');
        return await Router.routeApi(handler);
      }

      // 路由网页模板
      debug('route template');
      await Router.routeTemplate(handler);
    });
  }

  /**
   * 路由API
   * @param {*} handler
   */
  static async routeApi(handler) {

    const { route: board, params } = Router.findDefine(setting.board, handler.ctx.path);
    if (!board) {
      log.error(`Did not find the api : ${handler.ctx.path}`, handler.uid);
      throw new NotFound('Board not found');
    }

    log.info(`Route api   : ${handler.ctx.method} ${board.api}, action: ${board.class}#${board.action}`, handler.uid);

    handler.addParams(params);
    await Router.invokeApi(handler, board);
  }

  /**
   * 路由HTML模版
   * @param {*} handler
   */
  static async routeTemplate(handler) {

    const { ctx } = handler;
    const { route, params } = Router.findDefine(setting.route, ctx.path);

    if (!route) {
      log.error(`Did not find the route : ${handler.req.path}`);
      throw new NotFound('Route not found');
    }

    log.info(`Route html  : ${ctx.method} ${route.url}, template: ${route.template}`, handler.uid);

    handler.addParams(params);
    await Router.invokeRoute(handler, route);
  }

  /**
   * 调用指定的方法
   * @param {*} handler
   * @param {*} board
   */
  static async invokeApi(handler, board) {

    const { path: folder = '', class: clazz, action, kind } = board;
    const root = helper.isUnitTest() ? 'test' : '';

    debug('root: %s path: %s class: %s action: %s kind: %s', root, folder, clazz, action, kind);

    // 调用用户定义的controller
    const file = path.join(root, folder, CONST.PATH_CONTROLLER, clazz);
    let ctrl = helper.resolve(file) || {};
    let func = ctrl[action];

    debug('file:', file);

    // 逻辑API时，类与方法必须存在
    if (CONST.API_KIND_USER_LOGIC === kind && !func) {
      log.error(`Did not find the module : ${folder}#${clazz}#${action}`, handler.uid);
      throw new NotFound('Module not found');
    }

    // 调用datarider
    if (!func) {
      ctrl = rider[clazz] || {};
      func = ctrl[action];
    }

    // 指定的method没有找到
    if (!func) {
      log.error('Can not dispatch', handler.uid);
      throw new NotFound('Can not dispatch');
    }

    // 如果方法有两个参数则视为指定了callback回调方法（兼容旧的回调式调用方式）
    // Rider的接口定义是三个参数（handler，params，callbac），所以执行下面的async形式调用逻辑
    if (func.length === 2) {
      log.warn('Does not support callback method return value. action:', `${clazz}#${action}`, handler.uid);
      return new Promise((resolve, reject) => {
        func.call(ctrl, handler, (error, data) => {
          if (error) {
            log.error(error, handler.uid);
            return reject(error);
          }
          resolve();
          response.send(handler, data);
        });
      });
    }

    // 使用async的调用方式
    response.send(handler, await func.call(ctrl, handler));

  }

  /**
   * 画面路径匹配
   * @param {*} handler
   * @param {*} route
   */
  static async invokeRoute(handler, route) {

    // 画面用参数
    const parameter = {
      handler,
      req      : handler.req,
      res      : handler.res,
      user     : handler.user,
      info     : helper.applicationInfo(),
      env      : process.env,
      conf     : config,
      data     : {},
      i        : i18n.i,
      catalog  : i18n.catalog,
      csrftoken: handler.ctx.csrf
    };

    if (!route.class) {
      return Router.renderTemplate(handler, route, parameter);
    }

    // 查找自定义controller方法，如果存在就顶用该方法
    const root = helper.isUnitTest() ? 'test' : '';
    const file = path.join(root, CONST.PATH_CONTROLLER, route.class);
    const ctrl = helper.resolve(file) || {};
    const func = ctrl[route.action];

    if (!func) {
      log.error('Can not dispatch', handler.uid);
      throw new NotFound('Can not dispatch');
    }

    log.debug(`Inject     : ${route.class}#${route.action}`);

    // 如果方法有两个参数则视为指定了callback回调方法（兼容旧的回调式调用方式）
    if (func.length === 2) {
      log.warn('Does not support callback method return value.', handler.uid);
      return new Promise((resolve, reject) => {
        func.call(this, handler, (error, data) => {
          if (error) {
            log.error(error, handler.uid);
            return reject(error);
          }
          resolve();
          parameter.data = data;
          Router.renderTemplate(handler, route, parameter);
        });
      });
    }

    // 使用async的调用方式
    parameter.data = await func.call(this, handler);
    Router.renderTemplate(handler, route, parameter);
  }

  /**
   * 使用ejs模板渲染画面
   * 通过正则替换静态资源的URL
   *   src="/static/xxx" > src="/static/xxx?stamp=453fc7f4"
   *   src="//cdn.zf.link/static/xxx" > src="//cdn.zf.link/static/xxx?stamp=453fc7f4"
   *   href="/static/xxx" > href="/static/xxx?stamp=453fc7f4"
   * @param handler
   * @param define
   * @param parameter
   */
  static renderTemplate(handler, route, parameter) {

    // 如果body有值，则直接返回（自定义的 controllers 里已经做了http响应）
    if (handler.ctx.body) {
      return;
    }

    const { views = 'views', stamp = 'light' } = config.app || {};
    const root = helper.isUnitTest() ? 'test' : '';
    const file = path.join(process.cwd(), root, views, route.template);

    ejs.renderFile(file, { ...parameter, ...route.parameter }, (err, src) => {
      if (err) {
        log.error(err);
        throw new Error(err);
      }

      let html = src;
      html = html.replace(/( src=("|')[^"']*\/static\/[^"']*)/g, `$1?stamp=${stamp}`);
      html = html.replace(/( href=("|')[^"']*\/static\/[^"']*)/g, `$1?stamp=${stamp}`);

      response.sendHTML(handler, html);
    });

  }

  /**
   * @desc 从URL定义一览里，查找是否与指定URL完全匹配的项目。支持URL参数，如
   *
   *  定义的URL
   *   /admin/board/:id
   *  请求的URL
   *   /admin/board/552347b88f6e69d6d9ca9768
   *
   *  URL会被 / . _ 分割后进行匹配
   *   /WW_VERIFY_:id.TXT
   *
   * @param url
   * @param items
   * @returns {Object}
   */
  static findDefine(items, url) {

    const spliter = /\/|\.|_/;
    const request = url.split(spliter);

    for (const item of items) {

      // 用符号 / . _ 分割URL进行匹配，以支持类似 /WW_VERIFY_:id.TXT URL的匹配
      const define = (item.url || item.api).split(spliter);
      if (define.length !== request.length) {
        continue;
      }

      let matched = true;
      const params = define.reduce((memo, path, index) => {

        // 如果是 : 开头，则认为是URL路径参数
        if (path.charAt(0) === ':') {
          memo[path.split(':')[1]] = request[index];
        } else {
          matched = matched && (path === request[index]);
        }
        return memo;
      }, {});

      if (matched) {
        return { route: item, params };
      }
    }

    // 如果有通配符定义，返回该项
    const wildcard = items.find(item => item.url === '*');
    if (wildcard) {
      return { route: wildcard, params: {} };
    }

    return {};
  }

}

module.exports = Router;
