/**
 * Session
 */

const debug        = require('debug')('light:middleware:session');
const session      = require('koa-session');
const config       = require('../configuration');
const helper       = require('../helper');
const sessionstore = require('../db/sessionstore');
const CONST        = require('../constant');

class Session {

  static use(app) {

    const { sessionSecret = 'light.secret', sessionKey = 'light.sid' } = config.app || {};
    debug('session secret', sessionSecret);
    debug('session key', sessionKey);

    app.keys = [sessionSecret];
    app.use(async (ctx, next) => {

      // 如果时JWT请求，跳过session
      if (ctx[CONST.SYMBOL.JWT]) {
        debug('auth by jwt');
        return await next();
      }

      // 区分移动端与浏览器端timeout
      const { sessionTimeout = 24, sessionTimeoutMobile } = config.app || {};
      const { samesite = ['electron'] } = config.ignore || {};
      const client = ctx.get('x-client-type');
      const timeout = helper.isBrowser(ctx)
        ? sessionTimeout
        : sessionTimeoutMobile || sessionTimeout || 168;

      debug('client', client);
      debug('timeout: %d client: %s', timeout, client);
      await session({
        key     : sessionKey,
        rolling : true,                                       // cookie 有效期更新策略
        maxAge  : 60 * 60 * 1000 * timeout,                   // cookie 有效期间
        signed  : ctx.secure,                                 // cookie 加上签名
        sameSite: samesite.includes(client) ? 'none' : 'lax', // 开启严格的跨域cookie控制
        httpOnly: true,                                       // 不允许浏览器中的js代码来获取cookie
        store   : sessionstore.init({ db: ctx.db })           // session 外表存储
      }, app)(ctx, next);
    });

  }

  /**
   * 为了强化session的安全性，在登陆成功后重新生成新的sid，
   * 这里使用直接修改 _sessCtx 内部变量的方式，后续升级koa-session要注意该方式是否继续有效
   * @param {*} ctx
   */
  static renew(ctx) {
    const sid = helper.uuid();
    debug('renew session', sid);

    ctx.session._sessCtx.externalKey = sid;
  }

}

module.exports = Session;
