/**
 * @file ETL共同处理
 * @author r2space@gmail.com
 * @module common.js
 * @version 1.0.0
 */

'use strict';

const assert = require('assert').strict
  , debug    = require('debug')('light:model:migrate')
  , rider    = require('../datarider')
  , helper   = require('../../helper')
  ;


/**
 * 获取关联表的数据, 支持数组内容的关联
 * 是否是数组，根据传入的值的类型来判断，即Array.isArray为真时，按数组处理
 * @param handler
 * {
 *    params.data: {    // data里保存的数据是单行数据
 *      type: 'body',
 *      lang: [{en: '1', ja: '1'}, {en: '2', ja: '2'}]
 *    }
 * }
 * @param mapping
 * {
 *    key      : 'uid'
 *    variable : ''
 *    schema   : 'user'
 *    condition: { _id: '$uid', sex: '1' }
 *    fields   : ['name']
 *  }
 * @returns {*}
 */
exports.join = async function (handler, mapping) {

  if (!mapping.schema) {
    return;
  }

  const { data } = handler.params;
  data.options = {};

  debug('join', mapping.schema);
  await ctrl().join(handler, data, mapping);

  /**
   * 在ETL的导入功能里, 通常这个功能是用[名称]获取名称对应的[ID], [ID]会替换原有的[名称]保存
   * 旧的[名称]也需要保存起来, 为了处理过程中出错时, 用[名称]来提示哪个列有问题。
   * [名称]可以是用逗号分隔的数组, 这时需要遍历所有的值并保存原来的内容。
   *
   * 在ETL的导出功能里, 则相反, 一般是用[ID]获取名称。同时还支持[variable]功能, 通常这个功能用于ETL的导出。
   * 当variable被指定了, 那么取回来的值不会替换原来的值, 以方便后续项目参照
   * 获取来的值, 会以variable里指定的名称来保存。 在ETL的导出时, 如果指定了variable, 那么这里的值优先输出。
   */
  const option = Object.values(data.options[mapping.schema])
    , values = helper.get(option, mapping.fields[0]) // 关联只能选择单个字段
      .filter(item => typeof item !== 'undefined');

  if (values.length <= 0) {
    debug('no joined value');
    return;
  }

  /**
   * 当 key = 'lang.en' 时
   * 按 data = { lang: { en: [1, 2, 3] } } 格式保存数据
   * 以便后续可以用 helper.get(data, 'lang.en') 的统一方式获取到值
   *
   * 保存值时，优先使用 variable
   */
  const key = mapping.variable || mapping.key
    , [top, second] = key.split('.');
  if (second) {

    debug('set nested key:', key);
    data[top] = { [second]: values.length === 1 ? values[0] : values };
  } else {

    debug('set key:', key);
    data[top] = values.length === 1 ? values[0] : values;
  }

  delete data.options;
};

/**
 * 导入时，如果指定了uniqueKey，用该值获取数据库中对应的数据
 */
exports.findByUniqueKey = async function (importer, row) {

  debug('uniqueKey', importer.uniqueKey);
  if (!importer.uniqueKey || importer.uniqueKey.length <= 0) {
    return {};
  }

  const condition = importer.uniqueKey.reduce((memo, key) => {
    memo[key] = row[key];
    return memo;
  }, {});

  debug('conditions', condition);
  return await rider[importer.schema].get(importer.handler, condition);
};

exports.init = async function (controller, handler, model) {

  if (!controller) {
    debug('no initialize controller');
    return;
  }

  if (!controller.initialize) {
    debug('no initialize function');
  }

  debug('call custom initialize');
  return await controller.initialize.call(this, handler, model);
};

exports.before = async function (controller, handler, data) {

  if (!controller) {
    debug('no before controller');
    return;
  }

  if (!controller.before) {
    debug('no before function');
    return;
  }

  debug('call custom before');
  return await controller.before.call(this, handler, data);
};

exports.parse = async function (controller, handler, row) {

  if (!controller) {
    debug('no parse controller');
    return;
  }

  if (!controller.parse) {
    debug('no parse function');
    return;
  }

  debug('call custom parse');
  return await controller.parse.call(this, handler, row);
};

exports.after = async function (controller, handler, data) {
  if (!controller) {
    debug('no after controller');
    return;
  }

  if (!controller.after) {
    debug('no after function');
    return;
  }

  debug('call custom after');
  return await controller.after.call(this, handler, data);
};

exports.valid = async function (controller, handler, row) {
  if (!controller) {
    debug('no valid controller');
    return;
  }

  if (!controller.valid) {
    debug('no valid function');
    return;
  }

  debug('call custom after');
  return await controller.valid.call(this, handler, row);
};

exports.dump = async function (controller, handler, data) {

  if (!controller) {
    debug('no dump controller');
    return;
  }

  if (!controller.dump) {
    debug('no dump function');
    return;
  }

  debug('call custom dump');
  return await controller.dump.call(this, handler, data);
};

exports.end = async function (controller, handler, result) {

  if (!controller) {
    debug('no end controller');
    return;
  }

  if (!controller.end) {
    debug('no end function');
    return;
  }

  debug('call custom end');
  return await controller.end.call(this, handler, result);
};

function ctrl() {
  if (process.env.LIGHTSQLSERVER_HOST) {
    return require('../datarider/sqlserver');
  }

  if (process.env.LIGHTMYSQL_HOST) {
    return require('../datarider/mysql');
  }

  if (process.env.LIGHTMONGO_HOST || process.env.LIGHTMONGO_HOSTS) {
    return require('../datarider/mongo');
  }

  assert.fail('invalid database type');
}

