/**
 * @file Excel解析, 生成
 *  参考 https://github.com/SheetJS/sheetjs
 * @author r2space@gmail.com
 * @module lib.model.datamigrate.excel
 * @version 1.0.0
 */

'use strict';

const debug = require('debug')('light:model:migrate')
  , XLSX = require('xlsx');


/**
 * 抽取xlsx的内容, 以sheet为单位, 结果为二位数组
 * @param name    Excel文件路径
 * @param mapping 指定excel列与数据库字段的对应关系
 * @param type    字段对应方法 index | title
 * @param title   是否要表头
 *  要表头结果式为: [{ content: [], title: [] }]
 *  不要表头结果式为: [content]
 */
exports.parse = function (name, mapping = [], type = 'index', title = false) {

  const workbook = XLSX.readFile(name);

  // 遍历Sheet
  return workbook.SheetNames.map(name => {

    const worksheet = workbook.Sheets[name];
    debug('parse by', type);

    if (!worksheet['!ref']) {
      debug('empty worksheets');
      return null;
    }

    const range = XLSX.utils.decode_range(worksheet['!ref']);

    let keys = [];
    if (type === 'title') {

      // 根据 mapping，按照列标题设定列名称
      keys = Array(range.e.c + 1).fill()
        .map((dummy, col) => {

          // 第一行视为列标题
          const address = XLSX.utils.encode_cell({ c: col, r: 0 })
            , val = XLSX.utils.format_cell(worksheet[address])
            , item = mapping.find(item => item.title === val) || {};

          return item.variable || item.key;
        })
        .filter(item => item);

      debug('title keys:', keys);
    } else {

      // 根据 mapping，按照列顺序设定列名称
      keys = Array(range.e.c + 1).fill()
        .map((dummy, col) => {
          const item = mapping.find(item => item.col === col + 1);

          if (item) {
            return item.variable || item.key;
          }

          // 如果 mapping 里没有指定key, 则将第一行的值作为字段名
          const address = XLSX.utils.encode_cell({ c: col, r: 0 })
            , val = XLSX.utils.format_cell(worksheet[address]);

          return val || String(col + 1);
        });
      debug('index keys:', keys);
    }

    mapping.forEach(item => {

      // rule=fix 即，使用固定值的项目做数据导入处理
      if (item.sanitize && item.sanitize.rule === 'fix') {
        return;
      }

      // 将excel中没有出现的mappings规则设为无效（不进行数据导入处理）
      // 在importer的行parse时，确认valid = 0
      if (!keys.includes(item.key)) {
        item.valid = 0;
      }
    });

    // 转换 sheet 数据为 json

    const content = Array(range.e.r).fill()
      .map((dummy, row) => row + 1)
      .map(row => Array(range.e.c + 1)
        .fill()
        .reduce((memo, dummy, col) => {
          if (!keys[col]) {
            return memo;
          }

          const address = XLSX.utils.encode_cell({ c: col, r: row });
          memo[keys[col]] = XLSX.utils.format_cell(worksheet[address]);
          return memo;
        }, {}));

    // 获取Excel表头
    if (title) {
      return {
        content,
        title: Array(range.e.c + 1).fill()
          .map((dummy, col) => {
            const address = XLSX.utils.encode_cell({ c: col, r: 0 });
            return XLSX.utils.format_cell(worksheet[address]);
          })
      };
    }

    return content;
  });
};


/**
 * 导出数据到xlsx里
 * @param name
 * @param data
 */
exports.dump = function (name, data) {

  function Workbook() {
    if (!(this instanceof Workbook)) return new Workbook();
    this.SheetNames = ['data'];
    this.Sheets = { data: toCsrSpreadSheet(data) };
  }

  XLSX.writeFile(new Workbook(), name);
};


function toCsrSpreadSheet(data) {

  const range = { s: { c: 0, r: 0 }, e: { c: 0, r: data.length - 1 } }
    , sheet = Array(data.length)
      .fill()
      .reduce((memo, item, row) => (Array(data[row].length)
        .fill()
        .reduce((memo, item, index) => {
          range.e.c = range.e.c < index ? index : range.e.c;

          const cell = { v: data[row][index], t: 's' };
          if (cell.v === 'number') {
            cell.t = 'n';
          }

          if (cell.v === 'boolean') {
            cell.t = 'b';
          }

          memo[XLSX.utils.encode_cell({ c: index, r: row })] = cell;
          return memo;
        }, memo)
      ), {});

  sheet['!ref'] = XLSX.utils.encode_range(range);
  sheet['!cols'] = range.e.c;
  return sheet;
}
