/**
 * @file ETL处理，执行数据导出
 *
 *  自定义的controller，可以定义下面的方法，加载数据的时候会被调用
 *   dump
 *
 * @author r2space@gmail.com
 * @module lib.model.importer
 * @version 1.0.0
 */

'use strict';

const os      = require('os')
  , path      = require('path')
  , fs        = require('fs')
  , assert    = require('assert').strict
  , debug     = require('debug')('light:model:migrate')
  , excel     = require('./excel')
  , common    = require('./common')
  , rider     = require('../datarider')
  , file      = require('../file')
  , helper    = require('../../helper')
  , validator = require('../../validator')
  , logger    = require('../../log')
  , CONST     = require('../../constant')
  ;


class Exporter {

  /**
   * 构造函数
   * @param handler
   * @param options
   */
  constructor(handler, options = {}) {

    assert(options.name || options.schema, 'name or schema cannot be empty');

    this.original = handler;
    this.handler = handler.copy();
    this.init(options);
  }

  /**
   * 初始化参数
   * @param {*} etl
   */
  init(options) {
    this.name = options.name;
    this.type = options.type;
    this.schema = options.schema;
    this.mappings = options.mappings;
    if (options.class) {
      const root = helper.isUnitTest() ? 'test' : '';
      this.controller = helper.resolve(path.join(root, CONST.PATH_CONTROLLER, options.class));
    }
  }

  /**
   * 执行导出功能
   * @param {*} filename 明确指定输出文件
   */
  async exec(filename) {

    this.filename = filename;
    logger.debug('exporter exec.');
    debug('exec.');

    try {

      if (this.name) {

        debug('fetch etl by name:', this.name);

        const handler = this.handler.copy({ condition: { name: this.name } });
        handler.code = CONST.SYSTEM_DB_PREFIX;
        const etl = await rider.etl.get(handler);

        debug('elt mappings', etl);
        if (etl._id) {
          this.init(etl);
        }
      }

      const data = await this.transform();
      return await this.dump(data);
    } catch (err) {
      logger.error(err);
      throw err;
    }

  }


  /**
   * 转换数据
   */
  async transform() {

    debug('transform.');
    assert(rider[this.schema], 'rider is not defined');
    assert(rider[this.schema].list, 'rider list is not defined');

    debug('params', this.handler.params);
    const { items, totalItems } = await rider[this.schema].list(this.handler)
      , result = [];

    debug('total items', totalItems);
    for (const row of items) {
      await this.parse(row);
      result.push(this.handler.params.data);
    }

    return result;
  }


  /**
   * 处理每行数据
   * @param row
   * @param callback
   */
  async parse(row) {

    debug('parse.');

    for (const mapping of this.mappings) {

      // 获取关联数据
      this.handler.params.data = row;
      await common.join(this.handler, mapping);

      // 数据格式化，并将对象数据扁平化处理
      const key = mapping.variable || mapping.key;
      if (mapping.sanitize && mapping.sanitize.rule) {
        mapping.sanitize.key = `data.${key}`;

        debug('sanitize', mapping.sanitize);
        row[key] = await validator.format(this.handler, mapping.sanitize);
      } else {
        row[key] = helper.get(this.handler.params.data, key);
      }
    }

  }


  /**
   * 后期处理
   */
  async dump(data) {

    debug('dump.');

    // 只获取 col 被指定的项目, 并且以 col 的值排序
    const mapping = this.mappings
      .filter(item => typeof item.col !== 'undefined')
      .sort((val1, val2) => (val1.col > val2.col ? 1 : -1));

    // 添加标题栏
    data.unshift(mapping.reduce((memo, item) => {
      memo[item.variable || item.key] = item.title || item.key;
      return memo;
    }, {}));
    debug('titles', data[0]);

    // 尝试调用自定义方法
    await common.dump(this.controller, this.original, data);

    // 按照col的定义输出列, 当variable被定义, 那么他的值优先被使用
    // 如果col被重复定义, 那么后面的内容将最终得到输出
    const filename = this.filename || path.join(os.tmpdir(), helper.randomGUID8());
    excel.dump(filename, data.map(row => mapping.reduce((memo, item) => {
      memo[parseInt(item.col) - 1] = row[item.variable || item.key];
      return memo;
    }, [])));

    // 如果指定了filename，那么仅生成文件不上传gridfs
    if (this.filename) {
      debug('total', this.total);
      return { total: data.length - 1 };
    }

    // 写到 gridfs 里
    this.handler.params.data = {};
    this.handler.params.files = [{
      stream: fs.createReadStream(filename),
      name  : this.resultName(),
      type  : this.resultType()
    }];
    const [result] = await file.add(this.handler);

    // 删除临时文件
    fs.unlinkSync(filename);

    debug('total', data.length - 1, '_id', result._id);
    return { total: data.length - 1, _id: result._id };
  }

  resultName() {
    if (this.filename) {
      return this.filename;
    }
    if (this.type === 'excel') {
      return `${this.schema}.xlsx`;
    }
    if (this.type === 'csv') {
      return `${this.schema}.csv`;
    }
    return this.schema;
  }

  resultType() {
    if (this.type === 'excel') {
      return 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet';
    }
    if (this.type === 'csv') {
      return 'text/csv';
    }
    return 'application/octet-stream';
  }

}

module.exports = Exporter;
