/**
 * @file ETL处理，执行数据导入
 *
 *  数据的路径
 *   通过 clean - extract - transform - load 阶段，导入到指定collection中
 *
 *  自定义的controller，可以定义下面的方法，加载数据的时候会被调用
 *   init
 *   before
 *   parse
 *   after
 *   valid
 *   end
 *
 * @author r2space@gmail.com
 * @module lib.model.importer
 * @version 1.0.0
 */

'use strict';

const path          = require('path')
  , assert          = require('assert').strict
  , debug           = require('debug')('light:model:migrate')
  , excel           = require('./excel')
  , common          = require('./common')
  , { MemoryStore } = require('./store')
  , rider           = require('../datarider')
  , helper          = require('../../helper')
  , validator       = require('../../validator')
  , rule            = require('../../validator/rule')
  , setting         = require('../../setting')
  , logger          = require('../../log')
  , CONST           = require('../../constant')
  ;


class Importer {

  /**
   * 构造函数
   * @param handler
   * @param options
   */
  constructor(handler, options = {}) {

    this.log = [];    // 保存校验行数据时的异常信息
    this.titles = []; // 保存Excel的标题
    this.update = []; // 保存数据库中的待更新值（用uniqueKey值获取）

    this.meta = {
      count   : 0,
      total   : 0,
      success : 0,
      error   : [],
      inserted: [],
      updated : []
    };

    this.primitive = new MemoryStore();
    this.processed = new MemoryStore();

    this.handler = handler.copy();
    this.init(options);
  }


  init(options) {

    this.name = options.name;
    this.schema = options.schema;
    this.mappings = options.mappings;

    this.type = options.type;
    this.mappings = options.mappings || [];
    this.mappingType = options.mappingType;
    this.uniqueKey = options.uniqueKey;

    this.allowError = options.allowError;
    this.allowErrorMax = options.allowErrorMax;
    this.allowUpdate = options.allowUpdate;

    // 获取自定义类
    if (options.class) {
      const root = helper.isUnitTest() ? 'test' : '';
      this.controller = helper.resolve(path.join(root, CONST.PATH_CONTROLLER, options.class));
    }
  }


  /**
   * 执行导入功能
   * @param callback
   */
  async exec(callback) {

    debug('exec.');

    try {

      // 如果指定了name，name优先
      if (this.name) {
        debug('fetch etl by name:', this.name);

        const handler = this.handler.copy({ condition: { name: this.name } });
        handler.code = CONST.SYSTEM_DB_PREFIX;
        const etl = await rider.etl.get(handler);

        debug('elt mappings', etl);
        if (etl._id) {
          this.init(etl);
        }
      }

      await this.initialize();
      await this.extract();
      await this.transform();
      await this.load();
      await this.end();

      const result = {
        total   : this.meta.total,
        success : this.meta.success,
        inserted: this.meta.inserted,
        updated : this.meta.updated,
        error   : this.log
      };
      return callback ? callback(null, result) : Promise.resolve(result);
    } catch (err) {
      logger.error(err);
      return callback ? callback(err) : Promise.reject(err);
    }
  }


  /**
   * 导入前初始化
   */
  async initialize() {

    debug('initialize.');

    await this.primitive.drop();
    await this.processed.drop();
    await common.init(this.controller, this.handler, this.primitive);
  }


  /**
   * 从导入源加载数据
   */
  async extract() {

    debug('extract.');

    if (this.type === 'excel') {

      debug('file', this.handler.params.files[0].path);
      const sheets = excel.parse(this.handler.params.files[0].path, this.mappings, this.mappingType, true)
        , data = [];

      sheets.forEach(sheet => {
        if (sheet) {
          this.titles.push(sheet.title);
          sheet.content.forEach(row => data.push(row));
        }
      });

      // 尝试调用用户自定义的 befor 方法
      await common.before(this.controller, this.handler, data);

      // Excel的数据总行数
      this.meta.count = data.length;

      for (const row of data) {

        // TODO: {a: {b: [10]}} 还是 {a: [{b: 10}]}
        // 扁平化结构数据转对象 {a.b: '10'} => {a: { b: '10 }}
        // Object.keys(row).forEach(key => {
        //   if (key.includes('.')) {

        //     debug('has sub document', key);
        //     helper.set(row, key, row[key]);
        //     delete row[key];
        //   }
        // });
        await this.primitive.add(row);
      }
      return;
    }

    assert.fail(`${this.type} format is not yet supported`);
  }


  /**
   * 转换数据
   * 1. 校验表头
   * 2. 数据转换
   * @param callback
   */
  async transform() {

    debug('transform.');

    // 检查导入文件的头是否是正确的格式
    const option = setting.validator
      .find(item => item.group === this.name && item.rule === CONST.EXTEND.RULE.CHECKFIELD);
    if (option) {

      debug('check field');
      [this.handler.params.data] = this.titles;
      const error = await rule.checkField(this.handler, option);

      // 有表头校验错误，不再继续导入
      if (error) {
        debug('has check field error', error);
        return this.log.push(error);
      }
    }

    // 遍历所有数据
    const cursor = await this.primitive.cursor();
    for (const one of Array(this.meta.count).fill(1)) {

      this.meta.total = this.meta.total + one;

      // 逐行进行转换
      const parsed = await this.parse(await cursor.next());

      // 确认异常
      if (this.log.length > 0) {
        debug('allow error max: %s, current error count: %d', this.allowErrorMax, this.log.length);
        if (this.allowError && this.log.length < (this.allowErrorMax || 0)) {
          debug('continue parse');
          continue;
        }
        debug('stop parse');
        return;
      }

      // 添加数据
      await this.processed.add(parsed);
    }
  }


  /**
   * 处理每行数据
   * @param row
   */
  async parse(row) {

    debug('parse.');
    this.handler.params.data = row;

    // 类型转换
    for (const mapping of this.mappings) {

      // valid值在按列标题导入模式时，没有找到对应的列定义会被设定为0，参考excel.js
      if (mapping.valid === 0) {
        debug('in valid mapping', mapping);
        continue;
      }

      // 格式化
      if (mapping.sanitize && mapping.sanitize.rule) {
        const key = mapping.variable || mapping.key;
        mapping.sanitize.key = `data.${key}`;

        debug('sanitize', mapping.sanitize);
        const data = await validator.format(this.handler, mapping.sanitize);
        helper.set(row, key, data);
      }

      // 获取关联数据
      debug('get join data');
      await common.join(this.handler, mapping);
    }

    // 尝试调用开发者自定义的类型转换
    debug('try call custom parse');
    await common.parse(this.controller, this.handler, row);

    // 获取数据库中对应的数据
    const data = await common.findByUniqueKey(this, row);
    if (data._id) {
      debug('find update data', data._id);
      this.update.push({ index: this.meta.total, data });
    }

    // 数据校验
    let errs = await validator.check(this.handler, this.name, data._id ? 'update' : 'add');
    errs.forEach(item => item.row = this.meta.total); // 异常中添加行信息

    debug('has error', errs.length);
    this.log = this.log.concat(errs);

    // 尝试调用开发者自定义的数据校验
    debug('try call custom valid');
    errs = await common.valid(this.controller, this.handler, row);
    if (errs) {
      errs.forEach(item => item.row = this.total);

      debug('has custom error', errs.length);
      this.log = this.log.concat(errs);
    }
  }


  /**
   * 后期处理
   */
  async load() {

    debug('load.');

    // 有校验错误, 且 allowError = false 则不更新最终数据库
    if (!this.allowError && this.log.length > 0) {
      debug('has error, count %d, allow error: %s', this.log.length, this.allowError);
      return;
    }

    const data = await this.processed.list();

    // 尝试调用用户的方法
    await common.after(this.controller, this.handler, data);

    for (const row of data) {
      delete row._id;
      await this.add(row);
      this.meta.success = this.meta.success + 1;
    }
  }


  /**
   * 向最终表添加数据
   * @param row
   * @param callback
   */
  async add(row) {

    debug('add.');

    // 当允许更新的FLG被设定, 需要判断是否有 uniqueKey 相匹配的数据
    if (this.allowUpdate) {

      debug('allow update');

      // 判断有, 则更新, 没有就插入一条新数据
      const { data } = this.update.find(item => item.index === this.meta.success) || {};
      if (data) {
        debug('update id:', data._id);
        const res = await rider[this.schema].update(this.handler, { id: data._id, data });
        return this.meta.updated.push(res._id);
      }
    }

    const res = await rider[this.schema].add(this.handler, { data: row });
    return this.meta.inserted.push(res._id);
  }


  /**
   * 结束处理, 如果是自动生成的临时表, 则执行完以后自动删除
   */
  async end() {

    debug('end.');

    const result = {
      total   : this.meta.total,
      success : this.meta.success,
      inserted: this.meta.inserted,
      updated : this.meta.updated,
      error   : this.log
    };

    await common.end(this.controller, this.handler, result);
    await this.primitive.drop();
    await this.processed.drop();
  }

}

module.exports = Importer;
