/**
 * index.js
 *
 * 封装DataStore
 *  通过board里定义Class和Action，创建DataStore访问对象
 *  如 user.get() user.list() 等
 */

'use strict';

const assert  = require('assert').strict;
const path    = require('path');
const debug   = require('debug')('light:model:rider');
const light   = require('../light');
const setting = require('../../setting');
const CONST   = require('../../constant');
const helper  = require('../../helper');


class Rider {

  /**
   * 根据环境变量设置，获取特定数据库的Rider实例
   */
  static instance() {
    if (process.env.LIGHTSQLSERVER_HOST) {
      return require('./sqlserver');
    }

    if (process.env.LIGHTMYSQL_HOST) {
      return require('./mysql');
    }

    if (process.env.LIGHTMONGO_HOST || process.env.LIGHTMONGO_HOSTS) {
      return require('./mongo');
    }

    assert.fail('invalid database type');
  }

  /**
   * 初始化
   */
  static init() {

    if (this.initialized) {
      debug('rider has been initialized');
      return;
    }

    // 初始化系统表控制器
    light.bind();

    // 初始化用户表控制器
    this.bind();

    // 初始化标识，防止执行多次初始化
    this.initialized = true;
  }

  /**
   * 根据board定义，创建访问类
   * @returns {*}
   */
  static bind() {

    // 遍历所有board，以class名注册类，并添加board方法
    setting.board.forEach(item => {

      debug('bind rider', item.class, item.action);

      this[item.class] = this[item.class] || {};
      this[item.class][item.action] = (handler, _params, _callback) => {

        let params = _params;
        let callback = _callback;

        // params参数未指定
        if (typeof params === 'function') {
          callback = params;
          params = undefined;
        }

        // 添加参数
        const copy = handler.copy(params || handler.params);
        let clazz = this.instance();
        let func = clazz[CONST.BOARD_TYPE[item.type]];

        /**
         * 调用以下Api时，将调用转到file模块中
         * /api/file/add
         * /api/file/update
         * /api/file/remove
         * 优点：
         *  1. 通过api直接调用时，直接交给file模块处理
         *  2. 在controller中做完处理后，可以和其他api一样调用rider.file.add
         * 隐患：
         *  1. 若不提供这段代码，则无法通过api直接（上传、修改、删除）文件，
         *     需在controller中调用core提供的file模块功能进行（上传、修改、删除）
         */
        if (item.class === 'file' && CONST.FILE.METHOD.includes(item.action)) {
          clazz = helper.resolve(item.class, path.join(__dirname, '..'));
          func = clazz[item.action];
        }

        copy.api = item;

        const promise = func.call(clazz, copy);
        if (callback) {
          promise.then(callback.bind(null, null), callback);
        }
        return promise;
      };
    });

    // 遍历所有表，注册aggregate方法
    setting.struct.forEach(item => {

      this[item.schema] = this[item.schema] || {};
      this[item.schema].aggregate = (handler, pipeline, callback) => {

        const copy = handler.copy();
        copy.api = { schema: item.schema };

        const promise = this.instance().aggregate(copy, pipeline);
        if (callback) {
          promise.then(callback.bind(null, null), callback);
        }
        return promise;
      };
    });
  }

  static startTransaction(handler) {
    handler.transaction.use = true;
  }

  static commitTransaction(handler, callback) {
    const promise = this.instance().commit(handler);
    if (callback) {
      promise.then(callback.bind(null, null), callback);
    }
    return promise;
  }

  static abortTransaction(handler, callback) {
    const promise = this.instance().abort(handler);
    if (callback) {
      promise.then(callback.bind(null, null), callback);
    }
    return promise;
  }

}

module.exports = Rider;
