/**
 * mongo.js
 * 数据处理的Controller
 */

'use strict';

const assert    = require('assert').strict;
const debug     = require('debug')('light:model:rider');
const CONST     = require('../../constant');
const validator = require('../../validator');
const log       = require('../../log');
const Ctrl      = require('../../db/mongo/controller');
const operator  = require('../../db/mongo/operator');
const util      = require('../../db/mongo/util');
const helper    = require('../../helper');
const setting   = require('../../setting');

class Mongo {

  /**
   * @desc 添加数据<br>
   *  - 校验数据<br>
   *  - 获取structure，判断是否锁定<br>
   *  - 插入到数据库(转换类型，设定缺省值，过滤非定义字段)<br>
   *  - 剔除选择对象外的项目<br>
   *  - 获取附加项目<br>
   *
   * @param {Object} handler 上下文对象
   * @param {Object} handler.params 参数信息
   * @param {Object|Object[]} handler.params.data 添加的数据对象
   */
  static async add(handler) {

    const { api, struct } = this.struct(handler);

    assert(api, 'api cannot be empty');
    debug('add', api.schema);

    // 如果是继承表，那么设定type为schema名称
    if (struct.parent) {
      debug('has parent', struct.parent);
      Object.assign(handler.params.data, { type: api.schema });
    }

    // 处理数据中的预约语
    this.reserveAll(handler, 'add');

    // 插入数据
    const data = await new Ctrl(handler, api.schema).add();

    // 添加附加项
    await this.options(handler, data, api);

    // 格式化
    if (Array.isArray(data)) {
      for (const item of data) {
        await this.format(item, api, handler.timezone);
      }
    } else {
      await this.format(data, api, handler.timezone);
    }

    return data;
  }

  /**
   * 获取数据定义
   * @param {*} handler
   */
  static struct(handler) {

    const structs = handler.code === CONST.SYSTEM_DB_PREFIX ? setting.system.struct : setting.struct;

    return {
      api   : handler.api,
      struct: structs.find(item => item.schema === handler.api.schema)
    };
  }


  /**
   * @desc 获取数据
   *  <br>
   *  - 生成filter<br>
   *  - 生成排序<br>
   *  - 检索数据<br>
   *  - 获取附加项目<br>
   *
   * @param {Object} handler 上下文对象
   * @param {Object} handler.params 参数信息
   * @param {Object} handler.params.condition 获取数据的检索条件
   */
  static async list(handler) {

    debug('list', handler.params);

    const { api, struct } = this.struct(handler);
    assert(api, 'api cannot be empty');

    // 生成filter, sort
    handler.params.condition = this.filter(handler, api, struct);
    handler.params.sort = this.order(handler, api);
    handler.params.select = this.select(handler, api);

    debug('parsed', handler.params);

    // 检索数据
    const data = await new Ctrl(handler, api.schema).list();

    // 添加附加项
    await this.options(handler, data, api);

    // 格式化
    for (const item of data.items) {
      await this.format(item, api, handler.timezone);
    }

    return data;
  }

  /**
   * @desc 获取单条数据
   *  <br>
   *  - 参数里指定id，则用id与数据库的_id比较来获取数据。指定的condition会被忽略<br>
   *  - 如果没有id，会用condition检索数据，并返回第一条匹配的数据<br>
   *  - 生成filter<br>
   *  - 检索数据<br>
   *  - 获取附加项目<br>
   *
   * @param {Object} handler 上下文对象
   * @param {Object} handler.params 参数信息
   * @param {String} [handler.params.id] 数据唯一识别号
   * @param {Object} [handler.params.condition] 获取数据的检索条件
   */
  static async get(handler) {

    const { api, struct } = this.struct(handler);
    assert(api, 'api cannot be empty');

    // 生成filter
    handler.params.condition = this.filter(handler, api, struct);
    handler.params.select = this.select(handler, api);

    const data = await new Ctrl(handler, api.schema).get();

    // 添加附加项
    await this.options(handler, data, api);

    // 格式化
    await this.format(data, api, handler.timezone);

    return data;
  }


  /**
   * @desc 获取数据件数
   * @param {Object} handler 上下文对象
   * @param {Object} handler.params 参数信息
   * @param {Object} handler.params.condition 获取数据的检索条件
   */
  static async count(handler) {

    const { api, struct } = this.struct(handler);
    assert(api, 'api cannot be empty');

    // 生成filter
    handler.params.condition = this.filter(handler, api, struct);

    // 获取件数
    return new Ctrl(handler, api.schema).count();
  }

  /**
   * @desc 删除数据
   *  <br>
   *  - 用ID删除时，没有找到对象数据，则返回错误，如果删除成功返回删除的数据<br>
   *  - 用条件删除时，返回的是删除件数<br>
   *
   * @param {Object} handler 上下文对象
   * @param {Object} handler.params 参数信息
   * @param {String} [handler.params.id] 数据唯一识别号
   * @param {Object} [handler.params.condition] 删除数据的检索条件
   */
  static async remove(handler) {

    const { api, struct } = this.struct(handler);
    assert(api, 'api cannot be empty');

    // 生成filter
    handler.params.condition = this.filter(handler, api, struct);

    // 删除数据
    return new Ctrl(handler, api.schema).remove();
  }

  /**
   * @desc 更新数据
   *  <br>
   *  - 用ID更新时，没有找到对象数据，则返回错误，如果更新成功返回更新后的数据<br>
   *  - 用条件更新时，返回的是更新的件数<br>
   *
   * @param {Object} handler 上下文对象
   * @param {Object} handler.params 参数信息
   * @param {String} [handler.params.id] 数据唯一识别号
   * @param {Object} [handler.params.condition] 更新数据的检索条件
   * @param {Object} handler.params.data 更新的数据对象
   */
  static async update(handler) {

    const { api, struct } = this.struct(handler);
    assert(api, 'api cannot be empty');

    // 生成filter
    handler.params.condition = this.filter(handler, api, struct);

    // 处理数据
    this.reserveAll(handler, 'update');

    // 更新数据
    const data = await new Ctrl(handler, api.schema).update();

    // 更新多件时，返回的是件数
    if (typeof data === 'number') {
      return data;
    }

    // 添加附加项
    await this.options(handler, data, api);

    // 单项更新时，返回更新内容
    await this.format(data, api, handler.timezone);

    return data;
  }


  /**
   * @desc 替换数据中的预约语
   * @param handler
   * @param kind 类型 add | update
   */
  static reserveAll(handler, kind) {

    const { struct } = this.struct(handler);

    // 转换给定数据的关键词
    Object.keys(handler.params.data).forEach(key => {

      const val = handler.params.data[key];
      if (typeof val !== 'string') {
        return;
      }

      const parsed = this.reserve(handler, val);
      if (val !== parsed) {

        debug('set reserve value', key, val, parsed);
        handler.params.data[key] = parsed;
      }
    });

    // 只有添加数据时，才解析struct中定义的default值
    if (kind === 'update') {
      return;
    }

    // 转换数据定中有默认值，并且默认值指定了关键词，并且值没有设定的项目
    Object.keys(struct.items).forEach(key => {
      const defaults = struct.items[key].default;
      const hasResolve = ['$uid', '$sysdate', '$systime', '$corp'].includes(defaults);
      if (!hasResolve) {
        return;
      }

      const val = handler.params.data[key];
      if (typeof val === 'undefined' || val === '') {
        const parsed = this.reserve(handler, defaults);
        handler.params.data[key] = parsed;
        debug('set reserve value', key, val, parsed);
      }
    });
  }


  /**
   * @desc 排序<br>
   * @param handler 上下文对象
   * @param {Object} handler.params 参数信息
   * @param {String} handler.params.sorts | hanlder.params.order 动态排序时 包含排序字段
   * @param board 排序定义[[col1: 1], [col2: -1], [col3: 1]]
   * @returns {Array} 排序后数据
   */
  static order(handler, board) {

    // 如果，指定了排序方法，那么直接返回
    const sort = handler.params.sort || handler.params.order;
    if (sort) {
      debug('query sort', sort);
      return sort;
    }

    // 尝试使用board定义中的排序
    const sorts = (board.sorts || []).sort((v1, v2) => (v1.index > v2.index ? 1 : -1));
    debug('use board sort %j', sorts);
    return sorts.map(sort => [sort.key, sort.order.toLowerCase() === 'asc' ? 1 : -1]);
  }


  /**
   * @desc 生成条件字符串<br>
   *  free: 直接作为检索条件，在model层会对数据进行类型转换<br>
   *  遍历board的filter条件定义，生成检索条件<br>
   *   - 前端没有传递filter里定义的参数值，则尝试使用缺省值<br>
   *   - 缺省值是预约语的时候，用实际值代替<br>
   *   - 同一个条件组内的条件用and符合并<br>
   *   - 不同条件组间，用or符合并<br>
   *   - 没有明确指定valid，则添加valid=1条件
   * @ignore
   * @param {Object} handler
   * @param {Object} handler.params.free 自定义检索条件
   * @param {Object} handler.params.condition 检索条件
   * @param {Object} handler.params.filter 检索条件
   * @returns {Object} 条件对象
   */
  static filter(handler, board, struct) {

    // 如果定义了自由条件，则对检索条件框架不做任何处理
    if (handler.params.free) {
      log.warn('Free condition are not supported. Please use aggrigate.', handler.uid);
      return handler.params.free;
    }

    const filters = board.filters || [];

    // 没有指定任何条件并且有父类型，则在检索条件里默认加上type条件
    if (filters.length <= 0) {
      debug('no board filters');
      return this.valid(struct.parent ? { type: struct.schema } : {});
    }

    const condition = handler.params.condition || handler.params.filter || {};
    debug('condition', condition);

    let result = Object.values(filters.reduce((or, filter) => {

      or[filter.group] = or[filter.group] || {};
      const and = or[filter.group];

      let value = undefined;
      if (filter.parameter) {
        value = helper.get(condition, filter.parameter);
      }

      // 如果没能取到参数值，则使用缺省值
      if (typeof value === 'undefined' && filter.default) {
        debug('value is empty, try use default');
        value = this.reserve(handler, filter.default);
      }

      // 值没有定义，则不作为条件
      if (typeof value === 'undefined') {
        debug('value is empty');
        return or;
      }

      // 存在相同名的条件，则以and形式扩展
      const compare = operator.compare(filter.operator, filter.key, value);
      if (and[filter.key]) {
        Object.assign(and[filter.key], compare[filter.key]);
      } else {
        Object.assign(and, compare);
      }

      // 如果有父类型，那么检索条件里默认加上type条件
      if (struct.parent) {
        and.type = struct.schema;
      }

      return or;
    }, {}));

    if (result.length <= 0) {
      return this.valid({});
    }

    // 去除空的or条件 [{ name: 'light' }, {}] -> [{ name: 'light' }]
    result = result.filter(or => Object.keys(or).length > 0);
    if (result.length <= 0) {
      return this.valid({});
    }

    // 如果只有一个or条件的group，则去除or比较符
    if (result.length === 1) {
      return this.valid(result[0]);
    }

    return operator.compare('$or', null, this.valid(result));
  }


  /**
   * @desc 添加获取有效数据的条件，仅在没有指定valid条件时起作用<br>
   *  如果条件是[或]关系，则在每个[或]条件里添加valid=1的条件<br>
   *  如果是单个条件，则单纯添加valid=1的条件
   * @ignore
   * @param condition
   * @returns {*}
   */
  static valid(condition) {
    (Array.isArray(condition) ? condition : [condition]).forEach(item => {
      if (typeof item.valid === 'undefined') {
        item.valid = 1;
      }
    });
    return condition;
  }


  /**
   * @desc 去掉被选择项目以外的项目
   * 1. 只拷贝被选择的项目
   * 2. 如果定义了format或alias，则进行格式化和重命名
   * 3. 支持内嵌文档的拷贝，只支持两层结构
   *
   * @ignore
   * @param data
   * @param board
   * @param timezone
   * @returns {Object}
   */
  static async format(data = {}, board, timezone) {

    // 特殊处理: 当schema是structure时，返回所有字段。
    // 因为，structure的board无法完整描述schema本身的定义
    if (board.schema === CONST.SYSTEM_DB_STRUCTURE || Object.keys(data).length <= 0) {
      return data;
    }

    for (const column of board.selects) {

      // not selected
      if (!column.select) {
        continue;
      }

      // format
      if (column.format) {

        const [key, child] = column.key.split('.');
        const option = {
          key   : child,
          rule  : column.format,
          option: column.option,
          timezone
        };

        // 获取第一层文档值
        let value = helper.get(data, key);

        // 1.一些特殊的格式化，需要将数据原样传给格式化方法
        // 2. 没有子文档，将数据原样传给格式化方法
        // 3. 第一层文档值为非数组，将数据原样传给格式化方法
        if (['join'].includes(column.format) || !child || !Array.isArray(value)) {
          option.key = column.key;
          value = [data];
        }

        for (const item of value) {
          debug('format key: %s, format: %s', option.key, option.rule);
          const val = await validator.format({ params: item }, option);
          debug('formated %s', val);
          helper.set(item, option.key, val);
        }
      }

      // alias
      if (column.alias) {

        const [key, child] = column.key.split('.');
        const value = helper.get(data, key);

        // 第一层数据的alias
        if (!child || !Array.isArray(value)) {
          debug('alias %s to %s', key, column.alias);
          helper.set(data, column.alias, helper.get(data, key));
          continue;
        }

        // 如果是数组，那么设置每个元素
        for (const item of value) {
          debug('alias %s to %s', child, column.alias);
          debug(item, column.alias, helper.get(item, child));
          helper.set(item, column.alias, helper.get(item, child));
        }

      }

    }

    return data;
  }


  /**
   * 根据board的select定义，生成select条件
   * @param handler
   * @param board
   * @returns {*}
   */
  static select(handler, board) {

    // 获取board中的选择项
    const selects = board.selects.filter(item => item.select).map(item => item.key);

    // 没有指定select，使用board的定义
    if (!handler.params.select) {
      debug('use board selects %j', selects);
      return selects.join(',');
    }

    // 如果query中指定了select项目，优先使用query中的这些项目，但这些项目必须在board中选中
    const query = util.projection(handler.params.select);

    debug('query select: %s', query);
    return Object.keys(query)
      .filter(item => query[item] && selects.includes(item))
      .join(',');
  }


  /**
   * @desc 遍历所有的选择项目，获取options项的值
   * @ignore
   * @param {Context} handler
   * @param {Object} data
   * @param board
   */
  static async options(handler, data = {}, board = {}) {

    const selects = (board.selects || []).filter(select => select.schema);
    if (Object.keys(data).length <= 0 || selects.length <= 0) {
      debug('result data or board selects is empty');
      return;
    }

    data.options = {};
    for (const select of selects) {
      debug('join', select.schema);
      await this.join(handler, data, select, board.type);
    }
  }


  /**
   * @desc 关联表
   *  在数据中找到指定的key值，检索关联表，获取数据
   *  并将结果保存到data的options里
   *
   * @param handler
   * @param data 数据
   * @param item 关联信息
   *  {
   *    key: 'uid'
   *    schema: 'user'
   *    condition: { _id: '$uid', sex: '1' }
   *    fields: ['id', 'name']
   *  }
   */
  static async join(handler, data, item) {

    data.options[item.schema] = data.options[item.schema] || {};

    debug('schema %s', item.schema);
    debug('condition %j', item.conditions);

    // 已经获取的内容（不再重复获取，检索时剔除这些内容）
    const selected = Object.keys(data.options[item.schema]);
    const filter = { condition: {}, select: item.fields };

    let storeKey;

    // 指定检索条件
    Object.keys(item.conditions).forEach(key => {

      const val = item.conditions[key];

      // $符号开头的是变量，需要在data里取值作为真正的检索条件值
      if (val[0] !== '$') {
        debug('use fixed val', val);
        return filter.condition[key] = val;
      }

      // 去掉空的, 去掉已经获取的, 去掉重复的
      debug('use parameter', val);
      const values = Array.from(new Set(helper.get(data.items || [data], val.substr(1))
        .flat()
        .filter(item => {
          if (typeof item === 'undefined' || item === '' || item === null) {
            return false;
          }
          return !selected.includes(item);
        })));

      if (values.length > 0) {
        filter.condition[key] = { $in: values };
      }

      // 确定保存Option数据的key，找出与当前数据字段相同的条件的名称
      if (val.substr(1) === item.key) {
        storeKey = key;
      }
    });

    // 没有条件，不检索options项
    if (filter.condition.length <= 0) {
      return data;
    }

    debug('condition %j', filter.condition);
    debug('field %j', filter.select);
    debug('reference key', storeKey);

    // 查询（不获取件数）
    const { items } = await new Ctrl(handler.copy(filter), item.schema).list(false);
    items.forEach(row => {
      const key = typeof row[storeKey] === 'undefined' ? row._id.toHexString() : row[storeKey];
      debug('store key', key);
      return data.options[item.schema][key] = row;
    });
  }


  /**
   * @desc 获取预约语对应的值
   *  $uid 当前用户ID
   *  $sysdate 日期型，系统当前日期 没有时间信息
   *  $systime 日期型，系统当前日期 包含时间信息
   * @ignore
   * @param handler
   * @param keyword
   * @returns {*}
   */
  static reserve(handler, keyword) {

    if (keyword === '$uid') {
      return handler.uid;
    }

    if (keyword === '$sysdate') {
      return new Date();
    }

    if (keyword === '$systime') {
      return new Date();
    }

    if (keyword === '$corp') {
      return handler.corp;
    }

    return keyword;
  }


  /**
   * 提交事物
   * @param {Context} handler
   */
  static commit(handler) {
    return new Ctrl(handler).commit();
  }


  /**
   * 回滚事物
   * @param {Context} handler
   */
  static abort(handler) {
    return new Ctrl(handler).abort();
  }


  /**
   * 执行Aggregate
   * @param {Context} handler
   * @param {Object} pipeline
   */
  static aggregate(handler, pipeline) {
    return new Ctrl(handler, handler.api.schema).aggregate(pipeline);
  }

}

module.exports = Mongo;
