/* eslint-disable camelcase */
/**
 * file.js
 */

'use strict';

const assert       = require('assert').strict;
const debug        = require('debug')('light:model:file');
const { ObjectID } = require('mongodb');
const rider        = require('./datarider');
const light        = require('./light');
const log          = require('../log');
const CONST        = require('../constant');

class File {

  /**
   * 添加文件
   * 添加物理文件本身，且添加Meta信息
   * @param handler
   * @param callback
   */
  static async add(handler, callback) {
    const _callback = this.defaultCallback(callback);

    debug('copy handler');
    const _handler = handler.copy();
    try {

      // 开启事物
      debug('begin transaction');
      rider.startTransaction(_handler);

      // 如果code是light，那么使用system操作文件表
      const { data = {}, files = [] } = _handler.params;
      const now = new Date();

      // 没有文件本身（添加文件夹数据）
      if (files.length <= 0) {

        debug('save meta only');
        assert(data.filename, 'name cannot be empty');

        _handler.addParams({ data: { uploadDate: now, ...data } });
        const result = await this.datarider(_handler.code).add(_handler);

        // 提交事物
        debug('commit transaction');
        await rider.commitTransaction(_handler);
        return _callback(null, result);
      }

      // 删除data中文件表预留字段
      CONST.FILE.FIELD.forEach(field => delete data[field]);

      // 上传文件到gridfs
      debug('save file content');
      const { items } = await this.controller(_handler).fileToGrid();

      // 更新文件meta信息
      debug('save file metadata');
      const result = [];
      for (const item of items) {

        _handler.addParams({
          id  : item._id,
          data: {
            ...data,
            createAt: now,
            createBy: _handler.uid,
            valid   : CONST.VALID
          }
        });
        const file = await this.datarider(_handler.code).update(_handler);

        // 对应chunk增加基础字段 (TODO: 没有必要更新chunk)
        _handler.addParams({
          id  : undefined,
          data: {
            createAt: now,
            createBy: _handler.uid,
            valid   : CONST.VALID
          },
          condition: {
            files_id: file._id
          }
        });
        await this.controller(_handler, 'chunk').update();

        result.push(file);
      }
      debug('file count', result.length);

      // 提交事物
      debug('commit transaction');
      await rider.commitTransaction(_handler);
      return _callback(null, result);
    } catch (err) {
      debug('abort transaction');
      log.error(err);

      // 提交事物
      await rider.abortTransaction(_handler);
      return _callback(err);
    }
  }


  /**
   * 更新，GridFS文件不能更新文件，我们先添加一个文件，
   * 然后更新原文件与chunks表的关联的方式，来实现更新。这种方法会在留下files表和chunks表
   * 留下垃圾文件，这些文件会在后续的清理操作中清除。
   *
   * 注意该方法效率低，请尽量使用add方法。
   *
   * 处理详细
   *  1. 上传文件到gridfs
   *  2. 更新文件详细信息
   *  3. 删除原来的文件体（更新 files_id 为无效值）
   *  4. 更新新的 chunks，放入原来文件ID
   * @param handler
   * @param callback
   */
  static async update(handler, callback) {
    const _callback = this.defaultCallback(callback);
    const _handler = handler.copy();

    // 只能用ID进行文件更新
    assert(handler.params.id, 'can only be updated with ID');

    // 文件体不能是多个
    const { data = {}, files = [] } = _handler.params;
    assert(files.length <= 1, 'does not support updating multiple files');

    try {

      // 开启事物
      debug('begin transaction');
      rider.startTransaction(_handler);

      // 只更新附加信息
      if (files.length <= 0) {

        debug('no file content');
        const result = await this.datarider(_handler.code).update(_handler);

        debug('commit transaction');
        await rider.commitTransaction(_handler);
        return _callback(null, result);
      }

      // 删除data中文件表预留字段
      CONST.FILE.FIELD.forEach(field => delete data[field]);

      // 1. 上传文件到gridfs
      debug('save file content');
      const { items: [file] } = await this.controller(_handler, 'file').fileToGrid();
      const id = new ObjectID(_handler.params.id);
      const now = new Date();
      debug('new id', file._id, 'old id', id);

      // 2. 更新文件详细信息
      debug('update metadata');
      _handler.addParams({
        id,
        data: {
          ...data,
          length     : file.length,
          chunkSize  : file.chunkSize,
          uploadDate : file.uploadDate,
          filename   : file.filename,
          md5        : file.md5,
          contentType: file.contentType
        }
      });
      const result = await this.datarider(_handler.code).update(_handler);

      // 3. 删除新上传的文件数据
      _handler.addParams({
        id  : file._id,
        data: {
          createAt: now,
          createBy: _handler.uid,
          valid   : CONST.INVALID
        }
      });
      await this.datarider(_handler.code).update(_handler);

      // 4. 删除原来的文件体（更新 files_id 为无效值, 并修改valid为0）
      debug('remove old file chunks');

      _handler.addParams({
        id       : undefined,
        condition: { files_id: id },
        data     : {
          files_id: `${id.toHexString()}-${file._id.toHexString()}`,
          valid   : CONST.INVALID
        }
      });
      await this.controller(_handler, 'chunk').update();

      // 5. 更新新的 chunks，放入原来文件ID
      debug('update new file chunks');
      _handler.addParams({
        condition: { files_id: file._id },
        data     : {
          files_id: id,
          createAt: now,
          createBy: _handler.uid,
          valid   : CONST.VALID
        }
      });
      await this.controller(_handler, 'chunk').update();

      // 6. 提交事物
      debug('commit transaction');
      await rider.commitTransaction(_handler);

      return _callback(null, result);
    } catch (err) {
      log.error(err);
      debug('abort transaction');
      await rider.abortTransaction(_handler);

      return _callback(err);
    }
  }

  /**
   * 删除文件
   * @param {*} handler
   * @param {*} callback
   */
  static async remove(handler, callback) {
    const _handler = handler.copy();
    const _callback = this.defaultCallback(callback);

    try {

      // 开启事物
      debug('begin transaction');
      rider.startTransaction(_handler);

      // 只能用ID进行文件删除
      assert(_handler.params.id, 'can only be removed with ID');

      // 删除文件
      const result = await this.datarider(_handler.code).remove(_handler);

      _handler.addParams({
        id       : undefined,
        condition: { files_id: new ObjectID(handler.params.id) }
      });
      // 删除chunk
      await this.controller(_handler, 'chunk').remove();

      // 提交事物
      debug('commit transaction');
      await rider.commitTransaction(_handler);

      return _callback(null, result);
    } catch (error) {
      debug('abort transaction');
      await rider.abortTransaction(_handler);

      return _callback(error);
    }
  }

  /**
   * 获取默认callback
   * @param {*} callback
   */
  static defaultCallback(callback) {
    return callback || ((err, res) => {
      if (err) throw err;
      return res;
    });
  }

  /**
   * 支持多种数据库
   * @param handler
   */
  static controller(handler, table = 'file') {

    if (process.env.LIGHTMONGO_HOST || process.env.LIGHTMONGO_HOSTS) {
      const Controller = require('../db/mongo/controller');
      return new Controller(handler, table);
    }

    if (process.env.LIGHTSQLSERVER_HOST) {
      const Controller = require('../db/sqlserver/controller');
      return new Controller(handler, table);
    }
    if (process.env.LIGHTMYSQL_HOST) {
      const Controller = require('../db/mysql/controller');
      return new Controller(handler, table);
    }

  }

  /**
   * 获取system、rider实例
   * @param {*} code
   */
  static datarider(code) {

    // 如果code是light，那么使用system操作文件表
    if (code === CONST.SYSTEM_DB_PREFIX) {
      return light.instance();
    }

    return rider.instance();
  }

}

module.exports = File;
