/**
 * @file 系统设定相关
 * @author r2space@gmail.com
 * @module light.model.system
 * @version 1.0.0
 */

'use strict';

const assert  = require('assert').strict;
const debuger = require('debug');
const path    = require('path');
const debug   = debuger('light:model:light');
const CONST   = require('../constant');
const setting = require('../setting');
const helper  = require('../helper');
const config  = require('../../package.json');


class Light {

  /**
   * 根据环境变量设置，获取特定数据库的Rider实例
   */
  static instance() {
    if (process.env.LIGHTSQLSERVER_HOST) {
      return require('./datarider/sqlserver');
    }

    if (process.env.LIGHTMYSQL_HOST) {
      return require('./datarider/mysql');
    }

    if (process.env.LIGHTMONGO_HOST || process.env.LIGHTMONGO_HOSTS) {
      return require('./datarider/mongo');
    }

    assert.fail('invalid database type');
  }

  /**
   * 根据board定义，创建访问类
   * @returns {*}
   */
  static bind() {

    // 遍历所有board，以class名注册类，并添加board方法
    setting.system.board.forEach(item => {

      debug('bind light', item.class, item.action);

      this[item.class] = this[item.class] || {};
      this[item.class][item.action] = (handler, _params, _callback) => {

        let params = _params;
        let callback = _callback;

        // params参数未指定
        if (typeof params === 'function') {
          callback = params;
          params = undefined;
        }

        // 添加参数
        const copy = handler.copy(params || handler.params);
        let clazz = this.instance();
        let func = clazz[CONST.BOARD_TYPE[item.type]];

        /**
         * 调用以下Api时，将调用转到file模块中
         * /api/file/add
         * /api/file/update
         * /api/file/remove
         * 优点：
         *  1. 通过api直接调用时，直接交给file模块处理
         *  2. 在controller中做完处理后，可以和其他api一样调用rider.file.add
         * 隐患：
         *  1. 若不提供次代码，则无法通过api直接（上传、修改、删除）文件，
         *     需在controller中调用core提供的file模块功能进行（上传、修改、删除）
         */
        if (item.class === 'file' && CONST.FILE.METHOD.includes(item.action)) {
          clazz = helper.resolve(item.class, path.join(__dirname));
          func = clazz[item.action];
        }

        copy.code = CONST.SYSTEM_DB_PREFIX;
        copy.api = item;

        const promise = func.call(clazz, copy);
        if (callback) {
          promise.then(callback.bind(null, null), callback);
        }
        return promise;
      };
    });

  }

  /**
   * 应用版本确认
   * API - [GET] /api/system/version
   */
  static version() {
    return {
      core: config.version
    };
  }

  /**
   * 设置DEBUG级别
   * API - [GET] /api/system/debug
   * @param {*} handler
   */
  static debug = function (handler) {
    debuger.enable(handler.params.level);
    debug('debug level:', handler.params.level);
  }

}

module.exports = Light;
