/**
 * MQ
 */

'use strict';

const assert     = require('assert').strict
  , EventEmitter = require('events')
  , debug        = require('debug')('light:mq')
  , log          = require('./log')
  , Model        = require('./db/mongo/model')
  , CONST        = require('./constant');

class MQ extends EventEmitter {

  constructor(domain = process.env.APPNAME) {
    super();

    assert(domain, 'Domain cannot be empty');

    this.id = null;
    this.option = {};
    this.model = new Model(domain, CONST.SYSTEM_DB_PREFIX, CONST.SYSTEM_DB_MQ);
    debug('use domain', domain);
  }

  /**
   * 消化任务（消费者）
   * @param {*} option
   *  domain 数据库
   *  queue
   *  interval 轮询间隔
   */
  async subscribe(option = {}) {

    assert(option.queue, 'Queue name cannot be empty');

    this.option = option;
    const param = {
      interval: option.interval || CONST.MQ.INTERVAL,
      kind    : 'subscribe',
      filter  : { queue: option.queue, status: CONST.MQ.STATUS_WAITING, valid: CONST.VALID },
      data    : { status: CONST.MQ.STATUS_PROCESSING, updateAt: new Date() },
      fields  : 'queue param status',
      sort    : { createAt: -1 }
    };

    debug('subscribe queue', option.queue);

    try {
      const { error, data } = await this.watch(param);
      if (error) {
        return this.emit('error', error);
      }

      debug('process start', data._id);
      this.id = data._id;
      this.emit('process', data);
    } catch (err) {
      log.error(err);
      this.emit('error', err);
    }
  }

  /**
   * 标记任务成功
   * @param {*} result
   */
  async ack(result = {}) {
    if (!this.id) {
      return;
    }

    const data = {
      updateAt: new Date(),
      status  : CONST.MQ.STATUS_FINISHED,
      result
    };
    await this.model.update(this.id, data);
    debug('process successed (ack)', this.id);

    this.id = null;
    await this.subscribe(this.option);
  }

  /**
   * 标记任务失败
   * @param {*} result
   */
  async nack(result = {}) {
    if (!this.id) {
      return;
    }

    const data = {
      updateAt: new Date(),
      status  : CONST.MQ.STATUS_FAILED,
      result
    };
    await this.model.update(this.id, data);
    debug('process failed (nack)', this.id);

    this.id = null;
    await this.subscribe(this.option);
  }

  /**
   * 发布任务（）
   * @param {*} option
   */
  async publish(option) {

    assert(option.queue, 'Queue name cannot be empty');

    const now = new Date()
      , data = {
        queue   : option.queue,
        param   : option.param || {},
        status  : CONST.MQ.STATUS_WAITING,
        timeout : option.timeout || CONST.MQ.TIMEOUT,
        result  : {},
        valid   : CONST.VALID,
        createAt: now,
        createBy: option.uid || CONST.DEFAULT_USER,
        updateAt: now,
        updateBy: option.uid || CONST.DEFAULT_USER
      };

    // MQ表插入任务数据
    debug('publish to', option.queue);
    const { _id } = await this.model.add(data);

    // 如果设置了任务等待标识，轮询（默认 3 秒间隔）查询任务结束，默认超时时间为 5 分钟
    if (option.watch) {
      const param = {
        kind    : 'publish',
        interval: option.interval || CONST.MQ.INTERVAL,
        timeout : option.timeout || CONST.MQ.TIMEOUT,
        filter  : { _id, status: { $in: [CONST.MQ.STATUS_FINISHED, CONST.MQ.STATUS_FAILED] }, valid: CONST.VALID },
        fields  : 'queue status param result'
      };

      try {
        const { error, data } = await this.watch(param);
        if (error) {
          log.error(error);
          return this.emit('error', error);
        }

        // 处理失败
        if (data.status === CONST.MQ.STATUS_FAILED) {
          debug('process failed', option.queue, this.id);
          this.emit('error', data);
        }

        // 处理成功
        if (data.status === CONST.MQ.STATUS_FINISHED) {
          debug('process finished', option.queue, this.id);
          this.emit('finished', data);
        }

      } catch (err) {
        log.error(err);
        this.emit('error', err);
      }
    }

  }

  /**
   * 轮询数据库mq表，值发生变化时返回
   * @param {*} option domain, interval, timeout, filter, data, time
   */
  async watch(option) {

    assert(option.interval >= 0.5, 'To prevent high database load, interval is not allowed to be less than 0.5');
    option.time = option.time || 0;

    if (option.kind === 'publish' && option.time >= option.timeout) {
      debug('timeout');
      return { error: new Error('timeout') };
    }

    debug('keep watching', option.time, '(s)');

    try {
      const data = await this.model.get(option.filter, option.fields, option.sort, option.data);
      if (data) {
        return { data };
      }
    } catch (error) {
      log.error(error);
      return { error };
    }

    debug('interval', option.interval, '(s)');

    // Sleep
    await this.sleep(option.interval * 1000);
    option.time += option.interval;

    return await this.watch(option);

  }

  sleep(time) {
    return new Promise(resolve => {
      setTimeout(() => resolve(), time);
    });
  }

}

module.exports = MQ;
