/**
 * @file MQ
 * @module mq.js
 * @author r2space@gmail.com
 * @version 2019.08.29
 *
 * 文档
 * http://www.squaremobius.net/amqp.node/channel_api.html
 */

'use strict';

const amqp = require('amqplib');
const CONST = require('../constant');
const log = require('../log');

class RabbitMQ {

  /**
   * @desc 获取与MQ服务器的连接
   * @param {Object} option 连接参数
   */
  async connect(option) {

    const params = {
      protocol : 'amqp',
      hostname : option.host || process.env.LIGHTMQ_HOST || CONST.MQ.HOST,
      port     : option.port || process.env.LIGHTMQ_PORT || CONST.MQ.PORT,
      username : option.login || process.env.LIGHTMQ_USER || CONST.MQ.USER,
      password : option.password || process.env.LIGHTMQ_PASS || CONST.MQ.PASS,
      vhost    : option.vhost || '/',
      heartbeat: option.heartbeat || 3,
      locale   : 'en_US',
      frameMax : 0
    };

    log.debug(`connect to mq : ${params.hostname}:${params.port}`);
    log.debug(`heartbeat : ${params.heartbeat}`);

    return await amqp.connect(params);
  }


  /**
   * 连接Exchange，没有则创建
   *  option.exchange exchange名称
   */
  async exchange(option) {

    const connection = await this.connect(option);
    log.debug('mq connection succeeded', option.exchange);

    // 获取普通的channel
    const channel = option.confirm
      ? await connection.createConfirmChannel()
      : await connection.createChannel();
    log.debug('channel was created');

    // 创建或连接 exchange
    await channel.assertExchange(option.exchange, 'direct', {
      durable   : true, // 持久形队列，服务重新启动的时候，不会清除
      autoDelete: false // 当queue中得消息消费完以后，不自动删除
    });

    log.debug('successfully connected to exchange');
    return channel;
  }


  /**
   * 连接Queue，没有则创建
   */
  async queue(option) {

    const channel = await this.exchange(option);
    log.debug(`connect to queue : ${option.exchange}, ${option.queue}`);

    // 创建或连接 queue
    await channel.assertQueue(option.queue, {
      durable   : true,   // 持久形队列，服务重新启动的时候，不会清除
      autoDelete: false   // 当queue中得消息消费完以后，不自动删除
    });
    log.debug('successfully bind to exchange');

    // 绑定queue 和 exchange
    channel.bindQueue(option.queue, option.exchange, option.queue);
    return channel;
  }


  /**
   * 订阅
   *  option           MQ连接参数
   *  option.exchange  exchange名称 (必须)
   *  option.queue     queue名称 (必须)
   *  option.prefetch  一次处理的消息数量 默认为 1
   *  option.noAck     自动响应 默认为 false
   *  callback         接收消息回调
   *    result.message 发送的消息内容
   *    result.reply   消息体
   *    result.channel 连接
   */
  subscribe(option, callback) {

    log.debug(`start subscribe : ${option.exchange}`);
    this.queue({ heartbeat: 30, ...option })
      .then(async channel => {

        // 定义消息接收方法
        function accept(reply) {

          // 尝试调用JSON.parse
          let message = reply.content.toString();
          try {
            message = JSON.parse(message);
          } catch (err) {
            log.error(err);
          }
          callback(null, { reply, channel, message });
        }

        // 开始订阅（prefetch每次从queue获取的数量）
        await channel.prefetch(option.prefetch || 1);

        // noAck标记设为false时，不自动响应消息已经被处理，需要ack方法通知mq
        await channel.consume(option.queue, accept, { noAck: !!option.noAck });
        log.debug('consume ready');
      })
      .catch(err => callback(err));

  }


  /**
   * 发送消息
   *  option          MQ连接参数
   *  option.exchange exchange名称 (必须)
   *  option.queue    queue名称 (必须)
   *  data            发送的消息对象 string | json
   */
  publish(option, data) {

    log.debug(`start publish: ${option.exchange}, ${option.queue}`);

    const message = typeof data === 'string' ? data : JSON.stringify(data);
    log.debug(`content: ${message}`);

    return new Promise((resolve, reject) => {

      // confirm: true 生成的channel为ConfirmChannel（完成后会调用publish的回调方法）
      this.exchange({ heartbeat: 0, confirm: true, ...option })
        .then(channel => {

          const params = {
            mandatory   : true, // 消息不能被路由的时候，服务器返回错误消息，而不是仅仅删除该消息
            deliveryMode: 2     // 持久  1: 为非持久 2: 持久
          };

          channel.publish(option.exchange, option.queue, Buffer.from(message), params, async err => {
            await channel.close();
            await channel.connection.close();
            log.debug('close mq connection');

            if (err) {
              return reject(err);
            }

            log.debug('publish finished');
            resolve();
          });

        })
        .catch(err => reject(err));

    });

  }

}

module.exports = new RabbitMQ();
