/**
 * casbin mongodb adapter
 */

const { Helper } = require('casbin');
const debug = require('debug')('light:permission:adapter');
const Model = require('../db/mongo/model');
const CONST = require('../constant');


class Adapter {

  static init(option = {}) {
    return new Adapter(option);
  }

  constructor(option) {

    const table = option.table || 'permission';
    debug('init adapter', table);

    this.filtered = option.filtered;
    this.model = new Model(process.env.APPNAME, CONST.SYSTEM_DB_PREFIX, table);
  }

  /**
   * loadPolicy loads all policy rules from the storage.
   */
  async loadPolicy(model) {

    debug('load policy');
    await this.loadFilteredPolicy(model);
  }

  /**
   * loadPolicy loads filtered policy rules from the storage.
   */
  async loadFilteredPolicy(model, filter) {

    debug('load filtered policy', filter);

    // TODO: 确认filter的用法 - 现在即使调用了getFilteredPolicy这里的filter也时undefined
    const lines = await this.model.list(this.filtered ? filter : { valid: 1 });
    debug('policy count', lines.length);

    for (const line of lines) {
      this.parse(line, model);
    }

  }

  /**
   * savePolicy saves all policy rules to the storage.
   */
  async savePolicy({ model }) {

    debug('save policy');
    debug('drop permission collection');
    await this.model.dropCollection();

    const lines = [];

    for (const [ptype, ast] of model.get('p')) {
      for (const rule of ast.policy) {
        const line = this.policyLine(ptype, rule);
        lines.push(line);
      }
    }

    for (const [ptype, ast] of model.get('g')) {
      for (const rule of ast.policy) {
        lines.push(this.policyLine(ptype, rule));
      }
    }

    debug('total line count', lines.length);
    if (Array.isArray(lines) && lines.length > 0) {
      this.model.add(lines);
    }

    return true;
  }

  /**
   * addPolicy adds a policy rule to the storage.
   */
  async addPolicy(_sec, ptype, rule) {

    debug('add policy', ptype, rule);

    const data = this.policyLine(ptype, rule);
    const now = new Date();

    await this.model.add({
      valid   : 1,
      createAt: now,
      createBy: CONST.DEFAULT_USER,
      updateAt: now,
      updateBy: CONST.DEFAULT_USER,
      ...data
    });
  }

  async addPolicies(_sec, ptype, rules) {
    const now = new Date();
    const data = rules.map(rule => (
      {
        valid   : 1,
        createAt: now,
        createBy: CONST.DEFAULT_USER,
        updateAt: now,
        updateBy: CONST.DEFAULT_USER,
        ...this.policyLine(ptype, rule)
      }
    ));
    await this.model.add(data);
  }

  /**
   * removePolicy removes a policy rule from the storage.
   */
  async removePolicy(_sec, ptype, rule) {

    debug('remove policy', ptype, rule);
    await this.model.remove(this.policyLine(ptype, rule));
  }

  /**
   * removeFilteredPolicy removes policy rules that match the filter from the storage.
   */
  async removeFilteredPolicy(_sec, ptype, fieldIndex, ...fieldValues) {

    debug('remove filtered policy', ptype, fieldIndex, fieldValues);

    const line = { ptype };

    if (ptype === 'p') {
      if (fieldIndex <= 0 && 0 < (fieldIndex + fieldValues.length)) {
        line.subject = fieldValues[0 - fieldIndex];
      }
      if (fieldIndex <= 1 && 1 < (fieldIndex + fieldValues.length)) {
        line.code = fieldValues[1 - fieldIndex];
      }
      if (fieldIndex <= 2 && 2 < (fieldIndex + fieldValues.length)) {
        line.object = fieldValues[2 - fieldIndex];
      }
      if (fieldIndex <= 3 && 3 < (fieldIndex + fieldValues.length)) {
        line.type = fieldValues[3 - fieldIndex];
      }
      if (fieldIndex <= 3 && 3 < (fieldIndex + fieldValues.length)) {
        line.action = fieldValues[3 - fieldIndex];
      }
    }

    if (ptype === 'g') {
      if (fieldIndex <= 0 && 0 < (fieldIndex + fieldValues.length)) {
        line.subject = fieldValues[0 - fieldIndex];
      }
      if (fieldIndex <= 1 && 1 < (fieldIndex + fieldValues.length)) {
        line.parent = fieldValues[1 - fieldIndex];
      }
      if (fieldIndex <= 2 && 2 < (fieldIndex + fieldValues.length)) {
        line.code = fieldValues[2 - fieldIndex];
      }
    }

    debug('remove', line);
    await this.model.remove(line);
  }

  parse(line, model) {

    if (line.ptype === 'p') {
      Helper.loadPolicyLine(
        `${line.ptype}, ${[
          line.subject,
          line.code,
          line.object,
          line.type,
          line.action
        ].filter(val => val).join(', ')}`,
        model
      );
    }

    if (line.ptype === 'g') {
      Helper.loadPolicyLine(
        `${line.ptype}, ${[
          line.subject,
          line.parent,
          line.code
        ].filter(val => val).join(', ')}`,
        model
      );
    }
  }

  policyLine(ptype, rule) {

    const result = { ptype };

    if (ptype === 'p') {
      if (rule.length > 0) {
        [result.subject] = rule;
      }
      if (rule.length > 1) {
        [, result.code] = rule;
      }
      if (rule.length > 2) {
        [, , result.object] = rule;
      }
      if (rule.length > 3) {
        [, , , result.type] = rule;
      }
      if (rule.length > 4) {
        [, , , , result.action] = rule;
      }
    }

    if (ptype === 'g') {
      if (rule.length > 0) {
        [result.subject] = rule;
      }
      if (rule.length > 1) {
        [, result.parent] = rule;
      }
      if (rule.length > 2) {
        [, , result.code] = rule;
      }
    }

    return result;
  }

}

module.exports = Adapter;
