/**
 *
 */

const path    = require('path');
const assert  = require('assert').strict;
const casbin  = require('casbin');
const adapter = require('./adapter');
const debug   = require('debug')('light:permission');

class Enforcer {

  static async enforcer() {

    if (this._instance) {
      return this._instance;
    }

    const model = path.join(__dirname, 'model.conf');
    this._instance = await casbin.newEnforcer(model, adapter.init());

    return this._instance;
  }

  /**
   * 重新加载权限
   */
  static async reloadEnforcer() {
    this._instance = null;
    this.enforcer();
  }

  /**
   * 确认是否有权访问资源
   * @param {*} domain
   * @param {*} subject
   * @param {*} object
   * @param {*} type
   * @param {*} action
   */
  static async enforce(domain, subject, object, type, action) {

    debug('enforce', domain, subject, object, type, action);

    const instance = await this.enforcer();
    return await instance.enforce(subject, domain, object, type, action);
  }

  /**
   * 获取指定用户的所有权限
   * @param {*} uid
   * @param {*} domain
   * @param {*} type [option]
   */
  static async forUser(uid, domain, type) {

    debug('for user', uid, domain, type);

    assert(uid, 'uid cannot be empty');
    assert(domain, 'domain cannot be empty');

    const instance = await this.enforcer();
    const permissions = await instance.getImplicitPermissionsForUser(String(uid), domain);

    return permissions
      .map(item => ({ subject: item[0], object: item[2], type: item[3], action: item[4] }))
      .filter(item => (type ? item.type === type : true));
  }

  /**
   * 添加策略，每次操作单个策略
   * @param {*} policy 策略
   * {subjec, domain, object, type, action}
   */
  static async add(policy) {

    debug('add', policy);

    assert(policy.subject, 'subject cannot be empty');
    assert(policy.domain, 'domain cannot be empty');
    assert(policy.object, 'object cannot be empty');
    assert(policy.type, 'type cannot be empty');
    assert(policy.action, 'action cannot be empty');

    const instance = await this.enforcer();
    await instance.addPolicy(
      policy.subject,
      policy.domain,
      policy.object,
      policy.type,
      policy.action
    );
  }

  static async remove(policy) {

    debug('remove', policy);

    assert(policy.subject, 'subject cannot be empty');
    assert(policy.domain, 'domain cannot be empty');
    assert(policy.object, 'object cannot be empty');
    assert(policy.type, 'type cannot be empty');
    assert(policy.action, 'action cannot be empty');

    const instance = await this.enforcer();

    await instance.removePolicy(
      policy.subject,
      policy.domain,
      policy.object,
      policy.type,
      policy.action
    );
  }

  /**
   * subject会继承object拥有的权限
   * @param {*} policy
   *  subject 待授权对象，可以是一个用户/角色/组
   *  domain 租户
   *  object 角色或组
   */
  static async addGroup(policy) {

    debug('add group', policy);

    assert(policy.subject, 'subject cannot be empty');
    assert(policy.object, 'object cannot be empty');
    assert(policy.domain, 'domain cannot be empty');

    const instance = await this.enforcer();
    await instance.addGroupingPolicy(policy.subject, policy.object, policy.domain);

  }

  static async removeGroup(policy) {

    debug('remove group', policy);

    assert(policy.subject, 'subject cannot be empty');
    assert(policy.object, 'object cannot be empty');
    assert(policy.domain, 'domain cannot be empty');

    const instance = await this.enforcer();
    await instance.removeGroupingPolicy(policy.subject, policy.object, policy.domain);

  }

  // 测试用，生成缺省数据
  static async genDefault(enforcer) {

    if (this.done) {
      return;
    }

    this.done = true;

    const setting = require('../setting');
    const CONST = require('../constant');

    // const menu = setting.menu.map(item => ([
    //   CONST.ROLE.DEFAULT,
    //   CONST.DEFAULT_TENANT,
    //   item.url,
    //   'menu',
    //   CONST.ACTION.ACCESS
    // ]));

    // await enforcer.addPolicies(menu.filter(item => item[2]));

    const route = setting.menu.map(item => ([
      'MEMBER',
      CONST.DEFAULT_TENANT,
      `/${item.url}`,
      'route',
      'ALLOW'
    ]));

    await enforcer.addPolicies(route.filter(item => item[2] !== '/'));

    const api = setting.board.map(item => ([
      'MEMBER',
      CONST.DEFAULT_TENANT,
      item.api,
      'api',
      'ALLOW'
    ]));

    await enforcer.addPolicies(api);

  }

}

module.exports = Enforcer;
