/**
 * 按照标准的JSON格式返回数据
 * @module response
 * @author r2space@gmail.com
 * @version 1.0.0
 */

'use strict';


const debug = require('debug')('light:response');

class Response {

  static send(handler, data) {

    // 如果body有值，则直接返回，避免重新指定body导致无法现在文件问题
    if (typeof handler.ctx.body !== 'undefined') {
      return;
    }

    handler.ctx.type = 'application/json; charset=utf-8';
    handler.ctx.status = 200;
    handler.ctx.body = Response.createDataSchema(data);

    debug('response', handler.ctx.type);
    debug('response', handler.ctx.status);
  }

  static sendHTML(handler, html) {
    handler.ctx.type = 'text/html';
    handler.ctx.status = 200;
    handler.ctx.body = html;
  }

  static sendError(handler, error) {

    const status = error.status || 500;
    debug('status', status);
    debug('error code:', error.code);
    debug('error message:', error.message);
    debug('error detail:', error.errors);

    handler.ctx.type = 'application/json; charset=utf-8';
    handler.ctx.status = status;
    handler.ctx.body = Response.createErrorSchema(error.code, error.message, error.errors);
  }

  static sendFile(handler, status, file) {

    if ([400, 404, 406, 500].includes(status)) {
      handler.ctx.status = status;
      handler.ctx.body = null;
      return;
    }

    handler.ctx.set('Cache-Control', 'public, max-age=0');
    handler.ctx.set('Content-Type', file.contentType);
    handler.ctx.set('Content-Length', file.length);
    handler.ctx.set('Last-Modified', file.uploadDate.toUTCString());

    // 中文名称等filename需要用双引号引起来，否则会出现 Error: write EPIPE 异常
    this.setDisposition(handler, {
      attachment : handler.params.attachment,
      filename   : `"${file.filename}"`,
      contentType: file.contentType
    });

    handler.ctx.status = status;
    handler.ctx.body = file.stream;
  }

  /**
   * 设置响应头 Content-Disposition 信息，用于控制文件被下载后以什么样的方式显示
   *
   * option.contentType
   * option.attachment
   *  inline - 固定在浏览器内inline方式打开
   *  auto - 根据文件的contentType判断，如果是image，pdf则使用inline方式打开
   * option.filename
   */
  static setDisposition(handler, option) {

    // 可以通过参数的 attachment 来控制下载文件后的显示方式
    // 如果不是 inline，那么直接显示
    if (option.attachment === 'inline') {
      return handler.ctx.set('Content-Disposition', 'inline');
    }

    // 根据Content-Type自动判断
    if (handler.params.attachment === 'auto') {

      const contentType = option.contentType || '';

      // 图片和PDF，用inline方式
      if (contentType.startsWith('image')) {
        return handler.ctx.set('Content-Disposition', 'inline');
      }
      if (contentType === 'application/pdf') {
        return handler.ctx.set('Content-Disposition', 'inline');
      }
    }

    // 非inline方式返回文件时，需要设置文件的名字
    const userAgent = (handler.ctx.get['user-agent'] || '').toLowerCase();
    const isIE      = userAgent.indexOf('msie') >= 0 || userAgent.indexOf('trident') >= 0;
    const isChrome  = userAgent.indexOf('chrome') >= 0;
    const isFirefox = userAgent.indexOf('firefox') >= 0;

    // IE Chrome
    if (isIE || isChrome) {
      const disposition = `attachment; filename=${encodeURIComponent(option.filename)}`;
      return handler.ctx.set('Content-Disposition', disposition);
    }

    // Firefox
    if (isFirefox) {
      const disposition = `attachment; filename*="utf8''${encodeURIComponent(option.filename)}"`;
      return handler.ctx.set('Content-Disposition', disposition);
    }

    // Safari等其他非主流浏览器
    const disposition = `attachment; filename=${Buffer.from(option.filename).toString('binary')}`;
    return handler.ctx.set('Content-Disposition', disposition);

  }

  /**
   * @desc 通过以后的数据生成错误对象
   * @param {Number} code 错误代码
   * @param {String} message 错误消息
   * @param {Object} errorDetail 错误详细信息
   * @returns {*} 结构化了的错误结果
   */
  static createErrorSchema(code, message, errors) {
    return {
      apiVersion: '1.0',
      error     : { code, message, errors }
    };
  }

  /**
   * @desc 通过已有的数据生成简单的结果对象
   * @param {Object} data 数据
   * @returns {*} 结构化了的数据结果
   */
  static createDataSchema(data) {
    return {
      apiVersion: '1.0',
      data      : typeof data === 'undefined' ? {} : data
    };
  }

  /**
   * 生成标准的响应JSON对象
   */
  static createFullSchema() {
    return {
      apiVersion: '1.0',
      context   : 'string',
      id        : 'string',
      method    : 'string',
      params    : { id: 'string' },
      // ---------- 响应的数据 ----------
      data      : {
        kind            : 'string',
        fields          : 'string',
        etag            : 'string',
        id              : 'string',
        lang            : 'string',
        updated         : 'string',
        deleted         : 'boolean',
        // ---------- 分页信息 ----------
        currentItemCount: 'integer',
        itemsPerPage    : 'integer',
        startIndex      : 'integer',
        totalItems      : 'integer',
        pageIndex       : 'integer',
        totalPages      : 'integer',
        pageLinkTemplate: 'string',
        // ---------- 链接 ----------
        next            : {},
        nextLink        : 'string',
        previous        : {},
        previousLink    : 'string',
        self            : {},
        selfLink        : 'string',
        edit            : {},
        editLink        : 'string',
        // ---------- 表示一组数据 ----------
        items           : [{ vals: {} }]
      },
      // ---------- 提供错误的详细信息 ----------
      error: {
        code   : 'integer',
        message: 'string',
        errors : [{
          domain      : 'string',
          reason      : 'string',
          message     : 'string',
          location    : 'string',
          locationType: 'string',
          extendedHelp: 'string',
          sendReport  : 'string'
        }]
      }
    };
  }

}

module.exports = Response;
