/**
 * @file 提供密码验证，加密，简易登陆等相关方法
 * @module security.js
 * @author r2space@gmail.com
 * @version 1.0.0
 */

'use strict';

const assert         = require('assert').strict;
const crypto         = require('crypto');
const jwt            = require('jwt-simple');
const debug          = require('debug')('light:security');
const config         = require('./configuration');
const log            = require('./log');
const rider          = require('./model/datarider');
const CONST          = require('./constant');
const session        = require('./middleware/session');
const { BadRequest } = require('./error/http');

/**
 * @desc 基于Cookie，Session的简易登陆功能<br>
 *  将用户信息保存到session当中,用户ID保存到header中
 * @param {Object} handler 上下文对象
 *  params.code: 租户
 *  params.id: 账户
 *  params.password: 密码
 *  params.hmackey: 密码加密key
 * @param {Object} params 参数
 * @param {Function} callback 回调函数，验证成功返回用户信息
 */
exports.simpleLogin = async function (handler, _params, _callback) {

  let params = _params;
  let callback = _callback;

  if (typeof _params === 'function') {
    callback = _params;
    params = null;
  }

  const { id, password, hmackey, domain, code } = params || handler.params;
  debug('id: %s, password: %s, hmackey: %s', id, password, hmackey);

  assert(id, 'id cannot be empty');
  assert(password, 'password cannot be empty');

  handler.domain = domain || process.env.APPNAME;
  handler.code = code || CONST.DEFAULT_TENANT;

  debug('login by', id, handler.domain, handler.code);
  const user = await rider.user.get(handler, { condition: { id } });
  if (!user || !user._id) {
    log.debug('User does not exist.');
    const err = new BadRequest('Username or password incorrect.');
    return callback ? callback(err) : Promise.reject(err);
  }

  const pwd = exports.sha256(password, hmackey);
  debug('param password', pwd);
  debug('db password', user.password);

  if (user.password !== pwd) {
    log.debug('The user password is not correct.');
    const err = new BadRequest('Username or password incorrect.');
    return callback ? callback(err) : Promise.reject(err);
  }

  // 登陆成功后，生成新的Session(安全规则)
  debug('recreate session');
  session.renew(handler.ctx);

  debug('set user, domain, code to session');
  handler.session(CONST.SESSION.DOMAIN, handler.domain);
  handler.session(CONST.SESSION.CODE, handler.code);
  handler.session(CONST.SESSION.USER, user);

  debug('remove password plaintext');
  delete user.password;

  return callback ? callback(null, user) : Promise.resolve(user);
};

/**
 * @desc 注销，删除Session
 * @param {Object} req 请求对象
 * @param {Function} callback
 */
exports.simpleLogout = function (handler) {
  handler.ctx.session = null;
};

/**
 * @desc 字符串加密方法<br>
 *  algorithm - 'sha1', 'md5', 'sha256', 'sha512'
 *  encoding - 'hex', 'binary', 'base64'
 * @ignore
 * @param {String} str 加密对象字符串
 * @param {String} [hmackey] 密钥
 * @returns {String} 加密后字符串
 */
exports.sha256 = function (str, hmackey) {

  if (!str) {
    return '';
  }

  const key = hmackey || config.app.hmackey;
  debug('hmackey', key);

  return crypto.createHmac('sha256', key)
    .update(str)
    .digest('hex');
};

/**
 * @desc 加密方法<br>
 *   - 使用的Algorithm - 'aes192'<br>
 *   - 使用的encoding - 'hex'<br>
 *   - 内容的encoding - 'utf-8'<br>
 * @param {String} str 加密对象
 * @param {String} secret 密钥
 * @returns {String} 加密后的内容
 */
exports.encrypt = function (str, secret) {
  const cipher = crypto.createCipher('aes192', secret);
  const crypted = cipher.update(str, 'utf8', 'hex');

  return crypted + cipher.final('hex');
};

exports.encrypt2 = function (str, secret) {
  let tmp = crypto.createHash('md5');
  tmp.update(secret);

  const key = tmp.digest('hex');

  tmp = crypto.createHash('md5');
  tmp.update(secret + key);

  const iv = tmp.digest('hex');
  const data = Buffer.from(str, 'utf8').toString('binary');
  const cipher = crypto.createCipheriv('aes-256-cbc', key, iv.slice(0, 16));
  const encrypted = cipher.update(data, 'utf8', 'binary') + cipher.final('binary');

  return Buffer.from(encrypted, 'binary').toString('base64');
};

/**
 * @desc 解密方法
 *   - 使用的Algorithm - 'aes192'<br>
 *   - 使用的encoding - 'hex'<br>
 *   - 内容的encoding - 'utf-8'<br>
 * @param {String} str 解密对象
 * @param {String} secret 密钥
 * @returns {String} 解密后的内容
 */
exports.decrypt = function (str, secret) {
  const decipher = crypto.createDecipher('aes192', secret);
  const decrypted = decipher.update(str, 'hex', 'utf8');

  return decrypted + decipher.final('utf8');
};

exports.decrypt2 = function (str, secret) {

  // Convert urlsafe base64 to normal base64
  const original = str.replace(/\-/g, '+').replace(/_/g, '/');

  // Convert from base64 to binary string
  const encoded = Buffer.from(original, 'base64').toString('binary');

  // Create key from password
  let tmp = crypto.createHash('md5');
  tmp.update(secret);
  const key = tmp.digest('hex');

  // Create iv from password and key
  tmp = crypto.createHash('md5');
  tmp.update(secret + key);
  const iv = tmp.digest('hex');
  const decipher = crypto.createDecipheriv('aes-256-cbc', key, iv.slice(0, 16));

  return (decipher.update(encoded, 'binary', 'utf8') + decipher.final('utf8'));
};

exports.jwtEncode = function (data, key, algorithm = 'HS256') {
  return jwt.encode(
    {
      _id     : data.uid,
      domain  : data.domain,
      code    : data.code,
      timezone: data.timezone,
      expires : data.expires
    },
    key,
    algorithm
  );
};
