/**
 * 从数据库或配置文件加载配置项目，包括：
 *  config
 *  i8n
 *  struct
 *  board
 *  validator
 *  route
 *  menu
 *  tenant
 *  job
 *
 * 加载的内容保存到内存中，更新内容是需要加载
 * 配置文件加载工程根目录的 /setting/ 目录下的内容
 */

const fs     = require('fs');
const path   = require('path');
const debug  = require('debug')('light:setting');
const assert = require('assert').strict;
const config = require('./configuration');
const i18n   = require('./i18n');
const job    = require('./job');
const Model  = require('./db/mongo/model');
const CONST  = require('./constant');
const helper = require('./helper');
const log    = require('./log');

class Setting {

  /**
   *  加载配置项
   */
  static async init(kind = CONST.APP_KIND_SERVER) {

    // 文件加载系统表数据
    this.fromSystemFile();

    // 开发环境，如果配置存在并且设置了setting=local，从本地读取配置
    if (process.env.APPDEV === 'true') {
      debug('is dev');

      const exists = fs.existsSync(path.join(process.cwd(), 'setting'));
      if (exists && process.env.APPSETTING === 'local') {
        debug('setting exists. load from file');
        return this.fromFile();
      }

      // 从数据库加载
      debug('load from db');
      return await this.fromDB(kind);
    }

    // 生产环境，强制从文件加载数据
    debug('load from file');
    return this.fromFile();
  }

  static get struct() {

    // Admin应用时，把系统api也加载进来，可以路由系统api相关的请求
    return process.env.APPNAME === CONST.SYSTEM_DB
      ? this._struct.concat(this.system.struct)
      : this._struct;
  }

  static get board() {
    return process.env.APPNAME === CONST.SYSTEM_DB
      ? this._board.concat(this.system.board)
      : this._board;
  }

  /**
   * 替换board，仅允许在单元测试模式下使用
   */
  static set board(board) {
    assert(helper.isUnitTest(), 'Can only be used in unit test mode');
    this._board = board;
  }

  /**
   * 从数据库加载数据
   * @param {*} kind
   */
  static async fromDB(kind) {

    // config
    config.load(await this.fetch(
      CONST.SYSTEM_DB_CONFIG,
      'type key valueType value prdValue',
      { type: 1, key: 1 }
    ));
    log.debug(`Loading configuration : ${config.catalog.length}`);

    // i18n
    i18n.load(await this.fetch(CONST.SYSTEM_DB_I18N, 'type key lang', { type: 1, key: 1 }));
    log.debug(`Loading i18n : ${i18n.catalog.length}`);

    // job
    if ([CONST.APP_KIND_SERVER].includes(kind)) {
      job.load(await this.fetch(
        CONST.SYSTEM_DB_JOB,
        'count limit name run schedule status last extend step',
        {  name: 1 }
      ));
      log.debug(`Loading job : ${job.catalog.length}`);
    }

    // struct
    this._struct = await this.fetch(
      CONST.SYSTEM_DB_STRUCTURE,
      'schema parent items kind lock public description',
      { kind: 1, schema: 1 }
    );
    log.debug(`Loading structure : ${this._struct.length}`);

    // board
    this._board = await this.fetch(
      CONST.SYSTEM_DB_BOARD,
      'action api class description filters kind path schema selects sorts type script',
      { kind: 1, api: 1, type: 1 }
    );
    log.debug(`Loading board : ${this._board.length}`);

    // validator
    this.validator = await this.fetch(
      CONST.SYSTEM_DB_VALIDATOR,
      'group name rule key message strict option class action condition description sanitize',
      { group: 1, key: 1 }
    );
    log.debug(`Loading validator : ${this.validator.length}`);

    // route
    if ([CONST.APP_KIND_SERVER].includes(kind)) {
      this.route = await this.fetch(
        CONST.SYSTEM_DB_ROUTE,
        'url template class action descritpion',
        { url: 1 }
      );
      log.debug(`Loading route : ${this.route.length}`);
    }

    // menu
    if ([CONST.APP_KIND_SERVER].includes(kind)) {
      this.menu = await this.fetch(
        CONST.SYSTEM_DB_FUNCTION,
        'description kind menu order parent status type url icon',
        { kind: 1, order: 1 }
      );
      log.debug(`Loading menu : ${this.menu.length}`);
    }

    // tenant
    this.tenant = await this.fetch(
      CONST.SYSTEM_DB_TENANT,
      'code name description logo schema',
      { code: 1, name: 1 }
    );
    log.debug(`Loading tenant : ${this.tenant.length}`);

  }

  /**
   * 检索表
   * @param {*} table
   * @param {*} field
   * @param {*} sort
   */
  static async fetch(table, field, sort) {
    assert(process.env.APPNAME, 'domain cannot be empty');
    const model = new Model(process.env.APPNAME, CONST.SYSTEM_DB_PREFIX, table);
    return await model.list({ valid: 1 }, 0, 0, sort, field);
  }

  /**
   * 从文件加载数据(脱离Mongo运行时使用)
   *
   * TODO: 数据校验
   */
  static fromFile() {

    log.debug('Loading configuration from file');

    config.load(this.loadYml(CONST.SYSTEM_DB_CONFIG));
    i18n.load(this.loadYml(CONST.SYSTEM_DB_I18N));
    job.load(this.loadYml(CONST.SYSTEM_DB_JOB));

    this._struct = this.loadYml(CONST.SYSTEM_DB_STRUCTURE);
    this._board = this.loadYml(CONST.SYSTEM_DB_BOARD);
    this.validator = this.loadYml(CONST.SYSTEM_DB_VALIDATOR);
    this.route = this.loadYml(CONST.SYSTEM_DB_ROUTE);
    this.menu = this.loadYml(CONST.SYSTEM_DB_FUNCTION);
    this.tenant = this.loadYml(CONST.SYSTEM_DB_TENANT);

    log.debug(`Loading config : ${config.catalog.length}`);
    log.debug(`Loading i18n : ${i18n.catalog.length}`);
    log.debug(`Loading job : ${job.catalog.length}`);
    log.debug(`Loading struct : ${this._struct.length}`);
    log.debug(`Loading board : ${this._board.length}`);
    log.debug(`Loading validator : ${this.validator.length}`);
    log.debug(`Loading route : ${this.route.length}`);
    log.debug(`Loading menu : ${this.menu.length}`);

    debug('app table', this._struct.map(item => item.schema));
  }

  /**
   * 从文件加载数据
   */
  static fromSystemFile() {

    log.debug('Loading configuration from file');

    // config.load(this.loadSysYml(CONST.SYSTEM_DB_CONFIG));
    // i18n.load(this.loadSysYml(CONST.SYSTEM_DB_I18N));
    // job.load(this.loadSysYml(CONST.SYSTEM_DB_JOB));

    this.system = {
      struct   : this.loadSysYml(`${CONST.SYSTEM_DB_STRUCTURE}`),
      board    : this.loadSysYml(`${CONST.SYSTEM_DB_BOARD}`),
      validator: this.loadSysYml(CONST.SYSTEM_DB_VALIDATOR),
      route    : this.loadSysYml(CONST.SYSTEM_DB_ROUTE),
      menu     : this.loadSysYml(CONST.SYSTEM_DB_FUNCTION),
      tenant   : this.loadSysYml(CONST.SYSTEM_DB_TENANT)
    };

    // log.debug(`Loading sys config : ${config.catalog.length}`);
    // log.debug(`Loading sys i18n : ${i18n.catalog.length}`);
    // log.debug(`Loading sys job : ${job.catalog.length}`);
    log.debug(`Loading sys struct : ${this.system.struct.length}`);
    log.debug(`Loading sys board : ${this.system.board.length}`);
    log.debug(`Loading sys validator : ${this.system.validator.length}`);
    log.debug(`Loading sys route : ${this.system.route.length}`);
    log.debug(`Loading sys menu : ${this.system.menu.length}`);

    debug('system table', this.system.struct.map(item => item.schema));
  }

  /**
   * 加载 /setting/ 下的yaml格式配置文件
   * 如果是单元测试，加载 /test/setting/ 下的配置
   * @param {*} name 文件名称
   */
  static loadYml(name) {
    const folder = helper.isUnitTest() ? 'test' : '';
    return helper.yamlLoader(`${folder}/setting/${name}.yml`) || [];
  }

  static loadSysYml(name) {
    return helper.yamlLoader(`../setting/${name}.yml`, __dirname) || [];
  }

}

module.exports = Setting;
