/**
 * socket.js
 *
 *  Socket请求分发：
 *  客户端的socket请求会被分发到对应的类及方法上，规则与普通api相同
 *  /api/test/get  ->  请求分派到 /controllers/test#get 方法里。
 *
 *  socket客户端连接通过uid保存，在后台逻辑了可以通过uid向前台发送消息
 *  也可以通过handler.socket，发送消息给请求端
 */

'use strict';

const path         = require('path');
const assert       = require('assert').strict;
const qs           = require('qs');
const { Server }   = require('ws');
const debug        = require('debug')('light:socket');
const log          = require('./log');
const CONST        = require('./constant');
const Context      = require('./context');
const helper       = require('./helper');
const response     = require('./response');
const sessionstore = require('./db/sessionstore');
const { BadRequest, Unauthorized } = require('./error').http;

class Socket {

  static clients = [];

  /**
   * 启动 WebSocket 服务
   * @param server
   */
  static listen(server, db) {

    debug('listen');
    this.wss = new Server({ server });
    this.wss.on('connection', async (ws, req) => {

      debug('on connection', req.method, req.url);

      const socket = new Socket();
      socket.db = db;

      await socket.dispatch(ws, req);
    });
    this.wss.on('headers', (headers, req) => {
      debug('headers', headers);
      debug('request headers', req.headers);
    });
    this.wss.on('error', error => {
      debug('error', error);
      log.error(error);
    });

    log.info(`Light websocket listening on port ${process.env.APPPORT || 7000}`);
  }

  /**
   * 关闭服务
   */
  static close() {

    debug('close');
    if (this.wss) {
      return new Promise(resolve => {
        this.wss.close(() => resolve());
        this.wss = null;
      });
    }
    return Promise.resolve();
  }


  /**
   * 发送消息给客户端（客户端可能是多个）
   * @param {String | Array} id Option 接收消息的用户ID或ClientID
   * @param {Object} data 发送的消息
   * @returns {boolean} 如果为false，为未能找到已经建立连接的客户端
   */
  static emit(id, data = {}) {

    debug('emit to', id);

    assert(id, 'id cannot be empty');
    debug('client', Socket.clients.map(item => item.uid));

    const ids = Array.isArray(id) ? id : [id];
    const clients = Socket.clients
      .filter(client => ids.includes(client.uid) || ids.includes(client.cid));

    if (clients.length <= 0) {
      log.debug('client not found');
      return false;
    }

    clients.forEach(client => Socket.send(client.ws, data));
    return true;
  }


  /**
   * 分发请求
   * @param {Object} ws
   * @param {Object} req
   * @returns {*}
   */
  async dispatch(ws, req) {

    log.info('web socket connected.');

    const query = this.parseQuery(req);
    debug('query', query);

    // 只是建立连接（等待后台发起请求）
    if (query.path === '/') {
      return await this.bindConnect(ws, query, req.headers);
    }

    // 调用controller
    await this.routeApi(ws, query, req.headers);
  }


  /**
   * 客户端建立连接，url为空
   */
  async bindConnect(ws, query, headers) {

    debug('bind connect', query);

    // 根据cookie恢复session
    try {
      const handler = await this.parseHandler(query, headers);
      this.saveConnection(ws, {
        uid: handler.uid,
        cid: query.params.cid,
        key: headers['sec-websocket-key']
      });
    } catch (err) {
      log.error(err.message);
      Socket.sendError(ws, new Unauthorized('authentication failed'));
      ws.close();
    }
  }


  /**
   * 通过请求路径，解析对应的类及方法
   * @param {socket连接} ws
   * @param {请求参数} query
   *   url格式需为 /api/class/action?key=val
   */
  async routeApi(ws, query, headers) {

    debug('route api');

    const items = query.path.split('/');
    debug('path', items);
    if (items.length < 4) {
      log.debug('request api format is incorrect');
      return Socket.sendError(ws, new BadRequest('can not dispatch'));
    }

    // 校验action是否存在
    const [, , clazz, action] = items;
    const root = helper.isUnitTest() ? 'test' : '';
    const ctrl = helper.resolve(path.join(root, CONST.PATH_CONTROLLER, clazz)) || {};
    const func = ctrl[action];

    if (!func) {
      log.error('action not found', `${clazz}#${action}`);
      return Socket.sendError(ws, new BadRequest('can not dispatch'));
    }

    // 根据cookie恢复session，并生成handler
    let handler = null;
    try {
      handler = await this.parseHandler(query, headers);
      this.saveConnection(ws, {
        uid: handler.uid,
        cid: query.cid,
        key: headers['sec-websocket-key']
      });
    } catch (err) {
      log.error(err.message);
      Socket.sendError(ws, new Unauthorized('authentication failed'));
    }

    // 使用老式回调方式调用controller
    if (func.length === 2) {
      log.warn('Does not support callback method return value', handler.uid);
      return func.call(this, handler, (err, data) => {
        if (err) {
          log.error(err, handler.uid);
          return Socket.sendError(ws, err);
        }
        Socket.send(ws, data);
      });
    }

    // 使用async的调用方式
    try {
      const data = await func.call(this, handler);
      if (data !== CONST.NONE) {
        Socket.send(ws, data);
      }
    } catch (err) {
      log.error(err, handler.uid);
      Socket.sendError(ws, err);
    }
  }


  /**
   * 解析请求路径
   *  /api/test/get?key=val > { path: /api/test/get, params: {key: val} }
   * @param {Object} req
   */
  parseQuery(req) {
    const [url, params] = req.url.split('?');
    return { path: url, params: qs.parse(params) };
  }


  /**
   * 解析cookie，生成handler
   * @param {*} query
   * @param {*} headers
   */
  async parseHandler(query, headers) {

    debug('parse header');

    const sid = this.parseCookie(headers.cookie)['light.sid'];
    debug('sid', sid);

    // 加AuthIgnore，允许在未正常登陆的状态下使用socket
    if (!sid && query.params.auth === 'ignore') {
      log.info(`socket ignore auth. ${query.params.cid}`, null);
      return {};
    }

    // 未能获取session key
    if (!sid) {
      throw new BadRequest('cannot resolve light.sid in cookie');
    }

    debug('read session data');
    const store = sessionstore.init({ db: this.db });
    const data = await store.get(sid);
    debug('session', data);

    // 给定的cookie不正确，无法恢复session
    if (!data) {
      debug('unable to resume session');
      throw new BadRequest('unable to resume session');
    }

    // session已经超时
    if (data._expire < new Date().getTime()) {
      debug('session has timed out');
      throw new BadRequest('session has timed out');
    }

    // 没有认证
    const { user, domain, code } = data;
    if (!user || !user._id) {
      debug('not logged in');
      throw new BadRequest('not logged in');
    }

    debug('create handler');
    const handler = new Context(user._id, domain, code);
    handler.addParams(query.params);

    debug('params', handler.params);
    return handler;
  }


  /**
   * 解析cookie
   * @param {*} str
   */
  parseCookie(str = '') {
    debug('parse cookie', str);
    return str.split(/;|,/).reduce((memo, item) => {
      if (!item.includes('=')) {
        return memo;
      }

      const [key, val] = item.split('=');
      if (['path', 'expires'].includes(key.trim())) {
        return memo;
      }

      memo[key.trim()] = val.trim();
      return memo;
    }, {});
  }


  /**
   * 用用户ID保存socket连接
   * @param {*} ws
   * @param {*} uid
   * @param {*} cid 客户端表示 - 使用socket请求的header的
   */
  saveConnection(ws, { uid, cid, key }) {

    debug('save connetion', uid, cid, key);

    // 保存客户端连接
    let client = Socket.clients.find(item => item.key === key);
    if (!client) {
      debug('add client', uid, cid, key);
      client = { uid, cid, key, ws };
      Socket.clients.push(client);
    }

    // 断开时，清除客户端连接
    ws.on('close', () => {
      debug('remove client', client.uid, client.cid, client.key);
      Socket.clients = Socket.clients.filter(item => item.key !== client.key);
      log.info('web socket disconnected.', Socket.clients.length);
    });
  }


  /**
   * 发送数据
   * @param {*} ws
   * @param {*} data
   */
  static send(ws, data) {
    const params = response.createDataSchema(data);
    debug('send data', params);
    ws.send(JSON.stringify(params));
  }


  /**
   * 发送异常数据
   * @param {*} ws
   * @param {*} data
   */
  static sendError(ws, error) {
    const params = response.createErrorSchema(error.code || 500, error.message);
    debug('send error', params);
    ws.send(JSON.stringify(params));
  }

}

module.exports = Socket;
