/**
 * udp.js
 *
 * 接受udp请求后，会把啊请求路由到control中去，所以对请求数据结构有一定要求
 * 请求参数格式为JSON
 *  {
 *    query,
 *    params: { uid, domain, code, data: {} }
 *  }
 *
 * 注:
 *  可以指定接收的白名单，query匹配白名单列表的消息才会被分发
 *  要给特定的服务器发送消息，需明确指定ip，不指定ip则网段内所有监听者都能接收到消息
 *
 *  TODO: 如果需要在Application层面使用，需要考虑安全校验
 */

'use strict';

const dgram   = require('dgram');
const path    = require('path');
const debug   = require('debug')('light:udp');
const helper  = require('./helper');
const Context = require('./context');
const log     = require('./log');
const CONST   = require('./constant');
const config  = require('./configuration');

class Udp {

  /**
   * 启动 UDP 服务
   * @param {*} whitelist 限定接收消息的query类型，如果指定了自由在这个列表里的才会被分发到control里
   */
  static listen(whitelist = []) {

    const server = dgram.createSocket('udp4');
    server.bind(process.env.APPPORT || 7000);

    // 开始监听
    server.on('listening', () => {

      // 如果指定了多播地址，那么启用多播（在k8s环境下，pod间无法通过multicast通信）
      if (config.app.multicast) {
        debug('use multicast', config.app.multicast);

        server.setBroadcast(true);
        server.setMulticastTTL(128);
        server.addMembership(config.app.multicast);
      }

      debug(server.address());
      log.info(`Light udp listening on port ${server.address().port}`);
    });

    // 分发请求
    server.on('message', (buffer, info) => {
      try {
        debug('address', info.address, info.port);
        this.dispatch(buffer.toString(), info, whitelist);
      } catch (err) {
        log.error(err);
      }
    });

    // 异常处理
    server.on('error', err => log.error(err.stack));

  }

  /**
   * 发送upd请求
   * @param {String} query 与API格式相同，udp服务器接收到消息以后根据query，分派ccontroller
   * @param {JSON} params 具体参数，可包含 { uid, domain, code, data }
   * @param {String} address 接收消息的地址，没有指定则进行广播
   * @param {*} socket UDP实例，如果是频繁发送可以提前做好实例使用，也需要自己关闭连接
   */
  static async emit(query, params, address, socket) {

    debug('emit udp request', query, params);

    return new Promise((resolve, reject) => {

      const buffer = Buffer.from(JSON.stringify({ query, params }));
      debug('buffer length', buffer.length);

      const instance = socket ? socket : dgram.createSocket('udp4');
      instance.send(buffer, 0, buffer.length, process.env.APPPORT || 7000, address, (err, bytes) => {

        debug('close socket');
        if (!socket) {
          instance.close();
        }
        if (err) {

          debug('an error has occurred', err);
          return reject(err);
        }

        debug('sent successfully');
        resolve(bytes);
      });
    });
  }

  /**
   * 分发请求
   * @param {*} message
   */
  static async dispatch(message, info, whitelist) {

    debug('receive', message);
    debug('whitelist', whitelist);

    let parsed = {};
    try {
      parsed = JSON.parse(message);
    } catch (err) {
      return log.error('Incorrect message format');
    }

    debug('query', parsed);
    if (!parsed.query) {
      return log.error('Incorrect message query format');
    }

    // 检查请求是否包含在白名单中
    if (whitelist.length > 0) {
      if (!whitelist.includes(parsed.query)) {
        return debug('request are not included in the whitelis');
      }
    }

    // 解析请求路径
    const items = parsed.query.split('/');
    if (items.length < 4) {
      return log.debug('request api format is incorrect');
    }

    // 校验action是否存在
    const [, , clazz, action] = items;
    const root = helper.isUnitTest() ? 'test' : '';
    const ctrl = helper.resolve(path.join(root, CONST.PATH_CONTROLLER, clazz)) || {};
    const func = ctrl[action];

    if (!func) {
      return log.error('action not found', `${clazz}#${action}`);
    }

    debug('create context');
    const { uid, domain, code, data } = parsed.params;
    const handler = new Context(uid || CONST.DEFAULT_USER, domain, code);

    // 使用async的调用方式
    try {
      handler.addParams(data);
      await func.call(this, handler);
    } catch (err) {
      log.error(err, handler.uid);
    }
  }

}

module.exports = Udp;
