/**
 * @file 验证, 依赖validator模块
 * @author r2space@gmail.com
 * @module lib.validator.index
 * @version 1.0.0
 */

'use strict';

const assert   = require('assert').strict;
const debug    = require('debug')('light:validator');
const path     = require('path');
const CONST    = require('../constant');
const setting  = require('../setting');
const log      = require('../log');
const helper   = require('../helper');
const rule     = require('./rule');
const sanitize = require('./sanitize');

class Validator {

  static get rule() {

    if (this._rule) {
      return this._rule;
    }

    this._rule = {};

    // 系统提供的方法
    Object.keys(rule).forEach(item => this._rule[item] = rule[item]);

    // 调用用户定义的controller
    const root = helper.isUnitTest() ? 'test' : '';
    const custom = path.join(root, CONST.PATH_CONTROLLER, CONST.VALIDATOR.CUSTOM);
    const clazz = helper.resolve(custom);

    if (clazz) {
      Object.keys(clazz).forEach(item => {

        // 字符串 rule 开头的方法，视为校验方法
        if (!item.startsWith('rule')) {
          return;
        }
        const key = item.substr(4);
        this._rule[key] = clazz[key];
      });
    }

    return this._rule;
  }

  static get sanitize() {

    if (this._sanitize) {
      return this._sanitize;
    }

    this._sanitize = {};

    // 系统提供的方法
    Object.keys(sanitize).forEach(item => this._sanitize[item] = sanitize[item]);

    // 调用用户定义的controller
    const root = helper.isUnitTest() ? 'test' : '';
    const custom = path.join(root, CONST.PATH_CONTROLLER, CONST.VALIDATOR.CUSTOM);
    const clazz = helper.resolve(custom);

    if (clazz) {
      Object.keys(clazz).forEach(item => {

        // 字符串 sanitize 开头的方法，视为前处理方法
        if (!item.startsWith('sanitize')) {
          return;
        }
        const key = item.substr(8);
        this._sanitize[key] = clazz[key];
      });
    }

    return this._sanitize;
  }

  /**
   * 校验
   * @param handler
   * @param group    指定特定的校验组进行校验
   * @param scope    可选 校验范围 add | update
   */
  static async check(handler, group, scope) {

    const key = group || handler.req.path;
    assert(key, 'group cannot be empty');
    debug('check key:', key);

    const validator = setting.validator.filter(item => item.group === key);
    const result = [];

    debug('find validator', validator.length);
    for (const item of validator) {

      debug('check by', item.rule, 'name', item.name);

      // 如果指定了校验范围，只有范围相同时才进行检查
      if (item.scope && item.scope !== scope) {
        debug('scope does not match', scope, item.scope);
        continue;
      }

      const func = this.rule[item.rule];
      if (func) {
        const invalid = await func.call(this, handler, item);
        if (invalid) {
          result.push(invalid);
        }
        continue;
      }
      log.error(`Rule does not define : ${item.rule}`);
    }

    return result.length > 0 ? result : null;
  }

  /**
   * 格式化数据
   *  handler
   *  sanitize.key    获取数据的key
   *  sanitize.rule   格式化规则
   *  sanitize.option 格式化选项
   */
  static async format(handler, sanitize) {

    assert(sanitize.key, 'key cannot be empty');
    assert(sanitize.rule, 'rule cannot be empty');

    const func = this.sanitize[sanitize.rule];
    if (func) {

      // 返回处理后的值
      debug('format by', sanitize.rule);
      return await func.call(this, handler, sanitize);
    }

    // 原值返回
    debug('return original value');
    return helper.get(handler.param, sanitize.key);
  }

}

module.exports = Validator;
