/**
 * rule.js
 */

'use strict';

const assert      = require('assert').strict;
const validator   = require('validator');
const debug       = require('debug')('light:validator:rule');
const dayjs       = require('dayjs');
const helper      = require('../helper');
const log         = require('../log');
const CONST       = require('../constant');
const Model       = require('../db/mongo/model');
const { DBError } = require('../error/db');
const setting     = require('../setting');


/**
 * 必须, 判断给定的值是否存在
 * @param handler
 * @param rule
 * @returns {*}
 */
exports.required = function (handler, rule) {

  debug('rule key', rule.key);

  if (prepare(handler, rule)) {
    debug('passed');
    return;
  }

  if (!helper.isBlank(helper.get(handler.params, rule.key))) {
    debug('passed');
    return;
  }

  debug('failed');
  return { name: rule.name, message: rule.message };
};

/**
 * 检查值在指定的表里是否是唯一
 * @param handler
 * @param {object} rule
 */
exports.unique = async function (handler, rule) {

  debug('rule key', rule.key);

  if (prepare(handler, rule)) {
    debug('passed');
    return;
  }

  if (await count(handler, rule) <= 0) {
    debug('passed');
    return;
  }

  debug('failed');
  return { name: rule.name, message: rule.message, value: helper.get(handler.params, rule.key) };
};

/**
 * 检查指定的值是在表里存在, 不存在则错误, 支持数组格式的数据
 * @param {function} handler
 * @param {object} rule
 */
exports.exists = async function (handler, rule) {

  debug('rule key', rule.key);

  if (prepare(handler, rule)) {
    debug('passed');
    return;
  }

  if (await count(handler, rule) > 0) {
    debug('passed');
    return;
  }

  debug('failed');
  return { name: rule.name, message: rule.message, value: helper.get(handler.params, rule.key) };
};

/**
 * 给定的数值在检查用列表里存在
 * @param handler
 * @param rule
 * @returns {*}
 */
exports.in = function (handler, rule) {

  debug('rule key', rule.key);

  assert(rule.option, 'option cannot be empty');
  assert(Array.isArray(rule.option), 'options must be an array');

  if (prepare(handler, rule)) {
    debug('passed');
    return;
  }

  const val = helper.get(handler.params, rule.key);
  if (rule.option.includes(val)) {
    debug('passed');
    return;
  }

  debug('failed');
  return { name: rule.name, message: rule.message, value: val };
};

/**
 * 给定的数值在检查用列表里不存在
 * @param handler
 * @param rule
 * @returns {*}
 */
exports.nin = function (handler, rule) {

  debug('rule key', rule.key);

  assert(rule.option, 'option cannot be empty');
  assert(Array.isArray(rule.option), 'options must be an array');

  if (prepare(handler, rule)) {
    debug('passed');
    return;
  }

  const val = helper.get(handler.params, rule.key);
  if (!rule.option.includes(val)) {
    debug('passed');
    return;
  }

  debug('failed');
  return { name: rule.name, message: rule.message, value: val };
};

/**
 * 小于
 * @param handler
 * @param rule
 * @returns {*}
 */
exports.lt = function (handler, rule) {

  debug('rule key', rule.key);
  assert(rule.option, 'option cannot be empty');

  if (prepare(handler, rule)) {
    debug('passed');
    return;
  }

  const val = helper.get(handler.params, rule.key);
  if (val < rule.option) {
    debug('passed');
    return;
  }

  debug('failed');
  return { name: rule.name, message: rule.message, value: val };
};

/**
 * 大于
 * @param handler
 * @param rule
 * @returns {*}
 */
exports.gt = function (handler, rule) {

  debug('rule key', rule.key);

  if (prepare(handler, rule)) {
    debug('passed');
    return;
  }

  const val = helper.get(handler.params, rule.key);
  if (val > rule.option) {
    debug('passed');
    return;
  }

  debug('failed');
  return { name: rule.name, message: rule.message, value: val };
};

/**
 * 相等
 * @param data
 * @param rule
 * @returns {*}
 */
exports.equals = function (handler, rule) {

  debug('rule key', rule.key);

  if (prepare(handler, rule)) {
    debug('passed');
    return;
  }

  const val = helper.get(handler.params, rule.key);
  if (val === rule.option) {
    debug('passed');
    return;
  }

  return { name: rule.name, message: rule.message, value: val };
};

/**
 * 是日期
 */
exports.date = function (handler, rule) {

  debug('rule key', rule.key);

  if (prepare(handler, rule)) {
    debug('passed');
    return;
  }

  const val = helper.get(handler.params, rule.key);

  if (dayjs(val).isValid(val)) {
    debug('passed');
    return;
  }

  debug('failed');
  return { name: rule.name, message: rule.message, value: val };
};

/**
 * 是数字
 */
exports.numeric = function (handler, rule) {

  debug('rule key', rule.key);

  if (prepare(handler, rule)) {
    debug('passed');
    return;
  }

  const val = helper.get(handler.params, rule.key);
  if (typeof val === 'number') {
    debug('passed');
    return;
  }

  if (typeof val === 'string' && validator.isNumeric(val)) {
    debug('passed');
    return;
  }

  debug('failed');
  return { name: rule.name, message: rule.message, value: val };
};

/**
 * 是布尔
 */
exports.boolean = function (handler, rule) {

  debug('rule key', rule.key);

  if (prepare(handler, rule)) {
    debug('passed');
    return;
  }

  const val = helper.get(handler.params, rule.key);
  if (typeof val === 'boolean') {
    debug('passed');
    return;
  }
  if (typeof val === 'string' && validator.isBoolean(val)) {
    debug('passed');
    return;
  }

  debug('failed');
  return { name: rule.name, message: rule.message, value: val };
};

/**
 * 是JSON字符串
 */
exports.json = function (handler, rule) {

  debug('rule key', rule.key);

  if (prepare(handler, rule)) {
    debug('passed');
    return;
  }

  const val = helper.get(handler.params, rule.key);
  if (typeof val === 'string' && validator.isJSON(val)) {
    debug('passed');
    return;
  }

  debug('failed');
  return { name: rule.name, message: rule.message, value: val };
};

/**
 * 是数组
 */
exports.array = function (handler, rule) {

  debug('rule key', rule.key);

  if (prepare(handler, rule)) {
    debug('passed');
    return;
  }

  const val = helper.get(handler.params, rule.key);
  if (Array.isArray(val)) {
    debug('passed');
    return;
  }

  debug('failed');
  return { name: rule.name, message: rule.message, value: val };
};

/**
 * 是字符串
 */
exports.string = function (handler, rule) {

  debug('rule key', rule.key);

  if (prepare(handler, rule)) {
    debug('passed');
    return;
  }

  const val = helper.get(handler.params, rule.key);
  if (typeof val === 'string') {
    debug('passed');
    return;
  }

  debug('failed');
  return { name: rule.name, message: rule.message, value: val };
};

/**
 * 校验字符串长度超出
 * @param handler
 * @param rule
 * @returns {*}
 */
exports.min = function (handler, rule) {

  debug('rule key', rule.key);

  if (prepare(handler, rule)) {
    debug('passed');
    return;
  }

  const val = helper.get(handler.params, rule.key);
  const len = parseInt(rule.option) || Number.MAX_VALUE;

  if (val && val.length < len) {
    debug('passed');
    return;
  }

  debug('failed');
  return { name: rule.name, message: rule.message, value: val };
};

/**
 * 校验字符串长度不足
 * @param handler
 * @param rule
 * @returns {*}
 */
exports.max = function (handler, rule) {

  debug('rule key', rule.key);

  if (prepare(handler, rule)) {
    debug('passed');
    return;
  }

  const val = helper.get(handler.params, rule.key);
  const len = parseInt(rule.option) || -1;

  if (val && val.length > len) {
    debug('passed');
    return;
  }

  debug('failed');
  return { name: rule.name, message: rule.message, value: val };
};

/**
 * 校验字符串长度范围
 * @param handler
 * @param rule
 * @returns {*}
 */
exports.range = function (handler, rule) {

  debug('rule key', rule.key);
  assert(rule.option, 'option cannot be empty');
  assert(Array.isArray(rule.option), 'options must be an array');

  if (prepare(handler, rule)) {
    debug('passed');
    return;
  }

  const val = helper.get(handler.params, rule.key);
  const [min, max] = rule.option;

  if (validator.isLength(val, min, max)) {
    debug('passed');
    return;
  }

  debug('failed');
  return { name: rule.name, message: rule.message, value: val };
};

/**
 * 正确的邮箱
 */
exports.email = function (handler, rule) {

  debug('rule key', rule.key);

  if (prepare(handler, rule)) {
    debug('passed');
    return;
  }

  const val = helper.get(handler.params, rule.key);
  if (validator.isEmail(val)) {
    debug('passed');
    return;
  }

  debug('failed');
  return { name: rule.name, message: rule.message, value: val };
};

/**
 * 正确的URL
 */
exports.url = function (handler, rule) {

  debug('rule key', rule.key);

  if (prepare(handler, rule)) {
    debug('passed');
    return;
  }

  const val = helper.get(handler.params, rule.key);
  if (validator.isURL(val)) {
    debug('passed');
    return;
  }

  debug('failed');
  return { name: rule.name, message: rule.message, value: val };
};

/**
 * 正确的IP
 */
exports.ip = function (handler, rule) {

  debug('rule key', rule.key);

  const val = helper.get(handler.params, rule.key);
  const version = rule.option || 4;

  if (prepare(handler, rule)) {
    debug('passed');
    return;
  }

  if (validator.isIP(val, version)) {
    debug('passed');
    return;
  }

  debug('failed');
  return { name: rule.name, message: rule.message, value: val };
};

/**
 * 正确的颜色值
 */
exports.hexcolor = function (handler, rule) {

  debug('rule key', rule.key);

  if (prepare(handler, rule)) {
    debug('passed');
    return;
  }

  const val = helper.get(handler.params, rule.key);
  if (validator.isHexColor(val)) {
    debug('passed');
    return;
  }

  debug('failed');
  return { name: rule.name, message: rule.message, value: val };
};

/**
 * 正则匹配
 */
exports.matches = function (handler, rule) {

  debug('rule key', rule.key);
  assert(rule.option, 'option cannot be empty');

  if (prepare(handler, rule)) {
    debug('passed');
    return;
  }

  const val = helper.get(handler.params, rule.key);
  if (new RegExp(rule.option).test(val)) {
    debug('passed');
    return;
  }

  debug('failed');
  return { name: rule.name, message: rule.message, value: val };
};

/**
 * Excel导入时，校验表头
 *  option.type 匹配方式 index | title
 *  option.fields {
 *    value: 列名或列顺序
 *    checkField: 是否检查该列
 *  }
 */
exports.checkField = function (handler, rule) {

  debug('rule key', rule.key);

  assert(rule.option, 'option cannot be empty');
  assert(rule.option.type, 'option.type cannot be empty');
  assert(rule.option.fields, 'option.fields cannot be empty');

  const { data } = handler.params;

  if (!data || !Array.isArray(data)) {
    debug('failed');
    return { name: rule.name, message: rule.message };
  }

  const errors = rule.option.fields.reduce((memo, item, index) => {
    const invalid = rule.option.type === 'index'
      ? data[index] !== item.value  // by index
      : !data.includes(item.value); // by title

    if (item.checkField === 'true' && invalid) {
      memo.push(item.value);
    }
    return memo;
  }, []);

  if (errors.length <= 0) {
    debug('passed');
    return;
  }

  debug('failed');
  return { name: rule.name, message: rule.message, value: errors };
};

/**
 * 获取数据件数
 * @param {*} handler
 * @param {*} rule
 */
async function count(handler, rule) {

  assert(rule.option.schema, 'schema cannot be empty');
  assert(Array.isArray(rule.option.conditions), 'conditions must be an array');

  debug('fetch count');

  const condition = rule.option.conditions.reduce((memo, item) => {

    // 如果是 $ 操作符，说明是数据引用
    const value = item.value[0] === '$'
      ? helper.get(handler.params, item.value.substr(1))
      : item.value;

    const filter = {};

    // 如果指定了操作符则按操作符过滤
    if (item.operator) {
      filter[item.parameter] =  { [item.operator]: value };
    } else {
      // 未指定，如果值是数组类型的，使用 $in 操作符
      filter[item.parameter] =  Array.isArray(value) ? { $in: value } : value;
    }

    return Object.assign(memo, filter);
  }, { valid: 1 });

  debug('condition', condition);

  const model = new Model(
    handler.domain || process.env.APPNAME,
    handler.code || CONST.DEFAULT_TENANT,
    rule.option.schema,
    { tz: handler.timezone, struct: setting.struct }
  );

  try {
    return await model.count(condition);
  } catch (err) {
    log.error(err);
    throw new DBError('fetch count error');
  }
}

/**
 * 预处理参数校验参数
 * 1. 处理strict模式
 * 2. 处理有在特定条件下进行的校验
 * @param {*} handler
 * @param {*} rule
 */
function prepare(handler, rule) {

  // 非严格模式（如果值为空，则不校验）
  if (rule.strict === false) {

    debug('no strict');

    const val = helper.get(handler.params, rule.key);
    if (helper.isBlank(val)) {
      debug('value is empty');
      return true;
    }
  }

  // 条件校验（只有在特定条件下进行校验，条件保存在condition中）
  if (rule.condition && rule.condition.key) {

    const param = helper.get(handler.params, rule.condition.key);
    const value = rule.condition.parameter;

    debug('conditional operator', rule.condition.operator);
    debug('parameter', param);
    debug('value', value);

    switch (rule.condition.operator) {
      case '$eq':
        return param === value;
      case '$ne':
        return param !== value;
      case '$gt':
        return param > value;
      case '$gte':
        return param >= value;
      case '$lt':
        return param < value;
      case '$lte':
        return param <= value;
    }
  }

  return false;
}
