/**
 * @file 校验前处理
 * @author r2space@gmail.com
 * @module sanitize
 * @version 1.0.0
 */


'use strict';

const util   = require('util')
  , debug    = require('debug')('light:validator:sanitize')
  , dayjs    = require('dayjs')
  , timezone = require('dayjs/plugin/timezone')
  , numeral  = require('numeral')
  , helper   = require('../helper')
  ;

dayjs.extend(timezone);

/**
 * 将字符串分隔成数组
 * @param data
 * @param option
 * @returns {*}
 */
exports.split = function (handler, rule) {

  debug('rule key', rule.key);

  const data = helper.get(handler.params, rule.key)
    , spliter = rule.option || /[ ]|,/;

  debug('data', data);
  debug('spliter', spliter);

  // 只对字符串类型进行处理
  if (typeof data !== 'string') {
    debug('type is not a string');
    return data;
  }

  return data.split(spliter);
};

/**
 * 数组合并成字符串
 * @param data
 * @param option
 * @returns {*}
 */
exports.join = function (handler, rule) {

  debug('rule key', rule.key);

  const data = helper.get(handler.params, rule.key)
    , spliter = rule.option || ',';

  debug('data', data);
  debug('spliter', spliter);

  // 只对数组类型进行处理
  if (!Array.isArray(data)) {
    debug('type is not an array');
    return data;
  }

  // 分隔符
  return data.join(spliter);
};

/**
 * 用固定值替换
 * @param data
 * @param option
 * @returns {*}
 */
exports.fix = function (handler, rule) {

  debug('rule key', rule.key);
  debug('value', rule.option);

  return rule.option;
};

/**
 * 进行字符串变化
 * option的值为 '1:男, 2:女'，如果给定的option不满足该格式那么不进行转换
 * @param data
 * @param option
 * @returns {*}
 */
exports.mapping = function (handler, rule) {

  debug('rule key', rule.key);

  const data = helper.get(handler.params, rule.key);
  if (!rule.option) {
    debug('rule.option is empty');
    return data;
  }

  const pair = rule.option.split(',')
    .map(item => {
      const [key, value] = item.split(':');
      return { key: key.trim(), value: value.trim() };
    })
    .find(item => String(item.key) === String(data));

  debug('option pair', pair);
  return (pair && typeof pair.value !== 'undefined')
    ? pair.value
    : data;
};

/**
 * 格式化数字
 */
exports.numberFormat = function (handler, rule) {

  debug('rule key', rule.key);
  const data = helper.get(handler.params, rule.key);

  debug('format', rule.option || '0,0.00');

  if (typeof data !== 'string' && typeof data !== 'number') {
    debug('type is invalid');
    return data;
  }

  debug('data', data);
  return numeral(data).format(rule.option || '0,0.00');
};

/**
 * 字符串格式化
 */
exports.stringFormat = function (handler, rule) {

  debug('rule key', rule.key);
  const data = helper.get(handler.params, rule.key);

  debug('format', rule.option || '%s');
  debug('data', data);
  return util.format(rule.option || '%s', data);
};

/**
 * 日期格式化
 */
exports.dateFormat = function (handler, rule) {

  debug('rule key', rule.key);
  const data = helper.get(handler.params, rule.key);

  if (!data) {
    debug('data is empty');
    return data;
  }

  const timezone = rule.timezone || handler.timezone || 'UTC'
    , format = rule.option || 'YYYY-MM-DD';

  debug('timezone', timezone);
  debug('format', format);
  debug('data', data);

  return dayjs(data).tz(timezone)
    .format(format);
};

/**
 * 去左侧空白
 */
exports.ltrim = function (handler, rule) {

  debug('rule key', rule.key);
  const data = helper.get(handler.params, rule.key);

  if (!data || typeof data !== 'string') {
    debug('type is not a string');
    return data;
  }

  debug('data', data);
  return data.trimLeft();
};

/**
 * 去空白
 */
exports.trim = function (handler, rule) {

  debug('rule key', rule.key);
  const data = helper.get(handler.params, rule.key);

  if (!data || typeof data !== 'string') {
    debug('type is not a string');
    return data;
  }

  debug('data', data);
  return data.trim();
};

/**
 * 去右侧空白
 */
exports.rtrim = function (handler, rule) {

  debug('rule key', rule.key);
  const data = helper.get(handler.params, rule.key);

  if (!data || typeof data !== 'string') {
    debug('type is not a string');
    return data;
  }

  debug('data', data);
  return data.trimRight();
};

/**
 * 转布尔
 */
exports.toBoolean = function (handler, rule) {

  debug('rule key', rule.key);
  const data = helper.get(handler.params, rule.key);

  debug('data', data);
  if (typeof data === 'string') {
    return data !== '0' && data.toLocaleLowerCase() !== 'false' && data !== '';
  }

  return !!data;
};

/**
 * 转日期
 */
exports.toDate = function (handler, rule) {

  debug('rule key', rule.key);
  const data = helper.get(handler.params, rule.key);

  debug('data', data);
  if (rule.option) {
    debug('format', rule.option);
    return dayjs(data, rule.option).toDate();
  }
  return dayjs(data).toDate();
};

/**
 * 转数字
 */
exports.toNumber = function (handler, rule) {

  debug('rule key', rule.key);
  const data = helper.get(handler.params, rule.key);

  debug('data', data);
  const result = Number(data);
  return Number.isNaN(result) ? data : result;
};

/**
 * 转字符
 */
exports.toString = function (handler, rule) {

  debug('rule key', rule.key);
  const data = helper.get(handler.params, rule.key);

  if (typeof data === 'undefined') {
    return 'undefined';
  }

  if (typeof data === 'string' || data instanceof String) {
    return data;
  }

  return JSON.stringify(data);
};
