/**
 * helper.js
 */

'use strict';

const Koa        = require('koa');
const supertest  = require('supertest');
const setting    = require('../lib/setting');
const rider      = require('../lib/model/datarider');
const pool       = require('../lib/db/mongo/pool');
const middleware = require('../lib/middleware');
const Context    = require('../lib/context');
const CONST      = require('../lib/constant');

class Helper {

  /**
   * 创建测试用koa实例，并使用指定中间件
   * 设置以下三个变量
   *  - app
   *  - server
   *
   * @param use 启用的中间件
   * @param keepcookie 设置为false，可以在每次it间不保持cookie
   */
  static initKoa(use = []) {

    // KOA实例
    this.app = new Koa();

    // 初始化加载的中间件
    middleware.init(this.app, use.map(item => ({ name: item })));

    // 启动服务开始监听端口
    this.server = this.app.listen(process.env.APPPORT || 7000);
  }

  static async initSetting(local = true) {
    if (local) {
      process.env.APPLOCAL = true;
    }
    await setting.init();
  }

  static async initRider(local = true) {
    this.initSetting(local);
    rider.init();
  }

  static initHandler(option = {
    domain: process.env.APPNAME,
    code  : CONST.DEFAULT_TENANT,
    uid   : CONST.DEFAULT_USER
  }) {
    return new Context(option.domain, option.code, option.uid);
  }

  /**
   * 清除服务
   */
  static async clear() {

    // 清理数据库连接
    await pool.destroy();

    // 清除所有mock对象
    jest.resetAllMocks();

    // 关闭koa服务
    if (this.server) {
      this.server.close();
    }
  }

  /**
   * 模拟session store
   */
  static mockSessionStore() {

    this.store = {};
    jest.mock('../../../lib/db/sessionstore', () => ({
      init: () => ({
        get: async sid => (this.store[sid] || {}).data,
        set: async (sid, data, maxAge, { changed, rolling }) => {
          if (changed || rolling) {
            this.store[sid] = { sid, data, updatedAt: new Date() };
          }
          return data;
        },
        destroy: async id => delete this.store[id]
      })
    }));

  }

  /**
   * 等待
   * @param {*} time
   */
  static async sleep(time = 1000) {
    await new Promise(resolve => setTimeout(resolve, time));
  }

  /**
   * 返回请求类
   * @param {*} keepcookie
   */
  static request(keepcookie = true) {

    // 是否每次都保持cookie
    if (keepcookie) {
      return supertest.agent(this.server);
    }
    return supertest(this.server);
  }

}

module.exports = Helper;
