/**
 * controller.test.js
 */

'use strict';

const fs         = require('fs');
const path       = require('path');
const Controller = require('../../../../lib/db/mongo/controller');
const pool       = require('../../../../lib/db/mongo/pool');
const CONST      = require('../../../../lib/constant');

describe('Controller', () => {

  beforeAll(() => {
    this.handler = { domain: CONST.SYSTEM_DB, params: {}, strict: true };
  });

  afterAll(() => {
    pool.destroy(true);
  });

  // 添加数据非严格模式
  it('add one(no stract)', async () => {

    this.handler.uid = '000000000002';
    this.handler.strict = false;
    this.handler.params = {
      data: {
        val     : 1,
        valid   : 0,
        createAt: new Date('2001/01/01'),
        createBy: '000000000003',
        updateAt: new Date('2001/01/02'),
        updateBy: '000000000004'
      }
    };

    const ctrl = new Controller(this.handler, 'test')
      , result = await ctrl.add();
    expect(result.val).toBe(1);
    expect(result.valid).toBe(0);
    expect(result.createAt.getTime()).toBe(new Date('2001/01/01').getTime());
    expect(result.createBy).toBe('000000000003');
    expect(result.updateAt.getTime()).toBe(new Date('2001/01/02').getTime());
    expect(result.updateBy).toBe('000000000004');
  });

  // 添加数据严格模式
  it('add one(stract)', async () => {

    this.handler.uid = CONST.DEFAULT_USER;
    this.handler.strict = true;
    this.handler.params = { data: { val: 1 } };

    const ctrl = new Controller(this.handler, 'test')
      , result = await ctrl.add();

    expect(result.val).toBe(1);
    expect(result.valid).toBe(1);
    expect(result.createAt.toDateString()).toBe(new Date().toDateString());
    expect(result.createBy).toBe(CONST.DEFAULT_USER);
    expect(result.updateAt.toDateString()).toBe(new Date().toDateString());
    expect(result.updateBy).toBe(CONST.DEFAULT_USER);
  });

  // 添加多条数据
  it('add many', async () => {
    this.handler.params = { data: [{ val: 2 }, { val: 3 }] };
    const ctrl = new Controller(this.handler, 'test')
      , result = await ctrl.add();
    expect(result.length).toBe(2);
  });

  // 获取一条数据
  it('get', async () => {
    this.handler.params = { data: { val: 1 } };
    const ctrl1 = new Controller(this.handler, 'test')
      , { _id: id } = await ctrl1.add();

    this.handler.params = { id, select: '_id' };
    const ctrl2 = new Controller(this.handler, 'test')
      , result = await ctrl2.get();
    expect(result._id).toEqual(id);
  });

  // 获取一条数据异常 没有此id
  it('get (data not found)', async () => {
    this.handler.params = { id: '5ea8f3a4ed935b41b5a04001', select: '_id' };
    const ctrl = new Controller(this.handler, 'test')
      , result = await ctrl.get();
    expect(result).toBe(null);
  });

  // 获取一条数据异常 id为空
  it('get (empty id)', async () => {
    this.handler.params = {};
    const ctrl = new Controller(this.handler, 'test')
      , result = await ctrl.get();
    expect(result).toBe(null);
  });

  // 获取数据列表
  it('list', async () => {
    this.handler.params = {};
    const ctrl = new Controller(this.handler, 'test')
      , result = await ctrl.list();
    expect(result.totalItems).toBeGreaterThan(0);
    expect(result.items.length).toBeGreaterThan(0);
  });

  it('list (no count)', async () => {
    this.handler.params = {};
    const ctrl = new Controller(this.handler, 'test')
      , result = await ctrl.list(false);
    expect(result.totalItems).toBe(0);
    expect(result.items.length).toBeGreaterThan(0);
  });

  // 删除数据
  it('remove', async () => {
    this.handler.params = { data: { val: 1 } };
    const ctrl1 = new Controller(this.handler, 'test')
      , { _id: id } = await ctrl1.add();

    this.handler.params = { id };
    const ctrl2 = new Controller(this.handler, 'test')
      , result = await ctrl2.remove();
    expect(result).toBe(1);
  });

  // 更新数据
  it('update', async () => {
    this.handler.params = { data: { val: 1 } };
    const ctrl1 = new Controller(this.handler, 'test')
      , { _id: id } = await ctrl1.add();

    this.handler.params = { id, data: { val: 4 } };
    const ctrl2 = new Controller(this.handler, 'test')
      , result = await ctrl2.update();
    expect(result.val).toBe(4);
  });

  // 通过condition查询，添加数据
  it('upsert add by condition', async () => {
    this.handler.params = { condition: { val: 8 }, data: { val: 5 } };
    const ctrl = new Controller(this.handler, 'test')
      , result = await ctrl.upsert();
    expect(result.val).toBe(5);
  });

  // 通过condition查询，修改数据
  it('upsert update by condition', async () => {
    this.handler.params = { condition: { val: 5 }, data: { val: 55 } };
    const ctrl = new Controller(this.handler, 'test')
      , result = await ctrl.upsert();
    expect(result).toBeGreaterThan(0);
  });

  // 通过id查询，添加数据
  it('upsert add by id', async () => {
    this.handler.params = { id: '5ea8f3bc88baf241cc31ed00', data: { val: 6 } };
    const ctrl = new Controller(this.handler, 'test')
      , result = await ctrl.upsert();
    expect(result.val).toBe(6);
  });

  // 通过id查询，修改数据
  it('upsert update by id', async () => {
    this.handler.params = { data: { val: 1 } };
    const ctrl1 = new Controller(this.handler, 'test')
      , { _id: id } = await ctrl1.add();

    this.handler.params = { id, data: { val: 6 } };
    const ctrl = new Controller(this.handler, 'test')
      , result = await ctrl.upsert();
    expect(result.val).toBe(6);
  });

  // 聚合查询
  it('aggregate', async () => {
    const ctrl = new Controller({ domain: CONST.SYSTEM_DB }, 'test')
      , result = await ctrl.aggregate([
        { $match: { val: { $ne: 0 } } },
        { $group: { _id: '1', max: { $max: '$val' } } }
      ]);
    expect(result[0].max).toBeGreaterThan(1);
    expect(result[0]._id).toBe('1');
  });

  // 文件写入到GridFS异常，没有文件
  it('file to grid (empty file)', async () => {
    this.handler.params.files = [];
    const ctrl = new Controller(this.handler, 'test')
      , result = await ctrl.fileToGrid();
    expect(result.totalItems).toBe(0);
  });

  // Buffer写入GridFS
  it('file to grid (buffer)', async () => {
    this.handler.params = {
      files: [{
        buffer  : Buffer.from('string'),
        name    : 'LICENSE',
        metadata: { kind: 'file' },
        type    : 'application/octet-stream'
      }]
    };
    const ctrl = new Controller(this.handler, 'test')
      , result = await ctrl.fileToGrid();
    expect(result.totalItems).toBe(1);
    expect(result.items[0].metadata).toEqual({ kind: 'file' });
  });

  // Stream写入GridFS
  it('file to grid (stream)', async () => {
    this.handler.params = {
      files: [{
        name    : 'LICENSE',
        stream  : fs.createReadStream(path.resolve('LICENSE')),
        metadata: { kind: 'file' },
        type    : 'application/octet-stream'
      }]
    };
    const ctrl = new Controller(this.handler, 'test')
      , result = await ctrl.fileToGrid();
    expect(result.totalItems).toBe(1);
    expect(result.items[0].metadata).toEqual({ kind: 'file' });
  });

  // 文件写入GridFS
  it('file to grid (file)', async () => {
    this.handler.params = {
      files: [{
        name    : 'LICENSE',
        path    : 'LICENSE',
        metadata: { kind: 'file' },
        type    : 'text/plain'
      }]
    };
    const ctrl = new Controller(this.handler, 'test')
      , result = await ctrl.fileToGrid();
    expect(result.totalItems).toBe(1);
    expect(result.items[0].metadata).toEqual({ kind: 'file' });
  });

  // GridFS读取文件到Stream
  it('gridToStream', async done => {
    this.handler.params = { files: [{ name: 'LICENSE', path: 'LICENSE' }] };
    const ctrl1 = new Controller(this.handler, 'test')
      , { items } = await ctrl1.fileToGrid();

    this.handler.params = { id: items[0]._id };
    const ctrl2 = new Controller(this.handler, 'test')
      , result = await ctrl2.gridToStream();

    let str = '';
    result.stream.on('data', data => str += data);
    result.stream.on('end', () => {
      expect(result.length).toBe(process.platform === 'darwin' ? 1077 : 1097);
      done();
    });
    expect(result.length).toBe(process.platform === 'darwin' ? 1077 : 1097);
    expect(result.filename).toBe('LICENSE');
    expect(result.contentType).toBe('application/octet-stream');
  });

  // GridFS读取一段文件到Stream
  it('gridToStreamInRange', async done => {

    this.handler.params = { files: [{ name: 'LICENSE', path: 'LICENSE' }] };
    const ctrl1 = new Controller(this.handler, 'test')
      , { items } = await ctrl1.fileToGrid();

    this.handler.params = { id: items[0]._id };
    const ctrl2 = new Controller(this.handler, 'test')
      , result = await ctrl2.gridToStream(59, 87);

    let str = '';
    result.stream.on('data', data => str += data);
    result.stream.on('end', () => {
      expect(str.length).toBe(28);
      done();
    });
    expect(result.length).toBe(process.platform === 'darwin' ? 1077 : 1097);
    expect(result.filename).toBe('LICENSE');
    expect(result.streamRange).toEqual({ start: 59, end: 87 });
    expect(result.contentType).toBe('application/octet-stream');
  });

  // 获取数据库进程信息
  it('statsServer', async () => {
    const ctrl = new Controller({
        domain: 'admin',
        db    : {
          user: process.env.ADMIN_ACCOUNT,
          pass: process.env.ADMIN_PASSWORD
        }
      })
      , result = await ctrl.statsServer();
    expect(result.ok).toBe(1);
  });

  // 获取数据库信息
  it('statsDb', async () => {
    const ctrl = new Controller({
        domain: 'admin',
        db    : {
          user: process.env.ADMIN_ACCOUNT,
          pass: process.env.ADMIN_PASSWORD
        }
      })
      , result = await ctrl.statsDb();
    expect(result.ok).toBe(1);
  });

  // 添加用户
  it('addUser', async () => {
    const ctrl = new Controller({
        params: { data: { user: 'unit_test_user', pass: '123456', domain: 'light' } }
      })
      , result = await ctrl.addUser();
    expect(result.ok).toBe(1);
  });

  // 删除用户
  it('dropUser', async () => {
    const ctrl = new Controller({
        params: { condition: { user: 'unit_test_user', domain: 'light' } }
      })
      , result = await ctrl.dropUser();
    expect(result).toBeTruthy();
  });

  // 执行命令
  it('runCommand', async () => {
    const ctrl = new Controller({
        params: { condition: { command: { ping: 1 } } }
      })
      , result = await ctrl.runCommand();
    expect(result.ok).toBe(1);
  });

});
