/**
 * mapping.test.js
 */

'use strict';

const mp = require('../../../../lib/db/mongo/mapping');

describe('Mapping', () => {

  // 基础方法
  describe('basic', () => {

    // 获取定义
    it('find define', () => {

      let define = mp.find('a', {});
      expect(define).toEqual({});

      define = mp.find('a', { a: { type: 'number' } });
      expect(define).toEqual({ type: 'number' });

      define = mp.find('a.b', { a: { contents: { b: { type: 'number' } } } });
      expect(define).toEqual({ type: 'number' });

      define = mp.find('a.b.c', { a: { contents: { b: {} } } });
      expect(define).toEqual({});

      define = mp.find('a.c', { a: { contents: { b: {} } } });
      expect(define).toEqual({});

      define = mp.find('a.1.b', { a: { contents: { b: { type: 'number' } } } });
      expect(define).toEqual({ type: 'number' });

    });

    // 基本数据的转换
    it('basic data parse', () => {
      const define = {
        a: { type: 'number' },
        b: { type: 'object', contents: { b1: { type: 'number' } } },
        c: { type: 'array', contents: 'number' },
        d: { type: 'array', contents: { d1: { type: 'number' }, d2: { type: 'object' }, d3: { type: 'array' } } },
        e: { type: 'object', contents: { e1: { type: 'number' } } }
      };
      const data = {
        a     : '10',
        b     : { b1: '10' },
        c     : ['1', '2', '3'],
        d     : [{ d1: '10', d2: { d21: '10' }, d3: [{}] }],
        'e.e1': '10'
      };
      const result = {
        a     : 10,
        b     : { b1: 10 },
        c     : [1, 2, 3],
        d     : [{ d1: 10, d2: { d21: '10' }, d3: [{}] }],
        'e.e1': 10
      };
      expect(mp.parse(data, define)).toEqual(result);
    });

  });

  // 查询系
  describe('query', () => {

    // 比较操作符
    it('$eq', () => {
      const define = {
        a: { type: 'number' },
        b: { type: 'object', contents: { b1: { type: 'number' } } },
        c: { type: 'array', contents: 'number' },
        d: { type: 'object' },
        f: { type: 'array' }
      };
      const data = {
        a: { $eq: '10' },
        b: { $eq: { b1: '10' } },     // 指定了JSON元素类型
        c: { $eq: ['1', '2', '3'] },  // 指定了数组元素类型
        d: { $eq: { d1: '10' } },     // 没有元素的类型，不转换
        f: { $eq: ['1', '2'] },       // 没有元素的类型，不转换
        z: '10'                       // 因为没有定义，所以会被删除的项目
      };
      const result = {
        a: { $eq: 10 },
        b: { $eq: { b1: 10 } },
        c: { $eq: [1, 2, 3] },
        d: { $eq: { d1: '10' } },
        f: { $eq: ['1', '2'] }
      };
      expect(mp.parse(data, define)).toEqual(result);
    });

    it('$in', () => {
      const define = {
        a: { type: 'number' },
        b: { type: 'string' },
        c: { type: 'object', contents: { c1: { type: 'number' } } },
        d: { type: 'array', contents: 'number' },
        f: { type: 'array', contents: { f1: { type: 'number' } } }
      };
      const data = {
        a: { $in: ['1'] },
        b: { $in: [/^.*$/] },
        c: { $in: [{ c1: '1' }] },
        d: { $in: ['10'] },
        f: { $in: [{ f1: '10' }] }
      };
      const result = {
        a: { $in: [1] },
        b: { $in: [/^.*$/] },
        c: { $in: [{ c1: 1 }] },
        d: { $in: [10] },
        f: { $in: [{ f1: 10 }] }
      };
      expect(mp.parse(data, define)).toEqual(result);
    });

    // 逻辑操作符
    it('$and', () => {
      const define = {
        a: { type: 'number' },
        b: { type: 'number' },
        c: { type: 'number' },
        d: { type: 'number' },
        e: { type: 'array', contents: 'number' },
        f: { type: 'object', contents: { f1: { type: 'number' } } }
      };
      const data = {
        $and: [
          { a: '10' },
          { b: { $gt: '10' } },
          { $or: [{ c: '10' }, { d: '10' }] },
          { e: { $in: ['1', '2'] } },
          { f: { f1: '10' } }
        ]
      };
      const result = {
        $and: [
          { a: 10 },
          { b: { $gt: 10 } },
          { $or: [{ c: 10 }, { d: 10 }] },
          { e: { $in: [1, 2] } },
          { f: { f1: 10 } }
        ]
      };
      expect(mp.parse(data, define)).toEqual(result);
    });

    // 元素操作符
    it('$exists', () => {
      const define = {
        a: { type: 'number' },
        b: { type: 'string' },
        c: { type: 'object', contents: { c1: { type: 'number' } } }
      };
      const data = {
        a     : { $exists: 1 },
        b     : { $exists: null },
        'c.c1': { $exists: '1' }
      };
      const result = {
        a     : { $exists: true },
        b     : { $exists: null },
        'c.c1': { $exists: true }
      };
      expect(mp.parse(data, define)).toEqual(result);
    });

    it('$type', () => {
      const define = {
        a: { type: 'number' },
        b: { type: 'string' },
        c: { type: 'object', contents: { c1: { type: 'number' } } }
      };
      const data = {
        a     : { $type: 1 },
        b     : { $type: 'string' },
        'c.c1': { $type: ['string', 'number'] }
      };
      const result = {
        a     : { $type: 1 },
        b     : { $type: 'string' },
        'c.c1': { $type: ['string', 'number'] }
      };
      expect(mp.parse(data, define)).toEqual(result);
    });

    // 评估操作符
    it('$expr', () => {
      const define = {};
      const data = { $expr: { $gt: ['$a', '$b'] } };
      const result = { $expr: { $gt: ['$a', '$b'] } };
      expect(mp.parse(data, define)).toEqual(result);
    });

    it('$jsonSchema', () => {
      const define = {};
      const data = { $jsonSchema: { $required: ['a', 'b'] } };
      const result = { $jsonSchema: { $required: ['a', 'b'] } };
      expect(mp.parse(data, define)).toEqual(result);
    });

    it('$mod', () => {
      const define = {};
      const data = { $mod: ['3', '0'] };
      const result = {  $mod: [3, 0] };
      expect(mp.parse(data, define)).toEqual(result);
    });

    it('$text', () => {
      const define = {};
      const data = { $text: { $search: 'str' } };
      const result = { $text: { $search: 'str' } };
      expect(mp.parse(data, define)).toEqual(result);
    });

    it('$regex', () => {
      const define = {
        a: { type: 'string' },
        b: { type: 'string' },
        c: { type: 'string' },
        d: { type: 'string' },
        e: { type: 'string' }
      };
      const data = {
        a: { $regex: '^$' },
        b: { $regex: RegExp('^$') },
        c: { $regex: /^$/ },
        d: { $regex: /^$/i },
        e: { $regex: /^$/, $options: 'i' }
      };
      const result = {
        a: { $regex: /^$/i },
        b: { $regex: /^$/ },
        c: { $regex: /^$/ },
        d: { $regex: /^$/i },
        e: { $regex: /^$/, $options: 'i' }
      };
      expect(mp.parse(data, define)).toEqual(result);
    });

    // 数组操作符
    it('$all', () => {
      const define = {
        a: { type: 'array', contents: 'number' },
        b: { type: 'array', contents: { b1: { type: 'number' } } }
      };
      const data = {
        a: { $all: ['1', '2'] },
        b: { $all: [{ b1: '10' }] }
      };
      const result = {
        a: { $all: [1, 2] },
        b: { $all: [{ b1: 10 }] }
      };
      expect(mp.parse(data, define)).toEqual(result);
    });

    it('$elemMatch', () => {
      const define = {
        a: { type: 'array', contents: 'number' },
        b: { type: 'array', contents: { b1: { type: 'string' }, b2: { type: 'number' } } },
        c: { type: 'object', contents: { c1: { type: 'array', contents: 'number' } } }
      };
      const data = {
        a     : { $elemMatch: { $gt: '10', $lt: '20' } },
        b     : { $elemMatch: { b1: '1', b2: { $gt: '20' } } },
        'c.c1': { $elemMatch: { $gt: '10', $lt: '20' } }
      };
      const result = {
        a     : { $elemMatch: { $gt: 10, $lt: 20 } },
        b     : { $elemMatch: { b1: '1', b2: { $gt: 20 } } },
        'c.c1': { $elemMatch: { $gt: 10, $lt: 20 } }
      };
      expect(mp.parse(data, define)).toEqual(result);
    });

    it('$size', () => {
      const define = {
        a: { type: 'array', contents: 'number' },
        b: { type: 'array', contents: { b1: { type: 'array' } } }
      };
      const data = {
        a     : { $size: '2' },
        'b.b1': { $size: '10' }
      };
      const result = {
        a     : { $size: 2 },
        'b.b1': { $size: 10 }
      };
      expect(mp.parse(data, define)).toEqual(result);
    });

    // Projection
    it('$meta', () => {
      const define = {};
      const data = { $meta: 'textScore' };
      const result = { $meta: 'textScore' };
      expect(mp.parse(data, define)).toEqual(result);
    });

  });

  // 更新系
  describe('update', () => {

    it('$currentDate', () => {
      const define = {
        a: { },
        b: { },
        c: { }
      };
      const data = {
        $currentDate: {
          a: 'true',
          b: { $type: 'timestamp' },
          c: { $type: 'date' }
        }
      };
      const result = {
        $currentDate: {
          a: true,
          b: { $type: 'timestamp' },
          c: { $type: 'date' }
        }
      };
      expect(mp.parse(data, define)).toEqual(result);
    });

    it('$inc', () => {
      const define = {
        a: { },
        b: { },
        c: { },
        d: { type: 'object', contents: { d1: {} } }
      };
      const data = {
        $inc: {
          a     : '-2',
          b     : 'str',
          c     : 10,
          'd.d1': '1'
        }
      };
      const result = {
        $inc: {
          a     : -2,
          b     : null,
          c     : 10,
          'd.d1': 1
        }
      };
      expect(mp.parse(data, define)).toEqual(result);
    });

    it('$max', () => {
      const define = {
        a: { type: 'object', contents: { a1: { type: 'number' } } },
        b: { type: 'number' },
        c: { type: 'string' },
        d: { type: 'date' }
      };
      const data = {
        $max: {
          'a.a1': '-2',
          b     : '1',
          c     : 'str',
          d     : '2020-01-01'
        }
      };
      const result = {
        $max: {
          'a.a1': -2,
          b     : 1,
          c     : 'str',
          d     : new Date('2020-01-01')
        }
      };
      expect(mp.parse(data, define)).toEqual(result);
    });

    it('$mul', () => {
      const define = {
        a: { type: 'object', contents: { a1: { type: 'number' } } },
        b: { type: 'number' }
      };
      const data = {
        $mul: {
          'a.a1': '-2',
          b     : '1'
        }
      };
      const result = {
        $mul: {
          'a.a1': -2,
          b     : 1
        }
      };
      expect(mp.parse(data, define)).toEqual(result);
    });

    it('$rename', () => {
      const define = {
        a: { type: 'object', contents: { a1: { type: 'number' } } },
        b: { type: 'number' }
      };
      const data = {
        $rename: {
          'a.a1': 10,
          b     : '1'
        }
      };
      const result = {
        $rename: {
          'a.a1': '10',
          b     : '1'
        }
      };
      expect(mp.parse(data, define)).toEqual(result);
    });

    it('$pull', () => {
      const define = {
        a: {  // { a: { a1: [1, 2, 3] } }
          type    : 'object',
          contents: { a1: { type: 'array', contents: 'number' } }
        },
        b: {  // { b: [{ b1: 1, b2: 2 }, { b1: 2, b2: 3 }] }
          type    : 'array',
          contents: { b1: { type: 'number' }, b2: { type: 'number' } }
        },
        c: {  // { c: { c1: [{ c11: 1 }, { c11: 2 }] } }
          type    : 'object',
          contents: { c1: { type: 'array', contents: { c11: { type: 'number' } } } }
        },
        d: {  // { d: ['1', '2', '3'] }
          type    : 'array',
          contents: 'string'
        },
        e: {  // { e: [{e1: 8, e2: 'b'}, {e1: 7, e2: 'b'}, {e1: 8, e2: 'a'}] }
          type    : 'array',
          contents: { e1: { type: 'number' }, e2: { type: 'string' } }
        }
      };
      const data = {
        $pull: {
          'a.a1'    : { $gt: '1' },
          'b.b1'    : { $in: ['1', 2] },
          b         : { b1: '1' },
          'c.c1.c11': '1',
          d         : 10,
          e         : { $elemMatch: { e1: { $gt: '8' }, e2: 'B' } }
        }
      };
      const result = {
        $pull: {
          'a.a1'    : { $gt: 1 },
          'b.b1'    : { $in: [1, 2] },
          b         : { b1: 1 },
          'c.c1.c11': 1,
          d         : '10',
          e         : { $elemMatch: { e1: { $gt: 8 }, e2: 'B' } }
        }
      };
      expect(mp.parse(data, define)).toEqual(result);
    });

    it('$addToSet', () => {
      const define = {
        a: { type: 'array', contents: 'number' },
        b: { type: 'array', contents: { b1: { type: 'number' } } },
        c: { type: 'array', contents: 'number' }
      };
      const data = {
        $addToSet: {
          a: '1',
          b: { b1: '10' },
          c: { $each: ['1', '2'] }
        }
      };
      const result = {
        $addToSet: {
          a: 1,
          b: { b1: 10 },
          c: { $each: [1, 2] }
        }
      };
      expect(mp.parse(data, define)).toEqual(result);
    });

    it('$pullAll', () => {
      const define = {
        a: { type: 'array', contents: 'number' },
        b: { type: 'array', contents: { b1: { type: 'number' } } }
      };
      const data = {
        $pullAll: {
          a: ['1', '2'],
          b: [{ b1: '10' }]
        }
      };
      const result = {
        $pullAll: {
          a: [1, 2],
          b: [{ b1: 10 }]
        }
      };
      expect(mp.parse(data, define)).toEqual(result);
    });

    it('$position', () => {
      const define = {
        a: { type: 'array', contents: 'number' }
      };
      const data = {
        $push: {
          a: {
            $each    : ['1', '2'],
            $position: '-2'
          }
        }
      };
      const result = {
        $push: {
          a: {
            $each    : [1, 2],
            $position: -2
          }
        }
      };
      expect(mp.parse(data, define)).toEqual(result);
    });

    it('$sort', () => {
      const define = {
        a: { type: 'array', contents: 'number' },
        b: { type: 'array', contents: { b1: { type: 'string' } } }
      };
      const data = {
        $push: {
          a: { $each: ['1', '2'], $sort: '1' },
          b: { $each: [], $sort: { b1: '-1' } }
        }
      };
      const result = {
        $push: {
          a: { $each: [1, 2], $sort: 1 },
          b: { $each: [], $sort: { b1: -1 } }
        }
      };
      expect(mp.parse(data, define)).toEqual(result);
    });

    it('$bit', () => {
      const define = {
        a: { type: 'number' },
        b: { type: 'number' },
        c: { type: 'number' }
      };
      const data = {
        $bit: {
          a: { and: '1' },
          b: { or: '2' },
          c: { xor: '4' }
        }
      };
      const result = {
        $bit: {
          a: { and: 1 },
          b: { or: 2 },
          c: { xor: 4 }
        }
      };
      expect(mp.parse(data, define)).toEqual(result);
    });
  });

  describe('set default', () => {

    it('no original value', () => {
      const define = {
        a: { type: 'number', default: '1' },
        b: { type: 'array', default: '[1,2,3]' },
        c: { type: 'array', contents: { c1: { type: 'number', default: '1' } } },
        d: {
          type    : 'object',
          contents: {
            d1: { type: 'number', default: '1' },
            d2: {
              type    : 'object',
              contents: { d21: { type: 'number', default: '1' } }
            }
          }
        },
        e: { type: 'object', default: 'null' },
        f: { type: 'object', default: 'undefined' }
      };
      const data = {
        c: [{}, {}],
        d: { }
      };
      const result = {
        a: 1,
        b: [1, 2, 3],
        c: [{ c1: 1 }, { c1: 1 }],
        d: { d1: 1, d2: { d21: 1 } },
        e: null,
        f: undefined
      };
      expect(mp.defaults(data, define)).toEqual(result);
    });

    it('has original value', () => {
    });

  });

});
