/**
 * model.test.js
 */

'use strict';

const fs      = require('fs');
const Model   = require('../../../../lib/db/mongo/model');
const setting = require('../../../../lib/setting');
const CONST   = require('../../../../lib/constant');
const helper  = require('../../../helper');

describe('Model', () => {

  beforeAll(() => {
    this.model = new Model(
      CONST.SYSTEM_DB,
      CONST.DEFAULT_TENANT,
      'test',
      { transaction: { use: false } }
    );

    this.admin = new Model('admin', null, null, {
      user: process.env.ADMIN_ACCOUNT,
      pass: process.env.ADMIN_PASSWORD
    });
  });

  afterAll(() => helper.clear());

  // 获取表名
  it('code（no struct define）', () => {
    expect(new Model().code).toBe('');
    expect(new Model('test').code).toBe('');
    expect(new Model('test', null, 'test').code).toBe('tests');
    expect(new Model('test', 'default', 'test').code).toBe('default.tests');
    expect(new Model('test', null, 'board').code).toBe('light.boards');
  });

  // 获取表名（有表定义）
  it('code（has struct define）', () => {

    setting.struct = [{ schema: 'child', parent: 'parent', items: [] }];
    expect(new Model('test', 'default', 'child').code).toBe('default.parents');

    setting.struct = [{ schema: 'child', items: [] }];
    expect(new Model('test', 'default', 'child').code).toBe('default.children');
  });

  // 插入异常，数据为空
  it('add empty', done => {
    this.model.add()
      .catch(error => expect(error.message).toBe('Document cannot be empty'))
      .finally(done);
  });

  // 插入异常，数组为空
  it('add empty array', done => {
    this.model.add([])
      .catch(error => expect(error.message).toBe('Document cannot be empty'))
      .finally(done);
  });

  // 插入单行
  it('add one', async () => {
    const res = await this.model.add({ val: 1 });
    expect(res.val).toBe(1);
  });

  // 插入多行
  it('add many', async () => {
    const res = await this.model.add([{ val: 2 }, { val: 3 }]);
    expect(res.length).toBe(2);
  });

  // 获取单行异常，条件为空
  it('get by empty', done => {
    this.model.get()
      .catch(error => expect(error.message).toBe('Filter is incorrect'))
      .finally(done);
  });

  // 获取单行异常，条件为空对象
  it('get by empty object', done => {
    this.model.get({})
      .catch(error => expect(error.message).toBe('Filter is incorrect'))
      .finally(done);
  });

  // 获取单行异常，不正确的ID
  it('get by bad id', done => {
    this.model.get(1).catch(error => {
      expect(error.message).toBe('Filter is incorrect');
      done();
    });
  });

  // 获取单行
  it('get by object id', async () => {
    let res = await this.model.get(this.id);
    expect(res.val).toBe(1);

    res = await this.model.get(this.id.toString());
    expect(res.val).toBe(1);

    res = await this.model.get({ val: 1 });
    expect(res.val).toBe(1);
  });

  // 获取多行
  it('list', async () => {

    await this.model.add({ val: 1 });

    const res1 = await this.model.list({ val: 1 });
    expect(res1.length).toBeGreaterThan(0);

    const res2 = await this.model.list({ val: 1 }, 1);
    expect(res2.length).toBe(res1.length - 1);

    const res3 = await this.model.list({ val: 1 }, 1, 1);
    expect(res3.length).toBe(1);

    const res4 = await this.model.list({ val: 1 }, 0, Number.MAX_VALUE);
    expect(res4.length).toBe(res1.length);
  });

  // 获取多行（选择特定列）
  it('list (projection)', async () => {
    const [res1] = await this.model.list({ val: 1 }, 0, 1, null, '_id');
    expect(res1).not.toHaveProperty('val');
    expect(res1).toHaveProperty('_id');

    const [res2] = await this.model.list({ val: 1 }, 0, 1, null, '_id,val');
    expect(res2).toHaveProperty('val');
    expect(res2).toHaveProperty('_id');

    const [res3] = await this.model.list({ val: 1 }, 0, 1, null, '_id val');
    expect(res3).toHaveProperty('val');
    expect(res3).toHaveProperty('_id');

    const [res4] = await this.model.list({ val: 1 }, 0, 1, null, { _id: 1 });
    expect(res4).not.toHaveProperty('val');
    expect(res4).toHaveProperty('_id');
  });

  // 获取多行（排序）
  it('list (sort)', async () => {
    const [res1] = await this.model.list({ val: 1 }, 0, 2, '_id');
    expect(res1).toHaveProperty('_id');

    const [res2] = await this.model.list({ val: 1 }, 0, 1, [['_id', 1]]);
    expect(res2).toEqual(res1);

    const [res3] = await this.model.list({ val: 1 }, 0, 1, [['_id', -1]]);
    expect(res3._id > res1._id).toBeTruthy();

    const [res4] = await this.model.list({ val: 1 }, 0, 1, { _id: -1 });
    expect(res4).toEqual(res3);

    const [res5] = await this.model.list({ val: 1 }, 0, 1, { _id: 'desc' });
    expect(res5).toEqual(res3);

    const [res6] = await this.model.list({ val: 1 }, 0, 1, { _id: 1 });
    expect(res6).toEqual(res1);

    const [res7] = await this.model.list({ val: 1 }, 0, 1, { _id: 'asc' });
    expect(res7).toEqual(res1);
  });

  // 更新
  it('update', async () => {
    const { _id } = await this.model.add({ val: 1 });
    expect(_id).toBeTruthy();

    const res1 = await this.model.update(_id, { author: 'cells' });
    expect(res1.author).toBe('cells');

    const res2 = await this.model.update({ author: 'cells' }, { author: 'new' });
    expect(res2).toBe(1);
  });

  // 更新异常，条件为空
  it('update by empty filter', done => {
    this.model.update()
      .catch(error => expect(error.message).toBe('Filter is incorrect'))
      .finally(done);
  });

  // 更新异常，数据为空
  it('update by empty data', done => {
    this.model.update(1)
      .catch(error => expect(error.message).toBe('Document cannot be empty'))
      .finally(done);
  });

  // 删除
  it('remove', async () => {
    const { _id: id1 } = await this.model.add({ val: 1 });
    expect(id1).toBeTruthy();
    const res1 = await this.model.remove(id1, { val: 2 });
    expect(res1).toBe(1);

    const res2 = await this.model.get(id1);
    expect(res2.val).toBe(2);
    expect(res2.valid).toBe(0);

    const { _id: id2 } = await this.model.add({ val: 1 });
    expect(id2).toBeTruthy();
    const res3 = await this.model.remove({ val: 1 });
    expect(res3).toBeGreaterThan(1);
  });

  // 更新插入
  it('upsert', async () => {

    // 插入
    const res1 = await this.model.upsert({ val: 3 }, { val: 3 });
    expect(res1.val).toBe(3);

    // ID更新
    const res2 = await this.model.upsert(res1._id, { val: 4 });
    expect(res2.val).toBe(4);

    // 条件更新
    const res3 = await this.model.upsert({ val: 4 }, { val: 5 });
    expect(res3).toBe(1);
  });

  // 更新插入异常，条件为空
  it('upsert by empty filter', done => {
    this.model.upsert()
      .catch(error => expect(error.message).toBe('Filter is incorrect'))
      .finally(done);
  });

  // 更新插入异常，数据为空
  it('upsert by empty data', done => {
    this.model.upsert(1)
      .catch(error => expect(error.message).toBe('Document cannot be empty'))
      .finally(done);
  });

  // 更新插入异常，数据不能是数组
  it('upsert by array data', done => {
    this.model.upsert(1, [])
      .catch(error => expect(error.message).toBe('Document cannot be array'))
      .finally(done);
  });

  // 获取件数
  it('count', async () => {
    const { _id } = await this.model.add({ val: 1 });
    expect(_id).toBeTruthy();

    const result = await this.model.count(_id);
    expect(result).toBe(1);
  });

  // 去重
  it('distinct', async () => {
    const res = await this.model.distinct({}, 'val');
    expect(res).toContain(1);
  });

  // 去重异常，未指定key
  it('distinct by empty key', done => {
    this.model.distinct()
      .catch(error => expect(error.message).toBe('Distinct key is incorrect'))
      .finally(done);
  });

  // Aggregate
  it('aggregate', async () => {
    const res = await this.model.aggregate([
      { $match: { val: { $ne: 0 } } },
      { $group: { _id: '1', max: { $max: '$val' } } }
    ], {});
    expect(res[0].max).toBeGreaterThan(1);
    expect(res[0]._id).toBe('1');
  });

  // 执行命令
  it('execute command', async () => {
    const res = await this.model.command({ ping: 1 });
    expect(res.ok).toBe(1);
  });

  // 重命名
  it('rename collection', async () => {
    const res1 = await this.model.rename('tests.new');
    expect(res1).toBeTruthy();
    await this.model.rename('tests');
  });

  // 事物处理（单个model）
  it('transaction(single model)', async () => {
    const m1 = new Model(
      CONST.SYSTEM_DB,
      CONST.DEFAULT_TENANT,
      'test',
      { transaction: { use: true, session: null } }
    );
    const m2 = new Model(
      CONST.SYSTEM_DB,
      CONST.DEFAULT_TENANT,
      'test',
      { transaction: { use: false } }
    );

    // 添加两行数据
    await m1.add({ val: 100 });
    await m1.add({ val: 100 });

    // 事物内确认添加的数据
    const res1 = await m1.list({ val: 100 });
    expect(res1.length).toBe(2);

    // 回滚
    await m1.abort();

    // 确认数据没有被更新
    const res2 = await m2.list({ val: 100 });
    expect(res2.length).toBe(0);
  });

  // 事物处理（多个model使用同一个事物）
  it('transaction(multiple model)', async () => {

    const options = { transaction: { use: true, session: null } }
    const m1 = new Model(
      CONST.SYSTEM_DB,
      CONST.DEFAULT_TENANT,
      'test',
      options
    );
    const m2 = new Model(
      CONST.SYSTEM_DB,
      CONST.DEFAULT_TENANT,
      'test',
      options
    );
    const m3 = new Model(CONST.SYSTEM_DB, CONST.DEFAULT_TENANT, 'test');

    // m1, m2使用同一个session
    await m1.add({ val: 200 });
    await m2.add({ val: 200 });

    // m2提交
    await m2.commit();

    // 确认m1的操作也被提交
    const res = await m3.list({ val: 200 });
    expect(res.length).toBeGreaterThan(2);
  });

  // Buffer写到GridFS里
  it('buffer to grid', async () => {
    const result = await this.model.bufferToGrid({
      name    : 'buffer',
      buffer  : Buffer.from('string'),
      type    : 'text/plan',
      metadata: { kind: 'folder' }
    });
    expect(result._id).toBeTruthy();
    expect(result.length).toBe(6);
    expect(result.filename).toBe('buffer');
    expect(result.contentType).toBe('text/plan');
    expect(result.metadata).toEqual({ kind: 'folder' });
  });

  // 流写到GridFS里
  it('stream to grid', async () => {
    const result = await this.model.streamToGrid({
      name    : 'LICENSE',
      stream  : fs.createReadStream('LICENSE'),
      type    : 'text/plain',
      metadata: { kind: 'folder' }
    });
    expect(result._id).toBeTruthy();
    expect(result.length).toBe(process.platform === 'darwin' ? 1077 : 1097);
    expect(result.filename).toBe('LICENSE');
    expect(result.contentType).toBe('text/plain');
    expect(result.metadata).toEqual({ kind: 'folder' });
  });

  // 文件写到GridFS里
  it('file to grid', async () => {
    const result = await this.model.fileToGrid({
      name    : 'LICENSE',
      path    : 'LICENSE',
      type    : 'text/plain',
      metadata: { kind: 'folder' }
    });
    expect(result._id).toBeTruthy();
    expect(result.length).toBe(process.platform === 'darwin' ? 1077 : 1097);
    expect(result.filename).toBe('LICENSE');
    expect(result.contentType).toBe('text/plain');
    expect(result.metadata).toEqual({ kind: 'folder' });
  });

  // GridFS文件输出到物理文件
  it('grid to file', async () => {
    const { _id } = await this.model.fileToGrid({ name: 'LICENSE', path: 'LICENSE' })
      , result = await this.model.gridToFile(_id);

    expect(result._id).toBeTruthy();
    expect(fs.existsSync(result.path)).toBeTruthy();
    fs.unlinkSync(result.path);
  });

  // GridFS文件输出到物理文件异常，文件找不到时
  it('grid to file (file not found)', done => {
    this.model.gridToFile('000000000000', 'output')
      .catch(error => {
        expect(error.message).toBe('File not found');
        done();
      });
  });

  // GridFS文件输出到物理文件异常，文件id没有时
  it('grid to file (empty id)', done => {
    this.model.gridToFile()
      .catch(error => {
        expect(error.message).toBe('File id cannot be empty');
        done();
      });
  });

  // GridFS文件输出到物理文件，输出物理文件没有时
  it('grid to file (empty file)', done => {
    this.model.gridToFile('000000000000')
      .catch(error => {
        expect(error.message).toBe('File not found');
        done();
      });
  });

  // 文件读到内存
  it('grid to buffer', async () => {
    const { _id } = await this.model.fileToGrid({ name: 'LICENSE', path: 'LICENSE' })
      , result = await this.model.gridToBuffer(_id);
    expect(result.length).toBe(process.platform === 'darwin' ? 1077 : 1097);
    expect(result.filename).toBe('LICENSE');
    expect(result.contentType).toBe('application/octet-stream');
    expect(result.buffer.length).toBe(1077);
  });

  // 文件读到内存异常，文件id没有时
  it('grid to buffer (File id cannot be empty)', done => {
    this.model.gridToBuffer()
      .catch(error => {
        expect(error.message).toBe('File id cannot be empty');
        done();
      });
  });

  // 文件读到内存，文件找不到时
  it('grid to buffer (File not found)', done => {
    this.model.gridToBuffer('000000000000')
      .catch(error => {
        expect(error.message).toBe('File not found');
        done();
      });
  });

  // 文件读到Stream
  it('grid to stream', async done => {
    const { _id } = await this.model.fileToGrid({ name: 'LICENSE', path: 'LICENSE' })
      , result = await this.model.gridToStream(_id);

    let str = '';
    result.stream.on('data', data => str += data);
    result.stream.on('end', () => {
      expect(str.length).toBe(process.platform === 'darwin' ? 1077 : 1097);
      done();
    });

    expect(result.length).toBe(process.platform === 'darwin' ? 1077 : 1097);
    expect(result.filename).toBe('LICENSE');
    expect(result.contentType).toBe('application/octet-stream');
  });

  // 文件读到Stream异常，文件id没有时
  it('grid to stream (File id cannot be empty)', done => {
    this.model.gridToStream()
      .catch(error => {
        expect(error.message).toBe('File id cannot be empty');
        done();
      });
  });

  // 文件读到Stream异常，文件找不到时
  it('grid to stream (File not found)', done => {
    this.model.gridToStream('000000000000')
      .catch(error => {
        expect(error.message).toBe('File not found');
        done();
      });
  });

  // 分段读取返回Stream正常
  it('grid to stream in range', async done => {
    const { _id } = await this.model.fileToGrid({ name: 'LICENSE', path: 'LICENSE' })
      , result = await this.model.gridToStreamInRange(_id, 59, '87');

    let str = '';
    result.stream.on('data', data => str += data);
    result.stream.on('end', () => {
      expect(str.length).toBe(28);
      done();
    });
    expect(result.length).toBe(process.platform === 'darwin' ? 1077 : 1097);
    expect(result.filename).toBe('LICENSE');
    expect(result.streamRange).toEqual({ start: 59, end: 87 });
  });

  // 分段读取返回Stream异常，id为空时
  it('grid to stream in range (File id cannot be empty)', done => {
    this.model.gridToStreamInRange()
      .catch(error => {
        expect(error.message).toBe('File id cannot be empty');
        done();
      });
  });

  // 分段读取返回Stream异常，文件找不到时
  it('grid to stream in range (File not found)', done => {
    this.model.gridToStreamInRange('000000000000', 59, 87)
      .catch(error => {
        expect(error.message).toBe('File not found');
        done();
      });
  });

  // 分段读取返回Stream异常，开始位置不是数字型时
  it('grid to stream in range (Incorrect start value)', done => {
    this.model.gridToStreamInRange('id', '59a', 87)
      .catch(error => {
        expect(error.message).toBe('Incorrect start value : 59a');
        done();
      });
  });

  // 分段读取返回Stream异常，结束位置不是数字型时
  it('grid to stream in range (Incorrect end value)', done => {
    this.model.gridToStreamInRange('id', 59, '87a')
      .catch(error => {
        expect(error.message).toBe('Incorrect end value : 87a');
        done();
      });
  });

  // 获取数据库一览
  it('list all db', async () => {
    const result = await this.admin.listDb();
    expect(result.length).toBeGreaterThan(1);
  });

  // 获取表一览
  it('list all collection', async () => {
    const result = await this.model.listCollection();
    expect(result.map(item => item.name)).toContain('default.tests');
  });

  // 获取索引一览
  it('list all index', async () => {
    const result = await this.model.listIndex();
    expect(result.map(item => item.name)).toContain('_id_');
  });

  // 添加索引
  it('add index', async () => {
    const result = await this.model.addIndex(
      { val: -1, valid: 1 },
      { name: 'unit_test_index' }
    );
    expect(result).toBe('unit_test_index');
  });

  // 删除索引
  it('drop index', async () => {
    const result = await this.model.dropIndex('unit_test_index');
    expect(result.ok).toBeTruthy();
  });

  // 添加用户
  it('add user', async () => {
    const result = await this.model.addUser('unit_test_user', '123456');
    expect(result.ok).toBe(1);
  });

  // 删除用户
  it('drop user', async () => {
    const result = await this.model.dropUser('unit_test_user');
    expect(result).toBeTruthy();
  });

  // 删除表
  it('drop collection', async () => {
    const result = await this.model.dropCollection();
    expect(result).toBeTruthy();
  });

  // 删除表异常
  it('drop collection（no collection）', async () => {
    const result = await this.model.dropCollection();
    expect(result).toBeTruthy();
  });

  // 获取表信息
  it('stats collection', async () => {
    const result = await this.model.statsCollection();
    expect(result.ok).toBe(1);
  });

  // 获取数据库信息
  it('stats db', async () => {
    const result = await this.model.statsDb();
    expect(result.ok).toBe(1);
  });

  // 获取数据库进程信息
  it('stats server', async () => {
    const result = await this.admin.statsServer();
    expect(result.ok).toBe(1);
  });

  // 删除库
  it('drop db', async () => {
    await this.admin.addUser('unit_test_user', '123456', 'test');
    const model = new Model('test', null, null, {
        user: 'unit_test_user',
        pass: '123456'
      })
      , result = await model.dropDb();
    expect(result).toBeTruthy();
    await this.admin.dropUser('unit_test_user', 'test');
  });

});
