/**
 * operator.test.js
 */

'use strict';

const op = require('../../../../lib/db/mongo/operator');

describe('Operator', () => {

  describe('compare', () => {

    it('Query Selectors - Comparison', () => {
      expect(op.compare('$eq', 'k', 'v')).toEqual({ k: { $eq: 'v' } });
      expect(op.compare('$gt', 'k', 'v')).toEqual({ k: { $gt: 'v' } });
      expect(op.compare('$gte', 'k', 'v')).toEqual({ k: { $gte: 'v' } });
      expect(op.compare('$in', 'k', ['v'])).toEqual({ k: { $in: ['v'] } });
      expect(op.compare('$in', 'k', 'v, w')).toEqual({ k: { $in: ['v', 'w'] } });
      expect(op.compare('$lt', 'k', 'v')).toEqual({ k: { $lt: 'v' } });
      expect(op.compare('$lte', 'k', 'v')).toEqual({ k: { $lte: 'v' } });
      expect(op.compare('$ne', 'k', 'v')).toEqual({ k: { $ne: 'v' } });
      expect(op.compare('$nin', 'k', ['v'])).toEqual({ k: { $nin: ['v'] } });
      expect(op.compare('$nin', 'k', 'v, w')).toEqual({ k: { $nin: ['v', 'w'] } });
    });

    it('Query Selectors - Logical', () => {
      expect(op.compare('$and', '', [{ k: 'v' }])).toEqual({ $and: [{ k: 'v' }] });
      expect(op.compare('$not', '', [{ k: 'v' }])).toEqual({ $not: [{ k: 'v' }] });
      expect(op.compare('$nor', '', [{ k: 'v' }])).toEqual({ $nor: [{ k: 'v' }] });
      expect(op.compare('$or', '', [{ k: 'v' }])).toEqual({  $or: [{ k: 'v' }] });
    });

    it('Query Selectors - Element', () => {
      expect(op.compare('$exists', 'k', true)).toEqual({ k: { $exists: true } });
      expect(op.compare('$type', 'k', 'number')).toEqual({ k: { $type: 'number' } });
    });

    it('Query Selectors - Evaluation', () => {
      expect(op.compare('$mod', 'k', [4, 0])).toEqual({ k: { $mod: [4, 0] } });
      expect(op.compare('$regex', 'k', 'v')).toEqual({ k: { $regex: 'v' } });
      expect(op.compare('$text', '', { $search: 'v' })).toEqual({ $text: { $search: 'v' } });
    });

    it('Query Selectors - Array', () => {
      expect(op.compare('$all', 'k', ['v1', 'v2'])).toEqual({ k: { $all: ['v1', 'v2'] } });
      expect(op.compare('$elemMatch', 'k', { $gt: 1, $lt: 3 })).toEqual({ k: { $elemMatch: { $gt: 1, $lt: 3 } } });
      expect(op.compare('$size', 'k', 2)).toEqual({ k: { $size: 2 } });
    });

    it('Query Selectors - Projection', () => {
      expect(op.compare('$slice', 'k', 2)).toEqual({ k: { $slice: 2 } });
    });

  });

  describe.only('parse', () => {

    it('$inc', () => {
      const data = { a: '-2', 'b.c': '1', d: 'str', e: 10 };
      const result = { a: -2, 'b.c': 1, d: null, e: 10 };
      expect(op.parse('$inc', data)).toEqual(result);
    });

    it.only('$gt', () => {
      const define = {
        a: { type: 'number' }
      };
      const data = {
        a: { $gt: '10' }
      };
      const result = {
        a: { $gt: 10 }
      };
      expect(op.parse('$gt', data, define)).toEqual(result);
    });

    it('$max', () => {
      const define = {
        a: { type: 'object', contents: { b: { type: 'number' } } },
        c: { type: 'number' },
        d: { type: 'string' },
        e: { type: 'date' }
      };
      const data = {
        'a.b': '-2',
        c    : '1',
        d    : 'str',
        e    : '2020-01-01'
      };
      const result = {
        'a.b': -2,
        c    : 1,
        d    : 'str',
        e    : new Date('2020-01-01')
      };
      expect(op.parse('$max', data, define)).toEqual(result);
    });

    it('$pull', () => {
      const define = {
        a: {
          type    : 'object',
          contents: { b: { type: 'array', contents: 'number' } }
        },
        c: {
          type    : 'array',
          contents: { d: { type: 'string', e: { type: 'string' } } }
        },
        f: {
          type    : 'array',
          contents: 'string'
        },
        g: {
          type    : 'array',
          contents: 'string'
        }
      };
      const data = {
        $pull: {
          'a.b': { $gt: '1' },
          'c.d': { $in: ['a', 1] },
          f    : { $nin: 10 },
          g    : 10
        }
      };
      const result = {
        $pull: {
          'a.b': { $gt: 1 },
          'c.d': { $in: ['a', '1'] },
          f    : { $nin: ['10'] },
          g    : '10'
        }
      };
      expect(op.parse('$max', data, define)).toEqual(result);
    });

    it('all types', () => {
      const define = {
        a: { type: 'object' },
        b: { type: 'boolean' },
        c: { type: 'number' },
        d: { type: 'string' },
        e: { type: 'objectid' },
        f: { type: 'array' },
        g: { type: 'date' }
      };
      const data = {
        a: undefined,
        b: undefined,
        c: undefined,
        d: undefined,
        e: undefined,
        f: undefined,
        g: undefined
      };
      const result = {
        a: {},
        b: false,
        c: null,
        d: '',
        e: null,
        f: [],
        g: null
      };
      expect(op.parse('$max', data, define)).toEqual(result);
    });

    it('all operators - update', () => {
      const define = {
        a: { type: 'string' },
        b: { type: 'string' },
        c: { type: 'string' },
        d: { type: 'string' },
        e: { type: 'string' },
        f: { type: 'string' },
        g: { type: 'string' }
      };
      const data = {
        $inc        : { a: '1' },
        $max        : { b: 1 },
        $min        : {},
        $mul        : {},
        $pop        : {},
        $pull       : {},
        $pullAll    : {},
        $push       : {},
        $pushAll    : {},
        $rename     : {},
        $search     : {},
        $set        : {},
        $setOnInsert: {},
        $unset      : {}
      };
      const result = {};
      expect(op.parse('$max', data, define)).toEqual(result);
    });

    // ---- data
    // {
    //   $push: {
    //     quizzes: {
    //        $each: [ { wk: 5, score: 8 }, { wk: 6, score: 7 }, { wk: 7, score: 6 } ],
    //        $sort: { score: -1 },
    //        $slice: 3
    //     }
    //   }
    // }
    // { $pull: { fruits: { $in: [ "apples", "oranges" ] }, vegetables: "carrots" } }

    // ---- query
    // { field: { $gt: value } }
    // { $or: [ { quantity: { $lt: 20 } }, { price: 10 } ] }
    // { field: { $in: [<value1>, <value2>, ... <valueN> ] } }

  });

});
