/**
 * helper.test.js
 */

'use strict';

const dayjs  = require('dayjs');
const fs     = require('fs');
const path   = require('path');
const os     = require('os');
const helper = require('../../lib/helper');

describe('Helper', () => {

  it('get', () => {
    const result = helper.get([{ a: { b: 1 } }, { a: { b: 2 } }], 'a.b');
    expect(result).toEqual([1, 2]);
  });

  it('set', () => {
    const result = helper.set({}, 'a.b.c', 10);
    expect(result).toEqual({ a: { b: { c: 10 } } });
  });

  it('randomGUID4', () => {
    const res1 = helper.randomGUID4();
    const res2 = helper.randomGUID4();
    expect(res1).toHaveLength(4);
    expect(res1).not.toBe(res2);

    expect(helper.randomGUID8()).toHaveLength(8);
    expect(helper.randomGUID12()).toHaveLength(12);
  });

  it('uuid', () => {
    expect(helper.uuid()).toHaveLength(36);
  });

  it('ejsParser', () => {
    const file = `${process.cwd()}/test/controllers/helper.tmpl`;
    const res1 = helper.ejsParser(file, { a: 'a' });
    expect(res1).toBe('a');

    // 结果写到文件
    const resultFile = `${os.tmpdir()}/helper.tmp`;
    helper.ejsParser(file, { a: 'a' }, resultFile);
    expect(fs.readFileSync(resultFile, 'utf8')).toBe('a');

    // 删除
    fs.unlinkSync(resultFile);
  });

  it('ejsFormat', () => {
    const res1 = helper.ejsFormat('<%= a %>', { a: 'a' });
    expect(res1).toBe('a');
  });

  it('isBrowser', () => {
    const res1 = helper.isBrowser({ headers: { 'x-client-type': 'wechat' } });
    expect(res1).toBe(false);
    const res2 = helper.isBrowser({ headers: { 'x-client-type': 'iphone' } });
    expect(res2).toBe(false);
    const res3 = helper.isBrowser({ headers: { 'x-client-type': 'android' } });
    expect(res3).toBe(false);

    const res4 = helper.isBrowser({ headers: { } });
    expect(res4).toBe(false);
    const res5 = helper.isBrowser({ headers: { 'user-agent': 'mozilla' } });
    expect(res5).toBe(true);
    const res6 = helper.isBrowser({ headers: { 'user-agent': 'electron' } });
    expect(res6).toBe(false);
  });

  it('isMobile', () => {
    const res1 = helper.isMobile({ headers: { 'x-client-type': 'wechat' } });
    expect(res1).toBe(true);
    const res2 = helper.isMobile({ headers: { 'x-client-type': 'rn' } });
    expect(res2).toBe(true);

    const res3 = helper.isMobile({ headers: { 'light-react-native': 'true' } });
    expect(res3).toBe(true);

    const res4 = helper.isMobile({ headers: { } });
    expect(res4).toBe(false);
    const res5 = helper.isMobile({ headers: { 'user-agent': 'iPhone' } });
    expect(res5).toBe(true);
    const res6 = helper.isMobile({ headers: { 'user-agent': 'Android' } });
    expect(res6).toBe(true);
    const res7 = helper.isMobile({ headers: { 'user-agent': 'electron' } });
    expect(res7).toBe(true);
  });

  it('ip', () => {
    expect(helper.ip().length).toBeGreaterThan(0);
    expect(helper.ip().length).toBeGreaterThan(0);
  });

  it('applicationInfo', () => {
    const { version, host, application } = helper.applicationInfo();
    expect(version).toBe(require('../../package.json').version);
    expect(host).toBe(os.hostname());
    expect(application).toBe('light-core');
  });

  it('resolve', () => {
    const mod1 = helper.resolve('helper.test', __dirname);
    expect(mod1).toEqual({});

    const mod2 = helper.resolve('package.json');
    expect(mod2.name).toBe('light-core');

    const mod3 = helper.resolve('dummy');
    expect(mod3).toBeUndefined();
  });

  it('xmlParser', async () => {
    const xml1 = await helper.xmlParser('/test/controllers/helper.xml');
    expect(xml1).toEqual({ root: 'hello' });

    const xml2 = await helper.xmlParser('dummy');
    expect(xml2).toEqual({});
  });

  it('yamlLoader', async () => {
    const yaml1 = await helper.yamlLoader('/test/controllers/helper.yaml');
    expect(yaml1).toEqual({ root: { key: 'val' } });

    const yaml2 = await helper.yamlLoader('dummy');
    expect(yaml2).toBe(null);
  });

  it('yamlParser', async () => {
    const yaml = await helper.yamlParser('root:\n  key: val');
    expect(yaml).toEqual({ root: { key: 'val' } });
  });

  it('yamlDumper', async () => {
    const res = await helper.yamlDumper({ root: { key: 'val' } });
    expect(res.startsWith('root')).toBe(true);
  });

  it('zipFiles', async () => {
    // 空列表
    const out = `${os.tmpdir()}/helper.zip`;
    await helper.zipFiles([], out);
    expect(fs.statSync(out).size).toBe(136);
    fs.unlinkSync(out);

    // 文件列表
    const multi = [`${process.cwd()}/LICENSE`, `${process.cwd()}/test/favicon.ico`];
    await helper.zipFiles(multi, out);
    expect(fs.statSync(out).size).toBe(12583);
    fs.unlinkSync(out);

    // 文件流
    const single = [`${process.cwd()}/LICENSE`];
    await helper.zipFiles(single, fs.createWriteStream(out));
    expect(fs.statSync(out).size).toBe(774);
    fs.unlinkSync(out);
  });

  it('qrcode', async () => {

    // base64
    const base64 = await helper.qrcode({ text: 'lalal' });
    expect(base64).toHaveLength(1026);
    expect(base64.startsWith('data:image/png;base64,')).toBe(true);

    // image
    const out = `${os.tmpdir()}/helper.png`;
    await helper.qrcode({ text: 'lalal' }, out);
    expect(fs.statSync(out).size).toBe(752);
    fs.unlinkSync(out);

    // stream
    await helper.qrcode({ text: 'lalal' }, fs.createWriteStream(out));
    expect(fs.statSync(out).size).toBe(752);
    fs.unlinkSync(out);
  });

  it('rmdir', () => {
    const folder = `${os.tmpdir()}/dir1/dir2/dir3`;
    fs.mkdirSync(folder, { recursive: true });
    helper.rmdir(`${os.tmpdir()}/dir1`);
    expect(fs.existsSync(`${os.tmpdir()}/dir1`)).toBe(false);
  });

  it('mkdirp', () => {
    const folder = `${os.tmpdir()}/dir1/dir2/`;
    helper.mkdirp(folder);
    expect(fs.existsSync(folder)).toBe(true);
    helper.rmdir(folder);
  });

  it('format', () => {

    expect(helper.format(undefined)).toBeUndefined();
    expect(helper.format(null)).toBe(null);

    const date = dayjs('2016-01-01T12:00:00Z').toDate();

    // date
    expect(helper.format(date)).toBe(date);
    expect(helper.format(date, 'YYYY/MM/DD hh:mm:ss')).toBe('2016/01/01 12:00:00');
    expect(helper.format(date, 'YYYY/MM/DD hh:mm:ss', 'UTC')).toBe('2016/01/01 12:00:00');
    expect(helper.format(date, 'YYYY/MM/DD hh:mm:ss', 'Asia/Shanghai')).toBe('2016/01/01 08:00:00');
    expect(helper.format(dayjs(date), 'YYYY/MM/DD hh:mm:ss')).toBe('2016/01/01 12:00:00');

    // number
    expect(helper.format(10000.01, '0,0.0000')).toBe('10,000.0100');

    // string
    expect(helper.format('world', 'hello %s')).toBe('hello world');
  });

  it('tree', () => {

    let result = [];

    result = helper.tree('./');
    expect(result.find(item => path.basename(item.file) === 'helper.test.js')).toBeTruthy();

    // ignore
    result = helper.tree('./', ['*test.js']);
    expect(result.find(item => path.basename(item.file) === 'helper.test.js')).toBeFalsy();

    // include
    result = helper.tree('./', ['*', '!/.gitignore']);
    expect(result.length).toBe(1);
    expect(path.basename(result[0].file)).toBe('.gitignore');
  });

  it('fileMd5', () => {
    expect(helper.fileMd5('LICENSE')).toBe('fe98679655cd4f0bc8524f353987837b');
  });

  it('isMongo', () => {
    expect(helper.isMongoDB()).toBe(true);
  });

  it('isMySQL', () => {
    process.env.LIGHTMYSQL_HOST = '';
    expect(helper.isMySQL()).toBe(true);
    delete process.env.LIGHTMYSQL_HOST;
  });

  it('isSQLServer', () => {
    process.env.LIGHTSQLSERVER_HOST = '';
    expect(helper.isSQLServer()).toBe(true);
    delete process.env.LIGHTSQLSERVER_HOST;
  });

  it('isMaster', () => {
    process.env.APPMASTER = true;
    expect(helper.isMaster()).toBe(true);
    delete process.env.APPMASTER;

    process.env.APPSLOT = '1';
    expect(helper.isMaster()).toBe(true);
    process.env.APPSLOT = '2';
    expect(helper.isMaster()).toBe(false);
    delete process.env.APPSLOT;
  });

  it('isUnitTest', () => {
    expect(helper.isUnitTest()).toBe(true);
  });

  it('captcha', () => {
    const { code, png } = helper.captcha({ code: 1111 });
    expect(code).toBe(1111);
    expect(png.startsWith('data:image/png;base64,iVBORw0KGgoAAAAN')).toBe(true);
  });

});
