/**
 * job.test.js
 */

'use strict';

const job = require('../../lib/job');

describe('Job action', () => {

  beforeAll(() => {
    process.env.APPSLOT = 1;
    jest.mock('../controllers/job', () => ({
      test1: async handler => await handler.params,
      test2: async handler => await handler.params,
      error: async () => {
        throw new Error('action error');
      }
    }));
  });

  afterAll(() => {
    delete process.env.APPSLOT;
    delete process.env.APPCRON;
    delete process.env.APPLOCAL;
    jest.unmock('../controllers/job');
  });

  afterEach(() => {
    job.stopAll();
  });

  it('run action', async () => {

    // 模拟 job 的 update 方法确认执行结果
    jest.spyOn(job, 'update').mockImplementation(res => {
      expect(res.error).toBeUndefined();
      expect(res.id).toBe('1');
      expect(res.result).toEqual([{ data: 'param' }, { data: 'param' }]);
    });

    job.init([
      {
        id      : '1',
        schedule: '*/1 * * * * *',
        run     : 'STARTED',
        domain  : process.env.APPNAME,
        step    : [
          { type: 'action', class: 'job', action: 'test1', params: 'param' },
          { type: 'action', class: 'job', action: 'test2', params: 'param' }
        ]
      }
    ]);

    // 等待定时任务执行完
    await new Promise(resolve => setTimeout(resolve, 1000 * 2));

    // 两秒内执行了两次
    expect(job.update).toBeCalledTimes(2);
  }, 1000 * 3);

  it('run bad action', async () => {

    job.update = jest.fn(res => {
      expect(res.error).toBe('Error: action error');
    });

    job.init([
      {
        id      : '2',
        schedule: '*/1 * * * * *',
        run     : 'STARTED',
        domain  : process.env.APPNAME,
        step    : [{ type: 'action', class: 'job', action: 'error' }]
      }
    ]);

    await new Promise(resolve => setTimeout(resolve, 1000 * 1));
  });

  it('run script', async () => {

    jest.spyOn(job, 'update').mockImplementation(res => {
      expect(res.error).toBeUndefined();
      expect(res.result).toHaveLength(1);
    });

    job.init([
      {
        id      : '3',
        schedule: '*/1 * * * * *',
        run     : 'STARTED',
        domain  : process.env.APPNAME,
        step    : [{ type: 'script', script: 'ls' }]
      }
    ]);

    await new Promise(resolve => setTimeout(resolve, 1000 * 1));
  });

  it('run bad script', async () => {

    jest.spyOn(job, 'update').mockImplementation(res => {
      expect(res.error).toContain('Error: Command failed: bad command');
    });

    job.init([
      {
        id      : '4',
        schedule: '*/1 * * * * *',
        run     : 'STARTED',
        domain  : process.env.APPNAME,
        step    : [{ type: 'script', script: 'bad command' }]
      }
    ]);

    await new Promise(resolve => setTimeout(resolve, 1000 * 1));
  });

  it('run no class', async () => {
    jest.spyOn(job, 'update').mockImplementation(res => {
      expect(res.result).toEqual(['class not found']);
    });
    job.init([
      {
        id      : '5',
        schedule: '*/1 * * * * *',
        run     : 'STARTED',
        domain  : process.env.APPNAME,
        step    : [{ type: 'action', class: 'noclass' }]
      }
    ]);
    await new Promise(resolve => setTimeout(resolve, 1000 * 1));
  });

  it('run no action', async () => {
    jest.spyOn(job, 'update').mockImplementation(res => {
      expect(res.result).toEqual(['action not found']);
    });
    job.init([
      {
        id      : '6',
        schedule: '*/1 * * * * *',
        run     : 'STARTED',
        domain  : process.env.APPNAME,
        step    : [{ type: 'action', class: 'job', action: 'noaction' }]
      }
    ]);
    await new Promise(resolve => setTimeout(resolve, 1000 * 1));
  });

  it('run immediate', async () => {
    const res = await job.immediate({
      id  : '7',
      step: [{ type: 'action', class: 'job', action: 'test1', params: 'param' }]
    });
    expect(res).toEqual({ id: '7', result: [{ data: 'param' }] });
  });

  it('not master', async () => {
    delete process.env.APPSLOT;
    delete process.env.APPCRON;
    job.init([]);
  });

  it('not schedule', async () => {
    process.env.APPCRON = true;
    job.init([{}]);
  });

  it('not domain', async () => {
    process.env.APPCRON = true;
    job.init([{ schedule: '*/1 * * * * *' }]);
  });

  it('no update', async () => {
    jest.resetAllMocks();
    process.env.APPLOCAL = true;
    await job.update();
  });
});
