/**
 * log.test.js
 */

'use strict';

const dayjs        = require('dayjs');
const { ObjectID } = require('mongodb');
const log          = require('../../lib/log');
const CONST        = require('../../lib/constant');
const uid          = new ObjectID();

describe('Log', () => {

  afterEach(() => {
    delete process.env.APPDEV;
    delete process.env.APPLOG;
    jest.resetAllMocks();
  });

  it('debug (local = true)', () => {

    const spy = jest.spyOn(console, 'log');

    process.env.APPDEV = true;
    log.debug('message', uid);

    // e.g. [2020-02-16T02:06:57.939] [A] [DEBUG] message user /Users/lilin/developer/light/LightCoreNode/test/lib/log.test.js:26
    const [time, type, level, message, user, code] = spy.mock.calls[0][0].split(' ');
    expect(time).toMatch(/\[\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}.\d{3}\]/);
    expect(type).toBe('[A]');
    expect(level).toBe('[DEBUG]');
    expect(message).toBe('message');
    expect(user).toBe(uid.toHexString());
    expect(code).toMatch(/.*\/test\/lib\/log.test.js:\d{2}/);
  });

  it('debug (local = false)', () => {

    const spy = jest.spyOn(console, 'log');
    log.debug('message', uid);

    // e.g. [2020-02-16T02:06:30.932] [A] [DEBUG] /Users/lilin/developer/light/LightCoreNode/test/lib/log.test.js:44 - - user - message
    const [time, type, level, code, p1, p2, user, p3, message] = spy.mock.calls[0][0].split(' ');
    expect(time).toMatch(/\[\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}.\d{3}\]/);
    expect(type).toBe('[A]');
    expect(level).toBe('[DEBUG]');
    expect(code).toMatch(/.*\/test\/lib\/log.test.js:\d{2}/);
    expect(p1).toBe('-');
    expect(p2).toBe('-');
    expect(user).toBe(uid.toHexString());
    expect(p3).toBe('-');
    expect(message).toBe('message');
  });

  it('debug (level : error)', () => {
    process.env.APPLOG = CONST.LOG_LEVEL_ERROR;
    const func = jest.spyOn(console, 'log');
    log.debug('message', 'user');
    expect(func).not.toHaveBeenCalled();
  });

  it('debug (level : warn)', () => {
    process.env.APPLOG = CONST.LOG_LEVEL_WARN;
    const func = jest.spyOn(console, 'log');
    log.debug('message', 'user');
    expect(func).not.toHaveBeenCalled();
  });

  it('debug (level : info)', () => {
    process.env.APPLOG = CONST.LOG_LEVEL_INFO;
    const func = jest.spyOn(console, 'log');
    log.debug('message', 'user');
    expect(func).not.toHaveBeenCalled();
  });

  it('debug (level : debug)', () => {
    process.env.APPLOG = CONST.LOG_LEVEL_DEBUG;
    const func = jest.spyOn(console, 'log');
    log.debug('message', 'user');
    expect(func).toHaveBeenCalled();
  });

  it('debug (object)', () => {

    // object
    const spy = jest.spyOn(console, 'log');
    log.debug({ value: 1 }, uid);
    expect(spy.mock.calls[0][0].split(' ')[8]).toBe('{"value":1}');

    // array
    log.debug([{ value: 1 }], uid);
    expect(spy.mock.calls[1][0].split(' ')[8]).toBe('[{"value":1}]');

    // error
    log.debug(new Error('error'), uid);
    expect(spy.mock.calls[2][0].split(' ')[8]).toBe('Error:error');

    // dayjs
    log.debug(dayjs('2020-02-02'), uid);
    expect(spy.mock.calls[3][0].split(' ')[8]).toBe('2020-02-01T16:00:00.000Z');

    // null
    log.debug(null, uid);
    expect(spy.mock.calls[4][0].split(' ')[8]).toBe('null');

    // nan
    log.debug(NaN, uid);
    expect(spy.mock.calls[5][0].split(' ')[8]).toBe('null');

    // undefined
    log.debug(undefined, uid);
    expect(spy.mock.calls[6][0].split(' ')[8]).toBe('undefined');

    // objectid
    log.debug(uid);
    expect(spy.mock.calls[7][0].split(' ')[8]).toBe(uid.toHexString());

    // date
    log.debug(new Date('2020-02-03T00:00:00'));
    expect(spy.mock.calls[8][0].split(' ')[8]).toBe('2020-02-02T16:00:00.000Z');

    // boolean
    log.debug(false);
    expect(spy.mock.calls[9][0].split(' ')[8]).toBe('false');

    // simble
    log.debug(Symbol(1));
    expect(spy.mock.calls[10][0].split(' ')[8]).toBe('Symbol(1)');

    // bitint
    // eslint-disable-next-line no-undef
    log.debug(BigInt(100));
    expect(spy.mock.calls[11][0].split(' ')[8]).toBe('100');

  });

  it('debug (no user)', () => {
    const spy = jest.spyOn(console, 'log');
    log.debug('message');

    const params = spy.mock.calls[0][0].split(' ');
    expect(params[6]).toBe('-');
    expect(params[8]).toBe('message');
  });

  it('debug (multiple params)', () => {
    const spy = jest.spyOn(console, 'log');
    log.debug('msg1', 'msg2', 'msg3');

    const params = spy.mock.calls[0][0].split(' ');
    expect(params[6]).toBe('-');
    expect(params[8]).toBe('msg1');
    expect(params[9]).toBe('msg2');
    expect(params[10]).toBe('msg3');
  });

  it('debug (multiple params + uid)', () => {
    const spy = jest.spyOn(console, 'log');
    log.debug('msg1', 'msg2', 'msg3', uid);

    const params = spy.mock.calls[0][0].split(' ');
    expect(params[6]).toBe(uid.toHexString());
    expect(params[8]).toBe('msg1');
    expect(params[9]).toBe('msg2');
    expect(params[10]).toBe('msg3');
  });


  it('info', () => {
    const spy = jest.spyOn(console, 'info');
    log.info('info', uid);

    const [, type, level, , , , , , message] = spy.mock.calls[0][0].split(' ');
    expect(type).toBe('[A]');
    expect(level).toBe('[INFO]');
    expect(message).toBe('info');
  });

  it('warn', () => {
    const spy = jest.spyOn(console, 'warn');
    log.warn('warn', uid);

    const [, type, level, , , , , , message] = spy.mock.calls[0][0].split(' ');
    expect(type).toBe('[A]');
    expect(level).toBe('[WARN]');
    expect(message).toBe('warn');
  });

  it('error', () => {
    const spy = jest.spyOn(console, 'error');
    log.error('error', uid);

    const [, type, level, , , , , , message] = spy.mock.calls[0][0].split(' ');
    expect(type).toBe('[A]');
    expect(level).toBe('[ERROR]');
    expect(message).toBe('error');
  });

  it('audit', () => {
    const spy = jest.spyOn(console, 'info');
    log.audit('audit', uid);

    const [, type, level, , , , , , message] = spy.mock.calls[0][0].split(' ');
    expect(type).toBe('[A]');
    expect(level).toBe('[INFO]');
    expect(message).toBe('audit');
  });

  it('operation', () => {
    const spy = jest.spyOn(console, 'info');
    log.operation('operation', uid);

    const [, type, level, , , , , , message] = spy.mock.calls[0][0].split(' ');
    expect(type).toBe('[A]');
    expect(level).toBe('[INFO]');
    expect(message).toBe('operation');
  });

});
