/**
 * middleware/router.test.js
 */

'use strict';

const router   = require('../../../lib/middleware/router');
const config   = require('../../../lib/configuration');
const setting  = require('../../../lib/setting');
const response = require('../../../lib/response');
const rider    = require('../../../lib/model/datarider');
const pkg      = require('../../../package.json');
const util     = require('./util');

describe('Router', () => {

  beforeAll(() => {
    util.setupKoa(['router', 'helloword']);
  });

  afterAll(() => util.clear());
  afterEach(() => jest.resetAllMocks());

  // 路由到用户自定义的controller（test/controller/user）
  it('route api (user controller)', async () => {
    setting.board = [
      { api: '/api/user/null', class: 'user', action: 'null' },
      { api: '/api/user/undefined', class: 'user', action: 'undefined' },
      { api: '/api/user/zero', class: 'user', action: 'zero' },
      { api: '/api/user/object', class: 'user', action: 'object' },
      { api: '/api/user/throw', class: 'user', action: 'throw' },
      { api: '/api/user/throwcustom', class: 'user', action: 'throwcustom' },
      { api: '/api/user/error', class: 'user', action: 'error' },
      { api: '/api/user/callback', class: 'user', action: 'callback' },
      { api: '/api/user/callbackerror', class: 'user', action: 'callbackerror' },
      { api: '/api/user/noaction', class: 'user', action: 'noaction' },
      { api: '/api/user/logicapi', class: 'user', action: 'logicapi', kind: 1 }
    ];

    let res = await util.request.get('/api/user/null');
    expect(res.body.data).toBe(null);

    res = await util.request.get('/api/user/undefined');
    expect(res.body.data).toEqual({});

    res = await util.request.get('/api/user/zero');
    expect(res.body.data).toBe(0);

    res = await util.request.get('/api/user/object');
    expect(res.body.data).toEqual({ name: 'light' });

    res = await util.request.get('/api/user/throw');
    expect(res.body.error).toEqual({ code: 500, message: 'error' });

    res = await util.request.get('/api/user/throwcustom');
    expect(res.body.error).toEqual({ code: 400, message: 'Bad Request' });

    res = await util.request.get('/api/user/error');
    expect(res.body.error).toEqual({ code: 500, message: 'this.fn is not a function' });

    res = await util.request.get('/api/user/callback');
    expect(res.body.data).toBe('helloword');

    res = await util.request.get('/api/user/callbackerror');
    expect(res.body.error).toEqual({ code: 400, message: 'Bad Request' });

    res = await util.request.get('/api/user/noboard');
    expect(res.body.error).toEqual({ code: 404, message: 'Board not found' });

    res = await util.request.get('/api/user/noaction');
    expect(res.body.error).toEqual({ code: 404, message: 'Can not dispatch' });

    res = await util.request.get('/api/user/logicapi');
    expect(res.body.error).toEqual({ code: 404, message: 'Module not found' });

  });

  // 路由到系统提供的API（lib/model/system）
  it('route api (use system model)', async () => {
    setting.board = [
      { api: '/api/system/version', class: 'system', action: 'version' }
    ];
    const res = await util.request.get('/api/system/version');
    expect(res.body.data).toEqual({ core: pkg.version });
  });

  // 路由到router
  it('route api (use rider)', async () => {

    const json = require('./router.test.json');
    setting.board = json.board;
    setting.struct = json.struct;
    rider.init();

    const spy = jest.spyOn(rider.test, 'list')
      .mockImplementation(async () => ({ val: 1 }));
    const res = await util.request.get('/api/test/list');

    expect(res.body.data).toEqual({ val: 1 });
    expect(spy).toHaveBeenCalled();
  });

  // 渲染HTML模板
  it('route template', async () => {
    setting.route = [
      { url: '/simple', template: 'index.html' },
      { url: '/useclass', template: 'index.html', class: 'helloword', action: 'use' },
      { url: '/stopresponse', template: '/', class: 'helloword', action: 'stop' },
      { url: '/callback', template: 'index.html', class: 'helloword', action: 'callback' },
      { url: '/callbackerror', template: 'index.html', class: 'helloword', action: 'callbackerror' },
      { url: '/noclass', template: 'index.html', class: 'helloword', action: 'noclass' },
      { url: '/throw', template: 'index.html', class: 'helloword', action: 'throw' }
    ];
    let res = await util.request.get('/simple');
    expect(res.text).toContain('</a>');

    res = await util.request.get('/nourl');
    expect(res.body.error.message).toBe('Route not found');

    res = await util.request.get('/useclass');
    expect(res.text).toContain('helloword');

    res = await util.request.get('/stopresponse');
    expect(res.text).toContain('hello2');

    res = await util.request.get('/callback');
    expect(res.text).toContain('hello3');

    res = await util.request.get('/callbackerror');
    expect(res.body.error.message).toBe('Internal Server Error');

    res = await util.request.get('/noclass');
    expect(res.body.error.message).toBe('Can not dispatch');

    res = await util.request.get('/throw');
    expect(res.body.error.message).toBe('Not Implemented');
  });

  // 确认URL匹配api或route
  it('find define', () => {

    const define = [
      { id: 1, api: '/api/user/:id' },
      { id: 2, api: '/api/user/list.html' },
      { id: 3, api: '/WW_VERIFY_:id.TXT' },
      { id: 4, url: '/' },
      { id: 5, url: '/.well-known/apple-app-site-association' },
      { id: 6, url: '/:id/:name/:age.html' }
    ];

    expect(router.findDefine(define, '/api/user/0001'))
      .toEqual({ route: { id: 1, api: '/api/user/:id' }, params: { id: '0001' } });

    expect(router.findDefine(define, '/api/user/list.html').route)
      .toEqual({ id: 2, api: '/api/user/list.html' });

    expect(router.findDefine(define, '/WW_VERIFY_.TXT'))
      .toEqual({ route: { id: 3, api: '/WW_VERIFY_:id.TXT' }, params: { id: '' } });

    expect(router.findDefine(define, '/WW_VERIFY_001.TXT'))
      .toEqual({ route: { id: 3, api: '/WW_VERIFY_:id.TXT' }, params: { id: '001' } });

    expect(router.findDefine(define, '/'))
      .toEqual({ route: { id: 4, url: '/' }, params: {} });

    expect(router.findDefine(define, '/.well-known/apple-app-site-association'))
      .toEqual({
        route : { id: 5, url: '/.well-known/apple-app-site-association' },
        params: {}
      });

    expect(router.findDefine(define, '/001/user/18.html'))
      .toEqual({
        route : { id: 6, url: '/:id/:name/:age.html' },
        params: { id: '001', name: 'user', age: '18' }
      });

    expect(router.findDefine(define, '/WW_VERIFY.TXT')).toEqual({});
    expect(router.findDefine(define, '/index')).toEqual({});
    expect(router.findDefine(define, '/index.html')).toEqual({});
    expect(router.findDefine(define, '/.well-known')).toEqual({});

  });

  // 确认route设定了 * 通配符时，URL可以正确匹配
  it('find wildcard define', () => {
    const define = [
      { id: 1, url: '*' },
      { id: 2, url: '/api/user/list.html' }
    ];
    expect(router.findDefine(define, '/api/user/list.html'))
      .toEqual({ route: { id: 1, url: '*' }, params: {} });
  });

  // 确认模板正确渲染，EJS转换成功，URL被正确替换
  it('render template', () => {

    config.load([{ type: 'app', key: 'stamp', value: '00000001' }]);

    const sendHTML = jest.spyOn(response, 'sendHTML');
    const params = { user: { name: 'light' },  data: {} };

    router.renderTemplate({ ctx: {} }, { template: 'index.html' }, params);
    expect(sendHTML.mock.calls[0][1]).toContain('<h2>light</h2>');
    expect(sendHTML.mock.calls[0][1].match(/stamp=00000001/g)).toHaveLength(3);

    const sendError = jest.spyOn(response, 'sendError');
    router.renderTemplate({ ctx: {} }, { template: 'empty.html' }, {});
    expect(sendError).toHaveBeenCalled();
  });

});
