/**
 * middleware/session.test.js
 */

'use strict';

const util    = require('./util');
const user    = require('../../controllers/user');
const config  = require('../../../lib/configuration');
const setting = require('../../../lib/setting');

describe('Session', () => {

  beforeAll(() => {

    // 设置sessionstore的mock
    util.mockSessionStore();

    // 设置超时时间为250ms
    config.app = { sessionTimeout: 1 / (60 * 60 * 4) };

    // 设置路由
    setting.board = [
      { api: '/api/user/get', class: 'user', action: 'get' },
      { api: '/api/user/update', class: 'user', action: 'update' }
    ];
    util.setupKoa(['session', 'router'], false);
  });

  afterAll(() => util.clear());
  afterEach(() => jest.resetAllMocks());

  // 保存值到session（用于后续it时确认session是否超时）
  it('save session', async () => {

    const spy = jest.spyOn(user, 'get')
      .mockImplementation(({ ctx }) => ctx.session.user = { id: 1, name: 'light' });
    const res = await util.request.get('/api/user/get');

    // 保存cookie
    this.cookie = res.headers['set-cookie'].join(';');
    expect(spy).toHaveBeenCalled();
  });

  // 获取session中的值（每次访问后，有效期延长）
  it('load session', async () => {

    const spy = jest.spyOn(user, 'update');

    // 第一次休眠100ms
    await util.sleep(100);
    await util.request.get('/api/user/update').set('Cookie', this.cookie);

    expect(spy.mock.calls[0][0].ctx.session.user).toEqual({ id: 1, name: 'light' });

    // 第二次休眠100ms
    await util.sleep(100);
    await util.request.get('/api/user/update').set('Cookie', this.cookie);
    expect(spy.mock.calls[0][0].ctx.session.user).toEqual({ id: 1, name: 'light' });

    // 第三次休眠100ms
    await util.sleep(100);
    await util.request.get('/api/user/update').set('Cookie', this.cookie);
    expect(spy.mock.calls[0][0].ctx.session.user).toEqual({ id: 1, name: 'light' });
  });

  // 浏览器session超时
  it('session timeout browser', async () => {

    // 休眠300ms使session超时
    await util.sleep(300);
    const spy = jest.spyOn(user, 'get');
    await util.request
      .get('/api/user/get').set('Cookie', this.cookie)
      .set('User-Agent', 'Mozilla/5.0 (X11; Linux x86_64; rv:12.0) Gecko/20100101 Firefox/12.0');
    expect(spy.mock.calls[0][0].ctx.session.user).toBeUndefined();
  });

  // 移动端session超时
  it('session timeout mobile', async () => {

    // 200ms超时
    config.app = { sessionTimeoutMobile: 1 / (60 * 60 * 5) };

    const getSpy = jest.spyOn(user, 'get')
      .mockImplementation(({ ctx }) => ctx.session.user = { id: 1, name: 'light' });
    const res = await util.request.get('/api/user/get');
    this.cookie = res.headers['set-cookie'].join(';');
    expect(getSpy).toHaveBeenCalled();

    // 休眠250ms
    await util.sleep(250);

    const updateSpy = jest.spyOn(user, 'update');
    await util.request.get('/api/user/update').set('Cookie', this.cookie);
    expect(updateSpy.mock.calls[0][0].ctx.session.user).toBeUndefined();
  });

});
