/**
 * exporter.test.js
 */

'use strict';

const os       = require('os');
const fs       = require('fs');
const app      = require('../../../../app');
const Exporter = require('../../../../lib/model/datamigrate/exporter');
const CONST    = require('../../../../lib/constant');
const Context  = require('../../../../lib/context');
const pool     = require('../../../../lib/db/mongo/pool');

describe('Exporter', () => {

  beforeAll(async () => {
    process.env.APPLOCAL = true;
    await app.initClient();
  });

  afterAll(() => {
    pool.destroy(true);
  });

  it('parse', async () => {

    const handler = new Context(CONST.DEFAULT_USER);
    const data = { type: 1, lang: [{ en: 1, ja: 1 }] };
    const exporter = new Exporter(handler, {
      schema  : 'i18n',
      mappings: [{
        key       : 'type',
        schema    : 'test',
        conditions: { val: '$type' },
        fields    : ['_id'],
        sanitize  : { rule: 'join', option: ',' }
      }]
    });

    // 通过join关联，并通过sanitize合并
    await exporter.parse(data);
    expect(typeof data.type).toBe('string');
    expect(data.type.includes(',')).toBe(true);

  });

  it('dump', async () => {

    const handler = new Context(CONST.DEFAULT_USER);
    const option = {
      schema  : 'i18n',
      type    : 'excel',
      mappings: [
        { col: 1, key: 'a', title: 'aa' },
        { col: 2, key: 'b', title: 'bb' }
      ]
    };
    const exporter = new Exporter(handler, option);

    // save to file
    exporter.filename = `${os.tmpdir()}/exporter.xlsx`;
    await exporter.dump([{ a: 1, b: 2 }, { a: 3, b: 4 }]);
    expect(fs.statSync(exporter.filename).size).toBe(14738);
    fs.unlinkSync(exporter.filename);

    // save to gridfs
    exporter.filename = undefined;
    const res = await exporter.dump([{ a: 1, b: 2 }, { a: 3, b: 4 }]);
    expect(res.total).toBe(2);
    expect(res._id).toBeTruthy();

  });

  it('exec', async () => {

    const handler = new Context(CONST.DEFAULT_USER, process.env.APPNAME, CONST.SYSTEM_DB_PREFIX);

    handler.params.condition = { type: 'menu', keyword: 'i' };
    const exporter = new Exporter(handler, { name: 'i18n-exp' });
    const filename = `${os.tmpdir()}/exporter.xlsx`;

    await exporter.exec(filename);
    expect(fs.statSync(filename).size).toBe(18249);

  });

});
