/**
 * importer.test.js
 */

'use strict';

const app      = require('../../../../app');
const Importer = require('../../../../lib/model/datamigrate/importer');
const CONST    = require('../../../../lib/constant');
const Context  = require('../../../../lib/context');
const pool     = require('../../../../lib/db/mongo/pool');

describe('Importer', () => {

  beforeAll(async () => {
    process.env.APPLOCAL = true;
    await app.initClient();
  });

  afterAll(() => {
    pool.destroy(true);
  });

  it('initialize', async () => {
    const handler = new Context(CONST.DEFAULT_USER);
    const option = {};
    const importer = new Importer(handler, option);
    const primitive = jest.spyOn(importer.primitive, 'drop');
    const processed = jest.spyOn(importer.processed, 'drop');

    await importer.initialize();
    expect(primitive).toHaveBeenCalled();
    expect(processed).toHaveBeenCalled();
  });

  it('extract', async () => {
    const handler = new Context(CONST.DEFAULT_USER, process.env.APPNAME, CONST.SYSTEM_DB_PREFIX);
    const option = { type: 'excel' };

    handler.params.files = [{ path: 'test/controllers/i18n-by-index.xlsx' }];

    const importer = new Importer(handler, option);
    await importer.extract();
    expect(importer.meta.count).toBe(2);
    expect(importer.titles).toEqual([['type', 'key', 'lang']]);
    expect(importer.primitive.store.length).toBe(2);
  });

  it('transform', async () => {

    const handler = new Context(CONST.DEFAULT_USER, process.env.APPNAME, CONST.SYSTEM_DB_PREFIX);
    const option = { type: 'excel', name: 'i18n-imp' };

    // 加载excel，为 transform 方法做准备
    handler.params.files = [{ path: 'test/controllers/i18n-by-index.xlsx' }];
    const importer = new Importer(handler, option);
    await importer.extract();

    // 校验头信息错误
    importer.titles = [['type', 'key1']];
    await importer.transform();
    expect(importer.log[0].message).toBe('导入文件格式不匹配');

    // 生成一次解析错误
    importer.log = [];
    importer.titles = [['type', 'key', 'lang']];
    importer.allowError = true;
    importer.allowErrorMax = 1;
    jest.spyOn(importer, 'parse').mockImplementation(async data => {
      if (data.key === 'test') {
        importer.log.push({});
      }
      return data;
    });
    await importer.transform();
    expect(importer.processed.store.length).toBe(1);

  });

  it('parse', async () => {

    const handler = new Context(CONST.DEFAULT_USER);
    const importer = new Importer(handler, { type: 'excel' });

    importer.mappings = [
      // valid = 0
      { valid: 0 },
      // sanitize
      { sanitize: { rule: 'split', option: ',' }, key: 'a' },
      // join
      { schema: 'test', conditions: { val: '$b' }, fields: ['createAt'], key: 'b' },
      // uniqueKey
      {}
    ];

    importer.schema = 'test';
    // importer.uniqueKey = ['valid', 'key']; // TODO:
    importer.name = 'i18n-imp';
    const data = {
      a: '1,2,3',
      b: 1
    };
    await importer.parse(data);

    expect(data.a).toEqual(['1', '2', '3']);
    expect(data.b.length).toBeGreaterThan(0);

  });

  it('load', async () => {

    const handler = new Context(CONST.DEFAULT_USER);
    const importer = new Importer(handler, { type: 'excel' });

    importer.meta = { success: 0, inserted: [], updated: [] };
    importer.log = [1];
    importer.allowError = false;
    await importer.load();
    expect(importer.meta.success).toBe(0);

    importer.log = [];
    importer.processed = { list: () => [1, 2] };
    const spy = jest.spyOn(importer, 'add').mockImplementation(async data => data);
    await importer.load();
    expect(spy).toHaveBeenCalledTimes(2);

  });

  it('add', async () => {
    const handler = new Context(CONST.DEFAULT_USER);
    const importer = new Importer(handler, { type: 'excel' });

    importer.schema = 'test';
    await importer.add({ val: 20 });
    const [_id] = importer.meta.inserted;
    expect(_id).toBeTruthy();

    importer.meta.success = 1;
    importer.allowUpdate = true;
    importer.update = [{ index: 1, data: { val: 10, _id } }];
    await importer.add();
    expect(importer.meta.updated.length).toBe(1);
  });

  it('exec', async () => {
  });

});
