/**
 * index.test.js
 */

'use strict';

const rider   = require('../../../../lib/model/datarider/index');
const mongo   = require('../../../../lib/model/datarider/mongo');
const setting = require('../../../../lib/setting');
const CONST   = require('../../../../lib/constant');
const Context = require('../../../../lib/context');
const json    = require('./index.test.json');

describe('Rider', () => {

  beforeAll(() => {
    this.handler = new Context(CONST.DEFAULT_USER, process.env.APPNAME);
    setting.board = json.board;
    setting.struct = json.struct;
    rider.init();
  });

  afterEach(() => {
    jest.resetAllMocks();
  });

  afterAll(() => {
    // pool.destroy(true);
  });

  it('init', () => {
    rider.init();
    expect(rider.test).toHaveProperty('list');
    expect(rider.test).toHaveProperty('update');
    expect(rider.test).toHaveProperty('get');
    expect(rider.test).toHaveProperty('add');
    expect(rider.test).toHaveProperty('remove');
    expect(rider.test).toHaveProperty('count');
    expect(rider.test).toHaveProperty('aggregate');
  });

  it('rider method async', async () => {
    const spy = jest.spyOn(mongo, 'add').mockImplementation(async ({ params }) => {
      expect(params).toEqual({ data: { val: 1 } });
    });
    this.handler.params.data = { val: 1 };
    await rider.test.add(this.handler);
    expect(spy).toHaveBeenCalled();
  });

  it('rider method async error', async () => {
    const spy = jest.spyOn(mongo, 'add').mockImplementation(async () => {
      throw new Error('error');
    });
    try {
      await rider.test.add(this.handler);
    } catch (err) {
      expect(err.message).toBe('error');
    }
    expect(spy).toHaveBeenCalled();
  });

  it('rider method custom parameter', async () => {
    const spy = jest.spyOn(mongo, 'get').mockImplementation(async ({ params }) => {
      expect(params).toEqual({ data: { val: 1 } });
    });
    await rider.test.get(this.handler, { data: { val: 1 } });
    expect(spy).toHaveBeenCalled();
  });

  it('rider method callback', done => {
    const spy = jest.spyOn(mongo, 'add').mockImplementation(async ({ params }) => {
      expect(params).toEqual({ data: { val: 1 } });
    });
    rider.test.add(this.handler, { data: { val: 1 } }, () => done());
    expect(spy).toHaveBeenCalled();
  });

  it('rider method callback error', done => {
    const spy = jest.spyOn(mongo, 'add').mockImplementation(async () => {
      throw new Error('error');
    });
    rider.test.add(this.handler, err => {
      expect(err.message).toBe('error');
      done();
    });
    expect(spy).toHaveBeenCalled();
  });

  it('aggregate async', async () => {
    const spy = jest.spyOn(mongo, 'aggregate').mockImplementation(async (handler, pipeline) => {
      expect(pipeline).toEqual([{ $match: { val: { $ne: 0 } } }]);
    });
    await rider.test.aggregate(this.handler, [{ $match: { val: { $ne: 0 } } }]);
    expect(spy).toHaveBeenCalled();
  });

  it('aggregate callback', done => {
    const spy = jest.spyOn(mongo, 'aggregate').mockImplementation(async (handler, pipeline) => {
      expect(pipeline).toEqual([{ $match: { val: { $ne: 0 } } }]);
    });
    rider.test.aggregate(this.handler, [{ $match: { val: { $ne: 0 } } }], done);
    expect(spy).toHaveBeenCalled();
  });

  it('aggregate callback error', done => {
    const spy = jest.spyOn(mongo, 'aggregate').mockImplementation(async () => {
      throw new Error('error');
    });
    rider.test.aggregate(this.handler, [], err => {
      expect(err.message).toBe('error');
      done();
    });
    expect(spy).toHaveBeenCalled();
  });

  it('transcation', () => {
    rider.startTranscation(this.handler);
    expect(this.handler.transaction.use).toBe(true);
  });

  it('transcation commit', () => {
    const spy = jest.spyOn(mongo, 'commit').mockImplementation(async () => {});
    rider.commitTranscation(this.handler);
    expect(spy).toHaveBeenCalled();
  });

  it('transcation commit callback', done => {
    const spy = jest.spyOn(mongo, 'commit').mockImplementation(async () => {});
    rider.commitTranscation(this.handler, done);
    expect(spy).toHaveBeenCalled();
  });

  it('transcation abort', () => {
    const spy = jest.spyOn(mongo, 'abort').mockImplementation(async () => {});
    rider.abortTranscation(this.handler);
    expect(spy).toHaveBeenCalled();
  });

  it('transcation abort callback', done => {
    const spy = jest.spyOn(mongo, 'abort').mockImplementation(async () => {});
    rider.abortTranscation(this.handler, done);
    expect(spy).toHaveBeenCalled();
  });
});
