/**
 * mongo.test.js
 *
 * 需要设置数据库认证用环境变量:
 *   LIGHTDB_USER
 *   LIGHTDB_PASS
 */

'use strict';

const mongo = require('../../../../lib/model/datarider/mongo');
const setting = require('../../../../lib/setting');
const CONST   = require('../../../../lib/constant');
const Context = require('../../../../lib/context');
const pool    = require('../../../../lib/db/mongo/pool');

describe('Mongo', () => {

  beforeAll(async () => {
    process.env.APPLOCAL = true;
    await setting.init();
  });

  afterAll(() => {
    pool.destroy(true);
  });

  it('add', async () => {
    const handler = new Context(CONST.DEFAULT_USER);
    handler.api = setting.board.find(item => item.api === '/api/test/add');
    handler.params.data = { val: 1 };

    const data = await mongo.add(handler);
    expect(data._id).toBeTruthy();
    expect(data.valid).toBeTruthy();
    expect(data.createAt).toBeTruthy();
    expect(data.updateAt).toBeTruthy();
    expect(data.val).toBe(1);
  });

  it('add (many)', async () => {
    const handler = new Context(CONST.DEFAULT_USER);
    handler.api = setting.board.find(item => item.api === '/api/test/add');
    handler.params.data = [{ val: 2 }, { val: 3 }];

    const data = await mongo.add(handler);
    expect(data.length).toBe(2);
    expect(data[0].val).toBe(2);
    expect(data[1].val).toBe(3);
  });

  it('get (option)', async () => {

    const handler = new Context(CONST.DEFAULT_USER);
    handler.api = setting.board.find(item => item.api === '/api/test/add');

    // 添加被关联数据
    handler.params.data = { val: 10 };
    const { _id: id1 } = await mongo.add(handler);

    // 添加被检索数据
    handler.params.data = { val: 20, link: id1 };
    const { _id: id2 } = await mongo.add(handler);

    // 检索
    handler.api = setting.board.find(item => item.api === '/api/test/get');
    handler.params.id = id2;
    const data = await mongo.get(handler);

    expect(data.val).toBe(20);
    expect(data.options.test[data.link].val).toBe(10);
  });

  it('list (option)', async () => {
    const handler = new Context(CONST.DEFAULT_USER);
    handler.api = setting.board.find(item => item.api === '/api/test/list');

    const data = await mongo.list(handler);
    expect(Object.keys(data.options.test).length).toBeGreaterThanOrEqual(1);
  });

  it('list (free)', async () => {
    const handler = new Context(CONST.DEFAULT_USER);
    handler.api = setting.board.find(item => item.api === '/api/test/list');
    handler.params.free = { val: { $in: ['1', '2'] } };

    const data = await mongo.list(handler);
    expect(data.items.length).toBeGreaterThan(1);
  });

  it('update by id', async () => {
    const handler = new Context(CONST.DEFAULT_USER);

    handler.api = setting.board.find(item => item.api === '/api/test/add');
    handler.params.data = { val: 10 };
    const { _id } = await mongo.add(handler);

    handler.api = setting.board.find(item => item.api === '/api/test/update');
    handler.params.id = _id;
    handler.params.data = { val: 30 };

    const data = await mongo.update(handler);
    expect(data.val).toBe(30);
  });

  it('update by condition', async () => {
    const handler = new Context(CONST.DEFAULT_USER);

    handler.api = setting.board.find(item => item.api === '/api/test/add');
    handler.params.data = { val: 1010 };
    await mongo.add(handler);

    handler.api = setting.board.find(item => item.api === '/api/test/update');
    handler.params.data = { val: 40 };
    handler.params.condition = { val: 1010 };

    const data = await mongo.update(handler);
    expect(data).toBe(1);
  });

  it('count', async () => {
    const handler = new Context(CONST.DEFAULT_USER);
    handler.api = setting.board.find(item => item.api === '/api/test/count');
    handler.params.condition = { val: 40 };

    const data = await mongo.count(handler);
    expect(data).toBeGreaterThan(1);
  });

  it('remove by id', async () => {
    const handler = new Context(CONST.DEFAULT_USER);
    handler.api = setting.board.find(item => item.api === '/api/test/add');
    handler.params.data = { val: 1010 };
    const { _id } = await mongo.add(handler);

    handler.api = setting.board.find(item => item.api === '/api/test/remove');
    handler.params.id = _id;

    const data = await mongo.remove(handler);
    expect(data).toBe(1);
  });

  it('remove by condition', async () => {
    const handler = new Context(CONST.DEFAULT_USER);
    handler.api = setting.board.find(item => item.api === '/api/test/add');
    handler.params.data = { val: 2020 };
    await mongo.add(handler);

    handler.api = setting.board.find(item => item.api === '/api/test/remove');
    handler.params.condition = { val: 2020 };
    const data = await mongo.remove(handler);
    expect(data).toBe(1);
  });

  it('aggregate', async () => {

    const handler = new Context(CONST.DEFAULT_USER);
    const pipeline = [
      { $match: { val: { $ne: 0 } } },
      { $group: { _id: '1', max: { $max: '$val' } } }
    ];

    handler.api = { schema: 'test' };
    const [data] = await mongo.aggregate(handler, pipeline);
    expect(data._id).toBe('1');
    expect(data.max).toBeGreaterThan(1);
  });

  it('format', async () => {

    const board = {
      schema : '',
      selects: [
        { key: 'a', select: true, format: 'toString' },
        { key: 'b.b1', select: true, format: 'toNumber' },
        { key: 'c.c1', select: true, format: 'toBoolean' },
        { key: 'd', select: true, format: 'join', alias: 'dd' },
        { key: 'e.e1', select: true, alias: 'ee' }
      ]
    };
    const data = {
      a: 10,
      b: { b1: '11' },
      c: [{ c1: 'true' }, { c1: 'false' }],
      d: [1, 2, 3],
      e: [{ e1: 1 }, { e1: 2 }]
    };
    const res = await mongo.format(data, board);

    expect(res.a).toBe('10');
    expect(res.b).toEqual({ b1: 11 });
    expect(res.c).toEqual([{ c1: true }, { c1: false }]);
    expect(res.d).toBe('1,2,3');
    expect(res.dd).toBe('1,2,3');
    expect(res.e).toEqual([{ e1: 1, ee: 1 }, { e1: 2, ee: 2 }]);
  });

  it('select', async () => {

    const handler = { params: {} };
    const board = {
      selects: [
        { key: 'a', select: true },
        { key: 'b.b1', select: true },
        { key: 'c', select: false }
      ]
    };

    handler.params = { select: '' };
    expect(mongo.select(handler, board)).toBe('a,b.b1');

    handler.params = { select: 'a,c,d' };
    expect(mongo.select(handler, board)).toBe('a');

    handler.params = { select: { a: 0, 'b.b1': 1, c: 1 } };
    expect(mongo.select(handler, board)).toBe('b.b1');
  });

  it('options', async () => {

    const handler = new Context(CONST.DEFAULT_USER);
    const board = {
      selects: [
        {
          key       : 'name',
          schema    : 'test',
          conditions: { val: '$val' },
          fields    : 'val,valid'
        },
        {
          key       : 'val',
          schema    : 'test',
          conditions: { val: '$val', createBy: '$by', updateBy: CONST.DEFAULT_USER },
          fields    : 'val,valid'
        }
      ]
    };
    const data = {
      items: [
        { val: 1, by: CONST.DEFAULT_USER },
        { val: 2, by: CONST.DEFAULT_USER }
      ]
    };

    await mongo.options(handler, data, board);
    expect(data.options.test[1]).toBeTruthy();
    expect(data.options.test[2]).toBeTruthy();

  });

  it('filter', async () => {

    // use free
    const handler = { params: { free: { a: 1 } } };
    const board = { filters: [], schema: 'test' };
    const res1 = mongo.filter(handler, board);
    expect(res1).toEqual({ a: 1 });

    // no filter & has parent
    setting.struct = [{ schema: 'test', parent: 'user' }];
    handler.params = {};
    const res2 = mongo.filter(handler, board);
    expect(res2).toEqual({ type: 'test', valid: 1 });

    // single params
    board.filters = [{
      group    : '0000',
      key      : 'resource',
      operator : '$eq',
      parameter: 'keyword'
    }];
    handler.params = { condition: { keyword: 'hello' } };
    const res3 = mongo.filter(handler, board);
    expect(res3).toEqual({ type: 'test', resource: { $eq: 'hello' }, valid: 1 });

    // multi params & use default
    board.filters = [{
      group    : '0000',
      key      : 'resource',
      operator : '$eq',
      parameter: 'keyword'
    }, {
      group    : '0001',
      key      : 'name',
      operator : '$regexp',
      parameter: 'name'
    }, {
      default  : 'default',
      group    : '0001',
      key      : 'addr',
      operator : '$eq',
      parameter: ''
    }];
    handler.params = { condition: { keyword: 'hello', name: 'light' } };
    const res4 = mongo.filter(handler, board);
    expect(res4.$or[0]).toEqual({ type: 'test', resource: { $eq: 'hello' }, valid: 1 });
    expect(res4.$or[1]).toEqual({ type: 'test', name: { $regexp: 'light' }, addr: { $eq: 'default' }, valid: 1 });

  });

  it('order', async () => {
    const res1 = mongo.order({ params: { sort: 'a:1, b:-1' } });
    expect(res1).toBe('a:1, b:-1');

    const res2 = mongo.order({ params: {} }, {
      sorts: [
        { key: 'a', index: 2, order: 'desc' },
        { key: 'b', index: 1, order: 'asc' }
      ]
    });
    expect(res2).toEqual([['b', 1], ['a', -1]]);
  });

  it('reserveAll', async () => {
    const handler = {
      uid   : 'light',
      corp  : 'alphabets',
      params: {
        data: {
          a: 1,
          b: '$uid',
          c: '$sysdate',
          d: '$systime',
          e: '$corp'
        }
      }
    };
    mongo.reserveAll(handler);
    expect(handler.params.data.a).toBe(1);
    expect(handler.params.data.b).toBe('light');
    expect(handler.params.data.c instanceof Date).toBe(true);
    expect(handler.params.data.d instanceof Date).toBe(true);
    expect(handler.params.data.e).toBe('alphabets');
  });

});
