/**
 * file.test.js
 */

'use strict';

const fs       = require('fs');
const stream   = require('stream');
const App      = require('../../../app');
const file     = require('../../../lib/model/file');
const CONST    = require('../../../lib/constant');
const Context  = require('../../../lib/Context');
const pool     = require('../../../lib/db/mongo/pool');
const response = require('../../../lib/response');

describe('File', () => {

  this.id = null;
  this.handler = new Context(CONST.DEFAULT_USER, 'light', 'light');

  beforeAll(async () => {
    process.env.APPLOCAL = true;
    await App.initClient();
  });

  afterAll(() => {
    pool.destroy(true);
  });

  // 添加空文件，Promise方式
  it('add (folder)', async done => {
    this.handler.params.files = null;
    this.handler.params.data = {
      filename: 'folder',
      metadata: { description: 'description' }
    };

    // promise
    const result = await file.add(this.handler);
    expect(result._id).toBeTruthy();
    expect(result.length).toBe(0);

    // callback
    file.add(this.handler, ((err, result) => {
      expect(result._id).toBeTruthy();
      expect(result.length).toBe(0);
      done();
    }));
  });

  // 添加文件，Promise方式
  it('add single - promise', async () => {
    this.handler.params.data = { metadata: { description: '首次上传的描述' } };
    this.handler.params.files = [{
      path: 'LICENSE',
      name: 'LICENSE',
      type: 'text/plan'
    }];

    const [result] = await file.add(this.handler);
    expect(result._id).toBeTruthy();
    // expect(result.length).toBe(process.platform === 'darwin' ? 1077 : 1097);
    // expect(result.filename).toBe('LICENSE');
    // expect(result.contentType).toBe('text/plan');
  });

  // 添加文件，callback方式
  it('add multiple - callback', done => {

    this.handler.params.files = [{
      path: 'LICENSE',
      name: 'LICENSE',
      type: 'text/plan'
    },
    {
      stream: fs.createReadStream('LICENSE'),
      name  : 'LICENSE',
      type  : 'text/plan'
    },
    {
      buffer: fs.readFileSync('LICENSE'),
      name  : 'LICENSE',
      type  : 'text/plan'
    }];

    file.add(this.handler, (err, result) => {
      const [f1, f2, f3] = result;
      expect(f1._id).toBeTruthy();
      expect(f1.length).toBe(process.platform === 'darwin' ? 1077 : 1097);
      expect(f1.filename).toBe('LICENSE');
      expect(f1.contentType).toBe('text/plan');
      expect(f2._id).toBeTruthy();
      expect(f3._id).toBeTruthy();
      done();
    });

  });

  // 修改文件
  it('update', async done => {

    this.handler.params.data = { metadata: { description: 'description1' } };
    this.handler.params.files = [{
      path: 'LICENSE',
      name: 'LICENSE',
      type: 'text/plan'
    }];
    const item = await file.add(this.handler);

    this.handler.params.id = item[0]._id;
    this.handler.params.data = { metadata: { path: '/' } };
    this.handler.params.files = [{
      buffer: Buffer.from('string'),
      name  : 'LICENSE_UPDATE',
      type  : 'application/octet-stream'
    }];

    // promise
    const result = await file.update(this.handler);
    expect(result._id).toBeTruthy();

    // callback
    file.update(this.handler, (err, result) => {
      expect(result._id).toBeTruthy();
      done();
    });

  });

  it.only('download (no file)', async () => {
    const sendError = jest.spyOn(response, 'sendError').mockImplementation(() => jest.fn);
    this.handler.params.id = '000000000000';
    await file.download(this.handler);
    expect(sendError.mock.calls[0][1].code).toBe(404);
    expect(sendError.mock.calls[0][1].message).toBe('File not found');
  });

  // 下载文件
  it('download (range)', async () => {

    // 添加一个文件
    this.handler.params.files = [{
      buffer: Buffer.from('string'),
      name  : 'filename',
      type  : 'application/octet-stream'
    }];
    const files = await file.add(this.handler)
      , sendFile = jest.spyOn(response, 'sendFile').mockImplementation(() => jest.fn)
      , headerFunc = jest.fn();

    // range不正确
    sendFile.mockClear();
    headerFunc.mockClear();
    this.handler.params.id = files[0]._id;
    this.handler.ctx = { set: headerFunc };
    this.handler.req = { headers: { range: 'bytes=500-20' } };
    await file.download(this.handler);
    expect(sendFile.mock.calls[0][1]).toBe(406);
    expect(headerFunc.mock.calls).toEqual([['Accept-Ranges', 'bytes'], ['Content-Range', 'bytes */6']]);

    // 正确下载
    sendFile.mockClear();
    headerFunc.mockClear();
    this.handler.params.id = files[0]._id;
    this.handler.ctx = { set: headerFunc };
    this.handler.req = { headers: { range: 'bytes=0-2' } };
    await file.download(this.handler);
    expect(sendFile.mock.calls[0][1]).toBe(206);
    expect(sendFile.mock.calls[0][2]).toBeInstanceOf(stream.Readable);
    expect(headerFunc.mock.calls).toEqual([['Accept-Ranges', 'bytes'], ['Content-Range', 'bytes 0-2/6'], ['Content-Length', 3]]);
  });

  // 下载文件
  it.only('download (full)', async () => {

    // 添加一个文件
    this.handler.params.files = [{
      buffer: Buffer.from('string'),
      name  : 'filename',
      type  : 'application/octet-stream'
    }];

    const files = await file.add(this.handler)
      , sendFile = jest.spyOn(response, 'sendFile').mockImplementation(() => jest.fn)
      , headerFunc = jest.fn();

    this.handler.params.id = files[0]._id;
    this.handler.ctx = { set: headerFunc };
    this.handler.req = { headers: {} };
    await file.download(this.handler);
    expect(sendFile.mock.calls[0][1]).toBe(200);
    expect(sendFile.mock.calls[0][2].stream).toBeInstanceOf(stream.Readable);
  });

  // 读取文件Stream
  it('stream', async done => {

    this.handler.params.files = [{
      buffer: Buffer.from('string'),
      name  : 'filename',
      type  : 'application/octet-stream'
    }];
    const files = await file.add(this.handler);

    // promise
    this.handler.params.id = files[0]._id;
    const result = await file.stream(this.handler);
    expect(result).toBeInstanceOf(stream.Readable);

    // callback
    await file.stream(this.handler, (err, result) => {
      expect(result).toBeInstanceOf(stream.Readable);
      done();
    });

  });

});
