/**
 * socket.test.js
 */

'use strict';

const WebSocket = require('ws');
const config    = require('../../lib/configuration');
const setting   = require('../../lib/setting');
const user      = require('../controllers/user');
const util      = require('./middleware/util');

describe('Socket', () => {

  beforeAll(async () => {

    await setting.init();

    util.mockSessionStore();
    setting.board = [{ api: '/api/user/get', class: 'user', action: 'get' }];
    util.setupKoa(['session', 'router']);

    // 开始socket服务
    this.socket = require('../../lib/socket');
    this.socket.listen(util.server);
  });

  afterAll(() => util.clear());

  it('bad api', async () => {
    const data = await this.emit('/test');
    expect(data.error).toEqual({ code: 400, message: 'can not dispatch' });
  });

  it('no class', async () => {
    const data = await this.emit('/api/clazz/func');
    expect(data.error).toEqual({ code: 400, message: 'can not dispatch' });
  });

  it('no action', async () => {
    const data = await this.emit('/api/user/nofunc');
    expect(data.error).toEqual({ code: 400, message: 'can not dispatch' });
  });

  it('no cookie', async () => {
    const data = await this.emit('/api/user/get');
    expect(data.error).toEqual({ code: 401, message: 'authentication failed' });
  });

  it('not loggin in', async () => {
    const data = await this.emit('/api/user/get', {
      cookie: 'light.sid=53a0db60-73cf-4cf1-980a-23e34211aaa6; path=/; expires=Sat, 14 Mar 2020 03:43:07 GMT; httponly'
    });
    expect(data.error).toEqual({ code: 401, message: 'authentication failed' });
  });

  // 通过websocket调用api
  it('request api', async () => {

    // 设置超时时间为200ms
    config.app = { sessionTimeout: 1 / (60 * 60 * 5) };

    // 保存认证信息
    const spy = jest.spyOn(user, 'get').mockImplementation(({ ctx }) => ctx.session.user = { _id: '001' });
    const res = await util.request.get('/api/user/get');
    expect(res.error).toBeFalsy();

    expect(spy).toBeCalled();
    this.cookie = { cookie: res.header['set-cookie'].join(',') };

    let data = await this.emit('/api/user/update?id=002', this.cookie);
    expect(data.data).toEqual({});

    data = await this.emit('/api/user/callback',  this.cookie);
    expect(data.data).toBe('helloword');

    data = await this.emit('/api/user/callbackerror',  this.cookie);
    expect(data.error).toEqual({ code: 400, message: 'Bad Request' });

    data = await this.emit('/api/user/throwcustom',  this.cookie);
    expect(data.error).toEqual({ code: 400, message: 'Bad Request' });
  });

  // 与后台建立连接，接收后台发送的消息
  it('bind connection', async done => {

    // 建立连接
    const ws = this.connect('/?cid=client-id', this.cookie);

    // 模拟后台发送消息到前台
    ws.on('open', () => this.socket.emit('001', 'hello'));

    // 前台确认接收到消息
    ws.on('message', data => {
      expect(JSON.parse(data).data).toBe('hello');
      ws.close();
      done();
    });
  });

  // session超时
  it('request api timeout', async () => {

    await util.sleep(200);
    const data = await this.emit('/api/user/update?id=002', this.cookie);
    expect(data.error).toEqual({ code: 401, message: 'authentication failed' });
  });

  // session超时
  it('bind connection timeout', async done => {

    const ws = this.connect('/', this.cookie);
    ws.on('message', data => {
      expect(JSON.parse(data).error).toEqual({ code: 401, message: 'authentication failed' });
      ws.close();
      done();
    });
  });

  // 没有指定的用户
  it('emit no user', () => {
    expect(this.socket.emit('666', {})).toBeFalsy();
  });

  this.connect = (api, headers) => {
    const port = process.env.APPPORT || 7000;
    const uri = `ws://127.0.0.1:${port}${api}`;
    return new WebSocket(uri, [], { headers });
  };

  this.emit = (api, headers) => {
    const ws = this.connect(api, headers);
    return new Promise(resolve => {
      ws.on('message', data => {
        ws.close();
        resolve(JSON.parse(data));
      });
    });
  };

});
