/**
 * validator.test.js
 */

'use strict';

const dayjs     = require('dayjs');
const validator = require('../../lib/validator');
const setting   = require('../../lib/setting');
const pool      = require('../../lib/db/mongo/pool');
const Model     = require('../../lib/db/mongo/model');
const CONST     = require('../../lib/constant');

describe('Validator', () => {

  beforeAll(() => {
    this.model = new Model(CONST.SYSTEM_DB, CONST.DEFAULT_TENANT, 'test');

    this.check = async (params, valid, invalid) => {
      const rule = {
        group  : 'g',
        name   : 'n',
        key    : 'key',
        rule   : params.rule,
        message: 'msg',
        option : params.option
      };

      // 校验正常
      setting.validator = [rule];
      const res1 = await validator.check({ params: valid }, 'g');
      expect(res1).toBe(null);

      // 校验异常
      const res2 = await validator.check({ params: invalid }, 'g');
      expect(res2[0].name).toBe('n');
      expect(res2[0].message).toBe('msg');

      // 非严格模式
      rule.strict = false;
      const res3 = await validator.check({ params: {} }, 'g');
      expect(res3).toBe(null);
    };

    this.format = (param, data) => {
      const rule = {
        key     : 'key',
        rule    : param.rule,
        option  : param.option,
        timezone: param.timezone
      };

      return validator.format({ params: { key: data } }, rule);
    };
  });

  afterAll(() => {
    pool.destroy(true);
  });

  describe('rule', () => {

    it('required', async () => {
      await this.check({ rule: 'required' }, { key: 'a' }, {});
    });

    it('unique', async () => {
      await this.model.add({ val: 1, valid: 1 });
      await this.check(
        {
          rule  : 'unique',
          option: {
            schema    : 'test',
            conditions: [{ parameter: 'val', value: '$key' }]
          }
        },
        { key: 'a' },
        { key: 1 }
      );
    });

    it('exists', async () => {
      await this.model.add({ val: 1, valid: 1 });
      await this.check(
        {
          rule  : 'exists',
          option: {
            schema    : 'test',
            conditions: [{ parameter: 'val', value: '$key' }]
          }
        },
        { key: 1 },
        { key: 'a' }
      );
    });

    it('in', async () => {
      await this.check({ rule: 'in', option: ['a', 'b'] }, { key: 'a' }, { key: 'c' });
    });

    it('nin', async () => {
      await this.check({ rule: 'nin', option: ['a', 'b'] }, { key: 'c' }, { key: 'a' });
    });

    it('gt', async () => {
      await this.check({ rule: 'gt', option: 'b' }, { key: 'c' }, { key: 'a' });
    });

    it('lt', async () => {
      await this.check({ rule: 'lt', option: 10 }, { key: 9 }, { key: 11 });
    });

    it('equals', async () => {
      await this.check({ rule: 'equals', option: 'a' }, { key: 'a' }, { key: 'b' });
    });

    it('email', async () => {
      await this.check({ rule: 'email' }, { key: 'a@zf.link' }, { key: 'aaa' });
    });

    it('url', async () => {
      await this.check({ rule: 'url' }, { key: 'http://zf.link' }, { key: 'aaa' });
    });

    it('ip', async () => {
      await this.check({ rule: 'ip' }, { key: '1.2.3.4' }, { key: 'aaa' });
    });

    it('hexcolor', async () => {
      await this.check({ rule: 'hexcolor' }, { key: '0FF0FF' }, { key: 'PPP' });
    });

    it('date', async () => {
      await this.check({ rule: 'date' }, { key: '2020-01-01' }, { key: '2020-13-32' });
      await this.check({ rule: 'date' }, { key: new Date() }, { key: 'abc' });
      await this.check({ rule: 'date' }, { key: dayjs() }, { key: 'qqq' });
    });

    it('numeric', async () => {
      await this.check({ rule: 'numeric' }, { key: '1' }, { key: 'a' });
      await this.check({ rule: 'numeric' }, { key: 10 }, { key: 'b' });
    });

    it('boolean', async () => {
      await this.check({ rule: 'boolean' }, { key: 'true' }, { key: 0 });
      await this.check({ rule: 'boolean' }, { key: false }, { key: 1 });
    });

    it('json', async () => {
      await this.check({ rule: 'json' }, { key: '{"a": 1}' }, { key: '0' });
      await this.check({ rule: 'json' }, { key: '[{"a": 1}]' }, { key: 1 });
    });

    it('array', async () => {
      await this.check({ rule: 'array' }, { key: [] }, { key: {} });
    });

    it('string', async () => {
      await this.check({ rule: 'string' }, { key: '' }, { key: {} });
    });

    it('min', async () => {
      await this.check({ rule: 'min', option: 2 }, { key: 'a' }, { key: 'abc' });
      await this.check({ rule: 'min', option: 2 }, { key: [1] }, { key: [1, 2, 3] });
    });

    it('max', async () => {
      await this.check({ rule: 'max', option: 2 }, { key: 'abc' }, { key: 'a' });
      await this.check({ rule: 'max', option: 2 }, { key: [1, 2, 3] }, { key: [1] });
    });

    it('range', async () => {
      await this.check({ rule: 'range', option: [2, 4] }, { key: 'abc' }, { key: 'a' });
      await this.check({ rule: 'range', option: [2, 4] }, { key: 'ac' }, { key: 'abcde' });
    });

    it('matches', async () => {
      await this.check({ rule: 'matches', option: '^ab$' }, { key: 'ab' }, { key: 'a b' });
    });

    it('checkField', async () => {

      const rule = {
        group  : 'g',
        name   : 'n',
        key    : 'key',
        rule   : 'checkField',
        message: 'msg',
        option : {
          type  : 'index',
          fields: [
            {
              value     : '序号',
              checkField: 'true'
            },
            {
              value     : 'ID',
              checkField: 'true'
            }
          ]
        }
      };

      // 校验正常
      setting.validator = [rule];
      const res1 = await validator.check({ params: { data: ['序号', 'ID'] } }, 'g');
      expect(res1).toBe(null);

      // 校验异常
      const res2 = await validator.check({ params: { data: ['ID', '序号'] } }, 'g');
      expect(res2[0].name).toBe('n');
      expect(res2[0].message).toBe('msg');

      // 切换为用标题匹配
      rule.option.type = 'title';

      setting.validator = [rule];
      const res3 = await validator.check({ params: { data: ['ID', '序号'] } }, 'g');
      expect(res3).toBe(null);

      // 校验异常
      const res4 = await validator.check({ params: { data: ['ID'] } }, 'g');
      expect(res4[0].name).toBe('n');
      expect(res4[0].message).toBe('msg');

    });

  });

  describe('sanitize', () => {

    it('split', async () => {
      const res1 = await this.format({ rule: 'split' }, 'a b,c');
      expect(res1).toEqual(['a', 'b', 'c']);

      const res2 = await this.format({ rule: 'split', option: ',' }, 'a b,c');
      expect(res2).toEqual(['a b', 'c']);

      const res3 = await this.format({ rule: 'split' }, 1);
      expect(res3).toBe(1);
    });

    it('join', async () => {
      const res1 = await this.format({ rule: 'join' }, ['a', 'b']);
      expect(res1).toBe('a,b');

      const res2 = await this.format({ rule: 'join', option: '-' }, ['a', 'b']);
      expect(res2).toBe('a-b');

      const res3 = await this.format({ rule: 'join' }, 1);
      expect(res3).toBe(1);
    });

    it('fix', async () => {
      const res = await this.format({ rule: 'fix', option: 'ab' });
      expect(res).toBe('ab');
    });

    it('mapping', async () => {
      const res1 = await this.format({ rule: 'mapping', option: '1:a' }, 1);
      expect(res1).toBe('a');

      const res2 = await this.format({ rule: 'mapping', option: '1:a,2:b' }, '2');
      expect(res2).toBe('b');
    });

    it('numberFormat', async () => {
      const res1 = await this.format({ rule: 'numberFormat', option: '0.00' }, '10');
      expect(res1).toBe('10.00');

      const res2 = await this.format({ rule: 'numberFormat' }, 10000);
      expect(res2).toBe('10,000.00');

      const res3 = await this.format({ rule: 'numberFormat' }, {});
      expect(res3).toEqual({});
    });

    it('stringFormat', async () => {
      const res1 = await this.format({ rule: 'stringFormat', option: '1%d3' }, 2);
      expect(res1).toBe('123');

      const res2 = await this.format({ rule: 'stringFormat' }, 'abc');
      expect(res2).toBe('abc');

      const res3 = await this.format({ rule: 'stringFormat', option: '%j' }, { a: 1 });
      expect(res3).toBe('{"a":1}');
    });

    it('dateFormat', async () => {
      const res1 = await this.format({ rule: 'dateFormat', option: 'MM-DD' }, dayjs('2020-01-01'));
      expect(res1).toBe('12-31');

      const res2 = await this.format({
        rule    : 'dateFormat',
        option  : 'MM-DD',
        timezone: 'Asia/Shanghai'
      }, dayjs('2020-01-01'));
      expect(res2).toBe('01-01');

      const res3 = await this.format({ rule: 'dateFormat' }, dayjs('2020-01-01'));
      expect(res3).toBe('2019-12-31');
    });

    it('trim', async () => {
      const res1 = await this.format({ rule: 'trim' }, ' 1 ');
      expect(res1).toBe('1');

      const res2 = await this.format({ rule: 'ltrim' }, ' 1 ');
      expect(res2).toBe('1 ');

      const res3 = await this.format({ rule: 'rtrim' }, ' 1');
      expect(res3).toBe(' 1');
    });

    it('toBoolean', async () => {
      const res1 = await this.format({ rule: 'toBoolean' }, 'FALSE');
      expect(res1).toBe(false);

      const res2 = await this.format({ rule: 'toBoolean' }, '0');
      expect(res2).toBe(false);

      const res3 = await this.format({ rule: 'toBoolean' });
      expect(res3).toBe(false);

      const res4 = await this.format({ rule: 'toBoolean' }, true);
      expect(res4).toBe(true);

      const res5 = await this.format({ rule: 'toBoolean' }, {});
      expect(res5).toBe(true);
    });

    it('toDate', async () => {
      const res1 = await this.format({ rule: 'toDate' }, '2020-01-01');
      expect(res1).toEqual(dayjs('2020-01-01').toDate());

      const res2 = await this.format({ rule: 'toDate', option: 'YY-MM-DD' }, '20-01-01');
      expect(res2).toEqual(dayjs('2020-01-01').toDate());
    });

    it('toNumber', async () => {
      const res1 = await this.format({ rule: 'toNumber' }, '1');
      expect(res1).toBe(1);

      const res2 = await this.format({ rule: 'toNumber' }, 'a');
      expect(res2).toBe('a');
    });

    it('toString', async () => {
      const res1 = await this.format({ rule: 'toString' }, 1);
      expect(res1).toBe('1');

      const res2 = await this.format({ rule: 'toString' }, {});
      expect(res2).toBe('{}');

      const res3 = await this.format({ rule: 'toString' }, null);
      expect(res3).toBe('null');

      const res4 = await this.format({ rule: 'toString' }, true);
      expect(res4).toBe('true');
    });
  });

});
